#ifndef _EOS_
#define _EOS_

#include "mc.h"

using namespace mc;

namespace eos {

  // Basic equation of state config
  class EOSBasicConfig {
  public :
    bool EIon;
    // Constructors
    EOSBasicConfig();
    EOSBasicConfig(const bool& _EIon);
    EOSBasicConfig(const string& dataFile);
    // Load data
    void load(const string& dataFile);
    // Log data
    void log();
  };

  // Basic equation of state
  class EOSBasic : public EOS {
  protected :
    EOSBasicConfig config;
  public :
    // Constructors
    EOSBasic(EOSBasicConfig _config=EOSBasicConfig());
    EOSBasic(const string& dataFile);
    // Get gas pressure
    double PGas(Cell* cell, double T=noT) const;
    // Get radiation pressure
    double PRad(Cell* cell, double T=noT) const;
    // Get pressure
    virtual double P(Cell* cell, double T=noT) const;
    // Get ionization energy
    double EIon(Cell* cell) const;
    // Get gas energy (per mass)
    double EGas(Cell* cell, double T=noT) const;
    // Get radiation energy (per mass) in LTE
    double ERad(Cell* cell, double T=noT) const;
    // Get energy (per mass)
    virtual double E(Cell* cell, double T=noT) const;
  };
}

#endif
