#include <vector>
#include <math.h>
#include <string.h>
#include "mc-util.h"
#include "mc-math.h"
#include "mc-physics.h"
#include "mc.h"
#include "msolver.h"
#include "eos.h"

using namespace util;
using namespace math;
using namespace physics;
using namespace mc;
using namespace msolver;
using namespace eos;

int main (int argc, char** argv) {

  // Initialize variables
  double Rph,L;
  double rMax,tMin,tMax;
  int i,j,k,l;
  int n,nEscape,nEvent,nEScatter,nLScatter,nLAbs,nLEmission,nFFAbs,nFFEmission,nBFAbs,nBFEmission,nCCross;
  bool T_converged=false;
  bool L_converged=false;
  stringstream specfile,statfile;
  bool initiate;

  /* Check if enough number of arguments provided */
  if (argc<5) {
    printf("Missing argument\n");
    exit(-1);
  }

  /* Get mandatory arguments */
  string gridDataFile=string(argv[1]);
  double Lph=strtod(argv[2],NULL);
  double Tph=strtod(argv[3],NULL);
  double Lobs=strtod(argv[4],NULL);

  /* Calculate photospheric radius */
  Rph=pow(Lph/(4*pi*sigma*pow(Tph,4)),1.0/2.0);

  cout<<"Lph="<<Lph<<", Rph="<<Rph<<", Tph="<<Tph<<"\n";

  /* Set default values for optional arguments */
  int nTrial=20;
  int nPacket=1000000;
  int nPacketSmall=100000;
  //int nBin=10000;
  int nBin=5000; // Changed!
  int nMuBin=20; // New!
  int nBinSmall=100;
  int nPacketGamma=0;
  double tauLim=1e-3;
  double TTol=1e-2;
  double LTol=5e-2;
  double lambdaMin=1;
  double lambdaMax=25000;
  bool finalMC=false;
  double eScatterLim=10.0;
  bool seed=false;
  bool eScatter=true;
  bool lineScatter=true;
  bool gScatter=false;
  bool freeFreeAbs=false;
  bool boundFreeAbs=false;
  bool flourescence=false;
  bool collProc=false;
  bool nonThermalProc=false;
  bool twoPhotonProc=false;
  bool chargeTransferProc=false;
  bool chargeTransferEmit=false;
  bool transTable=false;
  bool markovTable=false;
  bool markovTableExt=false;
  bool markovTableExtSplit=false;
  double markovTableLim=0.0;
  double markovTableThermLim=0.0;
  double markovTableCTLim=0.0;
  bool collExciteTable=false;
  double collExciteTableLim=0.0;
  bool nonThermalExciteTable=false;
  bool etaTable=false;
  UpdateMethod updateMethod=urPacketHistory;
  bool boundFreeTable=false;
  double boundFreeTableMin=1;
  double boundFreeTableMax=1e5;
  int boundFreeTableNBin=1e4;
  bool boundFreeSTable=false;
  int boundFreeSTableNBin=1e3;
  RadEstimator radEstimator=rfDBB;
  RadEstimator radEstimatorBB=rfDBB;
  bool radEstimatorExp=false;
  BBSampling bbSampling=bbsEqualEnergy;
  BBFrame bbFrame=bbfMixed;
  bool bbLimbDark=true;
  OnTheSpot onTheSpot("");
  OnTheSpotNew onTheSpotNew("");
  FFOnTheSpot ffOnTheSpot(false,10.0);
  bool boost=false;
  bool alfaCorrect=false;
  bool alfaCorrectEmit=false;
  bool alfaSimpleEmit=false;
  bool alfaSPH=false;
  bool forceDBalance=false;
  bool forceDBalanceOSP=false;
  bool boundFreeBB=false;
  double boundFreeBBTauLim=1e3;
  bool convertGPacket=true;
  bool transformKappa=false;
  bool bbEPacket=false;
  string configFile="";
  SolverMethod solverMethod=sNebular;
  bool exportEta=false;
  bool exportSBF=false;
  bool transformRay=false;
  bool limitFreeFreeCRange=false;
  bool JLineFallback=false;
  double JLineFallbackFactor=1.1;
  int JLineFallbackCount=3;
  bool JLineFallbackDBB=false;
  bool trackI=false;
  bool trackHK=false;
  double boundFreeCutOff=0.0;
  double boundFreeCutOffRoss=0.0;
  bool matterDerivedMPI=false;
  bool LTEGroupMacro=false;
  bool exportKappaMC2=true;
  string gridTemplateFile="";
  bool dustAbs=false;

  /* Get optional arguments */
  for (i=5;i<argc;i++) {
    if (strcmp(argv[i],"--ntrial")==0 && i+1<argc) 
      {i++; nTrial=atoi(argv[i]);}
    else if (strcmp(argv[i],"--npacket")==0 && i+1<argc)
      {i++; nPacket=atoi(argv[i]);}
    else if (strcmp(argv[i],"--npacket_small")==0 && i+1<argc)
      {i++; nPacketSmall=atoi(argv[i]);}
    else if (strcmp(argv[i],"--nbin")==0 && i+1<argc)
      {i++; nBin=atoi(argv[i]);}
    else if (strcmp(argv[i],"--nbin_small")==0 && i+1<argc)
      {i++; nBinSmall=atoi(argv[i]);}
    else if (strcmp(argv[i],"--npacket_gamma")==0 && i+1<argc)
      {i++; nPacketGamma=atoi(argv[i]);}
    else if (strcmp(argv[i],"--taulim")==0 && i+1<argc)
      {i++; tauLim=(float)strtod(argv[i],NULL);}
    else if (strcmp(argv[i],"--ttol")==0 && i+1<argc)
      {i++; TTol=(float)strtod(argv[i],NULL);}
    else if (strcmp(argv[i],"--ltol")==0 && i+1<argc)
      {i++; LTol=(float)strtod(argv[i],NULL);}
    else if (strcmp(argv[i],"--lambda_min")==0 && i+1<argc)
      {i++; lambdaMin=(float)strtod(argv[i],NULL);}
    else if (strcmp(argv[i],"--lambda_max")==0 && i+1<argc)
      {i++; lambdaMax=(float)strtod(argv[i],NULL);}
    else if (strcmp(argv[i],"--final_mc")==0 && i+1<argc)
      {i++; if (strcmp(argv[i],"yes")==0) finalMC=true;}
    else if (strcmp(argv[i],"--escatter_lim")==0 && i+1<argc)
      {i++; eScatterLim=(float)strtod(argv[i],NULL);}
    else if (strcmp(argv[i],"--seed")==0 && i+1<argc)
      {i++; if (strcmp(argv[i],"yes")==0) seed=true;}
    else if (strcmp(argv[i],"--escatter")==0 && i+1<argc)
      {i++; if (strcmp(argv[i],"no")==0) eScatter=false;}
    else if (strcmp(argv[i],"--linescatter")==0 && i+1<argc)
      {i++; if (strcmp(argv[i],"no")==0) lineScatter=false;}
     else if (strcmp(argv[i],"--gscatter")==0 && i+1<argc)
      {i++; if (strcmp(argv[i],"yes")==0) gScatter=true;}
    else if (strcmp(argv[i],"--freefree_abs")==0 && i+1<argc)
      {i++; if (strcmp(argv[i],"yes")==0) freeFreeAbs=true;}
    else if (strcmp(argv[i],"--boundfree_abs")==0 && i+1<argc)
      {i++; if (strcmp(argv[i],"yes")==0) boundFreeAbs=true;}
    else if (strcmp(argv[i],"--flourescence")==0 && i+1<argc)
      {i++; if (strcmp(argv[i],"yes")==0) flourescence=true;}
    else if (strcmp(argv[i],"--coll_proc")==0 && i+1<argc)
      {i++; if (strcmp(argv[i],"yes")==0) collProc=true;}
    else if (strcmp(argv[i],"--non_thermal_proc")==0 && i+1<argc)
      {i++; if (strcmp(argv[i],"yes")==0) collProc=true;}
    else if (strcmp(argv[i],"--two_photon_proc")==0 && i+1<argc)
      {i++; if (strcmp(argv[i],"yes")==0) twoPhotonProc=true;}
    else if (strcmp(argv[i],"--charge_transfer_proc")==0 && i+1<argc)
      {i++; if (strcmp(argv[i],"yes")==0) chargeTransferProc=true;}
    else if (strcmp(argv[i],"--charge_transfer_emit")==0 && i+1<argc)
      {i++; if (strcmp(argv[i],"yes")==0) chargeTransferEmit=true;}
    else if (strcmp(argv[i],"--trans_table")==0 && i+1<argc)
      {i++; if (strcmp(argv[i],"yes")==0) transTable=true;}
    else if (strcmp(argv[i],"--markov_table")==0 && i+1<argc)
      {i++; if (strcmp(argv[i],"yes")==0) markovTable=true;}
    else if (strcmp(argv[i],"--markov_table_ext")==0 && i+1<argc)
      {i++; if (strcmp(argv[i],"yes")==0) markovTableExt=true;}
    else if (strcmp(argv[i],"--markov_table_lim")==0 && i+1<argc) 
      {i++; markovTableLim=(float)strtod(argv[i],NULL);}
    else if (strcmp(argv[i],"--markov_table_ext_split")==0 && i+1<argc)
      {i++; if (strcmp(argv[i],"yes")==0) markovTableExtSplit=true;}
    else if (strcmp(argv[i],"--markov_table_therm_lim")==0 && i+1<argc) 
      {i++; markovTableThermLim=(float)strtod(argv[i],NULL);}
    else if (strcmp(argv[i],"--markov_table_ct_lim")==0 && i+1<argc) 
      {i++; markovTableCTLim=(float)strtod(argv[i],NULL);}
    else if (strcmp(argv[i],"--coll_exc_table")==0 && i+1<argc)
      {i++; if (strcmp(argv[i],"yes")==0) collExciteTable=true;}
    else if (strcmp(argv[i],"--coll_exc_table_lim")==0 && i+1<argc) 
      {i++; collExciteTableLim=(float)strtod(argv[i],NULL);}
    else if (strcmp(argv[i],"--non_thermal_exc_table")==0 && i+1<argc)
      {i++; if (strcmp(argv[i],"yes")==0) nonThermalExciteTable=true;}
    else if (strcmp(argv[i],"--eta_table")==0 && i+1<argc)
      {i++; if (strcmp(argv[i],"yes")==0) etaTable=true;}
    else if (strcmp(argv[i],"--update_method")==0 && i+1<argc)
      {
	i++; 
	if (strcmp(argv[i],"on_the_fly")==0) updateMethod=urOnTheFly;
	// Disable ray list option for the moment
	//if (strcmp(argv[i],"on_the_fly_ray")==0) updateMethod=urOnTheFlyRay;
      }
    else if (strcmp(argv[i],"--boundfree_table")==0 && i+1<argc)
      {i++; if (strcmp(argv[i],"yes")==0) boundFreeTable=true;}
    else if (strcmp(argv[i],"--boundfree_table_min")==0 && i+1<argc)
      {i++; boundFreeTableMin==(float)strtod(argv[i],NULL);}
    else if (strcmp(argv[i],"--boundfree_table_max")==0 && i+1<argc)
      {i++; boundFreeTableMax=(float)strtod(argv[i],NULL);}
    else if (strcmp(argv[i],"--boundfree_table_nbin")==0 && i+1<argc)
      {i++; boundFreeTableNBin=atoi(argv[i]);}
    else if (strcmp(argv[i],"--boundfree_stable")==0 && i+1<argc)
      {i++; if (strcmp(argv[i],"yes")==0) boundFreeSTable=true;}
    else if (strcmp(argv[i],"--boundfree_stable_nbin")==0 && i+1<argc)
      {i++; boundFreeSTableNBin=atoi(argv[i]);}
    else if (strcmp(argv[i],"--rad_estimator")==0 && i+1<argc)
      {i++; if (strcmp(argv[i],"lucy")==0) radEstimator=rfLucy;}
    else if (strcmp(argv[i],"--rad_estimator_bb")==0 && i+1<argc)
      {i++; if (strcmp(argv[i],"lucy")==0) radEstimatorBB=rfLucy;}
    else if (strcmp(argv[i],"--rad_estimator_exp")==0 && i+1<argc)
      {i++; if (strcmp(argv[i],"yes")==0) radEstimatorExp=true;}
    else if (strcmp(argv[i],"--bb_sampling")==0 && i+1<argc)
      {i++; if (strcmp(argv[i],"log_frequency")==0) bbSampling=bbsLogFrequency;}
    else if (strcmp(argv[i],"--bb_frame")==0 && i+1<argc)
      {
	i++; 
	if (strcmp(argv[i],"rest")==0) bbFrame=bbfRest;
	if (strcmp(argv[i],"cmf")==0) bbFrame=bbfCMF;
      }
    else if (strcmp(argv[i],"--bb_limb_dark")==0 && i+1<argc)
      {i++; if (strcmp(argv[i],"no")==0) bbLimbDark=false;}
    else if (strcmp(argv[i],"--onthespot")==0 && i+1<argc)
      {i++; onTheSpot=OnTheSpot(argv[i]);}
    else if (strcmp(argv[i],"--onthespot_new")==0 && i+1<argc)
      {i++; onTheSpotNew=OnTheSpotNew(argv[i]);}
    else if (strcmp(argv[i],"--ff_onthespot")==0 && i+1<argc)
      {i++; if (strcmp(argv[i],"yes")==0) ffOnTheSpot.apply=true;}
    else if (strcmp(argv[i],"--ff_onthespot_taulim")==0 && i+1<argc)
      {i++; ffOnTheSpot.tauLim=(float)strtod(argv[i],NULL);}
    else if (strcmp(argv[i],"--boost")==0 && i+1<argc)
      {i++; if (strcmp(argv[i],"yes")==0) boost=true;}
    else if (strcmp(argv[i],"--alfa_correct")==0 && i+1<argc)
      {i++; if (strcmp(argv[i],"yes")==0) alfaCorrect=true;}
    else if (strcmp(argv[i],"--alfa_correct_emit")==0 && i+1<argc)
      {i++; if (strcmp(argv[i],"yes")==0) alfaCorrectEmit=true;}
    else if (strcmp(argv[i],"--alfa_simple_emit")==0 && i+1<argc)
      {i++; if (strcmp(argv[i],"yes")==0) alfaSimpleEmit=true;}
    else if (strcmp(argv[i],"--alfa_SPH")==0 && i+1<argc)
      {i++; if (strcmp(argv[i],"yes")==0) alfaSPH=true;}
    else if (strcmp(argv[i],"--force_dbalance")==0 && i+1<argc)
      {i++; if (strcmp(argv[i],"yes")==0) forceDBalance=true;}
    else if (strcmp(argv[i],"--force_dbalance_osp")==0 && i+1<argc)
      {i++; if (strcmp(argv[i],"yes")==0) forceDBalanceOSP=true;}
    else if (strcmp(argv[i],"--convert_gpacket")==0 && i+1<argc)
      {i++; if (strcmp(argv[i],"no")==0) convertGPacket=false;}
    else if (strcmp(argv[i],"--transform_kappa")==0 && i+1<argc)
      {i++; if (strcmp(argv[i],"yes")==0) transformKappa=true;}
    else if (strcmp(argv[i],"--bb_epacket")==0 && i+1<argc)
      {i++; if (strcmp(argv[i],"yes")==0) bbEPacket=true;}
    else if (strcmp(argv[i],"--boundfree_bb")==0 && i+1<argc)
      {i++; if (strcmp(argv[i],"yes")==0) boundFreeBB=true;}
    else if (strcmp(argv[i],"--boundfree_bb_taulim")==0 && i+1<argc)
      {i++; boundFreeBBTauLim=(float)strtod(argv[i],NULL);}
    else if (strcmp(argv[i],"--config_file")==0 && i+1<argc)
      {i++; configFile=string(argv[i]);}
    else if (strcmp(argv[i],"--msolver_method")==0 && i+1<argc)
      {
	i++; 
	if (strcmp(argv[i],"nlte")==0) solverMethod=sNLTE;
	if (strcmp(argv[i],"comb")==0) solverMethod=sComb;
      } 
    else if (strcmp(argv[i],"--export_eta")==0 && i+1<argc)
      {i++; if (strcmp(argv[i],"yes")==0) exportEta=true;}
    else if (strcmp(argv[i],"--export_sbf")==0 && i+1<argc)
      {i++; if (strcmp(argv[i],"yes")==0) exportSBF=true;}
    else if (strcmp(argv[i],"--transform_ray")==0 && i+1<argc)
      {i++; if (strcmp(argv[i],"yes")==0) transformRay=true;}
    else if (strcmp(argv[i],"--limit_freefree_crange")==0 && i+1<argc)
      {i++; if (strcmp(argv[i],"yes")==0) limitFreeFreeCRange=true;}
    else if (strcmp(argv[i],"--jline_fallback")==0 && i+1<argc)
      {i++; if (strcmp(argv[i],"yes")==0) JLineFallback=true;}
    else if (strcmp(argv[i],"--jline_fallback_factor")==0 && i+1<argc)
      {i++; JLineFallbackFactor=(float)strtod(argv[i],NULL);}
    else if (strcmp(argv[i],"--jline_fallback_count")==0 && i+1<argc) 
      {i++; JLineFallbackCount=atoi(argv[i]);}
    else if (strcmp(argv[i],"--jline_fallback_dbb")==0 && i+1<argc)
      {i++; if (strcmp(argv[i],"yes")==0) JLineFallbackDBB=true;}
    else if (strcmp(argv[i],"--track_I")==0 && i+1<argc)
      {i++; if (strcmp(argv[i],"yes")==0) trackI=true;}
    else if (strcmp(argv[i],"--track_HK")==0 && i+1<argc)
      {i++; if (strcmp(argv[i],"yes")==0) trackHK=true;}
    else if (strcmp(argv[i],"--boundfree_cutoff")==0 && i+1<argc)
      {i++; boundFreeCutOff=(float)strtod(argv[i],NULL);}
    else if (strcmp(argv[i],"--boundfree_cutoff_ross")==0 && i+1<argc)
      {i++; boundFreeCutOffRoss=(float)strtod(argv[i],NULL);}
    else if (strcmp(argv[i],"--matter_derived_mpi")==0 && i+1<argc)
      {i++; if (strcmp(argv[i],"yes")==0) matterDerivedMPI=true;}
    else if (strcmp(argv[i],"--lte_group_macro")==0 && i+1<argc)
      {i++; if (strcmp(argv[i],"yes")==0) LTEGroupMacro=true;}
    else if (strcmp(argv[i],"--export_kappa_mc2")==0 && i+1<argc)
      {i++; if (strcmp(argv[i],"no")==0) exportKappaMC2=false;}
    else if (strcmp(argv[i],"--grid_template_file")==0 && i+1<argc)
      {i++; gridTemplateFile=string(argv[i]);}
  }

  // Seed if requested
  if (seed) srand(time(NULL));

  // Create, configure and load grid
  cout<<"lambdaMin: "<<lambdaMin<<", lamdaMax: "<<lambdaMax<<"\n";
  FrequencyBin nu(c/(lambdaMax*AA),c/(lambdaMin*AA));
  cout<<"nu.min: "<<nu.min<<", nu.max: "<<nu.max<<"\n";
  LineSelector* lineSelector = new LineSelector(nu,tauLim);
  CellConfig* cellConfig = new CellConfig(true,neIteration,lineSelector,updateMethod,boundFreeTable,boundFreeTableMin,boundFreeTableMax,boundFreeTableNBin,boundFreeSTable,boundFreeSTableNBin,radEstimator,radEstimatorBB,radEstimatorExp,transTable,markovTable,markovTableExt,markovTableExtSplit,markovTableLim,markovTableThermLim,markovTableCTLim,nonThermalExciteTable,collExciteTable,collExciteTableLim,freeFreeAbs,boundFreeAbs,collProc,nonThermalProc,twoPhotonProc,chargeTransferProc,onTheSpot,onTheSpotNew,ffOnTheSpot,false,boundFreeBB,1e3,boundFreeBBTauLim,bbTCur,exportEta,transformRay,limitFreeFreeCRange,JLineFallback,JLineFallbackFactor,JLineFallbackCount,JLineFallbackDBB,trackI,trackHK,boundFreeCutOff,boundFreeCutOffRoss,boost,alfaCorrect,alfaCorrectEmit,alfaSimpleEmit,alfaSPH,forceDBalance,forceDBalanceOSP,chargeTransferEmit,etaTable,exportSBF,nu.min,nu.max,nBin,nMuBin,LTEGroupMacro,exportKappaMC2);
  BStateSelector* bStateSelector= new BStateSelector();
  TransBBSelector* transSelector = new TransBBSelector(FrequencyBin(0,1e99));
  GridConfig* gridConfig = new GridConfig(cellConfig,bStateSelector,transSelector,updateMethod,matterDerivedMPI);
  MStateSolver* mStateSolver;
  switch (solverMethod) {
  case sNLTE : { mStateSolver = new NLTESolver(configFile); break; }
  case sComb : { mStateSolver = new CombSolver(configFile); break; }
  deafult : { mStateSolver = new NebSolver(configFile); } } 
  EOS* eos = new EOSBasic(configFile);
  GridSS grid(mStateSolver,eos,gridConfig);
  if (gridTemplateFile.compare("")) grid.processGridTemplate(gridTemplateFile,gridDataFile);
  grid.load(gridDataFile,Rph,Lph);

  // Configure MC experiment
  PacketConfig* packetConfig = new PacketConfig(true,false,lineScatter,eScatter,false,gScatter,freeFreeAbs,boundFreeAbs,false,0,dustAbs,flourescence,collProc,nonThermalProc,twoPhotonProc,chargeTransferProc,transTable,updateMethod,boost,convertGPacket,-1,-1,false,1.0,false,1.0,false,transformKappa,bbEPacket);
  MCConfig* mCConfig = new MCConfig(packetConfig);
  PacketSampler* packetSampler = NULL;
  PacketStat* packetStat = NULL;
  MCExperiment* mCExperiment = NULL;

  vector<double> TOld(grid.cell.size(),0);
  rMax=((CellSS*)grid.cell[grid.cell.size()-1])->rMax();
  tMin=grid.cell[0]->tMin();
  tMax=grid.cell[0]->tMax();
  // Temperature loop
  for (i=0;i<nTrial;i++) {
    // Calculate initiate flag
    initiate=(i==0 and j==0);
    // Store old cell temperatures
    for (j=0;j<grid.cell.size();j++) TOld[j]=grid.cell[j]->T();
    // Update matter state
    cout<<"Update matter state\n";
    grid.updateMatterState(initiate);
    // Check cell temperature convergence
    cout<<"Check temperature convergence\n";
    if (!(i==0)) {
      T_converged=true;
      for (j=0;j<grid.cell.size();j++) {
	cout<<"Temperature error in cell "<<j<<" : "<<fabs((grid.cell[j]->T()-TOld[j])/grid.cell[j]->T())<<"\n";
	if (fabs((grid.cell[j]->T()-TOld[j])/grid.cell[j]->T())>TTol) { 
	  T_converged=false;
	  // break;
	}
      }
    }
    // End loop if temperature has converged
    if (T_converged) break;
    // Create new small Monte Carlo experiment
    cout<<"Create new small Monte Carlo experiment: Rph="<<Rph<<", Lph="<<Lph<<", Tph="<<Tph<<", number of packets: "<<nPacketSmall<<", number of bins: "<<nBinSmall<<"\n";
    if (mCExperiment!=NULL) delete(mCExperiment);
    if (packetSampler!=NULL) delete(packetSampler);
    if (packetStat!=NULL) delete(packetStat);
    packetSampler = new PacketSamplerBB(Rph,Lph,nPacketSmall,nBinSmall,nu,bbSampling,bbFrame,bbLimbDark,&grid,packetConfig);
    packetStat = new PacketStat(&grid);
    mCExperiment = new MCExperiment(packetSampler,packetStat,mCConfig);
    // If on-the-fly update method reset grid MC statistics
    if (updateMethod==urOnTheFly) grid.resetMCStatistics(mCExperiment->nu);
    // Run small Monte Carlo experiment
    cout<<"Run small Monte Carlo experiment\n";
    mCExperiment->run();
    // Get statistics
    //mCExperiment->getStatistics(n,nEscape,nEvent,nEScatter,nLScatter,nLAbs,nLEmission,nFFAbs,nFFEmission,nBFAbs,nBFEmission,nCCross,tMin,tMax,rMax);
    //cout<<"Escape fraction: "<<float(nEscape)/float(n)<<", electon scatterings: "<<float(nEScatter)/float(n)<<", line scatterings: "<<float(nLScatter)/float(n)<<", free-free absorptions: "<<float(nFFAbs)/float(n)<<", cell crossings: "<<float(nCCross)/float(n)<<"\n"; 

    // End loop if electron scatter limit is exceeded
    cout<<"nEscatter="<<float(nEScatter)/float(n)<<", eScatterLim="<<eScatterLim<<"\n";
    if (float(nEScatter)/float(n)>eScatterLim) break;
    // Update radiation state
    cout<<"Update radiation state\n";
    grid.updateRadiationState(*mCExperiment);
  }
  if (!T_converged) { cerr<<"Error: Temperature did not converge\n";}
  else cout<<"Temperature converged in "<<i+1<<" iterations\n";

  // Get luminosity
  L=packetStat->L();
  // Check if luminosity converged
  if (fabs(Lobs-L)/Lobs<LTol) {cout<<"Luminosity converged\n"; L_converged=true;}
  else cout<<"Luminosity did not converge: (Lobs-L)/Lobs="<<(Lobs-L)/Lobs<<", LTol="<<LTol<<"\n";
    
  // If temperature and luminosity converged
  if (T_converged and (L_converged or finalMC)) {
    // Create final Monte Carlo Experiment
    cout<<"Create final Monte Carlo experiment: Rph="<<Rph<<", Lph="<<Lph<<", Tph="<<pow(Lph/(4*pi*pow(Rph,2)*sigma),1.0/4.0)<<", number of packets: "<<nPacket<<", number of bins: "<<nBin<<"\n";
    if (mCExperiment!=NULL) delete(mCExperiment);
    if (packetSampler!=NULL) delete(packetSampler);
    if (packetStat!=NULL) delete(packetStat);
    packetSampler = new PacketSamplerBB(Rph,Lph,nPacket,nBin,nu,bbSampling,bbFrame,bbLimbDark,&grid,packetConfig);
    packetStat = new PacketStat(&grid,true,packetSampler->nu);
    mCExperiment = new MCExperiment(packetSampler,packetStat,mCConfig);
    // If on-the-fly update method reset grid MC statistics
    if (updateMethod==urOnTheFly) grid.resetMCStatistics(mCExperiment->nu);
    // Run final Monte Carlo experiment
    cout<<"Run final Monte Carlo experiment\n";
    mCExperiment->run();
  }
  
  // Allways export spectrum and statistics
  cout<<"Export spectrum and grid statistics\n";
  packetStat->exportSpectrum("spectrum.dat");
  grid.exportStatistics("grid-stat.dat",1.0);
}
