#ifndef _MC_PHYSICS_
#define _MC_PHYSICS_

#include <vector>
#include <math.h>

using namespace std;
using namespace math;

namespace physics {

  // Speed of light
  const double c=2.99792458e8;
  // Plancks constant
  const double h=6.62606896e-34;
  const double hBar=h/(2*pi);
  // Gravitational constant
  const double G=6.67384e-11;
  // Boltzman constant
  const double k=1.3806504e-23;
  // Stefan Boltzman constant
  const double sigma=5.670500e-8;
  // Radiation constant
  const double a=4*sigma/c;
  // Bohr radius
  const double a0=5.2917721092e-11;
  // Vacuum permittivity
  const double epsilon0=8.854187817e-12;
  // Atomic mass unit
  const double u=1.660538782e-27;
  // Electron mass
  const double me=9.10938215e-31;
  // Electron charge
  const double ee=1.60217657e-19;
  // Electron volt
  const double eV=1.6021765e-19;
  // Angstrom
  const double AA=1e-10;
  // erg
  const double erg=1e-7;
  // cm
  const double cm=1e-2;
  // gram
  const double gram=1e-3;
  // minute
  const double minute=60;
  // hour
  const double hour=60*minute;
  // day
  const double day=24*hour;
  // year
  const double year=365.25*day;
  // Thomson scattering cross section
  const double sigmaT=6.6524586e-29;
  // Solar radius
  const double RSun=6.96e8;
  // Solar mass
  const double MSun=1.9891e30;
  // Solar luminosity
  const double LSun=3.846e26;
  // Mega-barn
  const double Mb=1e-22;

  // Energi per mass and time for 56Ni gamma-ray decay (Jerkstrand et al. 2013)
  const double eGamma56Ni=3.86e6;
  // Mean frequeny for the 56Ni gamma-ray-decay (fix this);
  const double nuGamma56Ni=0.5e6*eV/h;
  // Effective compton scattering cross section for 56Ni gamma-rays (Jerkstrand et al. 2013)
  //const double sigmaGamma56Ni=0.030*(u/0.5)*pow(cm,2)*pow(gram,-1);
  const double sigmaGamma56Ni=0.030*pow(cm,2)*pow(gram,-1);
  // E-folding time for the 56Ni decay (Jerkstrand et al. 2013)
  const double t56Ni=8.77*day;
  // Total energy per mass and time for the 56Ni decay
  const double e56Ni=eGamma56Ni;

  // Energi per mass and time for 56Co gamma-ray decay (Jerkstrand et al. 2013)
  const double eGamma56Co=6.31e5;
  // Mean frequeny for the 56Co gamma-ray-decay (fix this);
  const double nuGamma56Co=1.0e6*eV/h;
  // Effective compton scattering cross section for 56Co gamma-rays (Jerkstrand et al. 2013)
  //const double sigmaGamma56Co=0.030*(u/0.5)*pow(cm,2)*pow(gram,-1);
  const double sigmaGamma56Co=0.030*pow(cm,2)*pow(gram,-1);
  // Energi per mass and time for 56Co positrons, electrons and X-rays (Jerkstrand et al. 2013)
  const double ePos56Co=2.23e4+6.59e2;
  // Mean frequeny for the 56Co positrons, electrons and X-rays (fix this);
  const double nuPos56Co=1.0e6*eV/h;
  // Effective compton scattering cross section for 56Ni positrons, electrons ans X-rays (infinite)
  const double sigmaPos56Co=1e99;
  // E-folding time for the 56Co decay (Jerkstrand et al. 2013)
  const double t56Co=111.5*day;
  // Total energy per mass and time for the 56Co decay
  const double e56Co=eGamma56Co+ePos56Co;

  // Energi per mass and time for 57Ni gamma-ray decay (fix this)
  const double eGamma57Ni=0.0;
  // Mean frequeny for the 57Ni gamma-ray-decay (fix this);
  const double nuGamma57Ni=1.37e6*eV/h;
  // Effective compton scattering cross section for 57Ni gamma-rays (fix this)
  const double sigmaGamma57Ni=0.0;
  // E-folding time for the 57Ni decay (Claes Lecture Notes)
  const double t57Ni=52.0*hour;
  // Total energy per mass and time for the 57Ni decay
  const double e57Ni=eGamma57Ni;

  // Energi per mass and time for 57Co gamma-ray decay (Jerkstrand et al. 2013)
  const double eGamma57Co=6.71e3;
  // Mean frequeny for the 57Co gamma-ray-decay (fix this);
  const double nuGamma57Co=0.122e6*eV/h;
  // Effective compton scattering cross section for 57Co gamma-rays (Jerkstrand et al. 2013)
  //const double sigmaGamma57Co=0.072*(u/0.5)*pow(cm,2)*pow(gram,-1);
  const double sigmaGamma57Co=0.072*pow(cm,2)*pow(gram,-1);
  // Energi per mass and time for 57Co X-rays (Jerkstrand et al. 2013)
  const double ePos57Co=8.79e2;
  // Mean frequeny for the 57Co X-rays (fix this);
  const double nuPos57Co=0.1e6*eV/h;
  // Effective compton scattering cross section for 57Co X-rays (infinite)
  const double sigmaPos57Co=1e99;
  // E-folding time for the 57Co decay (Jerkstrand et al. 2013)
  const double t57Co=392.1*day;
  // Total energy per mass and time for the 57Co decay
  const double e57Co=eGamma57Co+ePos57Co;

  // Energi per mass and time for 44Ti gamma-ray decay (Jerkstrand et al. 2013)
  const double eGamma44Ti=1.89e3;
  // Mean frequeny for the 44Ti gamma-ray-decay (fix this);
  const double nuGamma44Ti=1.157e6*eV/h;
  // Effective compton scattering cross section for 56Co gamma-rays (Jerkstrand et al. 2013)
  //const double sigmaGamma44Ti=0.040*(u/0.5)*pow(cm,2)*pow(gram,-1);
  const double sigmaGamma44Ti=0.040*pow(cm,2)*pow(gram,-1);
  // Energi per mass and time for 44Ti positrons (Jerkstrand et al 2013)
  const double ePos44Ti=4.82e2;
  // Mean frequeny for the 44Ti positron decay (fix this);
  const double nuPos44Ti=1.0e6*eV/h;
  // Effective compton scattering cross section for 44Ti positrons (infinite)
  const double sigmaPos44Ti=1e99;
  // E-folding time for the 44Ti decay (Claes Lecture Notes)
  const double t44Ti=87.0*year;
  // Total energy per mass and time for the 44Ti decay
  const double e44Ti=eGamma44Ti+ePos44Ti;

  // Number density of next ionization state (Saha equation)
  double nIonLTE (const double& n1, const double& ne, const double& T, const double& Eion, const double& g1, const double& g2);

  // Number density of next excitation state (Boltzman equation)
  double nBoundLTE (const double& n1, const double& T, const double& E1, const double& E2, const double& g1, const double& g2);
  
  // NLTE number densities of bound states ignoring ionization and recombination
  /*vector<double> physics::nBoundNLTE (const double& nIon, const double& T, const double& ne, const vector<vector<double> >& J, const vector<vector<double> >& A, const vector<vector<double> >& upsilon, const vector<vector<double> >& E, const vector<double>& g, const vector<vector<double> > beta);*/

  // Partition function
  double ULTE(const double& T, const vector<double>& E, const vector<double>& g);

  // Sobolev optical depth
  double tauSobolev(const double& Eul, const double& nl, const double& nu, const double& A, const double& gu, const double& gl, const double& drdv );

  // Sobolev optical depth lower state number derivative
  double dTauSobolevdNl(const double& Eul, const double& nl, const double& nu, const double& A, const double& gu, const double& gl, const double& drdv );

  // Sobolev optical depth upper state number derivative
  double dTauSobolevdNu(const double& Eul, const double& nl, const double& nu, const double& A, const double& gu, const double& gl, const double& drdv );

  // Free-free class of functions
  class FreeFree {
  public : 
    // Emissivity
    // Eq. 5.14b in Ribicki & Lightman * (1/4pi)
    static double eta(const double& ne, const double& n, const double& Z, const double& T, const double& nu);

    // Opacity (assuming thermal emission -> k_nu=j_nu/B_nu)
    // Eq. 5.18b in Ribicki & Lightman
    static double kappa(const double& ne, const double& n, const double& Z, const double& T, const double& nu, const double rho);

    // Cooling rate
    // Eq. 5.15b in Ribicki & Lightman
    static double cooling(const double& ne, const double& n, const double& Z, const double& T);

    // Heating rate (assuming diluted BB radiation field)
    static double heatingDBB(const double& ne, const double& n, const double& Z, const double& nu, const double& T, const double& TR, const double& W);

    // Frequency integrated heating rate (assuming diluted BB radiation field)
    static double totalHeatingDBB(const double& ne, const double& n, const double& Z, const double& T, const double& TR, const double& W, const double& tol=1e-6);

    // Frequency that gives the (fractional) frequency intergrated emissivity z over the intervall (0,nu)
    //static double etaNu(const double& z, const double& T);
    static double etaNu(const double& z, const double& T, double nuMin=0, double nuMax=1e99);
  };

  // Bound-free class of functions
  class BoundFree {
  public:
    // Photo-ionization cross-section (Hydrogenic approximation)
    static double sigmaRadH(const double& C, const double& Xi, const double& E, const double& nu);
    // Recombination coefficient (Spnotaneous; Hydrogenic approximation)
    static double alfaSPH(const double& C, const double& Xi0, const double& E, const double& g1, const double& g2, const double& T, double nuMin=0, double nuMax=1e99, bool modified=false);
    // Photo-ionization cross-section (Lucy test case)
    static double sigmaRadL(const double& C, const double& Xi0, const double& E, const double& nu);
    // Collision cross-section (calculated from collisional strength)
    static double sigmaColl(const double& Xi, const double& g1, const double& omega, const double& T, const double& ne);
    // Collisional cross-section (Seaton approximation)
    static double sigmaCollS(const double& C, const double& Xi, const double& sigmaRad0, const double& T, const double& ne);
    /*
    // Get opacity (per volume) from cross-section
    static double k(const double& sigma, const double& n);
    // Get opacity (per volume) from cross-section (corrected for stimulated emission)
    static double kST(const double& sigma, const double& n, const double& nIon, const double& ne, const double& g, const double& gIon, const double& Xi, const double& T, const double& nu);
    // Get emissivity from cross-section (spontaneous emission only)
    static double eta(const double& sigma, const double& nIon, const double& ne, const double& g, const double& gIon, const double& Xi, const double& T, const double& nu);
    // Get emissivity from cross-section (including stimulated emission)
    static double etaST(const double& sigma, const double& nIon, const double& ne, const double& g, const double& gIon, const double& Xi, const double& T, const double& J, const double& nu);
    // Frequency that gives the (fractional) frequency intergrated (spontaneous) emissivity z over the intervall (0,nu)
    static double etaNu(const double& z, const double& T, const double Xi0, const double& E);
    */
  };

  // Bound-bound class of functions
  class BoundBound {
  public:
    //static double gamma(const double& g, const double& u0);
    static double gammaNeu(const double& u0);
    static double gamma(const double& u0, const double& C);
    // Collision cross-section (calculated from collisional strength)
    static double sigmaColl(const double& Xi, const double& g1, const double& omega, const double& T, const double& ne);
    // Collisional cross-section (Regemorter approximation)
    static double sigmaCollR(const double& Xi, const double& g1, const double& g2, const double& A, const double& C, const double& T, const double& ne);
  };

  // Blackbody class of functions
  class BlackBody {
  private :
    double _T;
  public : 
    // Intensity
    static double B(const double& nu, const double& T);

    // Frequency integrated intensity
    static double B(const double& nu1, const double& nu2, const double& T, const double& tol=1e-6);

    // Total frequency integrated intensity
    static double B(const double& T);

    // Temperature derivative of intensity
    static double dBdT(const double& nu, const double& T);

    // Temperature derivative of frequency integrated intensity - New!
    static double dBdT(const double& nu1, const double& nu2, const double& T, const double& tol=1e-6);
    
    // Temperature derivative of total frequency integrated flux intensity
    static double dBdT(const double& T);

    // Peak intensity frequency
    static double nuMax(const double& T);

    // Frequency intervall corresponding to a given intergrated intensity
    static double nuB(const double& _B, const double& T, const double& nu1, const double& bTol=1e-6);

    // Flux (isotropic radiation from a sphere)
    static double F(const double& nu, const double& T);

    // Frequency integrated flux (isotropic radiation from a sphere)
    static double F(const double& nu1, const double& nu2, const double& T, const double& tol=1e-6);

    // Total frequency integrated flux (isotropic radiation from a sphere)
    static double F(const double& T);

    // Frequency intervall corresponding to a given intergrated flux
    static double nuF(const double& F, const double& T, const double& nu1, const double& bTol=1e-6);
  };

  // Calculate late new mass fractions in the decay chain of 56Ni
  double decay56Ni(double& X56Ni, double& X56Co, double& X56Fe, const double& t1, const double& t2);

  // Calculate late new mass fractions in the decay chain of 57Ni
  double decay57Ni(double& X57Ni, double& X57Co, double& X57Fe, const double& t1, const double& t2);

  // Calculate late new mass fractions in the decay chain of 44Ti
  double decay44Ti(double& X44Ti, double& X44Ca, const double& t1, const double& t2);
}

#endif
