#include <cstdlib>
#include <iostream>
#include <fstream>
#include <sstream>
#include <string>
#include "mc-util.h"

using namespace std;
using namespace util;

// Load parameter data
void util::pstring::load(const string& dataFile) {
  fstream file;
  stringstream datastream,linestream;
  string data,line,wsLine,item;

  // Open data file
  file.open(dataFile.c_str());
  // Error handling
  if (file.fail()) {cerr<<"pstring::load: Error1\n"; return;}
  
  // Read all data from file
  while (!file.eof()) {
    getline(file,line);
    // Error handling
    if (file.fail() && !file.eof()) {cerr<<"pstring:load: Error2\n"; return;}
    if (file.fail() && file.eof()) break;
    // Ignore lines starting with #
    linestream.str(line);
    linestream.clear();
    line="";
    while (!linestream.eof()) {
      linestream>>skipws>>item;
      line.append(item);
    }
    if (line[0]!='#') {
	// Check + and = operators
	data.append(line);
      }
  }

  // Strip white space from data
  datastream.str(data);
  while (!datastream.eof()) {
    datastream>>skipws>>item;
    (*this).append(item);
  }

  // Close data file
  file.close();

  return;
}

// Save paramater data
void util::pstring::save(const string& dataFile) {
  fstream file;

  // Open data file
  file.open(dataFile.c_str());
  // Error handling
  if (file.fail()) {cerr<<"pstring::save: Could not open file\n"; return;}

  // Write data to file
  file.write((*this).c_str(),(*this).size());
  // Error handling
  if (file.fail()) {cerr<<"pstring::save: Could not save data to file\n"; return;}

  // Close data file
  file.close();

  return;
}

// Get parameter data
pstring util::pstring::getData(const string& parameter) {
  pstring paramData;
  int nextpos,startpos,endpos,count;

  // Find paramater start position
  startpos=(*this).find(parameter+"=",0);
  // If paramameter found
  if (startpos!=npos) {
    // Calculate parameter data start position
    startpos+=(parameter.size()+1);
    switch ((*this)[startpos]) {
    case '{' :
      // Extract structured data
      nextpos=startpos;
      count=1;
      // Search for matching }
      while (nextpos!=npos && count!=0) {
	nextpos=(*this).find_first_of("{}",nextpos+1);
	if (nextpos!=npos) {
	    if ((*this)[nextpos]=='{') count++;
	    if ((*this)[nextpos]=='}') count--;

	  }
      }
      // If found
      if (count==0) {
	endpos=nextpos;
	// = operator does not work as we want
	paramData.assign((*this).substr(startpos,endpos-startpos+1));
      }
      break;
    case '[' :
      // Extract array data
      nextpos=startpos;
      count=1;
      // Search for matching ]
      while (nextpos!=npos && count!=0) {
	nextpos=(*this).find_first_of("[]",nextpos+1);
	if (nextpos!=npos) {
	    if ((*this)[nextpos]=='[') count++;
	    if ((*this)[nextpos]==']') count--;

	  }
      }
      // If found
      if (count==0) {
	endpos=nextpos;
	// = operator does not work as we want
	paramData.assign((*this).substr(startpos,endpos-startpos+1));
      }
      break;
    default :
      // Exctract value data
      endpos=(*this).find_first_of(",}]",startpos);
      // = operator does not work as we want
      if (endpos!=npos) paramData.assign((*this).substr(startpos,endpos-startpos));
      else paramData.assign((*this).substr(startpos,npos));
    }
  }
 
  // Return paramater data
  return paramData;
}
 
// Get paramater data (from array)
pstring util::pstring::getData(const string& parameter, const int& i) {
  pstring paramData;
  int nextpos,startpos,endpos,aCount,sCount,j;

  // Find paramater start position
  startpos=(*this).find(parameter+"=",0);
  // If paramameter found
  if (startpos!=npos) {
    // Calculate parameter data start position
    startpos+=(parameter.size()+1);
    switch ((*this)[startpos]) {
    case '[' :
      // Extract array data
      nextpos=startpos;
      aCount=1;
      // Search for matching ]
      while (nextpos!=npos && aCount!=0) {
	nextpos=(*this).find_first_of("[]",nextpos+1);
	if (nextpos!=npos) {
	  if ((*this)[nextpos]=='[') aCount++;
	  if ((*this)[nextpos]==']') aCount--;
	}
      }
      // If found
      if (aCount==0) {
	endpos=nextpos;
	nextpos=startpos;
	startpos=startpos+1;
	sCount=0;
	aCount=1;
	j=-1;
	// Search for the specified element
	while(nextpos!=npos && j!=i && aCount!=0) {
	  nextpos=(*this).find_first_of("[]{},",nextpos+1);
	  if ((*this)[nextpos]=='{') sCount++;
	  if ((*this)[nextpos]=='}') sCount--;
	  if ((*this)[nextpos]=='[') aCount++;
	  if ((*this)[nextpos]==']') aCount--;
	  if ((*this)[nextpos]==',' && sCount==0 && aCount==1) {
	    j++;
	    if (j!=i) startpos=nextpos+1;
	  }
	  if ((*this)[nextpos]==']' && aCount==0) j++;
	}
	// If found
	if (j==i) {
	  startpos=startpos;
	  endpos=nextpos-1;
	  // Check + and = operators
	  paramData.assign((*this).substr(startpos,endpos-startpos+1));
	}
      }
      break;
    default :
      cerr<<"pstring::GetData: Error1\n";
      break;
    }
  }  
  else cerr<<"pstring::GetData: Error2\n";

  // Return parameter data
  return paramData;
}

// Get size of array data
int util::pstring::getSize(const string& parameter) {
  int pos,aCount,sCount,j;
  
  // Find paramater start position
  pos=(*this).find(parameter+"=",0);
  // If paramameter found
  if (pos!=npos) {
    // Calculate parameter data start position
    pos+=(parameter.size()+1);
    switch ((*this)[pos]) {
    case '[' :
      sCount=0;
      aCount=1;
      j=0;
      // Search for the specified element
      while(pos!=npos && aCount!=0) {
	pos=(*this).find_first_of("[]{},",pos+1);
	if ((*this)[pos]=='{') sCount++;
	if ((*this)[pos]=='}') sCount--;
	if ((*this)[pos]=='[') aCount++;
	if ((*this)[pos]==']') aCount--;
	if ((*this)[pos]==',' && sCount==0 && aCount==1) j++;
	if ((*this)[pos]==']' && sCount==0 && aCount==0) j++;
      }
      if (sCount==0 && aCount==0) return j;
      else cerr<<"pstring::getSize: Error1\n";
      break;
    default :
      cerr<<"pstring::getSize: Error2\n";
      break;
    }
  }  
  else cerr<<"pstring::getSize: Error3\n";

  return 0;
}

void util::pstring::putData(const string& paramater, const pstring& value, const PDataType type) {
  pstring pData;
  if (type==pdBeginStruct) pData+="{";
  else if (type==pdEnd or type==pdEndStruct or type==pdContinue) pData+=",";
  pData+=value;
  if (type==pdEndStruct) pData+="}";

  return;
}

void util::pstring::putData(const pstring& value, const PDataType type) {
  pstring pData;
  if (type==pdBeginArray) pData+="[";
  else if (type==pdEnd or type==pdEndArray or type==pdContinue) pData+=",";
  pData+=value;
  if (type==pdEndArray) pData+="]";

  return;
}
