// Calculate temperature, radiative energy and flux corrections for new grid 
// (using the thermal energy equations for the gas and the energy and momentum equations for the radiation field)
void GridHD::calculateDTDEDF(vector<double>& dT, vector<double>& dE, vector<double>& dF) {
  //double cellIndexMin=MAX(cellIndexNLTEMin-1,0);
  int nCells=cell.size()-cellIndexMin;
  DP dd;
  Vec_INT indx(3*nCells);
  Vec_DP kk(0.0,3*nCells);
  Mat_DP jj(0.0,3*nCells,3*nCells);
  double t;
  int i,j;
  
  // For each cell
  for (i=cellIndexMin;i<=cellIndexMax;i++) {

    // Testing
    t=grid.cell[i]->t();

    // Calculate the Jacobian

    // Calculate the temperature derivatives

    // Gas thermal energy equation
    // Add (the derivative of) the dE/dt term
    //jj[i][i]+=cellHD->dEGasdT();
    jj[i][i]+=dEGasdT(i,center);

    // Add (the derivative of) the P*d(1/rho) term
    //jj[i][i]+=0.5*cellHD->dPGasdT()*(pow(cellHD->rho(),-1)-pow(cellHDOld->rho(),-1));
    jj[i][i]+=0.5*dPGasdT(i,center)*(pow(rho(i,center),-1)-pow(rho(i,center,old),-1));
    // Add (the derivative of) the (1/rho)*4*pi*j term
    //jj[i][i]+=pow(cellHD->rho(),-1)*4*pi*cellHD->djdT(center)*dt;
    jj[i][i]+=pow(rho(i,center),-1)*4*pi*djdT(i,center)*dt;
    // Add (the derivative of) the (1/rho)*c*(kappa*rho)*(E*rho) term
    //jj[i][i]+=-pow(cellHD->rho(),-1)*c*(cellHD->rho())*cellHD->dkappadT(center))*(cellHD->ERad()*cellHD->rho())*dt;
    //jj[i][i]+=-pow(rho(i,center),-1)*c*(rho(i,center))*dkappadT(i,center))*(ERad(i,center)*rho(i,center))*dt;

    // Radiation field energy equation
    // Add the c*rho*kappa*E term
    //jj[i+nCells][i]+=c*(cellDH->rho()*cellHD->dkappadT(center))*(cellHD->ERad()*cellHD->rho())*dt;
    //jj[i+nCells][i]+=c*(rho(i,center)*dkappadT(i,center))*(ERad(i,center)*rho(i,center))*dt;
    // Add the 4*pi*j term
    //jj[i+nCells][i]+=-4*pi*cellHD->djdT(center)*dt;
    jj[i+nCells][i]+=-4*pi*djdT(i,center)*dt;

    // Radiation field momentum equation
    //jj[i+2*nCells][i]+=(cellHD->rho(back)*cellHD->dkappadT(back))*cellHD->FRad(back)*dt;
    //jj[i+2*nCells][i]+=(rho(i,back)*dkappadT(i,back))*FRad(i,back)*dt;

    // Calculate the radiative energy derivatives

    // Gas thermal energy equation
    // Add the (1/rho)*c*(kappa*rho)*E term
    //jj[i][i+nCells]+=-pow(cellHD->rho(),-1)*c*(cellHD->kappa(center)*cellHD->rho())*dt;
    jj[i][i+nCells]+=-pow(rho(i,center),-1)*c*(kappa(i,center)*rho(i,center))*dt;

    // Radiation field energy equation
    // Add the dE/dt term
    jj[i+nCells][i+nCells]+=1;
    // If outermost cell, add the rho*4*pi*d(r^2*F)/dm term - Check this!
    //if (i==nCells-1) jj[i+nCells][i+nCells]+=cellHD->rho()*4*pi*pow(cellHD->r(front),2)*cellHD->hE()*cellHD->jE()*c*(dt/cellHD->dm(center));
    if (i==nCells-1) jj[i+nCells][i+nCells]+=rho(i,center)*4*pi*pow(r(i,front),2)*hE(i,center)*jE(i,center)*c*(dt/dm(i,center));
    // Add the (P+E)*d(1/rho)/dt term
    //jj[i+nCells][i+nCells]+=cellHD->rho()*(cellHD->fE()+1)*(pow(cellHD->rho(),-1)-pow(cellHDOld->rho(),-1));
    jj[i+nCells][i+nCells]+=rho(i,center)*(fE(i,center)+1)*(pow(rho(i,center),-1)-pow(rho(i,center),-1));
    // Add the (3P-E)*(v/r) term
    //jj[i+nCells][i+nCells]+=-(3*cellHD->fE()-1)*(cellHD->v(center,front)/cellHD->r(center))*dt;
    jj[i+nCells][i+nCells]+=-(3*fE(i,center)-1)*(v(i,center,front)/r(i,center))*dt; // v(i,center,front)??
    // Add the c*rho*kappa*E term
    //jj[i+nCells][i+nCells]+=c*(cellHD->rho()*cellHD->kappa(center))*dt;
    jj[i+nCells][i+nCells]+=c*(rho(i,center)*kappa(i,center))*dt;

    // Radiation field momentum equation
    // Add the c*rho*4*pi*r^2*dE/dm term
    //jj[i+2*nCells][i+nCells]+=c*cellHD->rho(back)*4*pi*pow(cellHD->r(back),2)*cellHD->fE()*(dt/cellHD->dm(back));
    jj[i+2*nCells][i+nCells]+=c*rho(i,back)*4*pi*pow(r(i,back),2)*fE(i,back)*(dt/dm(i,back));
    //if (i>0) jj[i+2*nCells][i-1+nCells]+=-c*cellHD->rho(back)*4*pi*pow(cellHD->r(back),2)*cellHDPrev->fE()*(dt/cellHD->dm(back));
    if (i>0) jj[i+2*nCells][i-1+nCells]+=-c*rho(i,back)*4*pi*pow(r(i,back),2)*fE(i,back,old)*(dt/dm(i,back));
    // Else? Fix!
    // Add the (3P-E)/r term
    //jj[i+2*nCells][i+nCells]+=c*0.5*(3*cellHD->fE()-1)*(dt/cellHD->r(back));
    jj[i+2*nCells][i+nCells]+=c*0.5*(3*fE(i,back)-1)*(dt/r(i,back));
    //if (i>0) jj[i+2*nCells][i-1+nCells]+=c*0.5*(3*cellHDPrev->fE()-1)*(dt/cellHD->r(back));
    if (i>0) jj[i+2*nCells][i-1+nCells]+=c*0.5*(3*fE(i,back)-1)*(dt/r(i,back));
    // Else? Fix!

    // Calculate the radiative flux derivatives

    // Radiation field energy equation
    // Add the rho*4*pi*d(r^2*F)/dm term
    //if (i<nCells-1) jj[i+nCells][i+1+2*nCells]+=cellHD->rho()*4*pi*pow(cellHD->r(front),2)*(dt/cellHD->dm(center));
    if (i<nCells-1) jj[i+nCells][i+1+2*nCells]+=rho(i,center)*4*pi*pow(r(i,front),2)*(dt/dm(i,center));
    //if (i+cellIndexMin>0) jj[i+nCells][i+2*nCells]+=-cellHD->rho()*4*pi*pow(cellHD->r(back),2)*(dt/cellHD->dm(center));
    if (i+cellIndexMin>0) jj[i+nCells][i+2*nCells]+=-rho(i,center)*4*pi*pow(r(i,back),2)*(dt/dm(i,center));
    // Else? Fix!

    // Radiation field momentum equation
    // Add the (1/c)*dF/dt term
    jj[i+2*nCells][i+2*nCells]+=(1/c);
    // Add the rho*kappa*F term
    //jj[i+2*nCells][i+2*nCells]+=(cellHD->rho(back)*cellHD->kappa(back))*dt;
    jj[i+2*nCells][i+2*nCells]+=(rho(i,back)*kappa(i,back))*dt;
    
    // Calculate the constant
    
    // Gas thermal energy equation
    // Add the dE/dt term
    //kk[i]+=-(cellHD->EGas()-cellHDOld->EGas());
    kk[i]+=-(EGas(i,center)-EGas(i,center,old));
    // Add the P*d(1/rho)/dt term
    //kk[i]+=-0.5*(cellHD->PGas()+cellHDOld->PGas())*(pow(cellHD->rho(),-1)-pow(cellHDOld->rho(),-1));
    kk[i]+=-0.5*(PGas(i,center)+PGas(i,center,old))*(pow(rho(i,center),-1)-pow(rho(i,center,old),-1));
    // Add the Q*d(1/rho)/dt term
    //if (configHD->pseudoViscosity) kk[i]+=-cellHD->Q()*(pow(cellHD->rho(),-1)-pow(cellHDOld->rho(),-1));
    // Removed: No hydrodynamics!
    // Add the (1/rho)*c*(kappa*rho)*E term
    //kk[i]+=pow(cellHD->rho(),-1)*c*(cellHD->kappa(center)*cellHD->rho())*(cellHD->ERad()*cellHD->rho())*dt;
    kk[i]+=pow(rho(i,center),-1)*c*(kappa(i,center)*rho(i,center))*(ERad(i,center)*cellHD->rho(i,center))*dt;
    // Add the (1/rho)*4*pi*j term
    //kk[i]+=-pow(cellHD->rho(),-1)*4*pi*cellHD->j(center)*dt;
    kk[i]+=-pow(rho(i,center),-1)*4*pi*j(i,center)*dt;
    // Add the radioactive heating term
    //kk[i]+=0.5*(cellHD->eRadio()+cellHDOld->eRadio())*dt;
    kk[i]+=0.5*(eRadio(i,center)+eRadio(i,center,old))*dt;
    // Add the explosion energy term
    //if (configHD->injectEnergy) kk[i]+=configHD->ETh*(exp(-t/configHD->t0)-exp(-(t+dt)/configHD->t0))*(exp(-(cellHD->m(back)-mCut)/(configHD->m0-mCut))-exp(-(cellHD->m(front)-mCut)/(configHD->m0-mCut)))/cellHD->dm(center);
    // Removed - No hydrodynamics!

    // Radiation field energy equation
    // Add the dE/dt term
    //kk[i+nCells]+=-(cellHD->ERad()*cellHD->rho()-cellHDOld->ERad()*cellHDOld->rho());
    kk[i+nCells]+=-(ERad(i,center)*rho(i,center)-ERad(i,center,old)*rho(i,center,old));
    // Add the rho*4*pi*d(r^2*F)/dm term
    //kk[i+nCells]+=-cellHD->rho()*4*pi*(pow(cellHD->r(front),2)*cellHD->FRad(front)-pow(cellHD->r(back),2)*cellHD->FRad(back))*(dt/cellHD->dm(center));
    //kk[i+nCells]+=-rho(i,center)*4*pi*(pow(r(i,front),2)*FRad(i,front)-pow(r(i,back),2)*FRad(i,back))*(dt/dm(i,center));
    //if (i<nCells-1) kk[i+nCells]+=-cellHD->rho()*4*pi*(pow(cellHD->r(front),2)*cellHD->FRad(front)-pow(cellHD->r(back),2)*cellHD->FRad(back))*(dt/cellHD->dm(center));
    if (i<nCells-1) kk[i+nCells]+=-rho(i,center)*4*pi*(pow(r(i,front),2)*FRad(i,front)-pow(r(i,back),2)*FRad(i,back))*(dt/dm(i,center));
    // If outermost cell rewrite F(front) in terms of E
    //else kk[i+nCells]+=-cellHD->rho()*4*pi*(pow(cellHD->r(front),2)*cellHD->hE()*cellHD->jE()*c*(cellHD->ERad()*cellHD->rho())-pow(cellHD->r(back),2)*cellHD->FRad(back))*(dt/cellHD->dm(center));
    else kk[i+nCells]+=-rho(i,??)*4*pi*(pow(r(i,front),2)*hE(i,??)*jE(i,??)*c*(ERad(i,??)*rho(i,??))-pow(r(i,back),2)*FRad(i,back))*(dt/dm(i,center)); // ??
    // Add the (P+E)*d(1/rho)/dt term
    //kk[i+nCells]+=-cellHD->rho()*(cellHD->fE()+1)*(cellHD->ERad()*cellHD->rho())*(pow(cellHD->rho(),-1)-pow(cellHDOld->rho(),-1));
    kk[i+nCells]+=-rho(i,center)*(fE(i,center)+1)*(ERad(i,center)*rho(i,center))*(pow(rho(i,center),-1)-pow(rho(i,center,old),-1)); // center??
    // Add the (3P-E)*(v/r) term
    //kk[i+nCells]+=(3*cellHD->fE()-1)*(cellHD->ERad()*cellHD->rho())*(cellHD->v(center,front)/cellHD->r(center))*dt;
    kk[i+nCells]+=(3*fE(i,center)-1)*(ERad(i,center)*rho(i,center))*(v(i,center,front)/r(i,center))*dt; // v(i,center,front)??
    // Add the c*rho*kappa*E term
    //kk[i+nCells]+=-c*(cellHD->rho()*cellHD->kappa(center))*(cellHD->ERad()*cellHD->rho())*dt;
    kk[i+nCells]+=-c*(rho(i,center)*kappa(i,center))*(ERad(i,center)*rho(i,center))*dt; // center??
    // Add the 4*pi*j term
    //kk[i+nCells]+=4*pi*cellHD->j(center)*dt;
    kk[i+nCells]+=4*pi*j(i,center)*dt;

    // Radiation field momentum equation
    // Add the (1/c)*dF/dt term
    //kk[i+2*nCells]+=-(1/c)*(cellHD->FRad(back)-cellHDOld->FRad(back));
    kk[i+2*nCells]+=-(1/c)*(FRad(i,back)-FRad(i,back,old));
    // Add the c*rho*4*pi*r^2*dP/dm term
    //if (i+cellIndexMin>0) kk[i+2*nCells]+=-c*cellHD->rho(back)*4*pi*pow(cellHD->r(back),2)*(cellHD->fE()*cellHD->ERad()*cellHD->rho()-cellHDPrev->fE()*cellHDPrev->ERad()*cellHDPrev->rho())*(dt/cellHD->dm(back));
    if (i+cellIndexMin>0) kk[i+2*nCells]+=-c*rho(i,back)*4*pi*pow(r(i,back),2)*(fE(i,??)*ERad(i,??)*rho(i,??)-fE(i,??,old)*ERad(i,??,old)*rho(i,??,old))*(dt/dm(i,back)); // ??
    // Else - Fix!
    // Add the c*(3P-E)/r term
    //if (i+cellIndexMin>0) kk[i+2*nCells]+=-c*0.5*((3*cellHD->fE()-1)*(cellHD->ERad()*cellHD->rho())+(3*cellHDPrev->fE()-1)*(cellHDPrev->ERad()*cellHDPrev->rho()))*(dt/cellHD->r(back));
    if (i+cellIndexMin>0) kk[i+2*nCells]+=-c*0.5*((3*fE(i,??)-1)*(ERad(i,??)*rho(i,??))+(3*fE(i,??,old)-1)*(ERad(i,??,old)*rho(i,??,old)))*(dt/r(i,back)); // ??
    // Else - Fix!
    // Add the rho*kappa*F term
    //kk[i+2*nCells]+=-(cellHD->rho(back)*cellHD->kappa(back))*cellHD->FRad(back)*dt;
    kk[i+2*nCells]+=-(rho(i,back)*kappa(i,back))*FRad(i,back)*dt;
  }

  // Solve the equation system for dT, dE and dF using LU decomposition
  ludcmp(jj,indx,dd);
  lubksb(jj,indx,kk);

  // Get dT, dE and dF
  for (i=0;i<nCells;i++) {
    dT[i]=kk[i];
    dE[i]=kk[i+nCells]/cell[i+cellIndexMin]->rho();
    dF[i]=kk[i+2*nCells];
  }
}

// Update Eddington factor 
// (by a formal solution to the steady-state radiative transfer equation)
// Ignoring frame transformation of the opacity and emissivity - Fix!
// Ignoring scattering - Fix!
void GridHD::updatefERT(const double& cellIndexMin) {
  int C=10;
  vector<double> s_back,s_center,s_front,dtau_back,dtau_center,dtau_front,j_cell,mu_cell,mu_max,mu_min,dtau_outer(cell.size()-cellIndexMin+C);
  vector< vector<double> > mu_back(cell.size()-cellIndexMin+C), mu_center(cell.size()-cellIndexMin+C), mu_front(cell.size()-cellIndexMin+C);
  vector<Vec_O_DP*> j(cell.size()-cellIndexMin+C);
  Vec_DP *aa,*bb,*cc,*kk;
  double J,K,p,s_inner,j_center_prev,s_center_prev,s_front_prev,j_outer,J_outer,H_outer,K_outer;
  int i,k,l,ii;

  cout<<"GridHD::updatefERT: number of cells: "<<cell.size()<<", cellIndexMin="<<cellIndexMin<<"\n";

  // For each ray (impact parameter)
  for (i=0;i<cell.size()-cellIndexMin+C;i++) {
    // Calculate index for the originating cell
    if (i>=C) ii=i+cellIndexMin-C;
    else ii=cellIndexMin;
    // Initialize ray vectors
    aa=new Vec_DP(cell.size()-ii);
    bb=new Vec_DP(cell.size()-ii);
    cc=new Vec_DP(cell.size()-ii);
    kk=new Vec_DP(cell.size()-ii);
    j[i]=new Vec_DP(cell.size()-ii);
    mu_back[i].resize(cell.size()-ii);
    mu_center[i].resize(cell.size()-ii);
    mu_front[i].resize(cell.size()-ii);
    s_back.resize(cell.size()-ii);
    s_center.resize(cell.size()-ii);
    s_front.resize(cell.size()-ii);
    dtau_back.resize(cell.size()-ii);
    dtau_center.resize(cell.size()-ii);
    dtau_front.resize(cell.size()-ii);
    // Calculate impact parameter
    //if (i>=C) p=((CellHD*) cell[ii])->r(center);
    //else p=(i+1)*((CellHD*) cell[ii])->r(back)/C;
    if (i>=C) p=r(ii,center);
    else p=(i+1)*r(,ii,back)/C;
    // Calculate intersection with inner region
    if (i>=(C-1)) s_inner=0;
    //else s_inner=pow(pow(((CellHD*) cell[ii])->r(back),2)-pow(p,2),0.5);
    else s_inner=pow(pow(r(ii,back),2)-pow(p,2),0.5);
    // For each cell intersected by the ray
    for (k=ii;k<cell.size();k++) {
      // Calculalate ray vector index
      l=k-ii;
      // Calculate distances along the ray at cell back, center and front
      //s_front[l]=pow(pow(((CellHD*) cell[k])->r(front),2)-pow(p,2),0.5)-s_inner;
      s_front[l]=pow(pow(r(k,front),2)-pow(p,2),0.5)-s_inner;
      if (l>0) s_back[l]=s_front[l-1];
      else 
	if (i>=(C-1)) s_back[l]=-s_front[l];
      //else s_back[l]=pow(pow(((CellHD*) cell[k])->r(back),2)-pow(p,2),0.5)-s_inner;
	else s_back[l]=pow(pow(r(k,back),2)-pow(p,2),0.5)-s_inner;
      //s_center[l]=pow(pow(((CellHD*) cell[k])->r(center),2)-pow(p,2),0.5)-s_inner;
      s_center[l]=pow(pow(r(k,center),2)-pow(p,2),0.5)-s_inner;
      // Calculate mu at cell back, center and front
      //mu_front[i][l]=(s_front[l]+s_inner)/((CellHD*) cell[k])->r(front);
      mu_front[i][l]=(s_front[l]+s_inner)/r(k,front);
      if (l>0) mu_back[i][l]=mu_front[i][l-1];
      else
	if (i>=(C-1)) mu_back[i][l]=-mu_front[i][l];
        //else mu_back[i][l]=(s_back[l]+s_inner)/((CellHD*) cell[k])->r(back);
	else mu_back[i][l]=(s_back[l]+s_inner)/r(k,back);
      //mu_center[i][l]=(s_center[l]+s_inner)/((CellHD*) cell[k])->r(center);
      mu_center[i][l]=(s_center[l]+s_inner)/r(k,center);
    }
    // For each cell intersected by the ray
    for (k=ii;k<cell.size();k++) {
      // Calculate ray vector index
      l=k-ii;
      // If not first nor last cell
      if ((k<cell.size()-1) and (k>ii)) {
	//cout<<"GridHD::updatefERT: processing intersecting cell: "<<k<<", 1\n";
	// Calculate delta tau at cell back, center and front
	//dtau_front[l]=((CellHD*) cell[k])->kappa(center)*((CellHD*) cell[k])->rho(center)*(s_front[l]-s_center[l])+((CellHD*) cell[k+1])->kappa(center)*((CellHD*) cell[k+1])->rho(center)*(s_center[l+1]-s_front[l]);
	dtau_front[l]=kappa(k,center)*rho(k,center)*(s_front[l]-s_center[l])+kappa(k+1,center)*rho(k+1,center)*(s_center[l+1]-s_front[l]);
	dtau_back[l]=dtau_front[l-1];
	dtau_center[l]=0.5*(dtau_front[l-1]+dtau_front[l]);
	// Calculate the j coefficient for the previous cell
	(*aa)[l]=(1/dtau_center[l])*(1/dtau_back[l]);
	// Calculate the j coeffiecient for the current cell
	(*bb)[l]=-(1/dtau_center[l])*(1/dtau_back[l]+1/dtau_front[l])-1;
	// Calculate the j coefficient for the next cell
	(*cc)[l]=(1/dtau_center[l])*(1/dtau_front[l]);
	// Calculate the constant
	//(*kk)[l]=-((CellHD*) cell[k])->S(center);
	(*kk)[l]=-S(k,center);
      }
      // Else
      else 
	// If last (outermost) cell
	// Making use of j=h at outer boundary - see Mihalas (1984) page 370-373
	if (k==cell.size()-1) {
	  // Calculate delta tau at cell back and center
	  dtau_back[l]=dtau_front[l-1];
	  //dtau_center[l]=((CellHD*) cell[k])->kappa(center)*((CellHD*) cell[k])->rho(center)*(s_front[l]-s_back[l]);
	  dtau_center[l]=kappa(k,center)*rho(k,center)*(s_front[l]-s_back[l]);
	  // Save delta tau in outermost cell for later use
	  dtau_outer[i]=dtau_center[l];
	  // Calculate the j coefficient for the previous cell
	  (*aa)[l]=(1/dtau_back[l]);
	  // Calculate the j coeffiecient for the current cell
	  (*bb)[l]=-(1/dtau_back[l])-(1+0.5*dtau_center[l])-(1/dtau_center[l]);
	  // Set the j coefficient for the next cell to zero
	  (*cc)[l]=0;
	  // Calculate the constant
	  //(*kk)[l]=-dtau_center[l]*((CellHD*) cell[k])->S(center);
	  (*kk)[l]=-dtau_center[l]*S(k,center);
	}
        // If first (innermost) cell
	else {
	  // If the impact parameter is larger or equal to the inner boundary
	  // Make use of symmetry around cell center
	  if (i>=(C-1)) {
	    // Calculate delta tau at cell back, center and front
	    //dtau_front[l]=((CellHD*) cell[k])->kappa(center)*((CellHD*) cell[k])->rho(center)*(s_front[l]-s_center[l])+((CellHD*) cell[k+1])->kappa(center)*((CellHD*) cell[k+1])->rho(center)*(s_center[l+1]-s_front[l]);
	    dtau_front[l]=kappa(k,center)*rho(k,center)*(s_front[l]-s_center[l])+kappa(k+1,center)*rho(k+1,center)*(s_center[l+1]-s_front[l]);
	    dtau_back[l]=dtau_front[l];
	    dtau_center[l]=0.5*(dtau_front[l-1]+dtau_front[l]);
	    // Set the j coefficient for the previous cell to zero
	    (*aa)[l]=0;
	    // Calculate the j coeffiecient for the current cell
	    (*bb)[l]=-(1/dtau_center[l])*(1/dtau_back[l]+1/dtau_front[l])-1;
	    // Calculate the j coefficient for the next cell (including the coefficient for the previous cell)
	    (*cc)[l]=(1/dtau_center[l])*(1/dtau_back[l]+1/dtau_front[l]);
	    // Calculate the constant
	    //(*kk)[l]=-((CellHD*) cell[k])->S(center);
	    (*kk)[l]=-S(k,center);
	  }
	  // Else
	  else 
	    // If the ray originates from the inner boundary of the grid 
	    if (ii==0) {
	      // Calculate delta tau at cell back, center and front
	      //dtau_front[l]=((CellHD*) cell[k])->kappa(center)*((CellHD*) cell[k])->rho(center)*(s_front[l]-s_center[l])+((CellHD*) cell[k+1])->kappa(center)*((CellHD*) cell[k+1])->rho(center)*(s_center[l+1]-s_front[l]);
	      dtau_front[l]=kappa(k,center)*rho(k,center)*(s_front[l]-s_center[l])+kappa(k+1,center)*rho(k+1,center)*(s_center[l+1]-s_front[l]);
	      //dtau_back[l]=2*((CellHD*) cell[k])->kappa(center)*((CellHD*) cell[k])->rho(center)*(s_center[l]);
	      dtau_back[l]=2*kappa(k,center)*rho(k,center)*(s_center[l]);
	      dtau_center[l]=0.5*(dtau_front[l-1]+dtau_front[l]);
	      // Set the j coefficient for the previous cell to zero
	      (*aa)[l]=0;
	      // Calculate the j coeffiecient for the current cell (including the coefficient for the previous cell)
	      (*bb)[l]=-(1/dtau_center[l])*(1/dtau_front[l])-1;
	      // Calculate the j coefficient for the next cell
	      (*cc)[l]=(1/dtau_center[l])*(1/dtau_front[l]);
	      // Calculate the constant
	      //(*kk)[l]=-((CellHD*) cell[k])->S(center);
	      (*kk)[l]=-S(k,center);
	    }
	    // Else
	    // Make use of the known value of j in the previous cell
	    else {
	      // Calculate j in previous cell from ERad
	      //j_center_prev=(c/4*pi)*((CellHD*) cell[k-1])->ERad()*((CellHD*) cell[k-1])->rho();
	      j_center_prev=(c/4*pi)*ERad(k-1,center)*rho(k-1,center); //center??
	      //if (((CellHD*) cell[k-1])->r(center)>p) s_center_prev=pow(pow(((CellHD*) cell[k-1])->r(center),2)-pow(p,2),0.5)-s_inner;
	      if (r(k-1,center)>p) s_center_prev=pow(pow(r(k-1,center),2)-pow(p,2),0.5)-s_inner;
	      else s_center_prev=0.0-s_inner;
	      //if (((CellHD*) cell[k-1])->r(front)>p) s_front_prev=pow(pow(((CellHD*) cell[k-1])->r(front),2)-pow(p,2),0.5)-s_inner;
	      if (r(k-1,front)>p) s_front_prev=pow(pow(r(k-1,front),2)-pow(p,2),0.5)-s_inner;
	      else s_front_prev=0.0-s_inner;
	      // Calculate delta tau at cell back, center and front
	      //dtau_front[l]=((CellHD*) cell[k])->kappa(center)*((CellHD*) cell[k])->rho(center)*(s_front[l]-s_center[l])+((CellHD*) cell[k+1])->kappa(center)*((CellHD*) cell[k+1])->rho(center)*(s_center[l+1]-s_front[l]);
	      dtau_front[l]=kappa(k,center)*rho(k,center)*(s_front[l]-s_center[l])+kappa(k+1,center)*rho(k+1,center)*(s_center[l+1]-s_front[l]);
	      //dtau_back[l]=((CellHD*) cell[k-1])->kappa(center)*((CellHD*) cell[k-1])->rho(center)*(s_front_prev-s_center_prev)+((CellHD*) cell[k])->kappa(center)*((CellHD*) cell[k])->rho(center)*(s_center[l]-s_front_prev);
	      dtau_back[l]=kappa(k-1,center)*rho(k-1,center)*(s_front_prev-s_center_prev)+kappa(k,center)*rho(k,center)*(s_center[l]-s_front_prev);
	      dtau_center[l]=0.5*(dtau_back[l]+dtau_front[l]);
	      // Set the j coefficient for the previous cell to zero
	      (*aa)[l]=0;
	      // Calculate the j coeffiecient for the current cell
	      (*bb)[l]=-(1/dtau_center[l])*(1/dtau_back[l]+1/dtau_front[l])-1;
	      // Calculate the j coefficient for the next cell
	      (*cc)[l]=(1/dtau_center[l])*(1/dtau_front[l]);
	      // Calculate the constant (including the j center term for the previous cell)
	      //(*kk)[l]=-((CellHD*) cell[k])->S(center)-(1/dtau_center[l])*(1/dtau_back[l])*j_center_prev;
	      (*kk)[l]=-S(k,center)-(1/dtau_center[l])*(1/dtau_back[l])*j_center_prev;
	    }
	}
    }
    // Solve the equation system for j
    // Error handling should be added here
    tridag(*aa,*bb,*cc,*kk,*(j[i]));
    // Clean up
    delete aa;
    delete bb;
    delete cc;
    delete kk;
    cout<<"HERE\n";
    for (k=ii;k<cell.size();k++) {
      // Calculate ray vector index
      l=k-ii;
    }
  }
  J_outer=0;
  H_outer=0;
  K_outer=0;
  // For each cell
  for (k=cellIndexMin;k<cell.size();k++) {
    J=0;
    K=0;
    j_cell.resize(k-cellIndexMin+C+1);
    mu_cell.resize(k-cellIndexMin+C+1);
    mu_max.resize(k-cellIndexMin+C+1);
    mu_min.resize(k-cellIndexMin+C+1);
    // For each ray (impact parameter) intersecting the cell
    for (i=k-cellIndexMin+C;i>=0;i--) {
      // Calculate index for the originating cell
      if (i>=C) ii=i+cellIndexMin-C;
      else ii=cellIndexMin;
      // Calculate ray vector index
      l=k-ii;
      // Get mu
      mu_cell[i]=mu_center[i][l];
      // Get j
      j_cell[i]=(*(j[i]))[l];
    }
    // For each ray (impact parameter) intersecting the cell
    for (i=k-cellIndexMin+C;i>=0;i--) {
      // Calculate index for the originating cell
      if (i>=C) ii=i+cellIndexMin-C;
      else ii=cellIndexMin;
      // Calculate mu min
      if (i<k-cellIndexMin+C) mu_min[i]=0.5*(mu_cell[i]+mu_cell[i+1]);
      else mu_min[i]=0.0;
      // Calculate mu max
      if (i>0) mu_max[i]=0.5*(mu_cell[i]+mu_cell[i-1]);
      else mu_max[i]=1.0;
      // Integrate J
      J+=j_cell[i]*(mu_max[i]-mu_min[i]);
      // Integrate K
      K+=j_cell[i]*pow(mu_cell[i],2)*(mu_max[i]-mu_min[i]);
      // If outermost cell
      if (k==cell.size()-1) {
	// Integrate J, H and K at outer boundary
	j_outer=j_cell[i]/(1+0.5*dtau_outer[i]);
	J_outer+=j_outer*(mu_max[i]-mu_min[i]);
	H_outer+=j_outer*mu_cell[i]*(mu_max[i]-mu_min[i]);
	K_outer+=j_outer*pow(mu_cell[i],2)*(mu_max[i]-mu_min[i]);
      }
    }
    // Calculate and update Eddington factor
    //((CellHD*) cell[k])->fE(K/J);
    fE(K/J,k);
    cout<<"GridHD::updatefERT: originating cell "<<ii<<", fE="<<(((CellHD*) cell[k])->fE())<<"\n";
    // If outermost cell 
    if (k==cell.size()-1) {
      // Calculate and update additional Eddington factors (see Mihalas 1984 page 374)
      //((CellHD*) cell[k])->hE(H_outer/J_outer);
      hE(H_outer/J_outer,k);
      //((CellHD*) cell[k])->jE(J_outer/J);
      jE(J_outer/J,k);
      cout<<"GridHD::updatefERT: outer boundary, fE="<<K_outer/J_outer<<", hE="<<H_outer/J_outer<<", jE="<<J_outer/J<<"\n";
    }
  }
  //for (i=0;i<cellIndexMin;i++) ((CellHD*) cell[i])->fE(1.0/3.0);
  for (i=0;i<cellIndexMin;i++) fE(1.0/3.0,i);
  // Cleanup
  for (i=0;i<cell.size()-cellIndexMin+C;i++) delete j[i];
}

// Evolve/update thermodynamical state of grid
void MomSolver::evolveThermo() {
  int i,j,nIter=config.convergence.nIter; // Changed!
  vector<double> dT(cellIndexMax+1,0);
  double expFactor;

  // Update matter state - Testing !!!
  //updateMatterState(dT);

  // Iteration loop
  for (n=0;n<nIter;n++) {
  
    // Solve for temperature
    solveTemperature(dT);

    // Update convergence criteria
    updateTConvergence(dT);

    //if (log()) cout<<"DiffSolver::evolveThermo: 2\n";

    // If converged break
    if (TConverged()) break;

    // Update matter state
    updateMatterState(dT);

    // Update radiation state
    updateRadiationState(dE,dF);
  }
  if (n==nIter and log()) cout<<"DiffSolver::evolveThermo: State did not converge\n";
  else if (log()) cout<<"DiffSolver::evolveThermo: State converged in "<<n<<" iterations\n";
}


// Evolve thermodynamical state of grid (using implicit finitie difference method)
void MomSolver::evolveThermo(const GridHD& gridOld, const double& dt, bool& TConverged) {
  Vec_DP dT(cell.size()),dTLTE(cell.size()),dTNLTE(cell.size());
  vector<double> dE(cell.size()),dF(cell.size());
  double maxDT,maxDE,maxDF,damping=0.8,nCheck;
  CellHD* cellHD;
  int i,j,cellIndexLTEMax,cellIndexNLTEMin;

  cout<<"GridHD::evolveThermo\n";

  // Initialize iteration counter
  i=0;
  // Do
  do {
    // Initialize convergence flag
    TConverged=true;
    // If split LTE/NLTE scheme requested
    if (((GridHDConfig*) config)->radScheme==radSplit) {
      // Get max and min cell indices for the LTE and NLTE schemes
      cellIndexLTEMax=cellIndexSplit();
      cellIndexNLTEMin=cellIndexLTEMax+1;
      cout<<"GridHD::evolveThermo: cellIndexLTEMax: "<<cellIndexLTEMax<<", cellIndexNLTEMin: "<<cellIndexNLTEMin<<", cell.size(): "<<cell.size()<<"\n";
      // Setup scheme to use for each cell
      for (j=0;j<cell.size();j++) {
	if (j<=cellIndexLTEMax) ((CellHD*) cell[j])->setRadScheme(radLTE);
	else ((CellHD*) cell[j])->setRadScheme(radNLTE);
      }
      // If there is a LTE region 
      if (cellIndexLTEMax>=0)
	// Calculate dT for the LTE region (using the total thermal energy equation)
	calculateDT(gridOld,dt,dTLTE,cellIndexLTEMax);
      // If there is a NLTE region
      if (cellIndexNLTEMin<cell.size()) {
	// If coupled NLTE scheme
	if (((GridHDConfig*) config)->NLTECoupled) {
	  // Calculate dT, dE and dF for the NLTE region
	  calculateDTDEDF(gridOld,dt,dTNLTE,dE,dF,cellIndexNLTEMin);
	}
	// Else
	else {
	  // Evolve the radiation field for the NLTE region
	  evolveRadiationField(gridOld,dt,cellIndexNLTEMin);
	  // Calculate dT for the NLTE region (using the gas thermal energy equation)
	  calculateDTGas(gridOld,dt,dTNLTE,cellIndexNLTEMin);
	  //calculateDTGasOld(gridOld,dt,dTNLTE,cellIndexNLTEMin);
	}
      }
      // Combine the dT vectors
      for (j=0;j<cellIndexNLTEMin;j++) dT[j]=dTLTE[j];
      for (j=cellIndexNLTEMin;j<cell.size();j++) dT[j]=dTNLTE[j-cellIndexNLTEMin];
      for (j=0;j<cell.size();j++) cout<<"GridHD::evolveThermo: cell "<<j<<", T="<<cell[j]->T()<<", dT="<<dT[j]<<", ERad="<<((CellHD*) cell[j])->ERad()<<", FRad(back)="<<((CellHD*) cell[j])->FRad(back)<<", FRad(front)="<<((CellHD*) cell[j])->FRad(front)<<", j="<<((CellHD*) cell[j])->j(center)<<"\n";
    }
    // Else
    else {
      // Calculate dT (using the total thermal energy equation)
      calculateDT(gridOld,dt,dT,cell.size()-1);
      for (j=0;j<cell.size();j++) cout<<"GridHD::evolveThermo: cell "<<j<<", T="<<cell[j]->T()<<", dT="<<dT[j]<<", ERad="<<((CellHD*) cell[j])->ERad()<<", j="<<((CellHD*) cell[j])->j(center)<<", dj/dT="<<((CellHD*) cell[j])->djdT(center)<<"\n";
    }
    // Update Eddington factor according to configuration
    if (((GridHDConfig*) config)->eddingtonFactor!=isotropic) updatefE(cellIndexNLTEMin);
    // Reset maximum dT
    maxDT=0;
    // Calculate maximum dT
    for (j=0;j<dT.size();j++) if (dT[j]>=maxDT) maxDT=dT[j];
    // If coupled NLTE scheme
    if (((GridHDConfig*) config)->NLTECoupled) {
      maxDE=0;
      maxDF=0;
      // Calculate maximum dE
      for (j=cellIndexNLTEMin;j<dE.size();j++) if (dE[j]>=maxDE) maxDE=dE[j];
      // Calculate maximum dF
      for (j=cellIndexNLTEMin;j<dF.size();j++) if (dF[j]>=maxDF) maxDF=dF[j];
    }
    // For each cell
    for (j=0;j<cell.size();j++) {
      // Set convergence flag to false and break if temperature negative or too large (probably means we need to decrease the time step)
      if (((CellHD*) cell[j])->T()+dT[j]*damping<0 or fabs(dT[j]*damping)>(((CellHD*) gridOld.cell[j])->T()*((GridHDConfig*) config)->dTMax*100)) {
	TConverged=false;
	cout<<"GridHD::evolveThermo: Negative temperature or correction too large in cell "<<j<<" "<<"\n";
	break;
      }
      // Update temperature (using damping and avoiding negative temperatures)
      ((CellHD*) cell[j])->T(MAX(cell[j]->T()+dT[j]*damping,1.0));
      // If coupled NLTE scheme
      if (((GridHDConfig*) config)->NLTECoupled and j>=cellIndexNLTEMin) {
	// Set convergence flag to false and break if radiative energy negative (probably means we need to decrease the time step)
	if (((CellHD*) cell[j])->ERad()+dE[j]*damping<0) {
	  TConverged=false;
	  cout<<"GridHD::evolveThermo: Negative radiative energy or flux in cell "<<j<<" "<<"\n";
	  break;
	}
	// Update radiative energy (using damping and avoiding negative energy)
	((CellHD*) cell[j])->setERad(MAX(((CellHD*) cell[j])->ERad()+dE[j]*damping,0.0));
	// Update radiative flux (using damping)
	if (j>0) {
	  //((CellHD*) cell[j])->setFRad(MAX(((CellHD*) cell[j])->FRad(back)+dF[j]*damping,0.0),back);
	  //((CellHD*) cell[j])->setFRad(MAX(((CellHD*) cell[j-1])->FRad(front)+dF[j]*damping,0.0),front);
	  ((CellHD*) cell[j])->setFRad(((CellHD*) cell[j])->FRad(back)+dF[j]*damping,back);
	  ((CellHD*) cell[j-1])->setFRad(((CellHD*) cell[j])->FRad(back),front);
	}
      }
    }
    // Break if convergence flag set to false
    if (not TConverged) break;
    // Update state of matter
    updateMatterState();
    // Update Eddington factor according to configuration
    if (((GridHDConfig*) config)->eddingtonFactor!=isotropic) updatefE(cellIndexNLTEMin);
    // If requested ignore outermost cell in convergence check
    if (((GridHDConfig*) config)->ignoreOutermostCell) nCheck=cell.size()-1;
    else nCheck=cell.size();
    // For each cell check convergence
    for (j=0;j<nCheck;j++){
      // If temperature error in cell above limit
      if (fabs(dT[j]/((CellHD*) cell[j])->T())>((GridHDConfig*) config)->TTol) {
	TConverged=false;
	cout<<"GridHD::evolveThermo: Temperature error in cell "<<j<<" "<<fabs(dT[j]/((CellHD*) cell[j])->T())<<" above the limit\n";
	break;
      }
      // If coupled NLTE scheme
      if (((GridHDConfig*) config)->NLTECoupled and j>=cellIndexNLTEMin) {
	// If radiative energy error in cell above limit
	if (fabs(dE[j]/((CellHD*) cell[j])->ERad())>((GridHDConfig*) config)->ETol) {
	  TConverged=false;
	  cout<<"GridHD::evolveThermo: Radiative energy error in cell "<<j<<" "<<fabs(dE[j]/((CellHD*) cell[j])->E())<<" above the limit\n";
	  break;
	}
	// If radiative flux error in cell above limit
	if (fabs(dF[j]/((CellHD*) cell[j])->FRad(back))>((GridHDConfig*) config)->FTol and j>0) {
	  TConverged=false;
	  cout<<"GridHD::evolveThermo: Radiative flux error in cell "<<j<<" "<<fabs(dF[j]/((CellHD*) cell[j])->FRad(back))<<" above the limit\n";
	  break;
	}
      }
    }
    // Increase iteration counter
    i++;
  }
  // While temperature not converged and number of iteration within limit
  while ((not TConverged) and (i<((GridHDConfig*) config)->maxTIter));

  if (i>=((GridHDConfig*) config)->maxTIter) cout<<"GridHD::evolveThermo: Max iterations exceeded\n";

  if (TConverged) cout<<"GridHD::evolveThermo: Temperature converged\n";
  else cout<<"GridHD::evolveThermo: Temperature did not converged\n";
}
