// Constructors
FrequencyBin::FrequencyBin(const double& _min, const double& _max, bool _edge, bool _minorEdge) : min(_min), max(_max), edge(_edge), minorEdge(_minorEdge) { }
FrequencyBin::FrequencyBin() : min(), max(), edge(false), minorEdge(false) {}

FrequencyBinIterator FrequencyBinList::find(const double& _nu, int iGuess) const {
  vector<FrequencyBin>& nu = *((vector<FrequencyBin>*)this);
  long iBegin,iMiddle,iEnd;

  if (iGuess!=noBin and iGuess<nu.size())
    if (_nu>=nu[iGuess].min and _nu<nu[iGuess].max) return nu.begin()+iGuess;
  
  if (nu.size()!=0) {
    iBegin=0; iEnd=nu.size()-1; iMiddle=(iBegin+iEnd)/2;
    if (_nu>=nu[iBegin].min && _nu<nu[iEnd].max) {
      while (iBegin!=iEnd) {
	if (_nu<nu[iMiddle].max) {
	  iEnd=iMiddle;
	  iMiddle=(iBegin+iEnd)/2;
	}
	else {
	  iBegin=iMiddle+1;
	  iMiddle=(iBegin+iEnd)/2;
	}
      }
      return nu.begin()+iBegin;
    }
  }
  
  return nu.end();
}

// Constructors
AngleBin::AngleBin(const double& _min, const double& _max) : min(_min), max(_max) { }
AngleBin::AngleBin() : min(), max() {}

AngleBinIterator AngleBinList::find(const double& _mu, int iGuess) const {
  vector<AngleBin>& mu = *((vector<AngleBin>*)this);
  long iBegin,iMiddle,iEnd;

  if (iGuess!=noBin and iGuess<mu.size())
    if (_mu>=mu[iGuess].min and _mu<mu[iGuess].max) return mu.begin()+iGuess;
  
  if (mu.size()!=0) {
    iBegin=0; iEnd=mu.size()-1; iMiddle=(iBegin+iEnd)/2;
    if (_mu>=mu[iBegin].min && _mu<mu[iEnd].max) {
      while (iBegin!=iEnd) {
	if (_mu<mu[iMiddle].max) {
	  iEnd=iMiddle;
	  iMiddle=(iBegin+iEnd)/2;
	}
	else {
	  iBegin=iMiddle+1;
	  iMiddle=(iBegin+iEnd)/2;
	}
      }
      return mu.begin()+iBegin;
    }
  }
  
  return mu.end();
}


enum NuType { ntVal, ntBin, nuBoth }

class Nu {
  FrequencyBinIterator bin;
  double val;
  NuType type;
  EdgeType edgeType;
  Nu(const double& _value) : val(_val), type(ntVal) { }
  Nu(const FrequencyBinIterator& _bin) : bin(_bin), type(ntBin) {}
  Nu(const FrequencyBinIterator& _bin, const double& _val) : bin(_bin), val(_val), type(ntBoth) { }
}

template <class F> NuTabulator<F>::NuTabulator(FrequencyBinIterator _nuBegin, FrequencyBinIterator _nuEnd, const F& _f, double _edgeWidth) : nuBegin(_nuBegin), nuEnd(_nuEnd), f(_f), edgeWidth(_edgeWidth) {
  int i;

  // If any frequency bins to tabulate
  if (nuEnd>nuBegin) {
    // Initialize function table
    // Remove +1, right?
    fTable.resize(nuEnd-nuBegin+1);
    fRedTable.resize(nuEnd-nuBegin+1);
    fBlueTable.resize(nuEnd-nuBegin+1);
    // Calculate function values at the beginning of all frequency bins
    for (i=0;i<(nuEnd-nuBegin);i++) {
      fTable[i]=f(nuEdge(nuBegin+i,noEdge));
      if ((nuBegin+i)->edge) fRedTable[i]=f(nuEdge(nuBegin+i,redEdge));
      else fRedTable[i]=fTable[i];
      if ((nuBegin+i)->edge) fBlueTable[i]=f(nuEdge(nuBegin+i,blueEdge));
      else fBlueTable[i]=fTable[i];
    }
    // Calculate funcion values at the end of the last frequency bin
    fTable.push_back(f(nuEdge((nuEnd-1)->max,noEdge)));
    fRedTable.push_back(f(nuEdge((nuEnd-1)->max,redEdge)));
    fBlueTable.push_back(f(nuEdge((nuEnd-1)->max,blueEdge)));
  }
  // Else
  else {
    // Reset function table
    fTable.resize(0);
    fRedTable.resize(0);
    fBlueTable.resize(0);
  }
}

template <class F> void NuTabulator<F>::init(FrequencyBinIterator _nuBegin, FrequencyBinIterator _nuEnd, const F& _f, double _edgeWidth) {
  int i;

  // Initialize
  nuBegin=_nuBegin;
  nuEnd=_nuEnd;
  f=_f;
  edgeWidth=_edgeWidth;

  // If any frequency bins to tabulate
  if (nuEnd>nuBegin){
    // Initialize function table
    // Remove +1, right?
    fTable.resize(nuEnd-nuBegin+1);
    fRedTable.resize(nuEnd-nuBegin+1);
    fBlueTable.resize(nuEnd-nuBegin+1);
    // Calculate function values at the beginning of all frequency bins
    for (i=0;i<(nuEnd-nuBegin);i++) {
      fTable[i]=f(nuEdge(nuBegin+i,noEdge));
      if ((nuBegin+i)->edge) fRedTable[i]=f(nuEdge(nuBegin+i,redEdge));
      else fRedTable[i]=fTable[i];
      if ((nuBegin+i)->edge) fBlueTable[i]=f(nuEdge(nuBegin+i,blueEdge));
      else fBlueTable[i]=fTable[i];
    }
    // Calculate function values at the end of the last frequency bin
    fTable.push_back(f(nuEdge((nuEnd-1)->max,noEdge)));
    fRedTable.push_back(f(nuEdge((nuEnd-1)->max,redEdge)));
    fBlueTable.push_back(f(nuEdge((nuEnd-1)->max,blueEdge)));
  }
  // Else
  else {
    // Reset function table
    fTable.resize(0);
    fRedTable.resize(0);
    fBlueTable.resize(0);
  }
}

template <class F> void NuTabulator<F>::initOMP(FrequencyBinIterator _nuBegin, FrequencyBinIterator _nuEnd, const F& _f, double _edgeWidth) {
  int i;

  // Initialize
  nuBegin=_nuBegin;
  nuEnd=_nuEnd;
  f=_f;
  edgeWidth=_edgeWidth;

  // If any frequency bins to tabulate
  if (nuEnd>nuBegin){
    // Calculate function values at the beginning of all frequency bin
    fTable=f(nuBegin,nuEnd,noEdge);
    fRedTable=f(nuBegin,nuEnd,redEdge);
    fBlueTable=f(nuBegin,nuEnd,blueEdge);
    // To be on the safe side - Replace zeros in red and blue tables
    for (i=0;i<(nuEnd-nuBegin);i++)
      if (not ((nuBegin+i)->edge)) { fRedTable[i]=fTable[i]; fBlueTable[i]=fTable[i]; }
    // Calculate function values at the end of the last frequency bin
    fTable.push_back(f(nuEdge((nuEnd-1)->max,noEdge)));
    fRedTable.push_back(f(nuEdge((nuEnd-1)->max,redEdge)));
    fBlueTable.push_back(f(nuEdge((nuEnd-1)->max,blueEdge)));
  }
  // Else
  else {
    // Reset function table
    fTable.resize(0);
    fRedTable.resize(0);
    fBlueTable.resize(0);
  }
}

template <class F> vector<double>::const_iterator NuTabulator<F>::find(const double& nu) const {
  long iBegin,iMiddle,iEnd;
  
  if (nuEnd>nuBegin) {
    iBegin=0; iEnd=(nuEnd-nuBegin)-1; iMiddle=(iBegin+iEnd)/2;
    if (nu>=(nuBegin+iBegin)->min && nu<(nuBegin+iEnd)->min) {
      while (iEnd-iBegin>1) {
	if (nu<(nuBegin+iMiddle)->min) {
	  iEnd=iMiddle;
	  iMiddle=(iBegin+iEnd)/2;
	}
	else {
	  iBegin=iMiddle;
	  iMiddle=(iBegin+iEnd)/2;
	}
      }
      if (nu==(nuBegin+iBegin)->min) return fTable.begin()+iBegin;
    }
    else if (nu==(nuBegin+iEnd)->min) return fTable.begin()+iEnd;
    else if (nu==(nuBegin+iEnd)->max) return fTable.begin()+iEnd+1;
  }
  
  return fTable.end();
}

template <class F> double NuTabulator<F>::operator ()(FrequencyBinIterator nuBin, EdgeType edgeType) const {
  int i;

  if (nuBin>=nuBegin and nuBin<=nuEnd) {
    if (edgeType==noEdge) return fTable[(nuBin-nuBegin)];
    else if (edgeType==redEdge) return fRedTable[(nuBin-nuBegin)];
    else return fBlueTable[(nuBin-nuBegin)];
  }
  else { cout<<"NuTabulator: Error: Outside tabulated frequency range, nuEnd-nuBegin: "<<nuEnd-nuBegin<<", nuBegin-nuBin: "<<nuBegin-nuBin<<", nuBin-nuEnd: "<<nuBin-nuEnd<<"\n"; return 0; }
}

template <class F> double NuTabulator<F>::operator ()(double nu, EdgeType edgeType) const {
  vector<double>::const_iterator fIterator=find(nu);
  int i=fIterator-fTable.begin();
    
  if (fIterator==fTable.end()) return f(nuEdge(nu,edgeType));
  else if (edgeType==noEdge) return fTable[i];
  else if (edgeType==redEdge) return fRedTable[i];
  else return fBlueTable[i];
}

template <class F> NuInterpolator<F>::NuInterpolator(FrequencyBinIterator _nuBegin, FrequencyBinIterator _nuEnd, const F& _f, double _edgeWidth) : nuBegin(_nuBegin), nuEnd(_nuEnd), f(_f), edgeWidth(_edgeWidth) { init(nuBegin,nuEnd,f,edgeWidth); }

template <class F> void NuInterpolator<F>::init(FrequencyBinIterator _nuBegin, FrequencyBinIterator _nuEnd, const F& _f, double _edgeWidth) {
  int i;

  // Initialize
  nuBegin=_nuBegin;
  nuEnd=_nuEnd;
  f=_f;
  edgeWidth=_edgeWidth;

  // If any frequency bins to tabulate
  if (nuEnd>nuBegin){
    // Initialize function table
    // Remove +1, right?
    fTable.resize(nuEnd-nuBegin+1);
    fRedTable.resize(nuEnd-nuBegin+1);
    fBlueTable.resize(nuEnd-nuBegin+1);
    // Calculate function values at the beginning of all frequency bins
    for (i=0;i<(nuEnd-nuBegin);i++) {
      fTable[i]=f(nuEdge(nuBegin+i,noEdge));
      if ((nuBegin+i)->edge) fRedTable[i]=f(nuEdge(nuBegin+i,redEdge));
      else fRedTable[i]=fTable[i];
      if ((nuBegin+i)->edge) fBlueTable[i]=f(nuEdge(nuBegin+i,blueEdge));
      else fBlueTable[i]=fTable[i];
    }
    // Calculate function values at the end of the last frequency bin
    fTable.push_back(f(nuEdge((nuEnd-1)->max,noEdge)));
    fRedTable.push_back(f(nuEdge((nuEnd-1)->max,redEdge)));
    fBlueTable.push_back(f(nuEdge((nuEnd-1)->max,blueEdge)));
  }
  // Else
  else {
    // Reset function table
    fTable.resize(0);
    fRedTable.resize(0);
    fBlueTable.resize(0);
  }
}

template <class F> void NuInterpolator<F>::initOMP(FrequencyBinIterator _nuBegin, FrequencyBinIterator _nuEnd, const F& _f, double _edgeWidth) {
  int i;

  // Initialize
  nuBegin=_nuBegin;
  nuEnd=_nuEnd;
  f=_f;
  edgeWidth=_edgeWidth;

  // If any frequency bins to tabulate
  if (nuEnd>nuBegin){
    // Calculate function values at the beginning of all frequency bin
    fTable=f(nuBegin,nuEnd,noEdge);
    fRedTable=f(nuBegin,nuEnd,redEdge);
    fBlueTable=f(nuBegin,nuEnd,blueEdge);
    // To be on the safe side - Replace zeros in red and blue tables
    for (i=0;i<(nuEnd-nuBegin);i++)
      if (not ((nuBegin+i)->edge)) { fRedTable[i]=fTable[i]; fBlueTable[i]=fTable[i]; }
    // Calculate function values at the end of the last frequency bin
    fTable.push_back(f(nuEdge((nuEnd-1)->max,noEdge)));
    fRedTable.push_back(f(nuEdge((nuEnd-1)->max,redEdge)));
    fBlueTable.push_back(f(nuEdge((nuEnd-1)->max,blueEdge)));
  }
  // Else
  else {
    // Reset function table
    fTable.resize(0);
    fRedTable.resize(0);
    fBlueTable.resize(0);
  }
}

template <class F> vector<double>::const_iterator NuInterpolator<F>::find(const double& nu) const {
  long iBegin,iMiddle,iEnd;
  
  if (nuEnd>nuBegin) {
    iBegin=0; iEnd=(nuEnd-nuBegin)-1; iMiddle=(iBegin+iEnd)/2;
    if (nu>=(nuBegin+iBegin)->min && nu<(nuBegin+iEnd)->min) {
       while (iEnd-iBegin>1) {
	if (nu<(nuBegin+iMiddle)->min) {
	  iEnd=iMiddle;
	  iMiddle=(iBegin+iEnd)/2;
	}
	else {
	  iBegin=iMiddle;
	  iMiddle=(iBegin+iEnd)/2;
	}
      }
      return fTable.begin()+iBegin;
    }
    else if (nu<=(nuBegin+iEnd)->max) return fTable.begin()+iEnd;
  }
  
  return fTable.end();
}

template <class F> double NuInterpolator<F>::operator ()(FrequencyBinIterator nuBin, EdgeType edgeType) const {
  int i;

  if (nuBin>=nuBegin and nuBin<=nuEnd) {
    if (edgeType==noEdge) return fTable[(nuBin-nuBegin)];
    else if (edgeType==redEdge) return fRedTable[(nuBin-nuBegin)];
    else return fBlueTable[(nuBin-nuBegin)];
  }
  else { cout<<"NuInterpolator: Error: Outside tabulated frequency range, nuEnd-nuBegin: "<<nuEnd-nuBegin<<", nuBegin-nuBin: "<<nuBegin-nuBin<<", nuBin-nuEnd: "<<nuBin-nuEnd<<"\n"; return 0; }
}

template <class F> double NuInterpolator<F>::operator ()(double nu) const {
  vector<double>::const_iterator fIterator=find(nu);
  int i=fIterator-fTable.begin();
  double _f,dfdx;
    
  if (fIterator==fTable.end()) return f(nu);
  else {
    dfdx=(fRedTable[i+1]-fBlueTable[i])/(nuEdge((nuBegin+i)->max,redEdge)-nuEdge((nuBegin+i)->min,blueEdge));
    _f=fRedTable[i]+dfdx*(nu-nuEdge((nuBegin+i)->min,blueEdge));
  }

  return _f;
}

template <class F> double NuInterpolator<F>::operator ()(FrequencyBinIterator nuBin, double nu) const {
  double _f,dfdx;

  if (nuBin>=nuBegin and nuBin<nuEnd) {
    dfdx=(fRedTable[nuBin-nuBegin+1]-fBlueTable[nuBin-nuBegin])/(nuEdge(nuBin->max,redEdge)-nuEdge(nuBin->min,blueEdge));
    _f=fRedTable[nuBin-nuBegin]+dfdx*(nu-nuEdge(nuBin->min,blueEdge));
  }
  else { cout<<"NuInterpolator: Error: Outside tabulated frequency range, nuEnd-nuBegin: "<<nuEnd-nuBegin<<", nuBegin-nuBin: "<<nuBegin-nuBin<<", nuBin-nuEnd: "<<nuBin-nuEnd<<"\n"; return 0; }

  return _f;
}
