// Constructor
BState::BState() : index(0), data(NULL), pIState(NULL), pX(0), pn(update), pD(0), LTERefState(NULL), LTEGroup(0), transPList(), markovPList(), downTransBB(0), upTransBB(0), downTransBF(0), upTransBF(0) { } // Changed! - Testing!

// Copy constructor
// data is owned by the cell and needs to be updated elsewhere if necessay
// transPList and markovPList are reset and needs to be updated elsewhere
// downTransBB and downTransBF have to be updated elsewhere
// upTransBB and upTransBF have to be updated elsewhere
// LTERefState has to be updated elsewhere
// LTEGroup has to be updated elsewhere
BState::BState(const BState& bState) : index(bState.index), data(bState.data), pIState(bState.pIState), pX(bState.pX), pn(bState.pn), pD(bState.pD), pTransPRBoost(bState.pTransPRBoost), pTransPCBoost(bState.pTransPCBoost), pTransPBoost(bState.pTransPBoost), pMarkovPBoost(bState.pMarkovPBoost), LTERefState(NULL), LTEGroup(0), transPList(), markovPList(), downTransBB(0), upTransBB(0), downTransBF(0), upTransBF(0) { } // Changed! - Testing!

// Assignment operator
// data is owned by the cell and needs to be updated elsewhere if necessay
// transPList and markovPList are reset and needs to be updated elsewhere
// downTransBB and downTransBF have to be updated elsewhere
// upTransBB and downTransBF have to be updated elsewhere
// LTEREfState has to be updated elsewhere
BState BState::operator=(const BState& bState) {
  index=bState.index;
  data=bState.data;
  pIState=bState.pIState;
  pX=bState.pX;
  pn=bState.pn; // New! - Testing!
  pD=bState.pD;
  pTransPRBoost=bState.pTransPRBoost;
  pTransPCBoost=bState.pTransPCBoost;
  pTransPBoost=bState.pTransPBoost;
  pMarkovPBoost=bState.pMarkovPBoost;
  transPList.resize(0);
  markovPList.resize(0);
  LTERefState=NULL;
  LTEGroup.resize(0);
}

// Testing!
BState::~BState() {
  // Testing!
  //if (iState()!=NULL) if (iState()->atom()!=NULL) if (iState()->atom()->cell()!=NULL) if (iState()->atom()->cell()->master()) cout<<"BState::~BState: cell: "<<iState()->atom()->cell()->index()<<", "<<iState()->name()<<":"<<getIndex()<<"\n";  
}

// Get index
int BState::getIndex() const {
  return index;
}

// Get ionization state back reference pointer
IState* BState::iState() const {
  return pIState;
}

// Get number fraction
double BState::X() const {
  return pX;
}

// Set number fraction
void BState::X(const double& _X) {
  pX=_X;
  return;
}

// Get number density
/*double BState::n() const {
  return pIState->n()*pX;
  }*/

// Get number density - New version - Testing!
double BState::n() {
  if (pn==update) {
    pn=pIState->n()*pX;
  }
  return pn;
  // Back to old version!
  //return pIState->n()*pX;
}

// Set number density
void BState::n(const double& _n) {
  pX=pX*(_n/n());
  return;
}

// Get energy above ground level
double BState::E() const {
  return data->E;
}

// Get energy above ground level
double BState::Etot() const {
  return data->Etot;
}

// Get statistical weight
double BState::g() const {
  return data->g;
}

// Get cumulative A value
double BState::A() const {
  return data->A;
}

// Is metastable
bool BState::isMetastable() const {
  double metastableLim=10;
  if (data->A>=metastableLim) { 
    return false; 
  }
  return true;
}

// Return LTE group master
BState* BState::LTEGroupMaster() {
  if (LTERefState!=NULL) return LTERefState;
  else return this;
}

// Bound state is part of LTE group - New!
bool BState::partOfLTEGroup() const {
  return (LTERefState!=NULL or LTEGroup.size()!=0);
}

// Draw bound state from LTE group - New!
BState* BState::drawFromLTEGroup(bool boundFree, bool coll, bool nonThermal, bool twoPhoton, bool chargeTransfer) {
  double z,P,PTotal,DTotal;
  int i;

  // If not master state - Draw bound state from master state LTE group
  if (LTERefState!=NULL) return LTERefState->drawFromLTEGroup(boundFree,coll,nonThermal,twoPhoton,chargeTransfer);

  // Initialize
  PTotal=0;

  // Pick a random number between 0 och 1
  z=randomize(0,1);

  // Get total energy flow out of the group
  DTotal=D(boundFree,coll,nonThermal,twoPhoton,chargeTransfer,true);

  // Get total energy flow out of the master state
  P=D(boundFree,coll,nonThermal,twoPhoton,chargeTransfer)/DTotal;

  // If drawn
  if (z<=(PTotal+P)) return this;
  else PTotal+=P;

  // For each slave state
  for (i=0;i<LTEGroup.size();i++) {

    // Get total energy flow out of the state
    P=LTEGroup[i]->D(boundFree,coll,nonThermal,twoPhoton,chargeTransfer)/DTotal;

    // If drawn
    if (z<=(PTotal+P)) return LTEGroup[i];
    else PTotal+=P;    
  }

  cout<<"BState::drawFromLTEGroup: Error: No bound state drawn\n";
  cout<<"BState::drawFromLTEGroup: z: "<<z<<", PTotal: "<<PTotal<<"\n";

  exit(-1);

  // This should not happen
  return NULL;
}

// Total energy flow out of the state in downward transitions (needed for macro-atom machinery)
double BState::ED(bool boundFree, bool coll, bool twoPhoton, bool chargeTransfer) const { // Changed!
  Cell* cell=iState()->atom()->cell(); // New!
  double _ED=0;
  int i;  

  for (i=0; i<downTransBB.size(); i++) {
    if (cell->config->LTEGroupMacro and downTransBB[i]->isLTEGroupInternal()) continue; // New!
    _ED+=downTransBB[i]->RD(twoPhoton)*downTransBB[i]->lowerState->Etot(); // Changed!
    if (coll) _ED+=downTransBB[i]->CD()*downTransBB[i]->lowerState->Etot();
  }
  if (boundFree) 
    for (i=0; i<downTransBF.size(); i++) {
      if (cell->config->LTEGroupMacro and downTransBF[i]->isLTEGroupInternal()) continue; // New!
      _ED+=downTransBF[i]->RD()*downTransBF[i]->lowerState->Etot();
      _ED+=downTransBF[i]->RDAdd()*downTransBF[i]->lowerState->Etot(); // New!
      if (coll) _ED+=downTransBF[i]->CD()*downTransBF[i]->lowerState->Etot();
      if (chargeTransfer) _ED+=downTransBF[i]->CTD()*downTransBF[i]->lowerState->Etot(); // New!
    }

  return _ED;
}

// Total energy flow out of the state in upward transitions (needed for macro-atom machinery)
double BState::EU(bool boundFree, bool coll, bool nonThermal, bool chargeTransfer) const { // Changed!
  Cell* cell=iState()->atom()->cell(); // New!
  double _EU=0;
  int i;

  for (i=0; i<upTransBB.size(); i++) {
    if (cell->config->LTEGroupMacro and upTransBB[i]->isLTEGroupInternal()) continue; // New!
    _EU+=upTransBB[i]->RU()*Etot();
    if (coll) _EU+=upTransBB[i]->CU()*Etot();
    if (nonThermal) _EU+=upTransBB[i]->CNU()*Etot();
  }
  if (boundFree) 
    for (i=0; i<upTransBF.size(); i++) {
      if (cell->config->LTEGroupMacro and upTransBF[i]->isLTEGroupInternal()) continue; // New!
      _EU+=upTransBF[i]->RU()*Etot();
      if (coll) _EU+=upTransBF[i]->CU()*Etot();
      if (nonThermal) _EU+=upTransBF[i]->CNU()*Etot();
      if (chargeTransfer) _EU+=upTransBF[i]->CTU()*Etot(); // New!
    }

  return _EU;
}

// Total radiative energy flow out of the state (needed for macro-atom machinery)
double BState::ER(bool boundFree, bool twoPhoton, bool chargeTransfer) const { // Changed!
  double _ER=0;
  bool chargeTransferEmit=iState()->atom()->cell()->config->chargeTransferEmit;
  int i;

  for (i=0; i<downTransBB.size(); i++) _ER+=downTransBB[i]->RD(twoPhoton)*downTransBB[i]->Xi(); // Changed!
  if (boundFree)
    for (i=0; i<downTransBF.size(); i++) {
      _ER+=downTransBF[i]->RD()*downTransBF[i]->Xi();
      // Testing!
      _ER+=downTransBF[i]->RDAdd()*downTransBF[i]->Xi(); // New!
      //if (chargeTransfer and chargeTransferEmit) _ER+=downTransBF[i]->CTD()*downTransBF[i]->Xi(); // New!
    }

  return _ER;
}

// Total collisional energy flow out of the state (needed for macro-atom machinery)
double BState::EC(bool boundFree) const {
  double _EC=0;
  int i;

  for (i=0; i<downTransBB.size(); i++) _EC+=downTransBB[i]->CD()*downTransBB[i]->Xi();
  if (boundFree) for (i=0; i<downTransBF.size(); i++) _EC+=downTransBF[i]->CD()*downTransBF[i]->Xi();

  return _EC;
}

// Total charge-transfer energy flow out of the state (needed for macro-atom machinery) - New!
double BState::ECT() const {
  double _ECT=0;
  int i;

  for (i=0; i<downTransBF.size(); i++) _ECT+=downTransBF[i]->CTD()*downTransBF[i]->Xi();

  return _ECT;
}

// Total energy flow out of the state (needed for macro-atom machinery)
double BState::D(bool boundFree, bool coll, bool nonThermal, bool twoPhoton, bool chargeTransfer, bool sumLTEGroup) { // Changed
  //double small=10*numeric_limits<double>::min();
  double DSum; // New!
  int i; // New!

  // New block!
  if (sumLTEGroup) { 
    if (LTERefState!=NULL) return LTERefState->D(boundFree,coll,nonThermal,twoPhoton,chargeTransfer,sumLTEGroup);
    else { 
      DSum=D(boundFree,coll,nonThermal,twoPhoton,chargeTransfer);
      for (i=0;i<LTEGroup.size();i++) DSum+=LTEGroup[i]->D(boundFree,coll,nonThermal,twoPhoton,chargeTransfer);
      return DSum;
    }
  }

  if (pD==update) {
    pD=ER(boundFree,twoPhoton,chargeTransfer); // Changed!
    if (coll) pD+=EC(boundFree);
    if (boundFree and chargeTransfer) pD+=ECT(); // New!
    pD+=ED(boundFree,coll,twoPhoton,chargeTransfer); // Changed!
    pD+=EU(boundFree,coll,nonThermal,chargeTransfer); // Changed!
    //if (pD==0) pD==small; // Testing! - Prevents division by zero when normalizing.
  }

  // Testing!
  if (!isnumber(pD)) cout<<"BState::D: "<<iState()->name()<<":"<<getIndex()<<", cell: "<<iState()->atom()->cell()->index()<<", Warning: Invalid rate: "<<pD<<"\n";
  if (pD==0) cout<<"BState::D: "<<iState()->name()<<":"<<getIndex()<<", cell: "<<iState()->atom()->cell()->index()<<", Warning: Zero rate\n";

  return pD;
}

// Total bound-free opacity
double BState::boundFreeKappa(const double& nu, bool cutOff) const {
  double kappa=0;
  int i;

  for (i=0; i<upTransBF.size(); i++) kappa+=upTransBF[i]->kappa(nu,cutOff);

  return kappa;
}

// Total bound-free opacity
double BState::boundFreeKappa(FrequencyBinIterator nu, EdgeType edgeType, bool cutOff) const {
  double kappa=0;
  int i;

  for (i=0; i<upTransBF.size(); i++) kappa+=upTransBF[i]->kappa(nu,edgeType,cutOff);

  return kappa;
}

// Total bound-free opacity number derivative
double BState::boundFreedKappadN(const double& nu) {
  double dKappadN=0;
  int i;

  for (i=0; i<upTransBF.size(); i++) dKappadN+=upTransBF[i]->dKappadN(nu,this);
  for (i=0; i<downTransBF.size(); i++) dKappadN+=downTransBF[i]->dKappadN(nu,this);

  return dKappadN;
}


// Total bound-free emissivity
double BState::boundFreeEta(const double& nu, bool cutOff) const {
  double eta=0;
  int i;

  for (i=0; i<downTransBF.size(); i++) eta+=downTransBF[i]->eta(nu,cutOff);

  return eta;
}

// Total bound-free emissivity
double BState::boundFreeEta(FrequencyBinIterator nu, EdgeType edgeType, bool cutOff) const {
  double eta=0;
  int i;

  for (i=0; i<downTransBF.size(); i++) eta+=downTransBF[i]->eta(nu,edgeType,cutOff);

  return eta;
}

// Total bound-free emissivity number derivative
double BState::boundFreedEtadN(const double& nu) {
  double dEtadN=0;
  int i;

  for (i=0; i<downTransBF.size(); i++) dEtadN+=downTransBF[i]->dEtadN(nu,this);

  return dEtadN;
}

// Total bound-free cooling rate
double BState::boundFreeCooling(double T) const {
  double C=0;
  int i;
  if (T==noT) T=iState()->atom()->cell()->T();

  for (i=0; i<downTransBF.size(); i++) C+=downTransBF[i]->cooling(T);

  return C;
}

// Total (spontaneous) bound-free cooling rate
double BState::boundFreeCoolingSP(double T) const {
  double C=0;
  int i;
  if (T==noT) T=iState()->atom()->cell()->T();

  for (i=0; i<downTransBF.size(); i++) C+=downTransBF[i]->coolingSP(T);

  return C;
}

// Total bound-free heating rate
double BState::boundFreeHeating(double T) const {
  double H=0;
  int i;
  if (T==noT) T=iState()->atom()->cell()->T();

  for (i=0; i<upTransBF.size(); i++) H+=upTransBF[i]->heating(T);

  if (!isnumber(H) or H<0) cout<<"BState::boundFreeHeating: "<<iState()->name()<<":"<<index<<", Warning: Invalid heating rate: "<<H<<"\n";

  return H;
}

// Total collisional cooling rate
double BState::collCooling(double T) const {
  double C=0;
  int i;
  if (T==noT) T=iState()->atom()->cell()->T();

  for (i=0; i<upTransBB.size(); i++) C+=upTransBB[i]->CU(T)*upTransBB[i]->Xi(); 
  for (i=0; i<upTransBF.size(); i++) C+=upTransBF[i]->CU(T)*upTransBF[i]->Xi();

  return C;
}

// Total collisional heating rate
double BState::collHeating(double T) const {
  double H=0;
  int i;
  if (T==noT) T=iState()->atom()->cell()->T();

  for (i=0; i<downTransBB.size(); i++) H+=downTransBB[i]->CD(T)*downTransBB[i]->Xi();
  for (i=0; i<downTransBF.size(); i++) H+=downTransBF[i]->CD(T)*downTransBF[i]->Xi();

  if (!isnumber(H) or H<0) cout<<"BState::collHeating: "<<iState()->name()<<":"<<index<<", Warning: Invalid heating rate: "<<H<<"\n";

  return H;
}

// Total charge-transfer cooling rate - New!
double BState::CTCooling(double T) const {
  double C=0;
  int i,j;
  if (T==noT) T=iState()->atom()->cell()->T();

  for (i=0; i<upTransBF.size(); i++) {
    for (j=0;j<upTransBF[i]->CTUReaction.size();j++) C+=upTransBF[i]->CTUReaction[j]->forwardCooling(T);
    for (j=0;j<upTransBF[i]->CTDReaction.size();j++) C+=upTransBF[i]->CTDReaction[j]->backwardCooling(T);
  }

  return C;
}

// Total charge-transfer heating rate - New!
double BState::CTHeating(double T) const {
  double H=0;
  int i,j;
  if (T==noT) T=iState()->atom()->cell()->T();

  for (i=0; i<downTransBF.size(); i++) {
    for (j=0;j<downTransBF[i]->CTDReaction.size();j++) H+=downTransBF[i]->CTDReaction[j]->forwardHeating(T);
    for (j=0;j<downTransBF[i]->CTUReaction.size();j++) H+=downTransBF[i]->CTUReaction[j]->backwardHeating(T);
  }

  return H;
}

// De-excite (implements the macro-atom machinery)
Trans* BState::deExcite(TransType& transType, BState* root, bool& deActivate, bool boundFree, bool coll, bool nonThermal, bool twoPhoton, bool chargeTransfer) { // Changed!
  deActivate=false;
  if (iState()->atom()->cell()->config->transTable)
    if (iState()->atom()->cell()->config->markovTable or iState()->atom()->cell()->config->markovTableExt or iState()->atom()->cell()->config->markovTableExtSplit) 
      return deExciteMarkovTable(transType,root,deActivate,boundFree,coll,nonThermal,twoPhoton,chargeTransfer); // Changed!
    else 
      return deExciteTable(transType,root,deActivate,boundFree,coll,nonThermal,twoPhoton,chargeTransfer); // Changed!
  else return deExciteNew(transType,root,deActivate,boundFree,coll,nonThermal,twoPhoton,chargeTransfer); // Changed!
}

// De-excite (implements the macro-atom machinery) - Recursive version
Trans* BState::deExciteRec(TransType& transType, BState* root, bool& deActivate, bool boundFree, bool coll, bool nonThermal, bool twoPhoton, bool chargeTransfer) { // Changed!
  double P,z;
  double PTotal;
  double PMax=1;
  Trans trans;
  bool repeat=true;
  bool chargeTransferEmit=iState()->atom()->cell()->config->chargeTransferEmit; // New!
  bool useLTEGroup=iState()->atom()->cell()->config->LTEGroupMacro; // New!
  BState* bState; // New!
  int i;

  // Override process flags with cell configuration values (as in table version)
  boundFree=iState()->atom()->cell()->config->boundFreeMacro;
  coll=iState()->atom()->cell()->config->collMacro;
  nonThermal=iState()->atom()->cell()->config->nonThermalMacro;
  twoPhoton=iState()->atom()->cell()->config->twoPhotonMacro; // New!
  chargeTransfer=iState()->atom()->cell()->config->chargeTransferMacro; // New!

  // If use LTE group and bound state part of group - New!
  if (useLTEGroup and partOfLTEGroup()) {
    // Draw bound state from group
    bState=drawFromLTEGroup(); 
    // Continue from the drawn bound state
    return bState->deExcite(transType,root,deActivate,boundFree,coll,nonThermal,twoPhoton,chargeTransfer);
  }
 
  // Initialize
  PTotal=0;

  // Pick a random number between 0 och PMax
  z=randomize(0,PMax);

  // For each downward bound-bound transition
  for (i=0; i<downTransBB.size(); i++) {
      
    // Calculate the probability for radiative de-excitation
    P=(downTransBB[i]->RD(twoPhoton)*downTransBB[i]->Xi())/D(boundFree,coll,nonThermal,twoPhoton,chargeTransfer); // Changed!

    // If drawn
    if (z<(PTotal+P)) {
      transType=radiativeBB;
      return downTransBB[i];
    }
    else PTotal+=P;

    // If collisional processes included
    if (coll) {

      // Calculate the probability for collisional de-excitation
      P=(downTransBB[i]->CD()*downTransBB[i]->Xi())/D(boundFree,coll,nonThermal,twoPhoton,chargeTransfer); // Changed!

      // If drawn
      if (z<(PTotal+P)) {
	transType=collisional;
	return downTransBB[i];
      }
      else PTotal+=P;
    }
  }

  // If bound-free processes included
  if (boundFree) {

    // For each downward bound-free transition
    for (i=0; i<downTransBF.size() and z>=PTotal; i++) {

      // Calculate the probability for radiative recombination
      P=(downTransBF[i]->RD()*downTransBF[i]->Xi())/D(boundFree,coll,nonThermal,twoPhoton,chargeTransfer); // Changed!
      // Testing!!!
      P+=(downTransBF[i]->RDAdd()*downTransBF[i]->Xi())/D(boundFree,coll,nonThermal,twoPhoton,chargeTransfer); // New!
      //if (chargeTransfer and chargeTransferEmit) P+=downTransBF[i]->CTD()*downTransBF[i]->Xi()/D(boundFree,coll,nonThermal,twoPhoton,chargeTransfer); // New! Maybe include in RD instead?

      // If drawn
      if (z<(PTotal+P)) {
	transType=radiativeBF;
	return downTransBF[i];
      }
      else PTotal+=P;

      // If collisional processed included
      if (coll) {
    
	// Calculate the probability for collisional recombination
	P=(downTransBF[i]->CD()*downTransBF[i]->Xi())/D(boundFree,coll,nonThermal,twoPhoton,chargeTransfer); // Changed!

	// If drawn
	if (z<(PTotal+P)) {
	  transType=collisional;
	  return downTransBF[i];
	}
	else PTotal+=P;
      }

      // If charge-transfer processes included - New!
      if (chargeTransfer) {
    
	// Calculate the probability for charge-transfer recombination
	P=(downTransBF[i]->CTD()*downTransBF[i]->Xi())/D(boundFree,coll,nonThermal,twoPhoton,chargeTransfer); // Changed!

	// If drawn
	if (z<(PTotal+P)) {
	  transType=chargeTransferBF;
	  return downTransBF[i];
	}
	else PTotal+=P;
      }
    }
  }
  
  // For each downward bound-bound transition
  for (i=0; i<downTransBB.size() and z>=PTotal; i++) {

    // If not (use LTE group and trans is internal to the group) - New!
    if (not (useLTEGroup and downTransBB[i]->isLTEGroupInternal())) {
    
      // Calculate the probability for an internal transition
      P=downTransBB[i]->RD(twoPhoton)*downTransBB[i]->lowerState->Etot()/D(boundFree,coll,nonThermal,twoPhoton,chargeTransfer); // Changed!
      if (coll) P+=downTransBB[i]->CD()*downTransBB[i]->lowerState->Etot()/D(boundFree,coll,nonThermal,twoPhoton,chargeTransfer); // Changed!
      
      // If drawn continue from the new bound state
      if (z<(PTotal+P)) {
	return downTransBB[i]->lowerState->deExcite(transType,root,deActivate,boundFree,coll,nonThermal,twoPhoton,chargeTransfer); // Changed!
      }
      else PTotal+=P;
    }
  }

  // If bound-free processes included
  if (boundFree) {

    // For each downward bound-free transition
    for (i=0; i<downTransBF.size() and z>=PTotal; i++) {

      // If not (use LTE group and trans is internal to the group) - New!
      if (not (useLTEGroup and downTransBF[i]->isLTEGroupInternal())) {
	
	// Calculate the probability for an internal transition
	P=downTransBF[i]->RD()*downTransBF[i]->lowerState->Etot()/D(boundFree,coll,nonThermal,twoPhoton,chargeTransfer); // Changed!
	P+=downTransBF[i]->RDAdd()*downTransBF[i]->lowerState->Etot()/D(boundFree,coll,nonThermal,twoPhoton,chargeTransfer); // New!
	if (coll) P+=downTransBF[i]->CD()*downTransBF[i]->lowerState->Etot()/D(boundFree,coll,nonThermal,twoPhoton,chargeTransfer); // Changed!
	if (chargeTransfer) P+=downTransBF[i]->CTD()*downTransBF[i]->lowerState->Etot()/D(boundFree,coll,nonThermal,twoPhoton,chargeTransfer); // New!
	
	// If drawn continue from the new bound state
	if (z<(PTotal+P)) {
	  return downTransBF[i]->lowerState->deExcite(transType,root,deActivate,boundFree,coll,nonThermal,twoPhoton,chargeTransfer); // Changed!
	}
	else PTotal+=P;
      }
    }
  }
  
  // For each upward bound-bound transition
  for (i=0; i<upTransBB.size() and z>=PTotal; i++) {

    // If not (use LTE group and trans is internal to the group) - New!
    if (not (useLTEGroup and upTransBB[i]->isLTEGroupInternal())) {
    
      // Calculate the probability for an internal transition
      P=upTransBB[i]->RU()*Etot()/D(boundFree,coll,nonThermal,twoPhoton,chargeTransfer); // Changed!
      if (coll) P+=upTransBB[i]->CU()*Etot()/D(boundFree,coll,nonThermal,twoPhoton,chargeTransfer); // Changed!
      if (nonThermal) P+=upTransBB[i]->CNU()*Etot()/D(boundFree,coll,nonThermal,twoPhoton,chargeTransfer); //Changed!
      
      // If drawn continue from the new bound state
      if (z<(PTotal+P)) {
	return upTransBB[i]->upperState->deExcite(transType,root,deActivate,boundFree,coll,nonThermal,twoPhoton,chargeTransfer); // Changed!
      }
      else PTotal+=P;
    }
  }

  // If bound-free processes included
  if (boundFree) {

    // For each upward bound-free transition
    for (i=0; i<upTransBF.size() and z>=PTotal; i++) {

      // If not (use LTE group and trans is internal to the group) - New!
      if (not (useLTEGroup and upTransBF[i]->isLTEGroupInternal())) {
	
	// Calculate the probability for an internal transition
	P=upTransBF[i]->RU()*Etot()/D(boundFree,coll,nonThermal,twoPhoton,chargeTransfer); // Changed!
	if (coll) P+=upTransBF[i]->CU()*Etot()/D(boundFree,coll,nonThermal,twoPhoton,chargeTransfer); // Changed!
	if (nonThermal) P+=upTransBF[i]->CNU()*Etot()/D(boundFree,coll,nonThermal,twoPhoton,chargeTransfer); // Changed!
	if (chargeTransfer) P+=upTransBF[i]->CTU()*Etot()/D(boundFree,coll,nonThermal,twoPhoton,chargeTransfer); // New!
	
	// If drawn continue from the new bound state
	if (z<(PTotal+P)) {
	  return upTransBF[i]->upperState->deExcite(transType,root,deActivate,boundFree,coll,nonThermal,twoPhoton,chargeTransfer); // Changed!
	}
	else PTotal+=P;
      }
    }
  }

  cout<<"BState::deExcite: Error: No Transition drawn\n";
  cout<<"BState::deExcite: z: "<<z<<", PTotal: "<<PTotal<<", E: "<<E()<<", D: "<<D(boundFree,coll,nonThermal,twoPhoton,chargeTransfer)<<", Etot: "<<Etot()<<", coll: "<<coll<<"\n"; // Changed!

  exit(-1);

  // This should not happen
  return NULL;
}

// De-excite (implements the macro-atom machinery) - New version
Trans* BState::deExciteNew(TransType& transType, BState* root, bool& deActivate, bool boundFree, bool coll, bool nonThermal, bool twoPhoton, bool chargeTransfer) { // Changed!
  double P,z;
  double PTotal;
  double PMax=1;
  BState* bState;
  bool repeat=true;
  bool chargeTransferEmit=iState()->atom()->cell()->config->chargeTransferEmit; // New!
  bool useLTEGroup=iState()->atom()->cell()->config->LTEGroupMacro; // New!
  int i;

  // Override process flags with cell configuration values (as in table version) - Add?!

  // Set the current bound state
  bState=root;

  // State loop
  while (repeat) {

    // If use LTE group and bound state part of group - Draw bound state from group
    if (useLTEGroup and bState->partOfLTEGroup()) bState=bState->drawFromLTEGroup();

    // Reset repeat flag
    repeat=false;

    // Reset total probability
    PTotal=0;

    // Pick a random number between 0 och PMax
    z=randomize(0,PMax);

    // For each downward bound-bound transition
    for (i=0; i<bState->downTransBB.size(); i++) {
      
      // Calculate the probability for radiative de-excitation
      P=(bState->downTransBB[i]->RD(twoPhoton)*bState->downTransBB[i]->Xi())/bState->D(boundFree,coll,nonThermal,twoPhoton,chargeTransfer); // Changed!

      // If drawn
      // Changed to <= to account for z=1
      if (z<=(PTotal+P)) {
	transType=radiativeBB;
	return bState->downTransBB[i];
      }
      else PTotal+=P;

      // If collisional processes included
      if (coll) {

	// Calculate the probability for collisional de-excitation
	P=(bState->downTransBB[i]->CD()*bState->downTransBB[i]->Xi())/bState->D(boundFree,coll,nonThermal,twoPhoton,chargeTransfer); // Changed!

	// If drawn
	// Changed to <= to account for z=1
	if (z<=(PTotal+P)) {
	  transType=collisional;
	  return bState->downTransBB[i];
	}
	else PTotal+=P;
      }
    }

    // If bound-free processes included
    if (boundFree) {

      // For each downward bound-free transition
      for (i=0; i<bState->downTransBF.size() and z>=PTotal; i++) {

	// Calculate the probability for radiative recombination
	P=(bState->downTransBF[i]->RD()*bState->downTransBF[i]->Xi())/bState->D(boundFree,coll,nonThermal,twoPhoton,chargeTransfer); // Changed!
	// Testing!
	P+=(bState->downTransBF[i]->RDAdd()*bState->downTransBF[i]->Xi())/bState->D(boundFree,coll,nonThermal,twoPhoton,chargeTransfer); // New!
	//if (chargeTransfer and chargeTransferEmit) P+=(bState->downTransBF[i]->CTD()*bState->downTransBF[i]->Xi())/bState->D(boundFree,coll,nonThermal,twoPhoton,chargeTransfer); // New! Maybe include in RD instead?

	// If drawn
	// Changed to <= to account for z=1
	if (z<=(PTotal+P)) {
	  transType=radiativeBF;
	  return bState->downTransBF[i];
	}
	else PTotal+=P;

	// If collisional processed included
	if (coll) {
	  
	  // Calculate the probability for collisional recombination
	  P=(bState->downTransBF[i]->CD()*bState->downTransBF[i]->Xi())/bState->D(boundFree,coll,nonThermal,twoPhoton,chargeTransfer); // Changed!

	  // If drawn
	  // Changed to <= to account for z=1
	  if (z<=(PTotal+P)) {
	    transType=collisional;
	    return bState->downTransBF[i];
	  }
	  else PTotal+=P;
	}

	// If charge-transfer processes included - New!
	if (chargeTransfer) {
	  
	  // Calculate the probability for charge-transfer recombination
	  P=(bState->downTransBF[i]->CTD()*bState->downTransBF[i]->Xi())/bState->D(boundFree,coll,nonThermal,twoPhoton,chargeTransfer); // Changed!

	  // If drawn
	  // Changed to <= to account for z=1
	  if (z<=(PTotal+P)) {
	    transType=chargeTransferBF;
	    return bState->downTransBF[i];
	  }
	  else PTotal+=P;
	}
      }
    }
  
    // For each downward bound-bound transition
    for (i=0; i<bState->downTransBB.size() and z>=PTotal; i++) {

      // If not (use LTE group and trans is internal to the group) - New!
      if (not (useLTEGroup and downTransBB[i]->isLTEGroupInternal())) {
      
	// Calculate the probability for an internal transition
	P=bState->downTransBB[i]->RD(twoPhoton)*bState->downTransBB[i]->lowerState->Etot()/bState->D(boundFree,coll,nonThermal,twoPhoton,chargeTransfer); // Changed!
	if (coll) P+=bState->downTransBB[i]->CD()*bState->downTransBB[i]->lowerState->Etot()/bState->D(boundFree,coll,nonThermal,twoPhoton,chargeTransfer); // Changed!
	
	// If drawn continue from the new bound state
	// Changed to <= to account for z=1
	if (z<=(PTotal+P)) {
	  bState=bState->downTransBB[i]->lowerState;
	  repeat=true;
	  break;
	}
	else PTotal+=P;
      }
    }
    if (repeat) continue;

    // If bound-free processes included
    if (boundFree) {

      // For each downward bound-free transition
      for (i=0; i<bState->downTransBF.size() and z>=PTotal; i++) {

	// If not (use LTE group and trans is internal to the group) - New!
	if (not (useLTEGroup and downTransBF[i]->isLTEGroupInternal())) {
      
	  // Calculate the probability for an internal transition
	  P=bState->downTransBF[i]->RD()*bState->downTransBF[i]->lowerState->Etot()/bState->D(boundFree,coll,nonThermal,twoPhoton,chargeTransfer); // Changed!
	  P+=bState->downTransBF[i]->RDAdd()*bState->downTransBF[i]->lowerState->Etot()/bState->D(boundFree,coll,nonThermal,twoPhoton,chargeTransfer); // New!
	  if (coll) P+=bState->downTransBF[i]->CD()*bState->downTransBF[i]->lowerState->Etot()/bState->D(boundFree,coll,nonThermal,twoPhoton,chargeTransfer); // Changed!
	  if (chargeTransfer) P+=bState->downTransBF[i]->CTD()*bState->downTransBF[i]->lowerState->Etot()/bState->D(boundFree,coll,nonThermal,twoPhoton,chargeTransfer); // New!
	  
	  // If drawn continue from the new bound state
	  // Changed to <= to account for z=1
	  if (z<=(PTotal+P)) {
	    bState=bState->downTransBF[i]->lowerState;
	    repeat=true;
	    break;
	  }
	  else PTotal+=P;
	}
      }
      if (repeat) continue;
    }
  
    // For each upward bound-bound transition
    for (i=0; i<bState->upTransBB.size() and z>=PTotal; i++) {
    
      // If not (use LTE group and trans is internal to the group) - New!
      if (not (useLTEGroup and upTransBB[i]->isLTEGroupInternal())) {

	// Calculate the probability for an internal transition
	P=bState->upTransBB[i]->RU()*bState->Etot()/bState->D(boundFree,coll,nonThermal,twoPhoton,chargeTransfer); // Changed!
	if (coll) P+=bState->upTransBB[i]->CU()*bState->Etot()/bState->D(boundFree,coll,nonThermal,twoPhoton,chargeTransfer); // Changed!
	if (nonThermal) P+=bState->upTransBB[i]->CNU()*bState->Etot()/bState->D(boundFree,coll,nonThermal,twoPhoton,chargeTransfer); // Changed!
	
	// If drawn continue from the new bound state
	// Changed to <= to account for z=1
	if (z<=(PTotal+P)) {
	  bState=bState->upTransBB[i]->upperState;
	  repeat=true;
	  break;
	}
	else PTotal+=P;
      }
    }
    if (repeat) continue;

    // If bound-free processes included
    if (boundFree) {

      // For each upward bound-free transition
      for (i=0; i<bState->upTransBF.size() and z>=PTotal; i++) {

	// If not (use LTE group and trans is internal to the group) - New!
	if (not (useLTEGroup and upTransBF[i]->isLTEGroupInternal())) {
	
	  // Calculate the probability for an internal transition
	  P=bState->upTransBF[i]->RU()*bState->Etot()/bState->D(boundFree,coll,nonThermal,twoPhoton,chargeTransfer); // Changed!
	  if (coll) P+=bState->upTransBF[i]->CU()*bState->Etot()/bState->D(boundFree,coll,nonThermal,twoPhoton,chargeTransfer); // Changed!
	  if (nonThermal) P+=bState->upTransBF[i]->CNU()*bState->Etot()/bState->D(boundFree,coll,nonThermal,twoPhoton,chargeTransfer); // Changed!
	  if (chargeTransfer) P+=bState->upTransBF[i]->CTU()*bState->Etot()/bState->D(boundFree,coll,nonThermal,twoPhoton,chargeTransfer); // New!
	  
	  // If drawn continue from the new bound state
	  // Changed to <= to account for z=1
	  if (z<=(PTotal+P)) {
	    bState=bState->upTransBF[i]->upperState;
	    repeat=true;
	    break;
	  }
	  else PTotal+=P;
	}
      }
      if (repeat) continue;
    }
  }

  cout<<"BState::deExciteNew: Error: No Transition drawn\n";
  cout<<"BState::deExciteNew: cell: "<<iState()->atom()->cell()->index()<<", bState: "<<bState->iState()->name()<<":"<<bState->getIndex()<<", z: "<<z<<", PTotal: "<<PTotal<<", z-PTotal: "<<(z-PTotal)<<", E: "<<E()<<", D: "<<D(boundFree,coll,nonThermal,twoPhoton,chargeTransfer)<<", Etot: "<<Etot()<<", coll: "<<coll<<"\n"; // Changed!

  // Testing!
  //exit(-1);

  // This should not happen
  return NULL;
}

// De-excite (implements the macro-atom machinery) - Table version
Trans* BState::deExciteTable(TransType& transType, BState* root, bool& deActivate, bool boundFree, bool coll, bool nonThermal, bool twoPhoton, bool chargeTransfer) { // Changed!
  double P,z;
  double PTotal;
  double PMax=1;
  BState* bState;
  bool repeat=true;
  bool useLTEGroup=iState()->atom()->cell()->config->LTEGroupMacro; // New!
  int i;

  // Set the current bound state
  bState=root;

  // State loop
  while (repeat) {

    // If use LTE group and bound state part of group - Draw bound state from group
    if (useLTEGroup and bState->partOfLTEGroup()) bState=bState->drawFromLTEGroup();

    // Reset repeat flag
    repeat=false;

    // Reset total probability
    PTotal=0;

    // Pick a random number between 0 och PMax
    z=randomize(0,PMax);

    // For each transition
    for (i=0; i<bState->transPList.size(); i++) {

      // Increase cumulative transition probability
      PTotal+=bState->transPList[i].P;

      // If drawn
      // Changed to <= to account for z=1
      if (z<=PTotal) {


	// If internal transition
	if (bState->transPList[i].type==internal) {

	  // Get next bound state
	  if (bState->transPList[i].upward) bState=bState->transPList[i].trans->upperState;
	  else bState=bState->transPList[i].trans->lowerState;

	  // Repeat
	  repeat=true;
	  break;
	}
	// Else - i.e. de-activate macro-atom
	else {
	  
	  // Return transition
	  transType=bState->transPList[i].type;
	  return bState->transPList[i].trans;
	}
      }
    }

    // If repeat flag set continue
    if (repeat) continue;
  }

  cout<<"BState::deExciteTable: Error: No Transition drawn\n";
  cout<<"BState::deExciteTable: cell: "<<iState()->atom()->cell()->index()<<", bState: "<<bState->iState()->name()<<":"<<bState->getIndex()<<", z: "<<z<<", PTotal: "<<PTotal<<": z-PTotal: "<<(z-PTotal)<<", transPlist.size: "<<bState->transPList.size()<<"\n";

  // Testing!
  //exit(-1);

  // This should not happen
  return NULL;
}

// De-excite (implements the macro-atom machinery) - Markov table version
Trans* BState::deExciteMarkovTable(TransType& transType, BState* root, bool& deActivate, bool boundFree, bool coll, bool nonThermal, bool twoPhoton, bool chargeTransfer) { // Changed!
  double PTotal,z;
  double PMax=1;
  BState* bState=NULL;
  CTReaction* cTReaction; // New!
  CTDirection cTDirection; // New!
  bool markovTableExt=iState()->atom()->cell()->config->markovTableExt;
  bool markovTableExtSplit=iState()->atom()->cell()->config->markovTableExtSplit;
  bool useLTEGroup=iState()->atom()->cell()->config->LTEGroupMacro; // New!
  int i;

  // Testing!
  Cell* cell=iState()->atom()->cell();

  // If use LTE group and not master state - De-excite from master state - New!
  if (useLTEGroup and LTERefState!=NULL) return LTERefState->deExcite(transType,root,deActivate,boundFree,coll,nonThermal,twoPhoton,chargeTransfer);

  // Set de-activation flag depending on configuration
  //if (iState()->atom()->cell()->config->markovTableExt) deActivate=true;
  //else deActivate=false;
  deActivate=false;

  // Set the current bound state
  bState=root;

  // Reset total probability
  PTotal=0;

  // Pick a random number between 0 och PMax
  z=randomize(0,PMax);

  // For each Markov transition
  for (i=0; i<markovPList.size(); i++) {
    
    // Increase cumulative transition probability
    PTotal+=markovPList[i].P;

    // If drawn 
    // Changed to <= to account for z=1
    if (z<=PTotal) {

      // If macro-atom de-activation - Set next bound state
      if (markovPList[i].nextState!=NULL) bState=markovPList[i].nextState;
      // Else (i.e. thermal pool de-activation) - Set transition type and return NULL
      else { if (markovTableExt) deActivate=true; transType=collisional; return NULL; }

      // Break
      break;
    }
  }

  // !!! But if no transitin was drawn bState==root !!! Fix !!!
  if (bState==NULL) {
    cout<<"BState::deExciteMarkovTable: Error: No Markov transition drawn\n";
    cout<<"BState::deExciteMarkovTable: index: "<<index<<", z: "<<z<<", PTotal: "<<PTotal<<"\n";
    
    //exit(-1);
    return NULL;
  }

  // Reset total probability
  PTotal=0;

  // Pick a random number between 0 och PMax
  z=randomize(0,PMax);

  // For each transition
  for (i=0; i<bState->transPList.size(); i++) {
    
    // Increase cumulative transition probability
    PTotal+=bState->transPList[i].P;

    // If drawn
    // Changed to <= to account for z=1
    if (z<=PTotal) {

      // If extended split Markov table and (collisional or charge-transfer) transition - Break loop
      if (markovTableExtSplit and (bState->transPList[i].type==collisional or bState->transPList[i].type==chargeTransferBF)) break; // Changed!
      
      // Return transition
      if (markovTableExt) deActivate=true;
      transType=bState->transPList[i].type;
      return bState->transPList[i].trans;
    }
  }
  
  // If extended split Markov table
  if (markovTableExtSplit and i<bState->transPList.size()) { // Changed!

    // Testing!
    BState* bStateOld=bState;

    // If collisional transition - New!
    if (bState->transPList[i].type==collisional) {
      // Get next bound state (via thermal pool)
      bState=iState()->atom()->cell()->deExciteMarkovTableTherm();
    }
    // Else (i.e. charge-transfer transition) - New!
    else {
      // Draw charge.transfer reaction
      //cTReaction=((TransBF*) bState->transPList[i].trans)->drawCTDReaction();
      cTReaction=((TransBF*) bState->transPList[i].trans)->drawCTReaction(cTDirection);
      if (cTReaction==NULL) cout<<"BState::deExciteMarkovTable: cell: "<<iState()->atom()->cell()->index()<<", "<<iState()->name()<<":"<<getIndex()<<", Warning: No Transition drawn\n";
      if (cTReaction==NULL) cout<<"BState::deExciteMarkovTable: cell: "<<iState()->atom()->cell()->index()<<", "<<iState()->name()<<":"<<getIndex()<<", bState: "<<bState->iState()->name()<<":"<<bState->getIndex()<<", trans: "<<bState->transPList[i].trans->upperState->iState()->name()<<":"<<bState->transPList[i].trans->upperState->getIndex()<<"->"<<bState->transPList[i].trans->lowerState->iState()->name()<<":"<<bState->transPList[i].trans->lowerState->getIndex()<<", type: "<<bState->transPList[i].type<<"\n";
      // Get next bound state (via charge-transfer)
      bState=iState()->atom()->cell()->deExciteMarkovTableCT(cTReaction,cTDirection);
    }

    //cout<<"BState:deExciteMarkovTable: bState: "<<bState<<"\n";

    // If thermal pool de-activation - Set transition type and return
    if (bState==NULL) { deActivate=true; transType=collisional; return NULL; }

    // Reset total probability
    PTotal=0;

    // Pick a random number between 0 and the radiative de-activation probability
    z=randomize(0,bState->PRTot);

    // For each transition
    for (i=0; i<bState->transPList.size(); i++) {

      // If radiative - Increase cumulative transition probability
      if (bState->transPList[i].type!=collisional and bState->transPList[i].type!=chargeTransferBF) PTotal+=bState->transPList[i].P; // Changed!
      
      // If drawn
      if (z<=PTotal) {
	     
      // Return transition
      deActivate=true; transType=bState->transPList[i].type;
      return bState->transPList[i].trans;
      }
    } 
  }

  cout<<"BState::deExciteMarkovTable: Error: No Transition drawn\n";
  cout<<"BState::deExciteMarkovTable: z: "<<z<<", PTotal: "<<PTotal<<"\n";

  // This should not happen
  return NULL;
}

// Update transition probability list
// Probably not entirely up-to-date - Check !!!
void BState::updateTransPList() {
  TransP transP;
  double n=0, PTotal=0;
  float PFloor=numeric_limits<float>::min(); // New!
  int i;
  bool boundFree=iState()->atom()->cell()->config->boundFreeMacro;
  bool coll=iState()->atom()->cell()->config->collMacro;
  bool nonThermal=iState()->atom()->cell()->config->nonThermalMacro;
  bool twoPhoton=iState()->atom()->cell()->config->twoPhotonMacro; // New!
  bool chargeTransfer=iState()->atom()->cell()->config->chargeTransferMacro; // New!
  bool chargeTransferEmit=iState()->atom()->cell()->config->chargeTransferEmit; // New!  
  bool useLTEGroup=iState()->atom()->cell()->config->LTEGroupMacro; // New!
  
  // Reset transition probability list
  transPList.resize(0);

  // For each downward bound-bound transition
  for (i=0; i<downTransBB.size(); i++) {

    // Initialize radiative de-activation
    transP.trans=downTransBB[i];
    transP.index=i; // New! - Removed!
    transP.type=radiativeBB;
    transP.upward=false;
    
    // Calculate the probability for radiative de-activation
    transP.P=(downTransBB[i]->RD(twoPhoton)*downTransBB[i]->Xi())/D(boundFree,coll,nonThermal,twoPhoton,chargeTransfer); // Changed!
    PTotal+=transP.P; n++;
    
    // Append entry to list
    transPList.push_back(transP);

    // If collisional processes included
    if (coll) {

      // Initialize collisional de-activation
      transP.trans=downTransBB[i];
      transP.index=i; // New! - Removed!
      transP.type=collisional;
      transP.upward=false;
      
      // Calculate the probability for collisional de-activation
      transP.P=(downTransBB[i]->CD()*downTransBB[i]->Xi())/D(boundFree,coll,nonThermal,twoPhoton,chargeTransfer); // Changed!
      PTotal+=transP.P; n++;
      	     
      // Append entry to list
      transPList.push_back(transP);
    }

    // If not (use LTE group and trans is internal to the group) - New!
    if (not (useLTEGroup and downTransBB[i]->isLTEGroupInternal())) {

      // Initialize internal transition
      transP.trans=downTransBB[i];
      transP.index=i; // New! - Removed!
      transP.type=internal;
      transP.upward=false;
      
      // Calculate the probability for an internal transition
      transP.P=downTransBB[i]->RD(twoPhoton)*downTransBB[i]->lowerState->Etot()/D(boundFree,coll,nonThermal,twoPhoton,chargeTransfer); // Changed!
      if (coll) transP.P+=downTransBB[i]->CD()*downTransBB[i]->lowerState->Etot()/D(boundFree,coll,nonThermal,twoPhoton,chargeTransfer); // Changed!
      PTotal+=transP.P; n++;
      
      // Append entry to list
      transPList.push_back(transP);
    }
  }

  // For each upward bound-bound transition
  for (i=0; i<upTransBB.size(); i++) {

    // If not (use LTE group and trans is internal to the group) - New!
    if (not (useLTEGroup and upTransBB[i]->isLTEGroupInternal())) {

      // Initialize internal transition
      transP.trans=upTransBB[i];
      transP.index=i; // New! - Removed!
      transP.type=internal;
      transP.upward=true;
      
      // Calculate the probability for an internal transition
      transP.P=upTransBB[i]->RU()*Etot()/D(boundFree,coll,nonThermal,twoPhoton,chargeTransfer); // Changed!
      if (coll) transP.P+=upTransBB[i]->CU()*Etot()/D(boundFree,coll,nonThermal,twoPhoton,chargeTransfer); // Changed!
      if (nonThermal) transP.P+=upTransBB[i]->CNU()*Etot()/D(boundFree,coll,nonThermal,twoPhoton,chargeTransfer); // Changed!
      PTotal+=transP.P; n++;
      
      // Append entry to list
      transPList.push_back(transP);
    }
  }
 
  // If bound-free processes included
  if (boundFree) {

    // For each downward bound-free transition
    for (i=0; i<downTransBF.size(); i++) {

      // Initialize radiative de-activation
      transP.trans=downTransBF[i];
      transP.index=i; // New! - Removed!
      transP.type=radiativeBF;
      transP.upward=false;
      
      // Calculate the probability for radiative de-activation
      transP.P=(downTransBF[i]->RD()*downTransBF[i]->Xi())/D(boundFree,coll,nonThermal,twoPhoton,chargeTransfer); // Changed!
      // Testing!
      transP.P+=(downTransBF[i]->RDAdd()*downTransBF[i]->Xi())/D(boundFree,coll,nonThermal,twoPhoton,chargeTransfer); // New!
      //if (chargeTransfer and chargeTransferEmit) transP.P+=(downTransBF[i]->CTD()*downTransBF[i]->Xi())/D(boundFree,coll,nonThermal,twoPhoton,chargeTransfer); // New! Maybe include in RD instead?
      PTotal+=transP.P; n++;

      // Append entry to list
      transPList.push_back(transP);
      
      // If collisional processes included
      if (coll) {

	// Initialize collisional de-activation
	transP.trans=downTransBF[i];
	transP.index=i; // New! - Removed!
	transP.type=collisional;
	transP.upward=false;
      
	// Calculate the probability for collisional de-activation
	transP.P=(downTransBF[i]->CD()*downTransBF[i]->Xi())/D(boundFree,coll,nonThermal,twoPhoton,chargeTransfer); // Changed!
	PTotal+=transP.P; n++;

	// Append entry to list
	transPList.push_back(transP);
      }

      // If charge-transfer processes included - New!
      if (chargeTransfer) {

	// Initialize charge-transfer de-activation
	transP.trans=downTransBF[i];
	transP.index=i; // New! - Removed!
	transP.type=chargeTransferBF;
	transP.upward=false;
      
	// Calculate the probability for charge-transfer de-activation
	transP.P=(downTransBF[i]->CTD()*downTransBF[i]->Xi())/D(boundFree,coll,nonThermal,twoPhoton,chargeTransfer);
	PTotal+=transP.P; n++;

	// Append entry to list
	transPList.push_back(transP);
      }

      // If not (use LTE group and trans is internal to the group) - New!
      if (not (useLTEGroup and downTransBF[i]->isLTEGroupInternal())) {

	// Initialize internal transition
	transP.trans=downTransBF[i];
	transP.index=i; // New! - Removed!
	transP.type=internal;
	transP.upward=false;
	
	// Calculate the probability for an internal transition
	transP.P=downTransBF[i]->RD()*downTransBF[i]->lowerState->Etot()/D(boundFree,coll,nonThermal,twoPhoton,chargeTransfer); // Changed!
	transP.P+=downTransBF[i]->RDAdd()*downTransBF[i]->lowerState->Etot()/D(boundFree,coll,nonThermal,twoPhoton,chargeTransfer); // New!
	if (coll) transP.P+=downTransBF[i]->CD()*downTransBF[i]->lowerState->Etot()/D(boundFree,coll,nonThermal,twoPhoton,chargeTransfer); // Changed!
	if (chargeTransfer) transP.P+=downTransBF[i]->CTD()*downTransBF[i]->lowerState->Etot()/D(boundFree,coll,nonThermal,twoPhoton,chargeTransfer); // New!
	PTotal+=transP.P; n++;
	
	// Append entry to list
	transPList.push_back(transP);
      }
    }

    // For each upward bound-free transition
    for (i=0; i<upTransBF.size(); i++) {

      // If not (use LTE group and trans is internal to the group) - New!
      if (not (useLTEGroup and upTransBF[i]->isLTEGroupInternal())) {
      
	// Initialize internal transition
	transP.trans=upTransBF[i];
	transP.index=i; // New! - Removed!
	transP.type=internal;
	transP.upward=true;
	
	// Calculate the probability for an internal transition
	transP.P=upTransBF[i]->RU()*Etot()/D(boundFree,coll,nonThermal,twoPhoton,chargeTransfer); // Changed!
	if (coll) transP.P+=upTransBF[i]->CU()*Etot()/D(boundFree,coll,nonThermal,twoPhoton,chargeTransfer); // Changed!
	if (nonThermal) transP.P+=upTransBF[i]->CNU()*Etot()/D(boundFree,coll,nonThermal,twoPhoton,chargeTransfer); // Changed!
	if (chargeTransfer) transP.P+=upTransBF[i]->CTU()*Etot()/D(boundFree,coll,nonThermal,twoPhoton,chargeTransfer); // New!
	PTotal+=transP.P; n++;
	
	// Append entry to list
	transPList.push_back(transP);
      }
    }
  }

  // Set transition probability floor to avoid division by zero in the normalization - New!
  for (i=0; i<transPList.size(); i++) transPList[i].P=MAX(transPList[i].P,PFloor);

  // Normalize transition probability list - New!
  PTotal=0; for (i=0; i<transPList.size(); i++) PTotal+=transPList[i].P;
  for (i=0; i<transPList.size(); i++) transPList[i].P=transPList[i].P/PTotal;

  // Sort transition probability list in descending probability
  transPList.sort();
}

// Reset derived quantities (to be updated)
void BState::reset(int dependence) {
  int i;
  for (i=0;i<upTransBF.size();i++) upTransBF[i]->reset(dependence);
  for (i=0;i<upTransBB.size();i++) upTransBB[i]->reset(dependence);
  //pD=update;
  if (dependence&(dpRad|dpT|dpN|dpNe)) pD=update;
  if (dependence&(dpN)) pn=update; // New! - Testing!
}

// Export data
void BState::exportStat(YAML::Emitter& data, const BState& bStateLTE, const double& tauLim) {
  int i;

  data<<YAML::BeginMap;

  // Export index
  data<<YAML::Key<<"index"<<YAML::Value<<getIndex();

  // Export number fraction
  data<<YAML::Key<<"X"<<YAML::Value<<X();

  // Export departure coefficient
  // Should be in terms of number density, right?
  data<<YAML::Key<<"b"<<YAML::Value<<X()/bStateLTE.X();

  // Export data for each (upward) bound-bound transition
  // Temporary removed - Need to be configurable
  //if (iState()->atom()->name().compare("He")==0 or iState()->atom()->name().compare("O")==0 or iState()->atom()->name().compare("Na")==0 or iState()->atom()->name().compare("C")==0 or iState()->atom()->name().compare("Ca")==0 or iState()->atom()->name().compare("Mg")==0) {
  if (iState()->atom()->name().compare("He")==0) {
    data<<YAML::Key<<"transBB"<<YAML::Value<<YAML::BeginSeq;
    for (i=0;i<upTransBB.size();i++) upTransBB[i]->exportStat(data,tauLim);
    data<<YAML::EndSeq;
  }

  // Export data for each (upward) bound-free transition
  // Temporary removed - Need to be configurable
  //if (iState()->atom()->name().compare("He")==0 or iState()->atom()->name().compare("O")==0 or iState()->atom()->name().compare("Na")==0 or iState()->atom()->name().compare("C")==0 or iState()->atom()->name().compare("Ca")==0 or iState()->atom()->name().compare("Mg")==0) {
  if (iState()->atom()->name().compare("He")==0) {
    data<<YAML::Key<<"transBF"<<YAML::Value<<YAML::BeginSeq;
    for (i=0;i<upTransBF.size();i++) upTransBF[i]->exportStat(data);
    data<<YAML::EndSeq;
  }

  data<<YAML::EndMap;
}

// Load bound state
void BState::loadState(const YAML::Node& data) {
  string upperIStateName; // New!
  int i,j,upperBStateIndex; // New!

  // Get number fraction
  if (data["X"]) pX=data["X"].as<double>();
  else {cout<<"BState::loadState: Error: Could not get X\n"; exit(-1);}

  // If provided // New!
  if (data["transBB"]) {
    // For each bound-bound transitions
    for (i=0;i<data["transBB"].size();i++) {
      // Get upper ionization state name and upper bound state index
      if (data["transBB"][i]["upper"]) { 
	upperIStateName=data["transBB"][i]["upper"][0].as<string>();
	upperBStateIndex=data["transBB"][i]["upper"][1].as<int>();
      }
      else { cout<<"BState::loadState: Error: Could not get transBB upper\n"; exit(-1);}
      // Find bound-bound transition (improve this part) and load state
      for (j=0;j<upTransBB.size();j++)
	if (upTransBB[i]->upperState->getIndex()==upperBStateIndex) { upTransBB[i]->loadState(data["transBB"][i]); break; }
    } 
  }

  // If provided // New!
  if (data["transBF"]) {
    // For each bound-bound transitions
    for (i=0;i<data["transBF"].size();i++) {
      // Get upper ionization state name and upper bound state index
      if (data["transBB"][i]["upper"]) { 
	upperIStateName=data["transBF"][i]["upper"][0].as<string>();
	upperBStateIndex=data["transBF"][i]["upper"][1].as<int>();
      }
      else { cout<<"BState::loadState: Error: Could not get transBF upper\n"; exit(-1);}
      // Find bound-bound transition (improve this part) and load state
      for (j=0;j<upTransBF.size();j++)  	
	if (upTransBF[i]->upperState->iState()->name().compare(upperIStateName)==0 and upTransBF[i]->upperState->getIndex()==upperBStateIndex) { upTransBF[i]->loadState(data["transBB"][i]); break; }
    } 
  }

  return;
}

// Save bound state data
void BState::save(YAML::Emitter& data) const {
  int i;

  // Begin bound state map
  data<<YAML::BeginMap;

  // Save index
  data<<YAML::Key<<"index"<<YAML::Value<<getIndex();

  // Save number fraction
  data<<YAML::Key<<"X"<<YAML::Value<<X();

  // End bound state map
  data<<YAML::EndMap;
}

// Save bound state
void BState::saveState(YAML::Emitter& data) {
  int i;

  // Begin bound state map
  data<<YAML::BeginMap;

  // Save number fraction
  data<<YAML::Key<<"X"<<YAML::Value<<X();

  // Save state for each (upward) non-thermal bound-bound transition
  data<<YAML::Key<<"transBB"<<YAML::Value<<YAML::BeginSeq;
  for (i=0;i<upTransBB.size();i++) 
    if (upTransBB[i]->active(pNonThermal)) upTransBB[i]->saveState(data);
  data<<YAML::EndSeq;

  // Save state for each (upward) non-thermal bound-free transition
  data<<YAML::Key<<"transBF"<<YAML::Value<<YAML::BeginSeq;
  for (i=0;i<upTransBF.size();i++) 
    if (upTransBF[i]->active(pNonThermal)) upTransBF[i]->saveState(data);
  data<<YAML::EndSeq;

  // Close bound state map
  data<<YAML::EndMap;
}

// Set back references
void BState::setBackReference(IState* iState) {
  // Set ionization state back reference pointer
  pIState=iState;
}

// *****************************************
// *** New functions for boost *************
// *****************************************
// Note: Does not support charge-transfer!

// Boost transition probabilities - New!
void BState::boostTransP() {

  // Boost radiative transition probabilities
  boostTransPR();

  // Bost collisional transition probabilities
  boostTransPC();
}

// Boost radiative transition probabilities - New!
void BState::boostTransPR() {
  int i;

  // Get cell
  Cell* cell=iState()->atom()->cell();

  // If use LTE group and not master state - Set dummy value and return - New!
  if (cell->config->LTEGroupMacro and LTERefState!=NULL) { pTransPRBoost=1.0; return; }

  // Reset radiative boost factor
  pTransPRBoost=0;

  // If radiative transition probability is non-zero
  if (PRTot!=0) {
    // Calculate radiative boost factor
    for (i=0;i<transPList.size();i++) {
      if (transPList[i].type!=collisional) pTransPRBoost+=(transPList[i].P/PRTot)*transPList[i].trans->etaBoost();
      // Testing
      if (cell->master() and iState()->name().compare("AlI")==0 and cell->index()==36) cout<<"BState::boostTransPR: cell: "<<cell->index()<<", "<<iState()->name()<<":"<<getIndex()<<", transPList["<<i<<"].P: "<<transPList[i].P<<", PRTot: "<<PRTot<<",trans: "<<transPList[i].trans->upperState->iState()->name()<<":"<<transPList[i].trans->upperState->getIndex()<<"->"<<transPList[i].trans->lowerState->iState()->name()<<":"<<transPList[i].trans->lowerState->getIndex()<<", etaBoost: "<<transPList[i].trans->etaBoost()<<"\n";
    }
  }
  // Else (i.e. radiative transition rate is zero) - Set radiative boost factor to 1
  else pTransPRBoost=1;

  if (cell->master() and cell->index()==36) cout<<"BState::boostTransPR: cell: "<<cell->index()<<", "<<iState()->name()<<":"<<getIndex()<<", pTransPRBoost: "<<pTransPRBoost<<"\n";
}

// Boost collisional transition probabilities - New!
void BState::boostTransPC() {
  int i;

  // Get cell
  Cell* cell=iState()->atom()->cell();

  // If use LTE group and not master state - Set dummy values and return - New!
  if (cell->config->LTEGroupMacro and LTERefState!=NULL) { pTransPCBoost=1.0; pTransPBoost=1.0; return; }

  // Reset collisional boost factor
  pTransPCBoost=0;

  // If collisional transition probability is non-zero
  if (PCTot!=0) {
    // Calculate collisional boost factor
    for (i=0;i<transPList.size();i++) {
      if (transPList[i].type==collisional) pTransPCBoost+=(transPList[i].P/PCTot)*cell->markovPThermBoost();
      // Testing!!!
      if (cell->master() and iState()->name().compare("AlI")==0 and cell->index()==36) cout<<"BState::boostTransPC: cell: "<<cell->index()<<", "<<iState()->name()<<":"<<getIndex()<<", transPList["<<i<<"].P: "<<transPList[i].P<<", PCTot: "<<PCTot<<",trans: "<<transPList[i].trans->upperState->iState()->name()<<":"<<transPList[i].trans->upperState->getIndex()<<"->"<<transPList[i].trans->lowerState->iState()->name()<<":"<<transPList[i].trans->lowerState->getIndex()<<", markovPThermBoost: "<<cell->markovPThermBoost()<<"\n";
    }
  }
  // Else (i.e. if collisional transition probability is zero) - Set collisional boost factor to 1
  else pTransPCBoost=1;

  if (cell->master() and cell->index()==36) cout<<"BState::boostTransPC: cell: "<<cell->index()<<", "<<iState()->name()<<":"<<getIndex()<<", pTransPCBoost: "<<pTransPCBoost<<"\n";

  // Calculate boost factor
  pTransPBoost=PRTot*transPRBoost()+PCTot*transPCBoost();

  // If transition list empty - Set boost factor to zero
  //if (transPList.size()==0) transPBoost=1;
  
  if (cell->master() and cell->index()==36) cout<<"BState::boostTransPC: cell: "<<cell->index()<<", "<<iState()->name()<<":"<<getIndex()<<", pTransPBoost: "<<pTransPBoost<<", PRTot: "<<PRTot<<", PCTot: "<<PCTot<<", pTransPRBoost: "<<pTransPCBoost<<"\n";

  // For each transition
  for (i=0;i<transPList.size();i++) {
    // Calculate adjusted transition probability
    if (transPList[i].type==collisional) transPList[i].P=transPList[i].P*cell->markovPThermBoost()/transPBoost();
    else transPList[i].P=transPList[i].P*transPList[i].trans->etaBoost()/transPBoost();
  }

  // Sort transition probability list in descending probability
  transPList.sort();

  // Calculate adjusted total transition probabilities
  PRTot=PRTot*transPRBoost()/transPBoost();
  PCTot=PCTot*transPCBoost()/transPBoost();
}

// Boost Markov transition probabilities - New!
void BState::boostMarkovP() {
  double PTot;
  MarkovPList markovPListWork;
  double markovTableLim=iState()->atom()->cell()->config->markovTableLim;
  int i;

  // Get cell
  Cell* cell=iState()->atom()->cell();

  // If use LTE group and not master state - Set dummy value and return - New!
  if (cell->config->LTEGroupMacro and LTERefState!=NULL) { pMarkovPBoost=1.0; return; }

  // Reset boost factor
  pMarkovPBoost=0;
  
  // Calculate boost factor
  for (i=0;i<markovPList.size();i++) {
    if (markovPList[i].nextState!=NULL) pMarkovPBoost+=markovPList[i].P*markovPList[i].nextState->transPBoost();
    else pMarkovPBoost+=markovPList[i].P*cell->radCoolingBoost();
    // Testing!!!
    //if (cell->master() and cell->index()==36) 
    //if (markovPList[i].nextState!=NULL) cout<<"BState::boostMarkovP: cell: "<<cell->index()<<", "<<iState()->name()<<":"<<getIndex()<<", markovPList["<<i<<"].P: "<<markovPList[i].P<<", nextState: "<<markovPList[i].nextState->iState()->name()<<":"<<markovPList[i].nextState->getIndex()<<", transPBoost: "<<markovPList[i].nextState->transPBoost()<<"\n";
    //else cout<<"BState::boostMarkovP: cell: "<<cell->index()<<", "<<iState()->name()<<":"<<getIndex()<<", markovPList["<<i<<"].P: "<<markovPList[i].P<<", radCoolingBoost: "<<cell->radCoolingBoost()<<"\n";
  }

  if (cell->master() and cell->index()==36) cout<<"BState::boostMarkovP: cell: "<<cell->index()<<", "<<iState()->name()<<":"<<getIndex()<<", pMarkovBoost: "<<pMarkovPBoost<<"\n";

  // For each Markov transition
  for (i=0;i<markovPList.size();i++) {
    // Calculate adjusted Markov transition probability
    if (markovPList[i].nextState!=NULL) markovPList[i].P=markovPList[i].P*markovPList[i].nextState->transPBoost()/markovPBoost();
    else markovPList[i].P=markovPList[i].P*cell->radCoolingBoost()/markovPBoost();
  }

  // Sort out Markov transitions above limit
  // Note, we should also correct the boost factor for this - Fix!!!
  markovPListWork=markovPList; markovPList.resize(0);
  for (i=0; i<markovPListWork.size(); i++) 
    if (markovPListWork[i].P>markovTableLim) markovPList.push_back(markovPListWork[i]);
    
  // Renormalize Markov transition probability list
  PTot=0;
  for (i=0; i<markovPList.size(); i++) PTot+=markovPList[i].P;
  for (i=0; i<markovPList.size(); i++) markovPList[i].P=markovPList[i].P/PTot;
  if (cell->master() and PTot<0.99) cout<<"BState::boostMarkovP: cell: "<<cell->index()<<", "<<iState()->name()<<":"<<getIndex()<<", Warning: Total probability of transitions sorted out: "<<(1.0-PTot)<<", "<<markovPList.size()<<" out of "<<markovPListWork.size()<<" transitions kept\n";

  // Sort markov transition probability list in descending probability
  markovPList.sort();
}

// Get radiative transition probability boost factor - New!
double BState::transPRBoost() { 
  // If LTE group and not master state - Return master state boost factor - New!
  if (iState()->atom()->cell()->config->LTEGroupMacro and LTERefState!=NULL) return LTERefState->transPRBoost();
  // Else - Return boost factor
  else return pTransPRBoost; 
}

// Get collisional transition probability boost factor - New!
double BState::transPCBoost() { 
  // If LTE group and not master state - Return master state boost factor - New!
  if (iState()->atom()->cell()->config->LTEGroupMacro and LTERefState!=NULL) return LTERefState->transPCBoost();
  // Else - Return boost factor
  else return pTransPCBoost; 
}

// Get transition probability boost factor - New!
double BState::transPBoost() { 
  // If LTE group and not master state - Return master state boost factor - New!
  if (iState()->atom()->cell()->config->LTEGroupMacro and LTERefState!=NULL) return LTERefState->transPBoost();
  // Else - Return boost factpr
  else return pTransPBoost; 
}

// Get Markov transition probability boost factor - New!
double BState::markovPBoost() { 
  // If LTE group and not master state - Return master state boost factor - New!
  if (iState()->atom()->cell()->config->LTEGroupMacro and LTERefState!=NULL) return LTERefState->markovPBoost();
  // Else - Return boost factor
  else return pMarkovPBoost; 
}

// *****************************************
// *** New functions for alfa correction ***
// *****************************************

// Testing!
bool BState::alfaSimple() const {
  return data->alfaSimple;
}

// Recombination correction flag
bool BState::alfaCorrect() const {
  return data->alfaCorrect;
}

// Recombination fraction (if any)
double BState::alfaFrac() const {
  return data->alfaFrac;
}

// Total spontaneus recombination coefficient - New!
double BState::alfaSPTotal(double T) {
  double _alfaSPTotal=noAlfa;

  // If not provided - Get temperature
  if (T==noT) T=iState()->atom()->cell()->T();

  // If recombination coefficient data available - Get it
  if (data->alfa!=NULL) _alfaSPTotal=(*data->alfa)(T);

  return _alfaSPTotal;
}

// Original spontaneous recombination coefficient - New!
double BState::alfaSPOriginal(double T) {
  double _alfaSPOriginal=0;
  int i;

  //Cell* cell=iState()->atom()->cell();
  //if (cell->index()==1) cout<<"BState::alfaOriginal: cell: "<<cell->index()<<", "<<iState()->name()<<":"<<getIndex()<<", Enter\n";

  // If not provided - Get temperature
  if (T==noT) T=iState()->atom()->cell()->T();

  // If total spontaneous recombination coeffecient not available
  if (alfaSPTotal()==noAlfa)
    // Sum original spontaneous recombination coefficient for all (upward) bound-free transitions
    // Note - The X factor below converts from (upper) bState to (upper) iState number density
    //for (i=0;i<upTransBF.size();i++) _alfaSPOriginal+=upTransBF[i]->alfaSP(T,false,true,false)*upTransBF[i]->upperState->X();
    for (i=0;i<upTransBF.size();i++) _alfaSPOriginal+=upTransBF[i]->alfaSP(T,false,true,false)*upTransBF[i]->upperState->X();
  // Else - Get total spontaneous recombination coefficient
  else _alfaSPOriginal=alfaSPTotal();

  return _alfaSPOriginal;
}

// Spontaneous recombination coefficient - New!
double BState::alfaSP(double T, bool includeOSP) {
  double _alfaSP=0;
  int i;

  //Cell* cell=iState()->atom()->cell();
  //if (cell->index()==1) cout<<"BState::alfaSP: cell: "<<cell->index()<<", "<<iState()->name()<<":"<<getIndex()<<", Enter\n";

  // If not provided - Get temperature
  if (T==noT) T=iState()->atom()->cell()->T();

  // Sum original spontaneous recombination coeffficient for all (upward) bound-free transitions
  // Note - The X factor below converts from (upper) bState to (upper) iState number density
  for (i=0;i<upTransBF.size();i++) {
    _alfaSP+=upTransBF[i]->alfaSP(T,false,includeOSP)*upTransBF[i]->upperState->X();
    //cout<<"BState::alfaSP: cell: "<<iState()->atom()->cell()->index()<<", "<<iState()->name()<<":"<<getIndex()<<"->"<<upTransBF[i]->upperState->iState()->name()<<":"<<upTransBF[i]->upperState->getIndex()<<", RD_SP: "<<upTransBF[i]->alfaSP(T,false,includeOSP)*upTransBF[i]->upperState->X()*iState()->atom()->cell()->ne()*upTransBF[i]->upperState->iState()->n()<<", alfaSP: "<<upTransBF[i]->alfaSP(T,false,includeOSP)*upTransBF[i]->upperState->X()<<"\n";
  }

  return _alfaSP;
}

// Additional (non-emissive) recombination coefficient - New!
double BState::alfaAdd() {
  double _alfaAdd=0;
  int i;

  //Cell* cell=iState()->atom()->cell();
  //if (cell->index()==1) cout<<"BState::alfaAdd: cell: "<<cell->index()<<", "<<iState()->name()<<":"<<getIndex()<<", Enter\n";

  // Sum original spontaneous recombination coeffficient for all (upward) bound-free transitions
  // Note - The X factor below converts from (upper) bState to (upper) iState number density
  for (i=0;i<upTransBF.size();i++) {
    _alfaAdd+=upTransBF[i]->alfaAdd()*upTransBF[i]->upperState->X();
    //cout<<"BState::alfaAdd: cell: "<<iState()->atom()->cell()->index()<<", "<<iState()->name()<<":"<<getIndex()<<"->"<<upTransBF[i]->upperState->iState()->name()<<":"<<upTransBF[i]->upperState->getIndex()<<", RDAdd: "<<upTransBF[i]->alfaAdd()*upTransBF[i]->upperState->X()*iState()->atom()->cell()->ne()*upTransBF[i]->upperState->iState()->n()<<", alfaAdd: "<<upTransBF[i]->alfaAdd()*upTransBF[i]->upperState->X()<<"\n";
  }

  return _alfaAdd;
}

// Update recombination correction
void BState::updateAlfaCorr(double alfaSPIState, double alfaDiIState) {
  double alfaSPIStateDist,alfaDiIStateDist,alfaSPDist,distFrac,gInvTot=0;
  int i;
  
  // If no recombination coefficients available
  if (alfaSPIState==noAlfa and alfaDiIState==noAlfa and data->alfa==NULL) {
    // Make dummy update of recombination correction (i.e. no correction)
    for (i=0;i<upTransBF.size();i++) upTransBF[i]->updateAlfaCorr(noAlfa,noAlfa,noAlfa);
    // Return
    return;
  }
    
  // Sum reciprocal statistical weight for the upper states of all (upward) radiative bound-free transitions - Changed!
  //for (i=0;i<upTransBF.size();i++) gInvTot+=1/upTransBF[i]->upperState->g();
  for (i=0;i<upTransBF.size();i++) if (upTransBF[i]->active(pRadiative)) gInvTot+=1/upTransBF[i]->upperState->g();

  //if (iState()->name().compare("FeII")==0 and getIndex()==0 and iState()->atom()->cell()->index()==22) cout<<"BState:updateAlfaCorr: cell: "<<iState()->atom()->cell()->index()<<", "<<iState()->name()<<":"<<getIndex()<<", alfaSPOriginal: "<<alfaSPOriginal()<<", alfaFrac: "<<alfaFrac()<<", alfaSPIState: "<<alfaSPIState<<", alfaDiIState: "<<alfaDiIState<<"\n";

  // For each (upward) bound-free transition
  for (i=0;i<upTransBF.size();i++) {

    // If radiative processes active - New!
    if (upTransBF[i]->active(pRadiative)) {

      // Initialise recombinations coefficients to distribute
      alfaSPIStateDist=noAlfa; alfaDiIStateDist=noAlfa; alfaSPDist=noAlfa;
      
      // Calculate fraction to distribute
      distFrac=((1/upTransBF[i]->upperState->g())/gInvTot);
      
      // If provided - Calculate (ionization state) spontaneous recombination coefficient to distribute
      // Note - The X factor below converts from (upper) iState to (upper) bState number density
      if (alfaSPIState!=noAlfa) alfaSPIStateDist=alfaSPIState*distFrac/upTransBF[i]->upperState->X();
      // If provided - Calculate (ionization state) di-electronic recombination coefficient to distribute
      if (alfaDiIState!=noAlfa) alfaDiIStateDist=alfaDiIState*distFrac/upTransBF[i]->upperState->X();
      // If available - Calculate spontaneous recombination coefficient to distribute
      if (data->alfa!=NULL) alfaSPDist=alfaSPTotal()*distFrac/upTransBF[i]->upperState->X();
      
      // Update transition recombination correction
      upTransBF[i]->updateAlfaCorr(alfaSPDist,alfaSPIStateDist,alfaDiIStateDist);
    }
    // Else - Make dummy update of recombination correction (i.e. no correction) - New! - Is this necessary?
    else upTransBF[i]->updateAlfaCorr(noAlfa,noAlfa,noAlfa);
  }
}

// *****************************************
// *****************************************
// *****************************************

// Get memory allocated for probability lists - New!
double BState::memPList() const {
  double _mem=0;
  int i,j;

  // Add memory allocated for transition probability list
  for (i=0;i<transPList.size();i++) _mem+=sizeof(TransP);

  // Add memory allocated for markov transition probability list
  for (i=0;i<markovPList.size();i++) _mem+=sizeof(MarkovP);

  return _mem;
}

// Get memory allocated - New!
double BState::mem() const {
  double _mem=sizeof(*this);
  int i,j;

  // Add memory allocated probability lists
  _mem+=memPList();

  return _mem;
}

// Constructor
IState::IState() : data(NULL), pAtom(NULL), pFreeFreeHeating(update), alfaCorrUpdate(true), pAlfaAdd(update), pCU(update), pCD(update), pCNU(update), pCTU(update), pCTD(update), pDCTdN(0), dCTdNUpdate(true), pAlfaSP(update), pAlfaSPOSP(update), pAlfaST(update), pAlfaSTOSP(update), pGamma(update), pGammaOSP(update), pROSP(update), pDROSPdN(0), dROSPdNUpdate(true) { omp_init_lock(&alfaCorrUpdateLock); omp_init_lock(&dCTdNUpdateLock); omp_init_lock(&dROSPdNUpdateLock); } // Changed!

// Copy constructor
// data is owned by the grid and needs to be updated elsewhere if necessay
// downTransBF have to be updated elsewhere
// Major edge reset and have to be updated elsewhere
IState::IState(const IState& iState) : data(iState.data), pIndex(iState.pIndex), pAtom(iState.pAtom), pX(iState.pX), pU(iState.pU), pFreeFreeHeating(iState.pFreeFreeHeating), alfaCorrUpdate(iState.alfaCorrUpdate), alfaCorrUpdateLock(iState.alfaCorrUpdateLock), pAlfaAdd(iState.pAlfaAdd), pCU(iState.pCU), pCD(iState.pCD), pCNU(iState.pCNU), pCTU(iState.pCTU), pCTD(iState.pCTD), pDCTdN(iState.pDCTdN), dCTdNUpdate(iState.dCTdNUpdate), dCTdNUpdateLock(iState.dCTdNUpdateLock), pAlfaSP(iState.pAlfaSP), pAlfaSPOSP(iState.pAlfaSPOSP), pAlfaST(iState.pAlfaST), pAlfaSTOSP(iState.pAlfaSTOSP), pGamma(iState.pGamma), pGammaOSP(iState.pGammaOSP), pROSP(iState.pROSP), pDROSPdN(iState.pDROSPdN), dROSPdNUpdate(iState.dROSPdNUpdate), dROSPdNUpdateLock(iState.dROSPdNUpdateLock), bState(iState.bState) { // Changed!
  int i,j;
  TransBB* transBB;
  TransBF* transBF;
  
  // For each bound state
  for (i=0;i<bState.size();i++) {

    // Update LTE reference state pointer - Removed!
    //if (iState.bState[i].LTERefState!=NULL) bState[i].LTERefState=&(bState[iState.bState[i].LTERefState->index]);
    //else bState[i].LTERefState=NULL;

    // For each downward bound-bound transition
    for (j=0;j<iState.bState[i].downTransBB.size();j++) {
      // Make copy of transition
      transBB = new TransBB(*(iState.bState[i].downTransBB[j]));
      // Set bound state pointers
      transBB->lowerState=&(bState[iState.bState[i].downTransBB[j]->lowerState->index]);
      transBB->upperState=&(bState[iState.bState[i].downTransBB[j]->upperState->index]);
      // Append to the transition lists
      transBB->lowerState->upTransBB.push_back(transBB);
      transBB->upperState->downTransBB.push_back(transBB);
    }
  }
}

// Assigment operator
// data is owned by the grid and needs to be updated elsewhere if necessay
IState& IState::operator=(const IState& iState) {
  int i,j;
  TransBB* transBB;
  TransBF* transBF;

  if (&iState!=this)  {
    data=iState.data;
    pIndex=iState.pIndex; // New!
    pAtom=iState.pAtom;
    pX=iState.pX;
    pU=iState.pU;
    pFreeFreeHeating=iState.pFreeFreeHeating;
    alfaCorrUpdate=iState.alfaCorrUpdate; // New!
    alfaCorrUpdateLock=iState.alfaCorrUpdateLock; // New!
    pAlfaAdd=iState.pAlfaAdd; // New!
    //pUpdateCounter=iState.pUpdateCounter; // Testing!
    pCU=iState.pCU; // New!
    pCD=iState.pCD; // New!
    pCNU=iState.pCNU; // New!
    pCTU=iState.pCTU; // New!
    pCTD=iState.pCTD; // New!
    pDCTdN=iState.pDCTdN; // New!
    dCTdNUpdate=iState.dCTdNUpdate; // New!
    dCTdNUpdateLock=iState.dCTdNUpdateLock; // New!
    pAlfaSP=iState.pAlfaSP; // New!
    pAlfaSPOSP=iState.pAlfaSPOSP; // New!
    pAlfaST=iState.pAlfaST; // New!
    pAlfaSTOSP=iState.pAlfaSTOSP; // New!
    pGamma=iState.pGamma; // New!
    pGammaOSP=iState.pGammaOSP; // New!
    pROSP=iState.pROSP; // New!
    pDROSPdN=iState.pDROSPdN; // New!
    dROSPdNUpdate=iState.dROSPdNUpdate; // New!
    dROSPdNUpdateLock=iState.dROSPdNUpdateLock; // New!
    bState=iState.bState;

    // Delete downward bound-bound transition list transitions
    for (i=0;i<bState.size();i++) {
      for (j=0;j<bState[i].downTransBB.size();j++) delete bState[i].downTransBB[j];
      // Reset the transition lists
      bState[i].downTransBB.resize(0);
      bState[i].upTransBB.resize(0);
    }

    // Delete downward bound-free transition list transitions
    for (i=0;i<bState.size();i++) {
      for (j=0;j<bState[i].downTransBF.size();j++) delete bState[i].downTransBF[j];
      // Reset the transition lists
      bState[i].downTransBF.resize(0);
      bState[i].upTransBF.resize(0);
    }

    // For each bound state
    for (i=0;i<bState.size();i++) {

      // Update LTE reference state pointer - Removed!
      //if (iState.bState[i].LTERefState!=NULL) bState[i].LTERefState=&(bState[iState.bState[i].LTERefState->index]);
      //else bState[i].LTERefState=NULL;

      // For each downward bound-bound transition
      for (j=0;j<iState.bState[i].downTransBB.size();j++) {
	// Make copy of transition
	transBB = new TransBB(*(iState.bState[i].downTransBB[j]));
	// Set bound state pointers
	transBB->lowerState=&(bState[iState.bState[i].downTransBB[j]->lowerState->index]);
	transBB->upperState=&(bState[iState.bState[i].downTransBB[j]->upperState->index]);
	// Append to the transition lists
	transBB->lowerState->upTransBB.push_back(transBB);
	transBB->upperState->downTransBB.push_back(transBB);
      }
    }
  }
}

// Destructor
// data is owned by the grid and has to deleted elsewhere if necessary
// downTransBF have to be updated elsewhere
IState::~IState() {
  int i,j;

  // Testing!
  //if (atom()!=NULL) if (atom()->cell()!=NULL) if (atom()->cell()->master()) cout<<"IState::~IState: cell: "<<atom()->cell()->index()<<", "<<name()<<"\n";

  // Delete downward bound-bound transition list transitions
  for (i=0;i<bState.size();i++)
    for (j=0;j<bState[i].downTransBB.size();j++) delete bState[i].downTransBB[j];

  // Delete downward bound-free transition list transitions
  for (i=0;i<bState.size();i++)
    for (j=0;j<bState[i].downTransBF.size();j++) delete bState[i].downTransBF[j];

  // Destroy alfa correction update lock (openmp)
  omp_destroy_lock(&alfaCorrUpdateLock);

  // Destroy charge-transfer number derivative update lock (openmp)
  omp_destroy_lock(&dCTdNUpdateLock);
}

// Get index - New!
int IState::index() const {
  return pIndex;
}

// Get atom back reference pointer
Atom* IState::atom() const {
  return pAtom;
}

// Get number fraction
double IState::X() const {
  return pX;
}

// Set number fraction
void IState::X(const double& _X) {
  pX=_X;
  return;
}

// Get number density
double IState::n() const {
  return pAtom->n()*pX;
}

// Set number denisty
void IState::n(const double& _n) {
  pX=pX*(_n/n());
  return;
}

// Get atom charge
double IState::C() const {
  return data->C;
}

// Get ionization potential from ground state
double IState::Xi0() const {
  return data->Xi0;
}

// Get partition function
double IState::U() const {
  return pU;
}

// Set partition function
void IState::U(const double& _U) {
  pU=_U;
}

// Total free-free opacity
double IState::freeFreeKappa(const double& nu) const {
  double kappa=0;
  double ne=atom()->cell()->ne();
  double T=atom()->cell()->T();
  double rho=atom()->cell()->rho();

  // If in on the spot region return infinity
  if (nu<atom()->cell()->freeFreeOnTheSpotNu()) return 1e99;

  // Calculate free-free opacity
  kappa=FreeFree::kappa(ne,n(),abs(C()),T,nu,rho);

  return kappa;
}

// Total free-free emissivity
double IState::freeFreeEta(const double& nu, double T) const {
  double eta=0;
  double ne=atom()->cell()->ne();
  if (T==noT) T=atom()->cell()->T();

  // If in on the spot region return zero
  if (nu<atom()->cell()->freeFreeOnTheSpotNu()) return 0.0;

  // Add free-free emissivity
  eta=FreeFree::eta(ne,n(),abs(C()),T,nu);

  return eta;
}

// Free-free cooling integrand 
class FreeFreeCoolingIntegrand {
public:
  const IState* iState;
  const double T;
  FreeFreeCoolingIntegrand(const IState* _iState, const double& _T) : iState(_iState), T(_T) {};
  double operator()(double nu) const { return 4*pi*iState->freeFreeEta(nu,T); }
};

// Total free-free cooling
double IState::freeFreeCooling(double T) const {
  double _C=0;
  int i,j;
  double ne=atom()->cell()->ne();
  if (T==noT) T=atom()->cell()->T();
  FreeFreeCoolingIntegrand integrand(this,T);
  // Testing
  int nIter;

  // If (limit cooling range or on-the-spot region) and radiation field updated
  //if ((atom()->cell()->config->limitFreeFreeCRange or atom()->cell()->freeFreeOnTheSpotNu()>0) and atom()->cell()->radUpdated) {
  // Testing - Exclude on-the-spot region
  if ((atom()->cell()->config->limitFreeFreeCRange or atom()->cell()->freeFreeOnTheSpotNu()>0 or atom()->cell()->onTheSpotNu().min!=nuInf) and atom()->cell()->radUpdated) {
    // Integrate free-free cooling rate
    double nuMin=MAX(atom()->cell()->nu[0].min,atom()->cell()->freeFreeOnTheSpotNu());
    //double nuMax=atom()->cell()->nu[atom()->cell()->nu.size()-1].max;
    // Testing - Exclude on-the-spot region
    double nuMax=MIN(atom()->cell()->nu[atom()->cell()->nu.size()-1].max,atom()->cell()->onTheSpotNu().min);
    _C=Integrator<FreeFreeCoolingIntegrand>::romberg(integrand,nuMin,nuMax,nIter,1e-2);
  }
  // Else - calculate total free-free cooling rate
  else _C=FreeFree::cooling(ne,n(),abs(C()),T);

  if (!isnumber(_C) or _C<0) cout<<"IState::freeFreeCooling: "<<name()<<", cell: "<<atom()->cell()->index()<<", Warning: Invalid cooling rate: "<<_C<<"\n";

  return _C;
}

// Free-free heating (per frequency and mean intensity)
double IState::freeFreeHeatingNu(const double& nu, double T) const {

  double H;
  if (T==noT) T=atom()->cell()->T();
  double ne=atom()->cell()->ne();

  H=4*pi*FreeFree::kappa(ne,n(),abs(C()),T,nu,1.0);

  if (!isnumber(H) or H<0) cout<<"IState::freeFreeHeatingNu: "<<name()<<", cell: "<<atom()->cell()->index()<<", nu: "<<nu<<", n: "<<n()<<", ne: "<<ne<<", T: "<<T<<", Warning: Invalid heating rate: "<<H<<"\n";

  return H;
}

// Free-free heating integrand
class FreeFreeHeatingIntegrand : public JIntegrand {
public:
  const IState* iState;
  const double T;
  FreeFreeHeatingIntegrand(const IState* _iState, const double& _T) : iState(_iState), T(_T) {};
  double operator()(double nu) const { return iState->freeFreeHeatingNu(nu,T); }
  double operator()(FrequencyBinIterator nu, EdgeType edgeType=noEdge) const { return iState->freeFreeHeatingNu(nuEdge(nu,edgeType),T); };
};

// Total free-free heating
//double IState::freeFreeHeating(double T) const {
double IState::freeFreeHeating(double T) {
  double H=0;

  // If update requested or temperature supplied
  if (pFreeFreeHeating==update or (T!=noT and T!=atom()->cell()->T())) {

    // Testing
    int nIter;
    double relTol=1e-2,absTol=0.0;
    //bool includeOSP=true,useIterator=true;
    // Testing - Exclude on-the-spot region
    bool includeOSP=false,useIterator=true;
    if (T==noT) T=atom()->cell()->T();
    FreeFreeHeatingIntegrand integrand(this,T);

    // Integrate free-free heating
    //if (atom()->cell()->index()==59) cout<<"IState::freeFreeHeating: cell: "<<atom()->cell()->index()<<", "<<name()<<"\n";
    H=atom()->cell()->JIntegrate(integrand,atom()->cell()->freeFreeOnTheSpotNu(),nuInf,nIter,relTol,absTol,includeOSP,useIterator);
  
    // If requested - Update free-free heating
    if (pFreeFreeHeating==update) pFreeFreeHeating=H;
  }
  // Else - Get free-free heating
  else H=pFreeFreeHeating;

  if (!isnumber(H) or H<0) cout<<"IState::freeFreeHeating: "<<name()<<", cell: "<<atom()->cell()->index()<<", Warning: Invalid heating rate: "<<H<<"\n";

  return H;
}

// Total bound-free opacity
double IState::boundFreeKappa(const double& nu, bool cutOff) const {
  double kappa=0;
  int i;

  // For each bound state
  for (i=0; i<bState.size(); i++) kappa+=bState[i].boundFreeKappa(nu,cutOff);

  return kappa;
}

// Total bound-free opacity
double IState::boundFreeKappa(FrequencyBinIterator nu, EdgeType edgeType, bool cutOff) const {
  double kappa=0;
  int i;

  // For each bound state
  for (i=0; i<bState.size(); i++) kappa+=bState[i].boundFreeKappa(nu,edgeType,cutOff);

  return kappa;
}

// Total bound-free opacity number derivative - New!
double IState::boundFreedKappadN(const double& nu) {
  double dKappadN=0;
  int i;

  for (i=0; i<bState.size(); i++) dKappadN+=bState[i].boundFreedKappadN(nu)*bState[i].X();

  return dKappadN;
}

// Total bound-free emissivity
double IState::boundFreeEta(const double& nu, bool cutOff) const {
  double eta=0;
  int i;

  // For each bound state
  for (i=0; i<bState.size(); i++) eta+=bState[i].boundFreeEta(nu,cutOff);

  return eta;
}

// Total bound-free emissivity
double IState::boundFreeEta(FrequencyBinIterator nu, EdgeType edgeType, bool cutOff) const {
  double eta=0;
  int i;

  // For each bound state
  for (i=0; i<bState.size(); i++) eta+=bState[i].boundFreeEta(nu,edgeType,cutOff);

  return eta;
}

// Total bound-free opacity number derivative - New!
double IState::boundFreedEtadN(const double& nu) {
  double dEtadN=0;
  int i;

  for (i=0; i<bState.size(); i++) dEtadN+=bState[i].boundFreedEtadN(nu)*bState[i].X();

  return dEtadN;
}

// Total bound-free cooling
double IState::boundFreeCooling(double T) const {
  double C=0;
  int i;
  if (T==noT) T=atom()->cell()->T();

  // For each bound state
  for (i=0; i<bState.size(); i++) C+=bState[i].boundFreeCooling(T);

  return C;
}

// Total (spontaneous) bound-free cooling
double IState::boundFreeCoolingSP(double T) const {
  double C=0;
  int i;
  if (T==noT) T=atom()->cell()->T();  

  // For each bound state
  for (i=0; i<bState.size(); i++) C+=bState[i].boundFreeCoolingSP(T);

  return C;
}

// Total bound-free heating
double IState::boundFreeHeating(double T) const {
  double H=0;
  int i;
  if (T==noT) T=atom()->cell()->T();
  
  // For each bound state
  for (i=0; i<bState.size(); i++) H+=bState[i].boundFreeHeating(T);
  
  return H;
}

// Total collisional cooling
double IState::collCooling(double T) const {
  double C=0;
  int i;
  if (T==noT) T=atom()->cell()->T();

  // For each bound state
  for (i=0; i<bState.size(); i++) C+=bState[i].collCooling(T);

  return C;
}

// Total collisional heating
double IState::collHeating(double T) const {
  double H=0;
  int i;
  if (T==noT) T=atom()->cell()->T();
  
  // For each bound state
  for (i=0; i<bState.size(); i++) H+=bState[i].collHeating(T);
  
  return H;
}

// Total charge-transfer cooling - New!
double IState::CTCooling(double T) const {
  double C=0;
  int i;
  if (T==noT) T=atom()->cell()->T();

  // For each bound state
  for (i=0; i<bState.size(); i++) C+=bState[i].CTCooling(T);

  return C;
}

// Total charge-transfer heating - New!
double IState::CTHeating(double T) const {
  double H=0;
  int i;
  if (T==noT) T=atom()->cell()->T();
  
  // For each bound state
  for (i=0; i<bState.size(); i++) H+=bState[i].CTHeating(T);
  
  return H;
}

// Get ionization state name
string IState::name() const {
  string atomName=pAtom->name();
  switch(int(C())) {
  case 0: return atomName+"I";
  case -1: return atomName+"II";
  case -2: return atomName+"III";
  case -3: return atomName+"IV";
  case -4: return atomName+"V";
  case -5: return atomName+"VI";
  case -6: return atomName+"VII";
  case -7: return atomName+"VIII";
  case -8: return atomName+"IX";
  case -9: return atomName+"X";
  default: return atomName+">X";
  }

  return ">X";
}

// Reset derived quantities (to be updated)
void IState::reset(int dependence) {
  int i;
  for (i=0;i<bState.size();i++) bState[i].reset(dependence);
  if (dependence&(dpT|dpN|dpNe)) pFreeFreeHeating=update;
  //if (dependence&(dpT|dpN)) alfaCorrUpdate=true; // New!
  //if (dependence&(dpT|dpN)) pAlfaAdd=update; // New!
  // Testing!
  if (dependence&dpT) alfaCorrUpdate=true; // New!
  if (dependence&dpT) pAlfaAdd=update; // New!
  if (dependence&(dpT|dpN|dpNe)) pCU=update; // New!
  if (dependence&(dpT|dpN|dpNe)) pCD=update; // New!
  if (dependence&(dpAll)) pCNU=update; // New! Hmm!?!
  if (dependence&(dpT|dpN)) pCTU=update; // New!
  if (dependence&(dpT|dpN)) pCTD=update; // New!
  if (dependence&(dpT|dpN)) dCTdNUpdate=true; // New!
  //if (dependence&(dpT)) pAlfaSP=update; // New!
  //if (dependence&(dpT)) pAlfaSPOSP=update; // New!
  // Add dependence on N - Testing!
  if (dependence&(dpT|dpN)) pAlfaSP=update; // New!
  if (dependence&(dpT|dpN)) pAlfaSPOSP=update; // New!
  // Testing!
  /*if (dependence&dpT) pAlfaSPOriginal=update;
  if (dependence&dpT) pAlfaSPModified=update;
  if (dependence&dpT) pAlfaSPOriginalOSP=update;
  if (dependence&dpT) pAlfaSPModifiedOSP=update;*/
  //if (dependence&(dpRad|dpT)) pAlfaST=update; // New!
  // Add dependence on N - Testing!
  if (dependence&(dpRad|dpT|dpN)) pAlfaST=update; // New!
  if (dependence&(dpRad|dpT|dpN|dpNe)) pAlfaSTOSP=update; // New!
  // Testing!
  /*if (dependence&(dpRad|dpT)) pAlfaSTOriginal=update;
  if (dependence&(dpRad|dpT)) pAlfaSTModified=update;
  if (dependence&(dpT|dpN|dpNe)) pAlfaSTOriginalOSP=update;
  if (dependence&(dpT|dpN|dpNe)) pAlfaSTModifiedOSP=update;*/
  // Add dependence on N - Testing!
  //if (dependence&(dpRad)) pGamma=update; // New!
  if (dependence&(dpRad|dpN)) pGamma=update; // New!
  if (dependence&(dpRad|dpT|dpN|dpNe)) pGammaOSP=update; // New!
  // Testing!
  /*if (dependence&dpRad) pGammaOriginal=update;
  if (dependence&dpRad) pGammaModified=update;
  if (dependence&(dpT|dpN|dpNe)) pGammaOriginalOSP=update;
  if (dependence&(dpT|dpN|dpNe)) pGammaModifiedOSP=update;*/
  if (dependence&(dpT|dpN|dpNe)) pROSP=update; // New!
  if (dependence&(dpT|dpN|dpNe)) dROSPdNUpdate=true; // New!
}

// Export data
void IState::exportStat(YAML::Emitter& data, const IState& iStateLTE, const double& tauLim) {
  int i;

  data<<YAML::BeginMap;

  // Export name
  data<<YAML::Key<<"name"<<YAML::Value<<name();

  // Export number fraction
  data<<YAML::Key<<"X"<<YAML::Value<<X();

  // Export departure coefficient
  // Should be in terms of number density, right?
  data<<YAML::Key<<"b"<<YAML::Value<<X()/iStateLTE.X();

  // Export bound-free heating-cooling - New!
  data<<YAML::Key<<"boundFree"<<YAML::Value<<YAML::BeginMap;
  data<<YAML::Key<<"H"<<YAML::Value<<boundFreeHeating();
  data<<YAML::Key<<"C"<<YAML::Value<<boundFreeCooling()<<YAML::EndMap;

  // Export collisional heating-cooling - New!
  data<<YAML::Key<<"coll"<<YAML::Value<<YAML::BeginMap;
  data<<YAML::Key<<"H"<<YAML::Value<<collHeating();
  data<<YAML::Key<<"C"<<YAML::Value<<collCooling()<<YAML::EndMap;
 
  // Export data for each bound state
  data<<YAML::Key<<"bState"<<YAML::Value<<YAML::BeginSeq;
  for (i=0;i<bState.size();i++) bState[i].exportStat(data,iStateLTE.bState[i],tauLim);
  data<<YAML::EndSeq;

  data<<YAML::EndMap;
}

// Load ionization state data
void IState::load(pstring& iStatePData, IStateDataList& iStateDataList, IState* iStateDown, TransBBSelector* transSelector, BStateSelector* bStateSelector, double C) {
  string value;
  IStateData* iStateData;
  IStateData* iStateDownData;
  BState selectedBState;
  TransBB* transBB;
  TransBF* transBF;
  int i,j,k;

  // If C not specified
  if (C==noC) {
    // Get C
    value=iStatePData.getData("C");
    // Error handling
    if (value.size()==0) {cout<<"IState::load: Error: Could not get C\n"; return;}
    //cout<<"C="<<value<<"\n";
    C=atof(value.c_str());
  }
  // Set ionization state data pointer
  iStateData=iStateDataList.find(C);
  // Error handling
  if (iStateData==NULL)  {cout<<"IState::load: Error: Could not find ionization state with C="<<C<<"\n"; return;}
  data=iStateData;
  // Get next lower ionization state data pointer
  iStateDownData=iStateDataList.find(C+1);

  // For each bound state
  for (i=0;i<data->bState.size();i++) {
    // Select bound state
    if (bStateSelector->use(data->bState[i])) {
      // Set bound state data pointer
      selectedBState.data=&(data->bState[i]);
      // Insert bound state in list
      bState.push_back(selectedBState);
      // Set bound state index
      bState[bState.size()-1].index=bState.size()-1;
    }
  }
  // For each (selected) bound state
  for (i=0;i<bState.size();i++) {
    // If LTE reference state pointer set - Removed!
    /*if (bState[i].data->LTERefState!=NULL) {
      // For each bound state
      for (k=0;k<bState.size();k++)
	// If match set pointer
	if (bState[k].data==bState[i].data->LTERefState) bState[i].LTERefState=&(bState[k]);
    }
    // Else - Set pointer to NULL
    else bState[i].LTERefState=NULL;*/
    // For each downward bound-bound transition
    for (j=0;j<bState[i].data->downTransBB.size();j++) {
      // Select transition
      if (transSelector->use(bState[i].data->downTransBB[j])) {
	// Create transition object
	transBB = new TransBB;
	// Set transition data pointer
	transBB->data=bState[i].data->downTransBB[j];
	// Get processes flags - New!
	transBB->getProcesses();
	// Set upper state pointer
	transBB->upperState=&(bState[i]);
	// For each bound state
        //pragma omp parallel for firstprivate(transBB)
	for (k=0;k<bState.size();k++)
	  // If match set lower state pointer
	  if (bState[k].data==bState[i].data->downTransBB[j]->lowerState) transBB->lowerState=&(bState[k]);
	// Insert (pointer to) transition in upper state downward bound-bound transition list
	transBB->upperState->downTransBB.push_back(transBB);
	// Insert (pointer to) transition in lower state upward bound-bound transition list
	transBB->lowerState->upTransBB.push_back(transBB);
      }
    }
    if (bState[i].data->downTransBF.size()>0 and (iStateDownData==NULL or iStateDown==NULL)) {cout<<"IState::load: Error: Downward bound-free transitions to unkown ionization state with C="<<C<<"\n"; return;}
    // For each downward bound-free transition
    for (j=0;j<bState[i].data->downTransBF.size();j++) {
      // Create transition object
      transBF = new TransBF;
      // Set transition data pointer
      transBF->data=bState[i].data->downTransBF[j];
      // Get processes flags - New!
      transBF->getProcesses();
      // Set upper state pointer
      transBF->upperState=&(bState[i]);
      // For each bound state
      //pragma omp parallel for firstprivate(transBF)
      for (k=0;k<iStateDown->bState.size();k++) {
	// If match set lower state pointer
	if (&(iStateDownData->bState[k])==bState[i].data->downTransBF[j]->lowerState) transBF->lowerState=&(iStateDown->bState[k]);
      }
      // Insert (pointer to) transition in upper state downward bound-free transition list
      transBF->upperState->downTransBF.push_back(transBF);
      // Insert (pointer to) transition in lower state upward bound-free transition list
      transBF->lowerState->upTransBF.push_back(transBF);
    }
  }
  
  return;
}

// Construct ionization state
void IState::construct(double _C, IStateDataList& iStateDataList, TransBBSelector* transSelector, BStateSelector* bStateSelector) {
  BState selectedBState;
  TransBB* transBB;
  int i,j,k;

  // Set ionization state data pointer
  data=iStateDataList.find(_C);
  // Error handling
  if (data==NULL)  {cout<<"IState::load: Error: Could not find ionization state with C="<<_C<<"\n"; exit(-1);}

  // For each bound state
  for (i=0;i<data->bState.size();i++) {
    // Select bound state
    if (bStateSelector->use(data->bState[i])) {
      // Set bound state data pointer
      selectedBState.data=&(data->bState[i]);
      // Insert bound state in list
      bState.push_back(selectedBState);
      // Set bound state index
      bState[bState.size()-1].index=bState.size()-1;
    }
  }
  // For each (selected) bound state
  for (i=0;i<bState.size();i++) {
    // If LTE reference state pointer set - Removed!
    /*if (bState[i].data->LTERefState!=NULL) {
      // For each bound state
      for (k=0;k<bState.size();k++)
	// If match set pointer
	if (bState[k].data==bState[i].data->LTERefState) bState[i].LTERefState=&(bState[k]);
    }
    // Else - Set pointer to NULL
    else bState[i].LTERefState=NULL;*/
    // For each downward bound-bound transition
    for (j=0;j<bState[i].data->downTransBB.size();j++) {
      // Select transition
      if (transSelector->use(bState[i].data->downTransBB[j])) {
	// Create transition object
	transBB = new TransBB;
	// Set transition data pointer
	transBB->data=bState[i].data->downTransBB[j];
	// Get processes flags - New!
	transBB->getProcesses();
	// Set upper state pointer
	transBB->upperState=&(bState[i]);
	// For each bound state
	for (k=0;k<bState.size();k++)
	  // If match set lower state pointer
	  if (bState[k].data==bState[i].data->downTransBB[j]->lowerState) transBB->lowerState=&(bState[k]);
	// Insert (pointer to) transition in upper state downward bound-bound transition list
	transBB->upperState->downTransBB.push_back(transBB);
	// Insert (pointer to) transition in lower state upward bound-bound transition list
	transBB->lowerState->upTransBB.push_back(transBB);
      }
    }
  }
  
  return;
}

// Reload ionization state bound-bound transitions
void IState::reloadTransBB(TransBBSelector* transSelector) {
  TransBB* transBB;
  int i,j,k;

  // Reset bound-bound transition lists here !!!

  // For each bound state
  for (i=0;i<bState.size();i++) {
    // For each downward bound-bound transition
    for (j=0;j<bState[i].data->downTransBB.size();j++) {
      // Select transition
      if (transSelector->use(bState[i].data->downTransBB[j])) {
	// Create transition object
	transBB = new TransBB;
	// Set transition data pointer
	transBB->data=bState[i].data->downTransBB[j];
	// Get processes flags - New!
	transBB->getProcesses();
	// Set upper state pointer
	transBB->upperState=&(bState[i]);
	// For each bound state
        //pragma omp parallel for firstprivate(transBB)
	for (k=0;k<bState.size();k++)
	  // If match set lower state pointer
	  if (bState[k].data==bState[i].data->downTransBB[j]->lowerState) transBB->lowerState=&(bState[k]);
	// Insert (pointer to) transition in upper state downward bound-bound transition list
	transBB->upperState->downTransBB.push_back(transBB);
	// Insert (pointer to) transition in lower state upward bound-bound transition list
	transBB->lowerState->upTransBB.push_back(transBB);
      }
    }
  }
  
  return;
}

// Load ionization state data - Not up to date!
void IState::loadYAML(const YAML::Node& iStateData, IStateDataList& iStateDataList, TransBBSelector* transSelector, BStateSelector* bStateSelector) {
  double _C;
  int i;

  // Get C
  if (iStateData["C"]) _C=iStateData["C"].as<double>();
  else {cout<<"IState::load: Error: Could not get C\n"; exit(-1);}

  // Construct ionization state
  construct(_C,iStateDataList,transSelector,bStateSelector);

  // Get number fraction (optional)
  if (iStateData["X"]) pX=iStateData["X"].as<double>();

  // Load each bound state (optional)
  //if (atomData["bState"]) for (i=0;i<bState.size();i++) bState[i].load(iStateData["bState"]);
  
  return;
}

// Load ionization state
void IState::loadState(const YAML::Node& data, bool loadBState) {
  int i;

  // Get number fraction
  if (data["X"]) pX=data["X"].as<double>();
  else {cout<<"IState::loadState: Error: Could not get X\n"; exit(-1);}

  // If requested
  if (loadBState)
    // Load each bound state
    if (data["bState"])
      if (bState.size()==data["bState"].size()) for (i=0;i<bState.size();i++) bState[i].loadState(data["bState"][i]);
      else {
	cout<<"IState::loadState: Warning: Wrong number of bound states, assuming ion to be in the ground state\n";
	for (i=0;i<bState.size();i++) bState[i].X((i==0)?1.0:1e-15);
      }
    else {cout<<"IState::loadState: Error: Could not get bound states\n"; exit(-1);}

  return;
}

// Save ionization state data
void IState::save(YAML::Emitter& data) const {
  int i;

  // Begin ionization state map
  data<<YAML::BeginMap;

  // Save C
  data<<YAML::Key<<"C"<<YAML::Value<<C();

  // Save number fraction
  data<<YAML::Key<<"X"<<YAML::Value<<X();

  // Begin bound state sequence
  data<<YAML::Key<<"bState"<<YAML::Value<<YAML::BeginSeq;

  // Save data for each bound state
  for (i=0;i<bState.size();i++) bState[i].save(data);

  // End bound state sequence
  data<<YAML::EndSeq;

  // End ionization state map
  data<<YAML::EndMap;
}

// Save ionization state
void IState::saveState(YAML::Emitter& data) {
  int i;

  // Begin ionization state map
  data<<YAML::BeginMap;

  // Save number fraction
  data<<YAML::Key<<"X"<<YAML::Value<<X();

  // Save data for each bound state
  data<<YAML::Key<<"bState"<<YAML::Value<<YAML::BeginSeq;
  for (i=0;i<bState.size();i++) bState[i].saveState(data);
  data<<YAML::EndSeq;

  // Close ionization state map
  data<<YAML::EndMap;
}

// Set back references
void IState::setBackReference(Atom* atom) {
  int i;

  // Set atom back reference pointer
  pAtom=atom;
  // For each bound state set back references
  for (i=0;i<bState.size();i++) bState[i].setBackReference(this);
}

// Find ionization state
IState* IStateList::find(const int& C) {
  int i;
  for (i=0;i<(*this).size();i++) {
    if ((*this)[i].C()==C) return &((*this)[i]);
  }
  return NULL;
}

// *****************************************
// *** New functions for alfa correction ***
// *****************************************

// Get recombination distribution method - New!
AlfaDistMethod IState::alfaDistMethod() const {
  return data->alfaDistMethod;
}

// Force recombination coefficient match - New!
bool IState::alfaForceEqual() const {
  return data->alfaForceEqual;
}

// Total spontaneus recombination coefficient - New!
double IState::alfaSPTotal(double T) {
  double _alfaSPTotal=noAlfa;

  // If not provided - Get temperature
  if (T==noT) T=atom()->cell()->T();

  // If recombination coefficient data available - Get it
  if (data->alfa!=NULL) _alfaSPTotal=(*data->alfa)(T);

  return _alfaSPTotal;
}

// Di-electronic recombination coefficient - New!
double IState::alfaDi(double T) {
  double _alfaDi=noAlfa;

  // If not provided - Get temperature
  if (T==noT) T=atom()->cell()->T();

  // If di-electronic recombination coefficient data available - Get it
  if (data->alfaDi!=NULL) _alfaDi=(*data->alfaDi)(T);  

  return _alfaDi;
}

// Original spontaneous recombination coefficient - New!
double IState::alfaSPOriginal(double T) {
  double _alfaSPOriginal=0;
  int i;

  //Cell* cell=atom()->cell();
  //if (cell->index()==1) cout<<"IState::alfaOriginal: cell: "<<cell->index()<<", "<<name()<<", Enter\n";

  // If not provided - Get temperature
  if (T==noT) T=atom()->cell()->T();

  // Sum original recombination coefficient for all bound states
  for (i=0; i<bState.size(); i++) _alfaSPOriginal+=bState[i].alfaSPOriginal(T);

  return _alfaSPOriginal;
}

// Additional (non-emissive) recombination coefficient - New!
double IState::alfaAdd() {
  int i;

  // If requested
  if (pAlfaAdd==update) {

    // Reset additional (non-emissive) recombination coefficient
    pAlfaAdd=0;

    // Sum additional (non-emissive) recombination coefficient for all bound states
    for (i=0; i<bState.size(); i++) pAlfaAdd+=bState[i].alfaAdd();
  }

  //cout<<"IState::alfaAdd: cell: "<<atom()->cell()->index()<<", "<<name()<<", alfaAdd: "<<pAlfaAdd<<"\n";

  return pAlfaAdd;
}

//int IState::updateCounter() { return pUpdateCounter; }

// Update recombination correction - New!
void IState::updateAlfaCorr() {
  vector<double> _alfaSPDist(bState.size(),noAlfa),_alfaDiDist(bState.size(),noAlfa);
  int i;

  // If update requested
  if (alfaCorrUpdate) {

    // Set update lock (to avoid race conditions)
    omp_set_lock(&alfaCorrUpdateLock);    

    // Testing!!!
    //cout<<"IState::updateAlfaCorr: cell: "<<atom()->cell()->index()<<", "<<name()<<", Locked\n";

    // If update requested
    if (alfaCorrUpdate) {
      
      // If recombination correction
      if (atom()->cell()->config->alfaCorrect) {
	
	// If total recombination coefficient available - Get distributed coefficient
	if (data->alfa!=NULL) _alfaSPDist=alfaSPDist();
	
	// If di-electronic recombination coefficient available - Get distributed coefficient
	if (data->alfaDi!=NULL) _alfaDiDist=alfaDiDist();      
      }
      
      // Update recombination correction for each bound state
      for (i=0;i<bState.size();i++) bState[i].updateAlfaCorr(_alfaSPDist[i],_alfaDiDist[i]);
      
      // Mark as updated
      alfaCorrUpdate=false;
    }

    // Testing!!!
    //cout<<"IState::updateAlfaCorr: cell: "<<atom()->cell()->index()<<", "<<name()<<", Unlock\n";

    // Unset update lock
    omp_unset_lock(&alfaCorrUpdateLock);
  }
}

// Get distributed total recombination coeffiecient - New!
vector<double> IState::alfaSPDist() {
  double alfaSPDistFrac,alfaSPDistG,scaleFactor,alfaSPKeep=0,alfaSPExclude=0,g=0,alfaFrac=0;
  vector<double> _alfaSPDist(bState.size(),noAlfa);
  int i,nDist=0;

  // For each bound state
  for (i=0;i<bState.size();i++) {
    // If alfa correction
    if (bState[i].alfaCorrect()) {
      // If total alfa fraction specified
      if (bState[i].alfaFrac()!=-1) {
	// Sum original alfa to exclude
	alfaSPExclude+=bState[i].alfaSPOriginal();
	// Sum total alfa fraction to distribute
	alfaFrac+=bState[i].alfaFrac();
      }
      // Else
      else {
	// If addition distribution method - Sum original alfa too keep
	if (alfaDistMethod()==gAdd) alfaSPKeep+=bState[i].alfaSPOriginal(); 
	// Sum statistical weight
	g+=bState[i].g();
	// Sum (bound state) distribution counter
	nDist++;
      }
    }
    // Else sum original alfa to keep
    else alfaSPKeep+=bState[i].alfaSPOriginal(); 
  }

  // Calculate total alfa to distribute based on specified fractions
  alfaSPDistFrac=alfaSPTotal()*alfaFrac;

  // For each bound state
  for (i=0;i<bState.size();i++) 
    // If specified - Calculate new (corrected) alfa based on specified fraction
    if (bState[i].alfaFrac()!=-1) _alfaSPDist[i]=alfaSPDistFrac*(bState[i].alfaFrac()/alfaFrac);

  //if (name().compare("FeII")==0 and atom()->cell()->index()==22) cout<<"IState::updateAlfaCorr: cell: "<<atom()->cell()->index()<<", "<<name()<<", alfaSPOriginal: "<<alfaSPOriginal()<<", alfaSPTotal: "<<alfaSPTotal()<<", alfaSPKeep: "<<alfaSPKeep<<", alfaSPDistFrac: "<<alfaSPDistFrac<<", alfaSPDistG: "<<alfaSPDistG<<"\n";
  
  // If total coefficient not yet distributed exceeds the original one kept and any states to distribue to
  if ((alfaSPTotal()-alfaSPDistFrac)>alfaSPKeep and nDist>0) {
    
    // Calculate total alfa to distribute based on statistical weight
    alfaSPDistG=(alfaSPTotal()-alfaSPDistFrac)-alfaSPKeep;
		
    // For each bound state
    for (i=0;i<bState.size();i++) {
      // If alfa correction
      if (bState[i].alfaCorrect()) {
	// If total alfa fraction not specified
	if (bState[i].alfaFrac()==-1) {
	  // Calculate new (corrected) alfa based on statistical weight and distribution method
	  if (alfaDistMethod()==gReplace) _alfaSPDist[i]=alfaSPDistG*(bState[i].g()/g);
	  else _alfaSPDist[i]=bState[i].alfaSPOriginal()+alfaSPDistG*(bState[i].g()/g);
	}
      }
    }
  }
  // Else if force recombination coefficient match
  else if (alfaForceEqual()) {

    // Calculate scale factor
    scaleFactor=(alfaSPTotal()-alfaSPDistFrac)/(alfaSPOriginal()-alfaSPExclude);
    // For each bound state
    for (i=0;i<bState.size();i++) 
      // If total alfa fraction not specified - Calculate new (corrected) alfa by scaling the original one
      if (bState[i].alfaFrac()==-1) _alfaSPDist[i]=scaleFactor*bState[i].alfaSPOriginal();
  }

  return _alfaSPDist;
}

// Get distributed di-electronic recombination coeffiecient - New!
vector<double> IState::alfaDiDist() {
  double alfaSPKeep,alfaDiDistG,g=0;
  vector<double> _alfaDiDist(bState.size(),noAlfa);
  int i;

  // For each bound state
  for (i=0;i<bState.size();i++) {
    // If alfa correction and total alfa fraction not specified - Sum statistical weight
    if (bState[i].alfaCorrect() and bState[i].alfaFrac()==-1) g+=bState[i].g();
  }
  
  // Calculate di-electronic alfa to distribute
  alfaDiDistG=alfaDi();
  
  // For each bound state
  for (i=0;i<bState.size();i++) {
    // If alfa correction and total alfa fraction not specified - Calculate di-electronic alfa based on the statistical weight
    if (bState[i].alfaCorrect() and bState[i].alfaFrac()==-1) _alfaDiDist[i]=alfaDiDistG*(bState[i].g()/g);
  }

  return _alfaDiDist;
}

// *****************************************
// *** New functions for bound-free rates **
// *****************************************

// Upward collisional rate - New!
double IState::CU(double T) {
  double _CU=0;
  Cell* cell=atom()->cell();
  int i,j;

  // If update requested or temperature supplied
  if (pCU==update or (T!=noT and T!=cell->T())) {

    // Sum upward collisional rate
    for (i=0;i<bState.size();i++)
      for (j=0;j<bState[i].upTransBF.size();j++) _CU+=bState[i].upTransBF[j]->CU(T);

    // If requested - Update
    if (pCU==update) pCU=_CU;
  }
  // Else - Get upward collisional rate
  else _CU=pCU;

  return _CU;
}

// Downward collisional rate - New!
double IState::CD(double T) {
  double _CD=0;
  Cell* cell=atom()->cell();
  int i,j;

  // If update requested or temperature supplied
  if (pCD==update or (T!=noT and T!=cell->T())) {
      
    // Sum downward collisional rate
    for (i=0;i<bState.size();i++)
      for (j=0;j<bState[i].downTransBF.size();j++) _CD+=bState[i].downTransBF[j]->CD(T);

    // If requested - Update
    if (pCD==update) pCD=_CD;
  }
  // Else - Get upward collisional rate
  else _CD=pCD;

  return _CD;
}

// Upward non-thermal rate - New!
double IState::CNU() {
  int i,j;

  // If update requested
  if (pCNU==update) {

    // Reset upward non-thermal rate
    pCNU=0;

    // Sum upward non-thermal rate
    for (i=0;i<bState.size();i++)
      for (j=0;j<bState[i].upTransBF.size();j++) pCNU+=bState[i].upTransBF[j]->CNU();
  }

  return pCNU;
}

// Upward charge-transfer rate - New!
double IState::CTU() {
  int i,j;

  // If update requested
  if (pCTU==update) {
      
    // Reset upward charge-transfer rate
    pCTU=0;

    // Sum upward charge-transfer rate
    for (i=0;i<bState.size();i++)
      for (j=0;j<bState[i].upTransBF.size();j++) pCTU+=bState[i].upTransBF[j]->CTU();
  }

  return pCTU;
}

// Downward charge-transfer rate - New!
double IState::CTD() {
  int i,j;

  // If update requested
  if (pCTD==update) {
    
    // Reset downward charge-transfer rate
    pCTD=0;

    // Sum downward charge-transfer rate
    for (i=0;i<bState.size();i++)
      for (j=0;j<bState[i].downTransBF.size();j++) pCTD+=bState[i].downTransBF[j]->CTD();
  }

  return pCTD;
}

// Collisional rate (out of the state) - New!
double IState::CC(double T) {
  double _CC=0;
  IState *lowerState,*upperState;
  Cell* cell=atom()->cell();

  // Get lower and upper state pointers
  // Note: C reversed! Right?
  lowerState=atom()->iState.find(C()+1);
  upperState=atom()->iState.find(C()-1);

  // Add collisional rate out of the state
  _CC+=(CU(T)+CD(T));
  // Subtract collisional rate from the lower state
  if (lowerState!=NULL) _CC-=lowerState->CU(T);
  // Subtract collisional rate from the upper state
  if (upperState!=NULL) _CC-=upperState->CD(T);

  return _CC;
}

// Non-thermal rate (out of the state) - New!
double IState::CN() {
  double _CN=0;
  IState* lowerState;
  Cell* cell=atom()->cell();

  // Get lower state pointer
  // Note: C reversed! Right?
  lowerState=atom()->iState.find(C()+1);

  // Add non-thermal rate out of the state
  _CN+=CNU();
  // Subtract non-thermal rate from the lower state
  if (lowerState!=NULL) _CN-=lowerState->CNU();

  return _CN;
}

// Charge-transfer rate (out of the state) - New!
double IState::CT() {
  double _CT=0;
  IState *lowerState,*upperState;
  Cell* cell=atom()->cell();

  // Get lower and upper state pointers
  // Note: C reversed! Right?
  lowerState=atom()->iState.find(C()+1);
  upperState=atom()->iState.find(C()-1);

  // Add charge-transfer rate out of the state
  _CT+=(CTU()+CTD());
  // Subtract charge-transfer rate from lower state
  if (lowerState!=NULL) _CT-=lowerState->CTU();
  // Subtract charge-transfer rate from upper state
  if (upperState!=NULL) _CT-=upperState->CTD();

  return _CT;
}

// Collisional rate (out of the state) number derivative - New!
double IState::dCCdN(IState* iState, double T, bool dNedN) {
  double _CC=0,CCUpper=0,CCLower=0,_dCCdN=0;
  IState *lowerState,*upperState;
  Cell* cell=atom()->cell();

  // Get lower and upper state pointers
  // Note: C reversed! Right?
  lowerState=atom()->iState.find(C()+1);
  upperState=atom()->iState.find(C()-1);

  // Get collisional rate out of the state
  _CC=CU(T)+CD(T);
  // Get collisional rate from lower state
  if (lowerState!=NULL) CCLower=-lowerState->CU(T);
  // Get collisional rate from upper state
  if (upperState!=NULL) CCUpper=-upperState->CD(T);

  // Get number derivative with respect to the specified state
  if (iState==this) _dCCdN=_CC/n();
  else if (iState==upperState and upperState!=NULL) _dCCdN=CCUpper/upperState->n();
  else if (iState==lowerState and lowerState!=NULL) _dCCdN=CCLower/lowerState->n();

  // If requested- Add dNedN term
  //if (dNedN) _dCCdN+=(_CC+CCUpper+CCLower)*(cell->dNedN(iState)/cell->ne());
  // Changed - Downward collisional rates quadratic in ne, right?
  if (dNedN) _dCCdN+=((CU(T)+CCLower)+2*(CD(T)+CCUpper))*(cell->dNedN(iState)/cell->ne());

  return _dCCdN;
}

// Non-thermal rate (out of the state) number derivative - New!
double IState::dCNdN(IState* iState) {
  double _CN=0,CNLower=0,_dCNdN=0;
  IState* lowerState;
  Cell* cell=atom()->cell();

  // Get lower state pointer
  // Note: C reversed! Right?
  lowerState=atom()->iState.find(C()+1);

  // Get non-thermal rate out of the state
  _CN=CNU();
  // Get non-thermal rate from lower state
  if (lowerState!=NULL) CNLower=-lowerState->CNU();

  // Get number derivative with respect to the specified state
  if (iState==this) _dCNdN=_CN/n();
  else if (iState==lowerState and lowerState!=NULL) _dCNdN=CNLower/lowerState->n();

  return _dCNdN;
}

// Charge-transfer rate (out of the state) number derivative - New!
double IState::dCTdN(IState* iState) {

  // If update requested
  if (dCTdNUpdate) {

    // Set update lock (to avoid race conditions)
    omp_set_lock(&dCTdNUpdateLock);

    // Testing!!!
    cout<<"IState::dCTdN: cell: "<<atom()->cell()->index()<<", "<<name()<<", Locked\n";

    // If requested - Update
    if (dCTdNUpdate) { updateDCTdN(); dCTdNUpdate=false; }

    // Testing!!!
    cout<<"IState::dCTdN: cell: "<<atom()->cell()->index()<<", "<<name()<<", Unlock\n";

    // Unset update lock
    omp_unset_lock(&dCTdNUpdateLock);
  }

  // Return charge-transfer rate (out of the state) number derivative
  return pDCTdN[iState->atom()->index()][iState->index()];
}

// Update charge-transfer rate (out of the state) number derivative - New!
void IState::updateDCTdN() {
  TransBF *upTrans,*downTrans;
  IState *lowerState,*upperState;
  Cell* cell=atom()->cell();
  int i,j,k;

  // Get lower and upper state pointers
  // Note: C reversed! Right?
  lowerState=atom()->iState.find(C()+1);
  upperState=atom()->iState.find(C()-1);
  
  // Reset DCTdN vector
  pDCTdN.resize(0);
  pDCTdN.resize(cell->atom.size());
  for (i=0;i<cell->atom.size();i++) pDCTdN[i].resize(cell->atom[i].iState.size(),0);

  // Get number derivative of charge-transfer rate out of the state
  pDCTdN[atom()->index()][index()]=(CTU()+CTD())/n();
  // Get number derivative of charge-transfer rate from the lower state
  if (lowerState!=NULL) pDCTdN[atom()->index()][lowerState->index()]=-lowerState->CTU()/lowerState->n();
  // Get number derivative of charge transfer rate from the upper state
  if (upperState!=NULL) pDCTdN[atom()->index()][upperState->index()]=-upperState->CTD()/upperState->n();

  // For each bound state
  for (i=0;i<bState.size();i++) {

    // For each upward bound-free transition
    for (j=0;j<bState[i].upTransBF.size();j++) {
      // Set pointer to upward bound-free transition
      upTrans=bState[i].upTransBF[j];

      // For each upward CT transition
      for (k=0;k<upTrans->CTUReaction.size();k++) {
	// Set pointer to upper state of corresponding downward CT transition
	upperState=upTrans->CTUReaction[k]->downTrans->upperState->iState();
	// Add contribution to number derivative with respect to this state
	pDCTdN[upperState->atom()->index()][upperState->index()]+=upTrans->CTUReaction[k]->rate()/upperState->n();
	// Needs update for reaction direction!!!
      }

      // For each downward CT reaction
      for (k=0;k<upTrans->CTDReaction.size();k++) {
	// Set pointer to lower state of corresponding upward bound-free transition
	lowerState=upTrans->CTDReaction[k]->upTrans->lowerState->iState();
	// Add contribution to number derivative with respect to this state
	pDCTdN[lowerState->atom()->index()][lowerState->index()]-=upTrans->CTDReaction[k]->rate()/lowerState->n();
	// Needs update for reaction direction!!!
      }
    }

    // For each downward bound-free transition
    for (j=0;j<bState[i].downTransBF.size();j++) {
      // Set pointer to downward bound-free transition
      downTrans=bState[i].downTransBF[j];

      // For each downward CT reaction
      for (k=0;k<downTrans->CTDReaction.size();k++) {
	// Set pointer to lower state of corresponding upward bound-free transition
	lowerState=downTrans->CTDReaction[k]->upTrans->lowerState->iState();
	// Add contribution to number derivative with respect to this state
	pDCTdN[lowerState->atom()->index()][lowerState->index()]+=downTrans->CTDReaction[k]->rate()/lowerState->n();
	// Needs update for reaction direction!!!
      }

      // For each upward CT reaction
      for (k=0;k<downTrans->CTUReaction.size();k++) {
	// Set pointer to upper state of corresponding downward bound-free transition
	upperState=downTrans->CTUReaction[k]->downTrans->upperState->iState();
	// Add contribution to number derivative with respect to this state
	pDCTdN[upperState->atom()->index()][upperState->index()]-=downTrans->CTUReaction[k]->rate()/upperState->n();
	// Needs update for reeaction direction!!!
      }
    }
  }
}

// Spontaneous recombination coefficient - New!
double IState::alfaSP(double T, bool includeOSP) {
  double _alfaSP=0;
  Cell* cell=atom()->cell();
  int i;

  //if (atom()->cell()->index()==36) cout<<"IState::alfaSP: cell: "<<atom()->cell()->index()<<", "<<name()<<", includeOSP: "<<includeOSP<<", Enter\n";

  // If update requested or temperature supplied
  if ((pAlfaSP==update and (not includeOSP)) or (pAlfaSPOSP==update and includeOSP) or (T!=noT and T!=cell->T())) {

    // If not supplied - Get temperature
    if (T==noT) T=atom()->cell()->T();

    // Sum recombination coefficient for all bound states
    for (i=0; i<bState.size(); i++) _alfaSP+=bState[i].alfaSP(T,includeOSP);

    // Sum recombination coefficient for all bound states - Testing! - Does not work! - Remove!
    //for (i=0; i<majorEdge.size(); i++) _alfaSP+=majorEdge[i]->alfaSP(T,false,includeOSP)*majorEdge[i]->upperState->X();

    // If requested - Update
    if (pAlfaSPOSP==update and includeOSP) pAlfaSPOSP=_alfaSP;
    else if (pAlfaSP==update and (not includeOSP)) pAlfaSP=_alfaSP;
  }
  // Else - Get spontaneous recombination coefficient
  else if (includeOSP) _alfaSP=pAlfaSPOSP;
  else _alfaSP=pAlfaSP;

  //if (atom()->cell()->index()==36) cout<<"IState::alfaSP: cell: "<<atom()->cell()->index()<<", "<<name()<<", Return\n";

  //cout<<"IState::alfaSP: cell: "<<atom()->cell()->index()<<", "<<name()<<", alfaSP: "<<_alfaSP<<"\n";

  return _alfaSP;
}

// Get (spontaneous) recombination rate coefficient - New! - Testing!
double IState::alfaSPNu(const double& _nu, double T, bool modified) {
  double _alfaSPNu=0;
  int i,j;

  // Sum stimulated recombination rate coefficient
  for (i=0;i<bState.size();i++) {
    for (j=0;j<bState[i].upTransBF.size();j++) {
      if (bState[i].upTransBF[j]->alfaSimple()) continue;
      if (atom()->onTheSpot() and bState[i].E()==0) continue;
      _alfaSPNu+=bState[i].upTransBF[j]->alfaSPNu(_nu,T,modified)*bState[i].upTransBF[j]->upperState->X();
    }
  }

  return _alfaSPNu;
}

// Get (spontaneous) recombination rate coefficient - New! - Testing!
double IState::alfaSPNu(FrequencyBinIterator _nu, EdgeType edgeType, double T, bool modified) {
  double _alfaSPNu=0;
  int i,j;

  // Sum stimulated recombination rate coefficient
  for (i=0;i<bState.size();i++) {
    for (j=0;j<bState[i].upTransBF.size();j++) {
      if (bState[i].upTransBF[j]->alfaSimple()) continue;
      if (atom()->onTheSpot() and bState[i].E()==0) continue;
      _alfaSPNu+=bState[i].upTransBF[j]->alfaSPNu(_nu,edgeType,T,modified)*bState[i].upTransBF[j]->upperState->X();
    }
  }

  return _alfaSPNu;
}

// (Spontaneous) recombination rate coefficient integrand - New! - Testing! - Removed!
/*class AlfaSPIntegrandIState {
public:
  IState* iState;
  const double T;
  const bool modified;
  AlfaSPIntegrand(IState* _iState, const double& _T, const bool& _modified) : iState(_iState), T(_T), modified(_modified) {}; // Changed!
  double operator()(double nu, EdgeType edgeType=noEdge) const { return iState->alfaSPNu(nuEdge(nu,edgeType),T,modified); }
  double operator()(FrequencyBinIterator nu, EdgeType edgeType=noEdge) const { return iState->alfaSPNu(nu,edgeType,T,modified); }
};

// Get (spontaneous) recombination rate coefficient - New! - Testing! - Removed!
// Note: The meaning of the OSP suffix is OSP included
double IState::alfaSP(double T, bool modified, bool includeOSP) {
  double pAlfaSP=0;

  // Set cell pointer
  Cell* cell=upperState->iState()->atom()->cell();

  // If update requested or temperature supplied
  if ((pAlfaSPOriginalOSP==update and (not modified) and includeOSP) or (pAlfaSPModifiedOSP==update and modified and includeOSP) or (pAlfaSPOriginal==update and (not modified) and (not includeOSP)) or (pAlfaSPModified==update and modified and (not includeOSP)) or (T!=noT and T!=cell->T())) {
  
    double relTol=1e-2,absTol=0;
    int nIterMin=1,nIterMax=4,nBinMax=16,nIter=0;
    if (T==noT) T=cell->T();
    bool useIterator=(T==cell->T())?true:false;
    AlfaSPIntegrand integrand(this,T,modified);

    // Set frequency bounds
    double nuMin=nuInf,nuMax=nuInf,nuMinOSP=cell->onTheSpotNu().min;
    for (i=0;i<bState.size();i++) for (j=0;j<bState[i].upTransBF.size();j++) if (bState[i].upTransBF[j]->nu()<nuMin) nuMin=bState[i].upTransBF[j]->nu();
    if (not includeOSP) nuMax=MIN(nuMax,nuMinOSP);
    
    // If any frequency range left to integrate over
    if (nuMax>nuMin) {
      // Calculate (spontaneous) recombination rate coefficient
      if (cell->nuOld.size()==0) pAlfaSP=Integrator<AlfaSPIntegrandIState>::romberg(integrand,nuMin,nuMax,nIter,tol);
      else {
	pAlfaSP=BFIntegrator<AlfaSPIntegrandIState>::trapezoidalPW(integrand,nuMin,nuMax,cell->nuOld,nIter,relTol,absTol,useIterator,"",nIterMin,nIterMaxn,nBinMax);
	if (nIter>nIterMax) cout<<"IState::alfaSP: "<<upperState->iState()->name()<<":"<<upperState->getIndex()<<"->"<<lowerState->iState()->name()<<":"<<lowerState->getIndex()<<", cell: "<<upperState->iState()->atom()->cell()->index()<<", Warning: Maximum number of iterations exceeded\n";
      }
    }
    // Else - Set (spontaneous) recombination rate coefficient to zero
    else pAlfaSP=0;

    // If requested - Update
    if (modified) {
      if (includeOSP and pAlfaSPModifiedOSP==update) pAlfaSPModifiedOSP=pAlfaSP;
      else if ((not includeOSP) and pAlfaSPModified==update) pAlfaSPModified=pAlfaSP;
    }
    else {
      if (includeOSP and pAlfaSPOriginalOSP==update) pAlfaSPOriginalOSP=pAlfaSP;
      else if ((not includeOSP) and pAlfaSPOriginal==update) pAlfaSPOriginal=pAlfaSP;
    }
  }
  // Else - Get (spontaneous) recombination rate
  else 
    if (modified) {
      if (includeOSP) pAlfaSP=pAlfaSPModifiedOSP;
      else pAlfaSP=pAlfaSPModified;
    }
    else {
      if (includeOSP) pAlfaSP=pAlfaSPOriginalOSP;
      else pAlfaSP=pAlfaSPOriginal;
    }

  return pAlfaSP;
  }*/

// Get stimulated recombination coefficient - New! - Removed! - Testing!
double IState::alfaST(double T, bool includeOSP) {
  double _alfaST=0;
  Cell* cell=atom()->cell();
  int i,j;

  // If update requested
  if ((pAlfaST==update and (not includeOSP)) or (pAlfaSTOSP==update and includeOSP) or (T!=noT and T!=cell->T())) {

    // If not supplied - Get temperature
    if (T==noT) T=atom()->cell()->T();

    // Sum stimulated recombination coefficient
    for (i=0;i<bState.size();i++)
      for (j=0;j<bState[i].upTransBF.size();j++) {
	_alfaST+=bState[i].upTransBF[j]->alfaST(T,false,includeOSP)*bState[i].upTransBF[j]->upperState->X();
      }

    // Sum stimulated recombination coefficient - Testing! - Does not work! - Remove!
    //for (i=0;i<majorEdge.size();i++) _alfaST+=majorEdge[i]->alfaST(T,false,includeOSP)*majorEdge[i]->upperState->X();

    // If requested - Update
    if (pAlfaSTOSP==update and includeOSP) pAlfaSTOSP=_alfaST;
    else if (pAlfaST==update and (not includeOSP)) pAlfaST=_alfaST;
  }
  // Else - Get simulated recombination coefficient
  else if (includeOSP) _alfaST=pAlfaSTOSP;
  else _alfaST=pAlfaST;

  return _alfaST;
}

// Get (stimulated) recombination rate coefficient - New! - Testing!
double IState::alfaSTNu(const double& _nu, double T, bool modified) {
  double _alfaSTNu=0;
  int i,j;

  // Sum stimulated recombination rate coefficient
  for (i=0;i<bState.size();i++) {
    for (j=0;j<bState[i].upTransBF.size();j++) {
      if (bState[i].upTransBF[j]->alfaSimple()) continue;
      if (atom()->onTheSpot() and bState[i].E()==0) continue;
      _alfaSTNu+=bState[i].upTransBF[j]->alfaSTNu(_nu,T,modified)*bState[i].upTransBF[j]->upperState->X();
    }
  }

  return _alfaSTNu;
}

// Get (stimulated) recombination rate coefficient - New! - Testing!
double IState::alfaSTNu(FrequencyBinIterator _nu, EdgeType edgeType, double T, bool modified) {
  double _alfaSTNu=0;
  int i,j;

  // Sum stimulated recombination rate coefficient
  for (i=0;i<bState.size();i++) {
    for (j=0;j<bState[i].upTransBF.size();j++) {
      if (bState[i].upTransBF[j]->alfaSimple()) continue;
      if (atom()->onTheSpot() and bState[i].E()==0) continue;
      _alfaSTNu+=bState[i].upTransBF[j]->alfaSTNu(_nu,edgeType,T,modified)*bState[i].upTransBF[j]->upperState->X();
    }
  }

  return _alfaSTNu;
}

// (Stimulated) recombination rate coefficient integrand - New! - Testing!
class AlfaSTIntegrandIState : public JIntegrand {
public:
  IState* iState;
  const double T;
  const bool modified;
  AlfaSTIntegrandIState(IState* _iState, const double& _T, const bool& _modified) : iState(_iState), T(_T), modified(_modified) {};
  // The virtual prefix is missing, right?
  double operator()(double nu) const { return iState->alfaSTNu(nu,T,modified); }
  virtual double operator()(FrequencyBinIterator nu, EdgeType edgeType) const { return iState->alfaSTNu(nu,edgeType,T,modified); }
};

// Get (stimulated) recombination rate coefficient - New! - Testing!
// Note: The meaning of the OSP suffix is OSP only
/*double IState::alfaST(double T, bool includeOSP) {
  int i,j;
  bool modified=false;

  //if (atom()->cell()->index()==36) cout<<"IState::alfaST: cell: "<<atom()->cell()->index()<<", "<<name()<<", includeOSP: "<<includeOSP<<", Enter\n";

  // Set cell pointer
  Cell* cell=atom()->cell();

  double _alfaST=0,_alfaSTOSP=0;
  double relTol=1e-2,absTol=0;
  int nIter=0;
  if (T==noT) T=cell->T();
  bool useIterator=(T==cell->T())?true:false;
  AlfaSTIntegrandIState integrand(this,T,modified);

  // Set frequency bounds
  double nuMin=nuInf,nuMax=nuInf;
  for (i=0;i<bState.size();i++) for (j=0;j<bState[i].upTransBF.size();j++) if (bState[i].upTransBF[j]->nu()<nuMin) nuMin=bState[i].upTransBF[j]->nu();

  // If update requested or temperature supplied
  if (pAlfaST==update or (T!=cell->T())) {

    // Sum stimulated recombination coefficient
    for (i=0;i<bState.size();i++)
      for (j=0;j<bState[i].upTransBF.size();j++) {
	_alfaST+=bState[i].upTransBF[j]->alfaST(T,modified,false)*bState[i].upTransBF[j]->upperState->X();
      }
    
    // Calculate (stimulated) recombination rate coefficient
    //_alfaST=cell->JIntegrate(integrand,nuMin,nuMax,nIter,relTol,absTol,false,useIterator,false);

    // If requested - Update
    if (pAlfaST==update) pAlfaST=_alfaST;
  }
  // Else - Get (stimulated) recombination rate coefficient
  else _alfaST=pAlfaST;
  
  // If include OSP region
  if (includeOSP) {

    // If update requested or temperature supplied
    if (pAlfaSTOSP==update or (T!=cell->T())) {

      // Adjust absolute tolerence
      absTol=MAX(absTol,fabs(_alfaST)*relTol);
      
      // Calculate (stimulated) recombination rate coefficient
      _alfaSTOSP=cell->JIntegrate(integrand,nuMin,nuMax,nIter,relTol,absTol,true,useIterator,true);
      
      // If requested - Update
      if (pAlfaSTOSP==update) pAlfaSTOSP=_alfaSTOSP;
    }
    // Else - Get (stimulated) recombination rate coefficient
    else _alfaSTOSP=pAlfaSTOSP;
  }

  //if (atom()->cell()->index()==36) cout<<"IState::alfaST: cell: "<<atom()->cell()->index()<<", "<<name()<<", Return\n";
   
  if (includeOSP) return _alfaST+_alfaSTOSP;
  else return _alfaST;
  }*/

// Get recombination coefficient - New!
double IState::alfa(double T, bool includeOSP) {
  return alfaSP(T,includeOSP)+alfaST(T,includeOSP)+alfaAdd();
  //return alfaSP(T,false,includeOSP)+alfaST(T,false,includeOSP)+alfaAdd(); // Changed! - Testing!
}

// Get photo-ionization coefficient - Removed! - Testing!
double IState::gamma(bool includeOSP) {
  double _gamma=0;
  int i,j;

  //if (atom()->cell()->index()==36 and name().compare("FeIII")==0) cout<<"IState::gamma: cell: "<<atom()->cell()->index()<<", "<<name()<<", Enter\n";

  // If update requested
  if ((pGamma==update and (not includeOSP)) or (pGammaOSP==update and includeOSP)) {

    // Sum photo-ionization coefficient
    for (i=0;i<bState.size();i++)
      for (j=0;j<bState[i].upTransBF.size();j++) {
	_gamma+=bState[i].upTransBF[j]->gamma(false,includeOSP)*bState[i].upTransBF[j]->lowerState->X();
      }

    // Sum photo-ionization coefficient - Testing! - Does not work! - Remove!
    //for (i=0;i<majorEdge.size();i++) _gamma+=majorEdge[i]->gamma(false,includeOSP)*majorEdge[i]->lowerState->X();

    // Update
    if (pGammaOSP==update and includeOSP) pGammaOSP=_gamma;
    else pGamma=_gamma;
  }
  // Else - Get simulated recombination coefficient
  else if (includeOSP) _gamma=pGammaOSP;
  else _gamma=pGamma;

  //if (atom()->cell()->index()==36 and name().compare("FeIII")==0) cout<<"IState::gamma: cell: "<<atom()->cell()->index()<<", "<<name()<<", Enter\n";

  return _gamma;
}

// Calculate photo-ionization rate coefficient - New! - Testing!
double IState::gammaNu(const double& _nu, bool modified) {
  double _gammaNu=0;
  int i,j;

  // Sum photo-ionization rate coefficient for each upward transition
  for (i=0;i<bState.size();i++) {
    for (j=0;j<bState[i].upTransBF.size();j++) {
      if (bState[i].upTransBF[j]->alfaSimple()) continue;
      if (atom()->onTheSpot() and bState[i].E()==0) continue;
      _gammaNu+=bState[i].upTransBF[j]->gammaNu(_nu,modified)*bState[i].X();
    }
  }

  return _gammaNu;
}

// Calculate photo-ionization rate coefficient - New! - Testing!
double IState::gammaNu(FrequencyBinIterator _nu, EdgeType edgeType, bool modified) {
  double _gammaNu=0;
  int i,j;

  // Sum photo-ionization rate coefficient for each upward transition
  for (i=0;i<bState.size();i++) {
    for (j=0;j<bState[i].upTransBF.size();j++) {
      if (bState[i].upTransBF[j]->alfaSimple()) continue;
      if (atom()->onTheSpot() and bState[i].E()==0) continue;
      _gammaNu+=bState[i].upTransBF[j]->gammaNu(_nu,edgeType,modified)*bState[i].X();
    }
  }

  return _gammaNu;
}

// Photo-ionization rate coefficient integrand - New! - Testing!
class GammaIntegrandIState : public JIntegrand {
public:
  IState* iState;
  const bool modified;
  GammaIntegrandIState(IState* _iState, const bool& _modified) : iState(_iState), modified(_modified) {};
  // The virtual prefix is missing, right?
  double operator()(double nu) const { return iState->gammaNu(nu,modified); }
  double operator()(FrequencyBinIterator nu, EdgeType edgeType=noEdge) const { return iState->gammaNu(nu,edgeType,modified); }
};

// Get photo-ionization coefficient - New! - Testing!
// Note: The meaning of the OSP suffix is OSP only
/*double IState::gamma(bool includeOSP) {
  bool modified=false;
  double _gamma=0,_gammaOSP=0;
  double relTol=1e-2,absTol=0;
  bool useIterator=true;
  int i,j,nIter=0;
  GammaIntegrandIState integrand(this,modified);

  //if (atom()->cell()->index()==36) cout<<"IState::gamma: cell: "<<atom()->cell()->index()<<", "<<name()<<", includeOSP: "<<includeOSP<<", Enter\n";

  // Set frequency bounds
  double nuMin=nuInf,nuMax=nuInf;
  for (i=0;i<bState.size();i++) for (j=0;j<bState[i].upTransBF.size();j++) if (bState[i].upTransBF[j]->nu()<nuMin) nuMin=bState[i].upTransBF[j]->nu();

  // If update requested
  if (pGamma==update) {

    // Sum photo-ionization coefficient
    for (i=0;i<bState.size();i++)
      for (j=0;j<bState[i].upTransBF.size();j++)
	_gamma+=bState[i].upTransBF[j]->gamma(modified,false)*bState[i].upTransBF[j]->lowerState->X();
    
    // Calculate photo-ionization coefficient    
    //_gamma=atom()->cell()->JIntegrate(integrand,nuMin,nuMax,nIter,relTol,absTol,false,useIterator,false);
    
    pGamma=_gamma;
  }
  else _gamma=pGamma;

  // If include OSP region
  if (includeOSP) {

    // If update requested
    if (pGammaOSP==update) {

      // Adjust absolute tolerence
      absTol=MAX(absTol,fabs(_gamma)*relTol);
      
      // Calculate photo-ionization coefficient
      _gammaOSP=atom()->cell()->JIntegrate(integrand,nuMin,nuMax,nIter,relTol,absTol,true,useIterator,true);

      pGammaOSP=_gammaOSP;
    }
    else _gammaOSP=pGammaOSP;
  }

  //if (atom()->cell()->index()==36) cout<<"IState::gamma: cell: "<<atom()->cell()->index()<<", "<<name()<<", Enter\n";

  if (includeOSP) return _gamma+_gammaOSP;
  else return _gamma;
  }*/

// Get OSP radiative rate (out of the state) - New! - Removed! - Testing!
double IState::ROSP() {
  int i,j;

  //if (atom()->cell()->index()==36 and name().compare("FeIII")==0) cout<<"IState::ROSP: cell: "<<atom()->cell()->index()<<", "<<name()<<", Enter\n";

  // If update requested
  if (pROSP==update) {

    // Reset OSP radiative rate
    pROSP=0;

    // Sum OSP radiative rate
    for (i=0;i<bState.size();i++) {
      for (j=0;j<bState[i].upTransBF.size();j++) pROSP+=bState[i].upTransBF[j]->ROSP();
      for (j=0;j<bState[i].downTransBF.size();j++) pROSP-=bState[i].downTransBF[j]->ROSP();
    }

    // Sum OSP radiative rate- Testing! - Note: Now upward rate only! - Does not work! - Remove!
    //for (i=0;i<majorEdge.size();i++) pROSP+=majorEdge[i]->ROSP();
    //for (i=0;i<majorEdge.size();i++) pROSP-=majorEdge[i]->ROSP();
  }

  //if (atom()->cell()->index()==36 and name().compare("FeIII")==0) cout<<"IState::ROSP: cell: "<<atom()->cell()->index()<<", "<<name()<<", Return\n";

  return pROSP;
}

// Net radiative rate integrand 1 - New! - Testing!
class ROSPIntegrandIState1 : public NuIntegrand {
public:
  IState* iState;
  const double ne;
  const double nLower;
  double nUpper;
  ROSPIntegrandIState1(IState* _iState) : iState(_iState), ne(_iState->atom()->cell()->ne()), nLower(_iState->n()) { IState* upperState=_iState->atom()->iState.find(_iState->C()-1); nUpper=upperState->n(); };
  // The virtual prefix is missing, right?
  double operator()(double nu) const { return iState->gammaNu(nu)*nLower-iState->alfaSTNu(nu)*ne*nUpper; }
  double operator()(FrequencyBinIterator nu, EdgeType edgeType=noEdge) const { return iState->gammaNu(nu,edgeType)*nLower-iState->alfaSTNu(nu,edgeType)*ne*nUpper; }
};

// Net radiative rate integrand 2 - New! - Testing!
class ROSPIntegrandIState2 : public NuIntegrand {
public:
  IState* iState;
  const double ne;
  double nUpper;
  ROSPIntegrandIState2(IState* _iState) : iState(_iState), ne(_iState->atom()->cell()->ne()) { IState* upperState=_iState->atom()->iState.find(_iState->C()-1); nUpper=upperState->n(); };
  // The virtual prefix is missing, right?
  double operator()(double nu) const { return iState->alfaSPNu(nu)*ne*nUpper; }
  double operator()(FrequencyBinIterator nu, EdgeType edgeType=noEdge) const { return iState->alfaSPNu(nu,edgeType)*ne*nUpper; }
};

// Get (upward) OSP radiative rate - New! - Testing!
// Note: Definition of ROSP changed!
/*double IState::ROSP() {
  int i,j;

  //if (atom()->cell()->index()==36) cout<<"IState::ROSP: cell: "<<atom()->cell()->index()<<", "<<name()<<", Enter\n";

  // If highest ionization state - Return zero
  IState* upperState=atom()->iState.find(C()-1);
  if (upperState==NULL) return 0.0;

  // If update requested
  if (pROSP==update) {
    
    // Set cell pointer
    Cell* cell=atom()->cell();
   
    double relTol=1e-2,absTol=0.0;
    bool useIterator=true;
    ROSPIntegrandIState1 integrand1(this);
    ROSPIntegrandIState2 integrand2(this);

    // Set frequency bounds
    double nuMin=nuInf,nuMax=nuInf;
    for (i=0;i<bState.size();i++) for (j=0;j<bState[i].upTransBF.size();j++) if (bState[i].upTransBF[j]->nu()<nuMin) nuMin=bState[i].upTransBF[j]->nu();
    
    // Calculate (upward) OSP radiative rate
    pROSP=cell->OSPIntegrate(integrand1,integrand2,nuMin,nuMax,relTol,absTol,useIterator);
  }

  //if (atom()->cell()->index()==36) cout<<"IState::ROSP: cell: "<<atom()->cell()->index()<<", "<<name()<<", Return\n";

  return pROSP;
  }*/

// Get OSP radiative rate (out of the state) number derivative - New!
// Note: Not updated if the dNedN flag is changed - Fix!!!
double IState::dROSPdN(IState* iState, bool dNedN) {

  // If update requested
  if (dROSPdNUpdate) { 

    // Set update lock (to avoid race conditions)
    omp_set_lock(&dROSPdNUpdateLock);

    // Testing!!!
    cout<<"IState::dROSPdN: cell: "<<atom()->cell()->index()<<", "<<name()<<", Locked\n";

    // If requested - Update
    if (dROSPdNUpdate) { updateDROSPdN(dNedN); dROSPdNUpdate=false; }

    // Testing!!!
    cout<<"IState::dROSPdN: cell: "<<atom()->cell()->index()<<", "<<name()<<", Unlock\n";

    // Unset update lock
    omp_unset_lock(&dROSPdNUpdateLock);
  }

  // Return OSP radiative rate (out of the state) number derivative
  return pDROSPdN[iState->atom()->index()][iState->index()];
}

// Update OSP radiative rate (out of the state) number derivative - New!
// Note: dNedN flag does not work - Fix or remove!!!
void IState::updateDROSPdN(bool dNedN) {
  TransBF *trans,*trans1,*trans2;
  IState *lowerState,*upperState;
  Cell* cell=atom()->cell();
  int i,j,k,ii,jj,kk,ll;

  // Get lower and upper state pointers
  // Note: C reversed! Right?
  lowerState=atom()->iState.find(C()+1);
  upperState=atom()->iState.find(C()-1);
  
  // Reset dROSPdN vector
  pDROSPdN.resize(0);
  pDROSPdN.resize(cell->atom.size());
  for (i=0;i<cell->atom.size();i++) pDROSPdN[i].resize(cell->atom[i].iState.size(),0);

  // For each upward bound-free transition of the cell
  for (ii=0;ii<cell->atom.size();ii++) {
    for (jj=0;jj<cell->atom[ii].iState.size();jj++) {
      for (kk=0;kk<cell->atom[ii].iState[jj].bState.size();kk++) {
	for (ll=0;ll<cell->atom[ii].iState[jj].bState[kk].upTransBF.size();ll++) {
	  // If on-the-spot transition
	  if (cell->atom[ii].iState[jj].bState[kk].upTransBF[ll]->onTheSpotNew()) {
	    trans1=cell->atom[ii].iState[jj].bState[kk].upTransBF[ll];
	    // For each upward bound-free transition of the ionization state
	    for (i=0;i<bState.size();i++) {
	      for (j=0;j<bState[i].upTransBF.size();j++) {
		// If on-the-spot transition
		if (bState[i].upTransBF[j]->onTheSpotNew()) {
		  trans2=bState[i].upTransBF[j];
		  // Add the contribution from the second transition to the derivative with respect to the lower and upper levels of the first transition
		  // Hmm. Check conversion to ion number density.
		  pDROSPdN[cell->atom[ii].index()][trans1->upperState->iState()->index()]+=trans2->dROSPdN(trans1->upperState)*trans1->upperState->X();
 		  pDROSPdN[cell->atom[ii].index()][trans1->lowerState->iState()->index()]+=trans2->dROSPdN(trans1->lowerState)*trans1->lowerState->X();
		}
	      }
	      // For each downward bound-free transition of ionization state
	      for (j=0;j<bState[i].downTransBF.size();j++) {
		// If on-the-spot transition
		if (bState[i].downTransBF[j]->onTheSpotNew()) {
		  trans2=bState[i].downTransBF[j];
		  // Add the contribution from the second transition to the derivative with respect to the lower and upper levels of the first transition
		  // Hmm. Check conversion to ion number density.
		  pDROSPdN[cell->atom[ii].index()][trans1->upperState->iState()->index()]-=trans2->dROSPdN(trans1->upperState)*trans1->upperState->X();
 		  pDROSPdN[cell->atom[ii].index()][trans1->lowerState->iState()->index()]-=trans2->dROSPdN(trans1->lowerState)*trans1->lowerState->X();
		}
	      }
	    }
	  }
	}
      }
    }
  }

  // For each bound state of the ionization state
  /*for (i=0;i<bState.size();i++) {
    // For each upward bound-free transition
    for (j=0;j<bState[i].upTransBF.size();j++) {
      // If not on-the-spot transition
      if (not bState[i].upTransBF[j]->onTheSpotNew()) {
	trans=bState[i].upTransBF[j];
	// Add the contribution to the derivative with respect to the lower and upper levels of the transition
	pDROSPdN[atom()->index()][trans->lowerState->iState()->index()]+=(trans->gamma(true)-trans->gamma(false))*trans->lowerState->X();
	pDROSPdN[atom()->index()][trans->upperState->iState()->index()]-=(trans->alfa(noT,true)-trans->alfa(noT,false))*trans->upperState->X();
      }
    }
    // For each downward bound-free transition of ionization state
    for (j=0;j<bState[i].downTransBF.size();j++) {
      // If not on-the-spot transition
      if (not bState[i].downTransBF[j]->onTheSpotNew()) {
	trans=bState[i].downTransBF[j];
	// Add the contribution to the derivative with respect to the lower and upper levels of the transition
	pDROSPdN[atom()->index()][trans->upperState->iState()->index()]+=(trans->alfa(noT,true)-trans->alfa(noT,false))*trans->upperState->X();
	pDROSPdN[atom()->index()][trans->lowerState->iState()->index()]-=(trans->gamma(true)-trans->gamma(false))*trans->lowerState->X();
      }
    }
    }*/

  // If include electron number density term
  // Not correct!
  /*if (dNedN) {
    // For each ionization state of the cell
    for (ii=0;ii<cell->atom.size();ii++) {
      for (jj=0;jj<cell->atom[ii].iState.size();jj++) {
	// Add electron number density term
	pDROSPdN[ii][jj]+=ROSP()*(cell->dNedN(&(cell->atom[ii].iState[jj]))/cell->ne());
      }
    }
    }*/
}

// Get radiative rate (out of the state) - New!
// Testing! includeOSP set to true in alfa and gamma calls! Removed!
double IState::R(bool includeOSP) {
  double _R=0;
  IState *lowerState,*upperState;
  Cell* cell=atom()->cell();

  //if (atom()->cell()->index()==36) cout<<"IState::R: cell: "<<atom()->cell()->index()<<", "<<name()<<", includeOSP: "<<includeOSP<<", Enter\n";

  // Get lower and upper state pointers
  // Note: C reversed! Right?
  lowerState=atom()->iState.find(C()+1);
  upperState=atom()->iState.find(C()-1);

  //cout<<"IState::R, cell: "<<cell->index()<<", "<<name()<<", gamma: "<<gamma(true)<<", alfa: "<<alfa(noT,true)<<", n: "<<n()<<", ne: "<<cell->ne()<<"\n";
  //if (lowerState!=NULL) cout<<"IState::R, cell: "<<cell->index()<<", "<<name()<<", lowerState: "<<lowerState->name()<<", gamma: "<<lowerState->gamma(true)<<", alfa: "<<lowerState->alfa(noT,true)<<", n: "<<lowerState->n()<<"\n";
  //if (upperState!=NULL) cout<<"IState::R, cell: "<<cell->index()<<", "<<name()<<", upperState: "<<upperState->name()<<", n: "<<upperState->n()<<"\n";

  // Add radiative rate out of the state
  _R+=gamma(false)*n();
  if (lowerState!=NULL) _R+=lowerState->alfa(noT,false)*n()*cell->ne();
  // Subtract radiative rate from the upper state
  if (upperState!=NULL) _R-=alfa(noT,false)*upperState->n()*cell->ne();
  // Subtract radiative rate from the lower state
  if (lowerState!=NULL) _R-=lowerState->gamma(false)*lowerState->n();

  // If requested - Include OSP contribution
  // Testing! Removed! Re-added!
  if (includeOSP) _R+=ROSP();
  // Changed to reflect new definition of ROSP! - Testing!
  //if (includeOSP) { _R+=ROSP(); if (lowerState!=NULL) _R-=lowerState->ROSP(); }

  //if (atom()->cell()->index()==36) cout<<"IState::R: cell: "<<atom()->cell()->index()<<", "<<name()<<", Return\n";

  return _R;
}

// Get radiative rate (out of the state) number derivative - New!
// Testing! includeOSP set to true in alfa and gamma calls!
double IState::dRdN(IState* iState, bool includeOSP, bool dNedN) {
  double _dRdN=0;
  IState *lowerState,*upperState;
  Cell* cell=atom()->cell();

  //if (atom()->cell()->index()==36) cout<<"IState::dRdN: cell: "<<atom()->cell()->index()<<", "<<name()<<", includeOSP: "<<includeOSP<<", Enter\n";

  // Get lower and upper state pointers
  // Note: C reversed! Right?
  lowerState=atom()->iState.find(C()+1);
  upperState=atom()->iState.find(C()-1);

  // If with respect to this state
  if (iState==this) {
    // Calculate derivative
    _dRdN+=gamma(true);
    if (lowerState!=NULL) _dRdN+=lowerState->alfa(noT,true)*cell->ne();
  }
  // Else if with respect to the upper state - Calculate derivative
  else if (iState==upperState and upperState!=NULL) _dRdN-=alfa(noT,true)*cell->ne();
  // Else if with respect of the lower state - Calculate derivative
  else if (iState==lowerState and lowerState!=NULL) _dRdN-=lowerState->gamma(true);

  // If include electron number density terms
  // Note: OSP contribution inluded here!
  if (dNedN) {
    // If upper state present - Add electron number density term
    if (upperState!=NULL) _dRdN-=alfa(noT,true)*upperState->n()*cell->dNedN(iState);
    // If lower state present - Add electron number density term
    if (lowerState!=NULL) _dRdN+=lowerState->alfa(noT,true)*n()*cell->dNedN(iState);
  }

  // If requested - Include OSP contribution
  // Note: dNedN flag does not work - The OSP contribution is included above!
  // Testing! Removed!
  //if (includeOSP) _dRdN+=dROSPdN(iState,dNedN);

  //if (atom()->cell()->index()==36) cout<<"IState::dRdN: cell: "<<atom()->cell()->index()<<", "<<name()<<", Return\n";
  
  return _dRdN;
}

// OSP radiative rate number derivative integrand 1
/*class DROSPdNIIntegrand1 : public NuDNIIntegrand {
public:
  IState *thisIstate,lowerIState,upperIState;
  Cell* cell;
  const double ne;
  const double nLower;
  const double nUpper;
  DROSPdNIIntegrand1(IState* _thisIState, IState* _lowerIState, IState* _upperIState) : thisIState(_thisIState), lowerIState(_lowerIState), upperIState(_upperIState) {
    ne=thisIState->atom()->cell->ne();
    n=thisIState->n();
    nLower=(lowerState!=NULL)?lowerState->n():0; 
    nUpper=(upperState!=NULL)?upperState->n();0;
    cell=thisIState->atom()->cell();
  };
  virtual double F(double nu) const {
    double _F=0;    
    _F=iState->gammaNu(nu)*n;
    if (upperState!=NULL) _F-=thisIState->alfaSTNu(nu)*ne*nUpper;
    if (lowerState!=NULL) _F-=lowerIState->gammaNu(nu)*nLower;
    return _F;
  }
  virtual double dFdN(double nu, IState* iState) const { 
    double _dFdN=0;
    if (iState==thisIState) {
      _dFdN+=trans->gammaNu(nu);
      if (lowerState!=NULL) _dFdN+=lowerIState->alfaSTNu(nu)*ne;
    }
    if (iState==upperIState) _dFdN-=thisIState->alfaSTNu(nu)*ne;
    if (iState==lowerIState) _dFdN-=lowerState->gammaNu(nu);
    if (dNedN) {
      if (upperState!=NULL) _dFdN-=alfaSTNu()*nUpper*cell->dNedN(iState);
      if (lowerState!=NULL) _dFdN+=lowerState->alfaSTNu()*n*cell->dNedN(iState);
    }
    return _dFdN;
  }
  };*/

// OSP radiative rate number derivative integrand 2
/*class DROSPdNIIntegrand2 : public NuDNIIntegrand {
public:
  IState *thisIState,lowerIState,upperIState;
  Cell* cell;
  const double ne;
  const double nLower;
  const double nUpper;
  DROSPdNIIntegrand2(IState* _thisIState, IState* _lowerIState, IState* _upperIState) : thisIState(_thisIState), lowerIState(_lowerIState), upperIState(_upperIState) {
    ne=thisIState->atom()->cell()->ne();
    nLower=(lowerIState!=NULL)?lowerIState->n():0;
    nUpper=(upperIState!=NULL)?upperIState->n():0;
    cell=thisIState->atom()->cell();
  };
  virtual double F(double nu) const { 
    double _F=0;
    _F-=thisIState->alfaSPNu(nu)*ne*nUpper;
    if (lowerState!=NULL) _F+=lowerState->alfaSPNu(nu)*ne*n;
    return _F;
  }
  virtual double dFdN(double nu, IState* iState) const { 
    double _dFdN=0;
    if (iState==thisIState and lowerIState!=NULL) _dFdN+=lowerIState->alfaSPNu(nu)*ne;
    if (iState==upperIState) _dFdN-=thisIState->alfaSPNu(nu);
    if (dNedN) {
      if (upperState!=NULL) _dFdN-=alfaSPNu(nu)*nUpper*cell->dNedN(iState);
      if (lowerState!=NULL) _dFdN+=lowerState->alfaSPNu(nu)*n*cell->dNedN(iState);      
    }
    return _dFdN;
  }
  };*/

// Get OSP radiative rate number derivative
/*double IState::dROSPdN(IState* iState) {
  double _dROSPdN;

  // Set cell pointer
  Cell* cell=atom()->cell();
   
  double relTol=1e-2;
  DROSPdNIIntegrand1 integrand1(this,lowerState,upperState);
  DROSPdNIIntegrand2 integrand2(this,lowerState,upperState);
    
  // Calculate net radiative rate in the on-the-spot range
  _dROSPdN=cell->dROSPdNIntegrate(integrand1,integrand2,iState,nu(),nuInf,relTol);

  if (!isnumber(_dROSPdN)) cout<<"IState::dROSPdN: "<<name()<<", cell: "<<upperState->iState()->atom()->cell()->index()<<", Warning: Invalid rate: "<<_dROSPdN<<"\n";

  return _dROSPdN;
  }*/

// *****************************************
// *****************************************
// *****************************************

// Get memory allocated for bound-bound transitions - New!
double IState::memTransBB() const {
  double _mem=0;
  int i,j;

  // For each bound state
  for (i=0;i<bState.size();i++) {
    // Add memory allocated for each (upward) bound-bound transition
    for (j=0;j<bState[i].upTransBB.size();j++) _mem+=2*sizeof(TransBB*)+bState[i].upTransBB[j]->mem();
  }

  return _mem;
}

// Get memory allocated - New!
double IState::mem() const {
  double _mem=sizeof(*this);
  int i,j;

  // Add memory allocated for each bound state
  for (i=0;i<bState.size();i++) _mem+=bState[i].mem();
  
  // Add memory allocated for bound-bound transitions
  _mem+=memTransBB();

  // Add memory allocated for pDCTdN array
  for (i=0;i<pDCTdN.size();i++)
    for (j=0;j<pDCTdN[i].size();j++) _mem+=sizeof(double);

  // Add memory allocated for pDROSPdN array
  for (i=0;i<pDROSPdN.size();i++)
    for (j=0;j<pDROSPdN[i].size();j++) _mem+=sizeof(double);

  return _mem;
}

// Get memory allocated for probaility lists - New!
double IState::memPList() const {
  double _mem=0;
  int i;

  // For each bound state
  for (i=0;i<bState.size();i++) {
    // Add memory allocated for probability lists
    _mem+=bState[i].memPList();
  }

  return _mem;
}

// Update bound-free cut-off frequencies - New!
// This function does not work - Remove this and majorEdge!
/*void IState::updateBoundFreeCutOff(bool checkEta, double boundFreeCutOff) {
  EdgeList edge;
  int i,j;

  // Reset major edge list
  majorEdge.resize(0);

  // Add all transitions to edge list
  for (i=0; i<bState.size(); i++)
    for (j=0; j<bState[i].upTransBF.size(); j++) edge.push_back(bState[i].upTransBF[j]);

  // Initiate related vectors
  vector<double> kappaEdge(edge.size(),0),etaEdge(edge.size(),0);
  vector<bool> minorEdge(edge.size(),false);
    
  // If no cut-off - Add all transitions to major edge list and return
  if (boundFreeCutOff<=0) { majorEdge=edge; return; }

  // For each edge
  for (i=edge.size()-1;i>=0;i--) {
    // For each lower frequency edge
    for (j=edge.size()-1;j>=i;j--) {
      // If major edge
      if (not minorEdge[j]) {
	// Sum opacity and emissivity
	kappaEdge[i]+=edge[j]->kappa(nuEdge(edge[i]->nu(),blueEdge)); 
	etaEdge[i]+=edge[j]->eta(nuEdge(edge[i]->nu(),blueEdge));
      }
    }
    if (atom()->cell()->index()==36) cout<<"IState::updateBoundFreeCutOff: cell: "<<atom()->cell()->index()<<", "<<name()<<", i: "<<i<<", j: "<<j<<", kappa: "<<edge[i]->kappa(nuEdge(edge[i]->nu(),blueEdge))<<", kappaEdge: "<<kappaEdge[i]<<" eta: "<<edge[i]->eta(nuEdge(edge[i]->nu(),blueEdge))<<", etaEdge: "<<etaEdge[i]<<"\n";
    // If opacity and (optionally) emissivity below cut-off value
    if (edge[i]->kappa(nuEdge(edge[i]->nu(),blueEdge))<boundFreeCutOff*kappaEdge[i] and ((not checkEta) or edge[i]->eta(nuEdge(edge[i]->nu(),blueEdge))<boundFreeCutOff*etaEdge[i])) {
      if (atom()->cell()->index()==36) cout<<"IState::updateBoundFreeCutOff: cell: "<<atom()->cell()->index()<<", "<<name()<<", i: "<<i<<", j: "<<j<<", Edge marked as minor"<<"\n";
    // If opacity and (optionally) emissivity below cut-off value
      // Mark edge as minor
      minorEdge[i]=true;
    }  
    // Else - Add to major edge list
    else majorEdge.push_back(edge[i]);    
  }

  if (atom()->cell()->index()==36) cout<<"IState::updateBoundFreeCutOff: cell: "<<atom()->cell()->index()<<", "<<name()<<", Marked "<<(edge.size()-majorEdge.size())<<" out of "<<edge.size()<<" edges as minor using a bound-free cut-off of "<<boundFreeCutOff<<"\n";
  }*/

// Constructor - New!
CTReaction::CTReaction() : data(NULL), pCell(NULL), upTrans(NULL), downTrans(NULL) {};

// Destructor - New!
CTReaction::~CTReaction() {};

// Get index
int CTReaction::index() const {
  return pIndex;
}

// Get cell back reference pointer
Cell* CTReaction::cell() const {
  return pCell;
}

// Reaction coefficient - New!
double CTReaction::alfa(double T) {
  if (T==noT) T=cell()->T();

  double _alfa=(*data->alfa)(T);

  return _alfa;
}

// Number density type - New!
NStateType CTReaction::nType() {
  return data->nType;
}

// Reaction energy gain (in forward direction) - New!
double CTReaction::Xi() {
  return downTrans->Xi()-upTrans->Xi();
}

// Reaction rate - New!
double CTReaction::rate(double T) {
  
  if (T==noT) T=cell()->T();

  //cout<<"CTReaction::rate, cell: "<<cell()<<", data: "<<data<<", downTrans: "<<downTrans<<", upTrans: "<<upTrans<<"\n";
  //if (cell()!=NULL) cout<<"CTReaction::rate, cell: "<<cell()->index()<<"\n";
  //if (cell()!=NULL and data!=NULL) cout<<"CTReaction::rate, cell: "<<cell()->index()<<", index: "<<index()<<"\n";

  double _rate;  
  if (nType()==nBState) _rate=alfa(T)*downTrans->upperState->n()*upTrans->lowerState->n();
  else _rate=alfa(T)*downTrans->upperState->iState()->n()*upTrans->lowerState->iState()->n();

  //cout<<"CTReaction::rate, cell: "<<cell()->index()<<", index: "<<index()<<", Return\n";

  return _rate;
}

// Reaction rate (forward direction) - New!
double CTReaction::forwardRate(double T) {
  double _rate;

  // If not provided - Get temperature    
  if (T==noT) T=cell()->T();

  // Calculate forward rate
  if (nType()==nBState) _rate=alfa(T)*downTrans->upperState->n()*upTrans->lowerState->n();
  else _rate=alfa(T)*downTrans->upperState->iState()->n()*upTrans->lowerState->iState()->n();

  return _rate;
}

// Reaction rate (backward direction) - New!
double CTReaction::backwardRate(double T) {
  double nnInitial,nnFinal,ggInitial,ggFinal,xLTE,_rate;

  // If not provided - Get temperature
  if (T==noT) T=cell()->T();

  // Calculate number density products for the initial and final state (forward direction)
  nnInitial=downTrans->upperState->n()*upTrans->lowerState->n();
  nnFinal=downTrans->lowerState->n()*upTrans->upperState->n();

  // Calculate statistical weight products for the initial and final state (forward direction)
  ggInitial=downTrans->upperState->g()*upTrans->lowerState->g();
  ggFinal=downTrans->lowerState->g()*upTrans->upperState->g();

  // Calculate number density product (initial/final) LTE ratio
  xLTE=(ggInitial/ggFinal)*exp(-Xi()/(k*T));

  // Calculate backward rate using detailed balance
  _rate=(nnFinal/nnInitial)*xLTE*forwardRate(T);

  //cout<<"CTReaction::backwardRate: nnInitial: "<<nnInitial<<", nnFinal: "<<nnFinal<<", xLTE: "<<xLTE<<" T: "<<T<<", Xi: "<<Xi()<<", forwardRate: "<<forwardRate(T)<<", backwardRate: "<<_rate<<"\n";
  
  return _rate;
}

// Cooling rate - New!
double CTReaction::cooling(double T) {
  double C=forwardCooling(T)+backwardCooling(T);
  return C;
  //return MAX(rate(T)*(upTrans->Xi()-downTrans->Xi()),0);
}

// Cooling rate (forward direction) - New!
double CTReaction::forwardCooling(double T) {
  double C=MAX(-forwardRate(T)*Xi(),0);
  return C;
}

// Cooling rate (backward direction) - New!
double CTReaction::backwardCooling(double T) {
  double C=MAX(backwardRate(T)*Xi(),0);
  return C;
}

// Heating rate - New!
double CTReaction::heating(double T) {
  double H=forwardHeating(T)+backwardHeating(T);
  return H;
  //return MAX(rate(T)*(downTrans->Xi()-upTrans->Xi()),0);
}

// Heating rate (forward direction) - New!
double CTReaction::forwardHeating(double T) {
  double H=MAX(forwardRate(T)*Xi(),0);
  return H;
}

// Heating rate (backward direction) - New!
double CTReaction::backwardHeating(double T) {
  double H=MAX(-backwardRate(T)*Xi(),0);
  return H;
}

// Heating channel probability
double CTReaction::PHeating(CTDirection cTDirection) {
  double P;
  if (cTDirection==ctForward) P=MAX(Xi(),0)/downTrans->Xi();
  else P=MAX(-Xi(),0)/upTrans->Xi();
  return P;
}

// Draw absorption channel - New!
MCChannel CTReaction::absorptionChannel(CTDirection cTDirection) {
  MCChannel channel;
  double z,P;

  // Pick random number between 0 and 1
  z=randomize(0,1);
  // Calculate probability for heating
  //P=heating()/downTrans->Xi(); // This must be wrong, right!
  //if (cTDirection==ctForward) P=MAX(Xi(),0)/downTrans->Xi();
  //else P=MAX(-Xi(),0)/upTrans->Xi();
  P=PHeating(cTDirection);
  // If drawn - Select heating channel
  if (P<=z) channel=cHeating;
  // Else - Select ionization channel
  else channel=cIonization;

  return channel;
}

// Find bound-free transition matching the CT transition data - New!
// This is really far to complicated - Improve!!!
TransBF* CTReaction::findTrans(const CTTransData& transData, vector<Atom>& atomList) {
  TransBF* trans=NULL;
  Atom* atom=NULL;
  IState* upperIState=NULL;
  BState* upperBState=NULL;
  IState* lowerIState=NULL;
  BState* lowerBState=NULL;
  int i;

  for (i=0;i<atomList.size();i++)
    if (transData.atom->Z==atomList[i].Z()) atom=&(atomList[i]);
  if (atom==NULL) { cout<<"CTReaction::findTrans: atom Z="<<transData.atom->Z<<" not found\n"; return trans; }

  upperIState=atom->iState.find(transData.upperIState->C);
  if (upperIState==NULL) { cout<<"CTReaction::findTrans: atom Z="<<transData.atom->Z<<", upperIState C="<<transData.upperIState->C<<" not found\n"; return trans; }

  for (i=0;i<upperIState->bState.size();i++)
    if (transData.upperBState->index==upperIState->bState[i].data->index) {upperBState=&(upperIState->bState[i]); break; }
  if (upperBState==NULL) { cout<<"CTReaction::findTrans: atom Z="<<transData.atom->Z<<", upperIState C="<<transData.upperIState->C<<", upperBState index="<<transData.upperBState->index<<", not found\n"; return trans; }

  lowerIState=atom->iState.find(transData.lowerIState->C);
  if (lowerIState==NULL) { cout<<"CTReaction::findTrans: atom Z="<<transData.atom->Z<<", lowerIState C="<<transData.lowerIState->C<<" not found\n"; return trans; }

  for (i=0;i<lowerIState->bState.size();i++)
    if (transData.lowerBState->index==lowerIState->bState[i].data->index) {lowerBState=&(lowerIState->bState[i]); break; }
  if (lowerBState==NULL) { cout<<"CTReaction::findTrans: atom Z="<<transData.atom->Z<<", lowerIState C="<<transData.lowerIState->C<<", lowerBState index="<<transData.lowerBState->index<<", not found\n"; return trans; }

  for (i=0;i<upperBState->downTransBF.size();i++) 
    if (upperBState->downTransBF[i]->lowerState==lowerBState) { trans=upperBState->downTransBF[i]; break; }
  if (trans==NULL) cout<<"CTReaction::findTrans: Bound-free transition not found, atom Z="<<transData.atom->Z<<", upperIState C="<<transData.upperIState->C<<", upperBState index="<<transData.upperBState->index<<", lowerIState C="<<transData.lowerIState->C<<", lowerBState index="<<transData.lowerBState->index<<", upperBState.index: "<<upperBState->getIndex()<<", lowerBState.index: "<<lowerBState->getIndex()<<", upperBState.downTrans.size: "<<upperBState->downTransBF.size()<<"\n";
  
  return trans;
}

// Construct charge-transfer reaction
bool CTReaction::construct(CTReactionData* reactionData, vector<Atom>& atomList) {
  bool constructed=true;

  // Set atom data pointer
  data=reactionData;

  //cout<<"CTReaction::construct: upTrans: ["<<data->upTrans.atom->Z<<",["<<data->upTrans.upperIState->C<<","<<data->upTrans.upperBState->index<<"],["<<data->upTrans.lowerIState->C<<","<<data->upTrans.lowerBState->index<<"]], downTrans: ["<<data->downTrans.atom->Z<<",["<<data->downTrans.upperIState->C<<","<<data->downTrans.upperBState->index<<"],["<<data->downTrans.lowerIState->C<<","<<data->downTrans.lowerBState->index<<"]]\n";

  // Find upward transition
  upTrans=findTrans(data->upTrans,atomList);
  if (upTrans==NULL) { cout<<"CTReaction::construct: upTrans not found\n"; return false; }

  // Find downward transition
  downTrans=findTrans(data->downTrans,atomList);
  if (downTrans==NULL) { cout<<"CTReaction::construct: downTrans not found\n"; return false; }

  //cout<<"CTReaction::construct: upTrans: ["<<data->upTrans.atom->Z<<",["<<data->upTrans.upperIState->C<<","<<data->upTrans.upperBState->index<<"],["<<data->upTrans.lowerIState->C<<","<<data->upTrans.lowerBState->index<<"]], downTrans: ["<<data->downTrans.atom->Z<<",["<<data->downTrans.upperIState->C<<","<<data->downTrans.upperBState->index<<"],["<<data->downTrans.lowerIState->C<<","<<data->downTrans.lowerBState->index<<"]], Constructed\n";

  return constructed;
}

// Set back references
void CTReaction::setBackReference(Cell* cell) { 
  int i;
   
  // Set cell back reference pointer
  pCell=cell;
}

// Get memory allocated - New!
double CTReaction::mem() const {
  double _mem=sizeof(*this);

  return _mem;
}

// Is LTE group internal
bool Trans::isLTEGroupInternal() const {
  int i;

  return (lowerState->LTEGroupMaster()==upperState->LTEGroupMaster());

  /*if (lowerState->LTERefState!=NULL) {
    if (lowerState->LTERefState==upperState) return true;
    for (i=0;i<lowerState->LTERefState->LTEGroup.size();i++)
      if (lowerState->LTERefState->LTEGroup[i]==upperState) return true;
  }

  if (upperState->LTERefState!=NULL) {
    if (upperState->LTERefState==lowerState) return true;
    for (i=0;i<upperState->LTERefState->LTEGroup.size();i++)
      if (upperState->LTERefState->LTEGroup[i]==lowerState) return true;
      }
      
      return false;*/
} 

// Destructor - Testing!
TransBB::~TransBB() {
  // Testing!
  //if (upperState!=NULL and lowerState!=NULL) if (upperState->iState()!=NULL) if (upperState->iState()->atom()!=NULL) if (upperState->iState()->atom()->cell()!=NULL) if (upperState->iState()->atom()->cell()->master()) cout<<"TransBB::~TransBB: cell: "<<upperState->iState()->atom()->cell()->index()<<", "<<upperState->iState()->name()<<":"<<upperState->getIndex()<<"->"<<lowerState->getIndex()<<"\n";
}

// Get transition type
TransClass TransBB::type() const {
  return transClassBB;
}

// Get transition energy
double TransBB::Xi() const {
  return data->Xi;
}

// Get transition frequency
double TransBB::nu() const {
  return Xi()/h;
}

// Get downward radiative rate - Changed!
double TransBB::RD(bool twoPhoton) {
  double pRD;

  pRD=R1D();
  if (twoPhoton) pRD+=R2D();

  return pRD;
}

// Get line downward radiative rate - New!
double TransBB::R1D() {

  // If update requested
  if (pR1D==update) {
    if (A()!=0) pR1D=A()*beta()*upperState->n(); // Changed!
    else pR1D=0; // New!
  }

  // Testing!
  if (!isnumber(pR1D)) cout<<"TransBB::R1D: "<<lowerState->iState()->name()<<":"<<lowerState->getIndex()<<"->"<<upperState->getIndex()<<", cell: "<<upperState->iState()->atom()->cell()->index()<<", Warning: Invalid rate: "<<pR1D<<"\n";
 
  return pR1D;
}

// Get two-photon downward radiative rate - New!
double TransBB::R2D() {

  // If update requested
  if (pR2D==update) {
    pR2D=A2()*upperState->n();
  }

  // Testing!
  if (!isnumber(pR2D)) cout<<"TransBB::R2D: "<<lowerState->iState()->name()<<":"<<lowerState->getIndex()<<"->"<<upperState->getIndex()<<", cell: "<<upperState->iState()->atom()->cell()->index()<<", Warning: Invalid rate: "<<pR2D<<"\n";

  return pR2D;
}

// Get upward radiative rate
double TransBB::RU(bool preventNegative) { // Changed!

  // If update requested
  if (pRU==update) {
    // Update upward radiative rate
    if (A()!=0) pRU=(BU()*lowerState->n()-BD()*upperState->n())*beta()*lowerState->iState()->atom()->cell()->Jb(nu()); // Changed!
    else pRU=0; // New!
  }
 
  // If requested - Prevent negative rate (due to level inversion) - New!
  if (preventNegative) pRU=MAX(pRU,0.0);

  // Testing!
  if (!isnumber(pRU)) cout<<"TransBB::RU: "<<lowerState->iState()->name()<<":"<<lowerState->getIndex()<<"->"<<upperState->getIndex()<<", cell: "<<upperState->iState()->atom()->cell()->index()<<", Warning: Invalid rate: "<<pRU<<"\n";

  return pRU;
}

// Get net (upward) radiative rate
double TransBB::RNet(bool twoPhoton) {
  return RU(false)-RD(twoPhoton); // Changed!
}

// Get net (upward) radiative rate number derivative
double TransBB::dRNetdN(BState* bState, bool twoPhoton, bool _dBetadN) {
  double _dRNetdN;

  // If non-zero Einstein A coefficient
  if (A()!=0) {

    // Calculate explicit derivative
    if (bState==lowerState) _dRNetdN=BU()*lowerState->iState()->atom()->cell()->Jb(nu())*beta();
    else if (bState==upperState) _dRNetdN=(A()-BD()*lowerState->iState()->atom()->cell()->Jb(nu()))*beta();
    else _dRNetdN=0;

    // If (implicit) beta term requested - Calculate and add
    if (_dBetadN and (bState==lowerState or bState==upperState)) _dRNetdN+=(RNet(false)/beta())*dBetadN(bState);
  }

  // If include two-photon processes - Calculate and add
  if (twoPhoton and bState==upperState) _dRNetdN+=A2();

  return _dRNetdN;
}

// Get downward collisional rate
double TransBB::CD(double T) {
  double _CD;
  
  // Set cell pointer
  Cell* cell=upperState->iState()->atom()->cell();

  // If update requested or temperature supplied
  if (pCD==update or (T!=noT and T!=cell->T())) {

    // Get temperature if not supplied
    if (T==noT) T=cell->T();

    // Calculate LTE number density ratio
    double xLTE=nBoundLTE(1.0,T,lowerState->E(),upperState->E(),lowerState->g(),upperState->g());

    // If non-zero lower-state number density - Calculate collisional rate using detailed balance
    if (lowerState->n()!=0) _CD=CU(T)*(upperState->n()/lowerState->n())*(1/xLTE);
    // Try this to avoid division also by xLTE=0 - Testing!
    //if (lowerState->n()!=0 and xLTE!=0) _CD=CU(T)*(upperState->n()/lowerState->n())*(1/xLTE);
    // Else - Calculate collisional rate assuming LTE
    else _CD=CU(T);

    // If requested - Update
    if (pCD==update) pCD=_CD;
  }
  // Else - Get collisional rate
  else _CD=pCD;

  if (!isnumber(_CD) or _CD<0) cout<<"TransBB::CD: "<<lowerState->iState()->name()<<":"<<lowerState->getIndex()<<"->"<<upperState->getIndex()<<", cell: "<<upperState->iState()->atom()->cell()->index()<<", Warning: Invalid rate: "<<_CD<<", nLower: "<<lowerState->n()<<", nUpper: "<<upperState->n()<<", T: "<<upperState->iState()->atom()->cell()->T()<<"\n";

  return _CD;
}

// Get upward collisional rate
double TransBB::CU(double T) {
  double _CU;

  // Set cell pointer
  Cell* cell=upperState->iState()->atom()->cell();
  
  // If update requested or temperature supplied
  if (pCU==update or (T!=noT and T!=cell->T())) {

    // Get temperature if not supplied
    if (T==noT) T=cell->T();

    // Calculate collisional rate
    _CU=sigmaColl(T)*lowerState->n();

    // If requested - Update
    if (pCU==update) pCU=_CU;
  }
  // Else - Get collisional rate
  else _CU=pCU;

  if (!isnumber(_CU) or _CU<0) cout<<"TransBB::CU: "<<lowerState->iState()->name()<<":"<<lowerState->getIndex()<<"->"<<upperState->getIndex()<<", cell: "<<upperState->iState()->atom()->cell()->index()<<", Warning: Invalid rate: "<<_CU<<"\n";

  return _CU;
}

// Get net (upward) collisional rate
double TransBB::CNet(double T) {
  return CU()-CD();
}

// Get updward non-thermal collisional rate
double TransBB::CNU() const {

  if (!isnumber(pCNU) or pCNU<0) cout<<"TransBB::CNU: cell: "<<upperState->iState()->atom()->cell()->index()<<", "<<lowerState->iState()->name()<<":"<<lowerState->getIndex()<<"->"<<upperState->getIndex()<<", Warning: Invalid rate: "<<pCNU<<"\n";

  return pCNU;
}

// Set updward non-thermal collisional rate
void TransBB::CNU(const double& _CNU) {

  pCNU=_CNU;
}

// Get non-thermal flag - New! - Removed!
/*bool TransBB::nonThermal() const {
  return pNonThermal;
  }*/

// Set non-thermal flag - New! - Removed!
/*void TransBB::nonThermal(const bool& _nonThermal) {
  pNonThermal=_nonThermal;
  }*/

// Get processes flags - New!
void TransBB::getProcesses() {
  pProcesses=data->processes;
}

// Process is active - New!
bool TransBB::active(const ProcessType& processType) {
  return pProcesses&processType;
}

// Activate process - New!
void TransBB::activate(const ProcessType& processType) {
  pProcesses=pProcesses|processType;
}

// De-activate process - New!
void TransBB::deActivate(const ProcessType& processType) {
  pProcesses=pProcesses^processType;
}

// Get total downward rate (per particle)
// Including stimulated emission
double TransBB::gammaD(bool collisional, bool twoPhoton) { // Changed!

  double pGammaD;
  if (A()!=0) pGammaD=(A()+BD()*lowerState->iState()->atom()->cell()->Jb(nu()))*beta(); // Changed!
  else pGammaD=0; // New!
  if (collisional) pGammaD+=CD()/upperState->n();
  if (twoPhoton) pGammaD+=A2(); // New!

  if (!isnumber(pGammaD) or pGammaD<0) cout<<"TransBB::gammaD: "<<upperState->iState()->name()<<":"<<upperState->getIndex()<<"->"<<lowerState->getIndex()<<", Warning: Invalid rate: "<<pGammaD<<"\n";
  //if (!isnumber(pGammaD) or pGammaD<0) cout<<"TransBB::gammaD: "<<upperState->iState()->name()<<":"<<upperState->getIndex()<<"->"<<lowerState->getIndex()<<", Warning: Invalid rate: "<<pGammaD<<", nu: "<<nu()<<", lambda: "<<(c/nu())/AA<<", nUpper: "<<upperState->n()<<", Jb: "<<upperState->iState()->atom()->cell()->Jb(nu())<<", A: "<<A()<<", BD: "<<BD()<<"\n";

  return pGammaD;
}

// Get total upward rate (per particle)
// Excluding stimulated emission
double TransBB::gammaU(bool collisional, bool nonThermal) {

  double pGammaU;
  if (A()!=0) pGammaU=BU()*beta()*lowerState->iState()->atom()->cell()->Jb(nu()); // Changed!
  else pGammaU=0; // New!
  if (collisional) pGammaU+=CU()/lowerState->n();
  if (nonThermal) pGammaU+=CNU()/lowerState->n();
  
  if (!isnumber(pGammaU) or pGammaU<0) cout<<"TransBB::gammaU: "<<lowerState->iState()->name()<<":"<<lowerState->getIndex()<<"->"<<upperState->getIndex()<<", Warning: Invalid rate: "<<pGammaU<<"\n";
  //if (!isnumber(pGammaU) or pGammaU<0) cout<<"TransBB::gammaU: "<<lowerState->iState()->name()<<":"<<lowerState->getIndex()<<"->"<<upperState->getIndex()<<", Warning: Invalid rate: "<<pGammaU<<", nu: "<<nu()<<", lambda: "<<(c/nu())/AA<<", nLower: "<<lowerState->n()<<", Jb: "<<lowerState->iState()->atom()->cell()->Jb(nu())<<", BU: "<<BU()<<", beta: "<<beta()<<"\n";
  
  return pGammaU;
}

// Get Einstein spontaneous de-excitation (emission) coefficient
double TransBB::A() const {
  return data->A;
}

// Get two-photon Einstein spontaneous de-excitation (emission) coefficient - New!
double TransBB::A2() const {
  return data->A2;
}

// Get Einstein stimulated de-excitation (emission) coefficient
double TransBB::BD() const {
  if (A()!=0) return (pow(c,2)/(2*h*pow(nu(),3)))*A(); // Changed!
  else return 0; // New!
}

// Get Einstein excitation (absorption) coefficient
double TransBB::BU() const {
  if (A()!=0) return BD()*(upperState->g()/lowerState->g()); // Changed!
  else return 0; // New!
}

// Get collisional cross-section
// Not a cross-section - rate per particle !!!
double TransBB::sigmaColl(double T) const {

  double sigma;
  if (T==noT) T=upperState->iState()->atom()->cell()->T();
  double ne=upperState->iState()->atom()->cell()->ne();
  double C=upperState->iState()->C();

  // If collisional strength available - Calculate cross-section
  if (data->omega!=NULL) sigma=BoundBound::sigmaColl(Xi(),lowerState->g(),(*data->omega)(T),T,ne);
  // Else - Calculate cross-section using Regermorter approximation
  else sigma=BoundBound::sigmaCollR(Xi(),lowerState->g(),upperState->g(),A(),C,T,ne);

  if (!isnumber(sigma) or sigma<0) cout<<"TransBB::sigmaColl: "<<lowerState->iState()->name()<<":"<<lowerState->getIndex()<<"->"<<upperState->iState()->name()<<":"<<upperState->getIndex()<<", Warning: Invalid rate: "<<sigma<<"\n";

  return sigma;
}

// Get optical depth
double TransBB::tau(bool preventNegative) const { // Changed!
  double _tau;

  // If A zero - return zero - New!
  if (A()==0) return 0;

  // Calculate Sobolev optical depth
  _tau=tauSobolev(Xi(),lowerState->n(),upperState->n(),A(),upperState->g(),lowerState->g(),1/lowerState->iState()->atom()->cell()->dvdr());

  // If requested - Prevent negative optical depth - New!
  if (preventNegative) _tau=MAX(_tau,0.0);

  return _tau;
}

// Get optical depth number derivative
double TransBB::dTaudN(BState* bState, bool preventNegative) const {
  double _dTaudN;

  // If A zero - return zero
  if (A()==0) return 0;

  // Add use of preventNegative here?

  // Calculate Sobolev optical depth number derivative
  if (bState==lowerState) _dTaudN=dTauSobolevdNl(Xi(),lowerState->n(),upperState->n(),A(),upperState->g(),lowerState->g(),1/lowerState->iState()->atom()->cell()->dvdr());
  else if (bState==upperState) _dTaudN=dTauSobolevdNu(Xi(),lowerState->n(),upperState->n(),A(),upperState->g(),lowerState->g(),1/lowerState->iState()->atom()->cell()->dvdr());
  else _dTaudN=0;

  return _dTaudN;
}

// Get escape fraction
double TransBB::beta() const {
  double _tau,_beta,tau_min=1e-6;

  // If A zero - return one - New!
  if (A()==0) return 1;

  _tau=tau();
  if (_tau>tau_min) _beta=(1-exp(-_tau))/_tau;
  else _beta=1;

  return _beta;
}

// Get escape fraction number derivative
double TransBB::dBetadN(BState* bState) const {
  double _tau,_dBetadN,tau_min=1e-6;

  // If A zero - return zero
  if (A()==0) return 0;

  if (bState==upperState or bState==lowerState) {
    _tau=tau();
    if (_tau>tau_min) _dBetadN=exp(-_tau)*(_tau-exp(-_tau)+1)/pow(_tau,2)*dTaudN(bState);
    else _dBetadN=0;
  }
  else _dBetadN=0;

  return _dBetadN;
}

// Draw emission type
int TransBB::etaType() {

  // If nonzero two-photon rate
  if (A2()!=0) {

    // Set cell pointer
    Cell* cell=upperState->iState()->atom()->cell();

    // Draw a random number between 0 and 1
    double z=randomize(0,1);

    // Calculate probability for two-photon emission
    double P=R2D()/(R1D()+R2D());

    // If boost - Adjust two-photon emission probability - New!
    if (cell->config->boost) P=P*eta2Boost()/etaBoost();

    // If drawn - Return bound-bound two-photon process
    if (z<=P) return pBoundBound|pTwoPhoton;
    // Else - Return bound-bound radiative (line) process
    else return pBoundBound|pRadiative;
  }
  // Else - Return bound-bound radiative (line) process
  else return pBoundBound|pRadiative;
}

// Get two-photon emissivity - New!
double TransBB::eta2(const double& _nu) {
  double pEta;

  // If nonzero two-photon rate - Calculate two-photon emissivity
  if (A2()!=0) pEta=(h*_nu/(4*pi))*(*data->eta2)(_nu)*A2()*upperState->n();
  // Else - Set emissivity to zero
  else pEta=0;

  if (!isnumber(pEta) or pEta<0) cout<<"TransBB::eta: "<<upperState->iState()->name()<<":"<<upperState->getIndex()<<"->"<<lowerState->iState()->name()<<":"<<lowerState->getIndex()<<", cell: "<<upperState->iState()->atom()->cell()->index()<<", Warning: Invalid emissivity: "<<pEta<<", nu: "<<_nu<<"\n";

  return pEta;
}

// Get two-photon emissivity cumulative distribution function - New!
double TransBB::eta2F(double _nu) {

  // If emissivity provided - Return cumulative distribution function
  if (data->eta2!=NULL) return data->eta2->F(_nu);
  // Else - Return one
  else return 1.0;
}

// Get two-photon emissivity cumulative distribution function (inverse version) - New!
double TransBB::eta2FInv(double F) {

  // If emissivity provided - Return cumulative distribution function (inverse version)
  if (data->eta2!=NULL) return data->eta2->FInv(F);
  // Else - Return transition frequency
  else return nu();
}

// Draw two-photon emission frequency - New!
double TransBB::eta2Nu() {

  // Get cell
  Cell* cell=upperState->iState()->atom()->cell();

  // If boost mode - Use boost version
  if (cell->config->boost) return eta2NuBoost();
  // Else (i.e. normal mode) - Use normal version
  else return eta2NuNormal();
}

// Draw two-photon emission frequency (intervall version) - New!
double TransBB::eta2Nu(double nuMin, double nuMax) {
  double _nu,z;

  // Draw random number betwen F(nuMin) and F(nuMax)
  z=randomize(eta2F(nuMin),eta2F(nuMax));

  // Get the corresponding frequency
  _nu=eta2FInv(z);

  return _nu;
}

// Draw two-photon emission frequency (normal version) - New!
double TransBB::eta2NuNormal() {
  double _nu,z;

  // Draw random number betwen 0 and 1
  z=randomize(0,1);

  // Get the corresponding frequency
  _nu=eta2FInv(z);

  return _nu;
}

// Draw two-photon emission frequency (boost version) - New!
double TransBB::eta2NuBoost() {
  double z,PTot=0;
  int i;

  // Get cell
  Cell* cell=upperState->iState()->atom()->cell();

  // Draw random number betwen 0 and 1
  z=randomize(0,1);

  // For each boost bin
  for (i=0;i<cell->nuBoost.size();i++) {

    // If bin drawn - Return emission frequency
    if (z<=PTot+eta2P[i]) return eta2Nu(cell->nuBoost[i].min,cell->nuBoost[i].max);
    // Else increase accumulated probability
    else PTot+=eta2P[i];
  }

  cout<<"TransBB::eta2NuBoost: Error: No frequency drawn\n";
  cout<<"TransBB::eta2NuBoost: z: "<<z<<", PTot: "<<PTot<<"\n";

  // This should never happen!
  return 0.0;
}

// Boost emissivity - New!
void TransBB::boostEta() {

  // Boost line emissivity
  boostEta1();

  // If non-zero rate - Boost two-photon emissivity
  if (A2()!=0) boostEta2();
}


// Boost line emissivity - New!
void TransBB::boostEta1() {
  int i,j;

  // Get cell
  Cell* cell=upperState->iState()->atom()->cell();

  // Set boost factor
  pEta1Boost=cell->boostFactor(nu());
}

// Boost two-photon emissivity - New!
void TransBB::boostEta2() {
  double PTot=0;
  int i;

  // Get cell
  Cell* cell=upperState->iState()->atom()->cell();

  // Resize emission probability vector
  eta2P.resize(cell->nuBoost.size());

  // For each boost bin
  for (i=0;i<cell->nuBoost.size();i++) {
    // Calculate emission probability
    if (cell->nuBoost[i].min<nu()) eta2P[i]=eta2F(MIN(cell->nuBoost[i].max,nu()))-eta2F(cell->nuBoost[i].min);
    else eta2P[i]=0;
    // Sum emission probability
    PTot+=eta2P[i];
  }

  // Normalize emission probability (to be on the safe side)
  for (i=0;i<cell->nuBoost.size();i++) eta2P[i]=eta2P[i]/PTot;

  // Reset boost factor
  pEta2Boost=0;

  // Calculate boost factor
  for (i=0;i<cell->nuBoost.size();i++) pEta2Boost+=eta2P[i]*cell->boost[i];

  // Adjust emission probability for each boost bin
  for (i=0;i<cell->nuBoost.size();i++) eta2P[i]=eta2P[i]*cell->boost[i]/eta2Boost();
}

// Boost opacity - New!
void TransBB::boostKappa() {
}

// Get emissivity boost factor - New!
double TransBB::etaBoost() {

  // If nonzero two-photon rate - Calculate and return emissivity boost factor
  if (A2()!=0) return (R1D()*eta1Boost()+R2D()*eta2Boost())/(R1D()+R2D());
  // Else - return line emissivity boost factor
  else return eta1Boost();
}

// Get line emissivity boost factor - New!
double TransBB::eta1Boost() { return pEta1Boost; }

// Get two-photon emissivity boost factor - New!
double TransBB::eta2Boost() { return pEta2Boost; }

// Get opacity boost factor - New!
double TransBB::kappaBoost() { return upperState->markovPBoost(); }

// Reset derived quantities (to be updated)
void TransBB::reset(int dependence) {
  if (dependence&(dpN)) pR1D=update; // New!
  if (dependence&(dpN)) pR2D=update; // New!
  if (dependence&(dpRad|dpN)) pRU=update;
  if (dependence&(dpT|dpN|dpNe)) pCD=update;
  if (dependence&(dpT|dpN|dpNe)) pCU=update;
}

// Export data
void TransBB::exportStat(YAML::Emitter& data, const double& tauLim) {

  // If optical depth above limit
  if (tau()>=tauLim) {

    data<<YAML::Flow<<YAML::BeginMap;

    // Export upper-state 
    data<<YAML::Key<<"upper"<<YAML::Value;
    data<<YAML::BeginSeq<<upperState->iState()->name()<<upperState->getIndex()<<YAML::EndSeq;

    // Export wavelength
    data<<YAML::Key<<"lambda"<<YAML::Value<<(c/nu())/AA;

    // Export optical depth
    data<<YAML::Key<<"tau"<<YAML::Value<<tau();

    // Export (collisional) heating
    data<<YAML::Key<<"H"<<YAML::Value<<CD()*Xi();

    // Export (collisional) cooling
    data<<YAML::Key<<"C"<<YAML::Value<<CU()*Xi();

    // Export radiative upward rate
    data<<YAML::Key<<"RU"<<YAML::Value<<RU();

    // Export (total) radiative downward rate - Changed!
    data<<YAML::Key<<"RD"<<YAML::Value<<RD(true);

    // Export radiative downward line rate - New!
    data<<YAML::Key<<"R1D"<<YAML::Value<<R1D();

    // Export radiative downward two-photon rate - New!
    data<<YAML::Key<<"R2D"<<YAML::Value<<R2D();

    // Export collisional upward rate
    data<<YAML::Key<<"CU"<<YAML::Value<<CU();

    // Export collisional downward rate
    data<<YAML::Key<<"CD"<<YAML::Value<<CD();

    // Export non-thermal (upward) rate
    data<<YAML::Key<<"CNU"<<YAML::Value<<CNU();

    data<<YAML::EndMap;
  }
  else cout<<"TransBB::exportStat: tau("<<tau()<<") < tauLim("<<tauLim<<")\n";
}

// Save state
void TransBB::saveState(YAML::Emitter& data) {

  data<<YAML::Flow<<YAML::BeginMap;
  
  // Export upper-state 
  data<<YAML::Key<<"upper"<<YAML::Value;
  data<<YAML::BeginSeq<<upperState->iState()->name()<<upperState->getIndex()<<YAML::EndSeq;
  
  // Export non-thermal (upward) rate
  data<<YAML::Key<<"CNU"<<YAML::Value<<CNU();
  
  data<<YAML::EndMap;
}

// Load state
void TransBB::loadState(const YAML::Node& data) {

  // Load non-thermal (upward rate)
  if (data["CNU"]) pCNU=data["CNU"].as<double>();
}

// Get memory allocated - New!
double TransBB::mem() const {
  double _mem=sizeof(*this);

  // Add memory allocated for eat2P vector
  _mem+=eta2P.size()*sizeof(double);

  return _mem;
}

// Constructors
LineList::LineList() { }
LineList::LineList(int size) : vector<TransBB*>(size) { }

// Find line with largest transition energy smaller than Xi
// (Assuming the list is sorted in reverse Xi)
LineIterator LineList::find(const double& Xi) const {
  vector<TransBB*>& line = *((vector<TransBB*>*)this);
  long iBegin,iMiddle,iEnd;

  if (line.size()!=0) {
    iBegin=0; iEnd=line.size()-1; iMiddle=(iBegin+iEnd)/2;
    if (Xi>=line[iBegin]->Xi()) {
      return line.begin();
    }
    else {
      if (Xi<line[iEnd]->Xi()) {
	return line.end();
      }
      else {
	while (iBegin!=iEnd) {
	  if (Xi>=line[iMiddle]->Xi()) {
	    iEnd=iMiddle;
	    iMiddle=(iBegin+iEnd)/2;
	  }
	  else {
	    iBegin=iMiddle+1;
	    iMiddle=(iBegin+iEnd)/2;
	  }
	}
	return line.begin()+iBegin;
      }
    }
  }

  return line.end();
}

// Sort line list in reverse Xi
void LineList::sort() {
  LineList list;
  TransBB* line;
  LineIterator lineIterator;
  int i;

  for (i=0;i<(*this).size();i++) {
    line=(*this)[i];
    lineIterator=list.find(line->Xi());
    list.insert(lineIterator,line);
  }
  (*this).swap(list);
  
  return;
}

// Constructors
EdgeList::EdgeList() { }
EdgeList::EdgeList(int size) : vector<TransBF*>(size) { }

// Find ionization edge with largest transition energy smaller than Xi
// (Assuming the list is sorted in reverse Xi)
EdgeIterator EdgeList::find(const double& Xi) const {
  vector<TransBF*>& edge = *((vector<TransBF*>*)this);
  long iBegin,iMiddle,iEnd;

  if (edge.size()!=0) {
    iBegin=0; iEnd=edge.size()-1; iMiddle=(iBegin+iEnd)/2;
    if (Xi>=edge[iBegin]->Xi()) {
      return edge.begin();
    }
    else {
      if (Xi<edge[iEnd]->Xi()) {
	return edge.end();
      }
      else {
	while (iBegin!=iEnd) {
	  if (Xi>=edge[iMiddle]->Xi()) {
	    iEnd=iMiddle;
	    iMiddle=(iBegin+iEnd)/2;
	  }
	  else {
	    iBegin=iMiddle+1;
	    iMiddle=(iBegin+iEnd)/2;
	  }
	}
	return edge.begin()+iBegin;
      }
    }
  }

  return edge.end();
}

// Sort ionization edge list in reverse Xi
void EdgeList::sort() {
  EdgeList list;
  TransBF* edge;
  EdgeIterator edgeIterator;
  int i;

  for (i=0;i<(*this).size();i++) {
    edge=(*this)[i];
    edgeIterator=list.find(edge->Xi());
    list.insert(edgeIterator,edge);
  }
  (*this).swap(list);
  
  return;
}

// Copy constructor - New!
// data is owned by the grid and needs to be updated elsewhere if necessay
// CTDReaction and CTUReaction are reset - Have to be updated elsewhere
TransBF::TransBF(const TransBF& trans) : data(trans.data), pRD(trans.pRD), pRU(trans.pRU), pdRDdN(trans.pdRDdN), pdRUdN(trans.pdRUdN), pCD(trans.pCD), pCU(trans.pCU), pCNU(trans.pCNU), pCTD(trans.pCTD), pCTU(trans.pCTU), pProcesses(trans.pProcesses), pCoolingSP(trans.pCoolingSP), pGammaOriginal(trans.pGammaOriginal), pGammaModified(trans.pGammaModified), pGammaOriginalOSP(trans.pGammaOriginalOSP), pGammaModifiedOSP(trans.pGammaModifiedOSP), pGammaSTOriginal(trans.pGammaSTOriginal), pGammaSTModified(trans.pGammaSTModified), pGammaSTOriginalOSP(trans.pGammaSTOriginalOSP), pGammaSTModifiedOSP(trans.pGammaSTModifiedOSP), pAlfaSPOriginal(trans.pAlfaSPOriginal), pAlfaSPModified(trans.pAlfaSPModified), pAlfaSPOriginalOSP(trans.pAlfaSPOriginalOSP), pAlfaSPModifiedOSP(trans.pAlfaSPModifiedOSP), pAlfaSTOriginal(trans.pAlfaSTOriginal), pAlfaSTModified(trans.pAlfaSTModified), pAlfaSTOriginalOSP(trans.pAlfaSTOriginalOSP), pAlfaSTModifiedOSP(trans.pAlfaSTModifiedOSP), pOnTheSpotNew(trans.pOnTheSpotNew), pROSP(trans.pROSP), pGOSP(trans.pGOSP), pRNet(trans.pRNet), pPhi(trans.pPhi), pNuCutOff(trans.pNuCutOff), pEtaBoost(trans.pEtaBoost), pKappaBoost(trans.pKappaBoost), etaP(trans.etaP), etaPTable(trans.etaPTable), CTDReaction(0), CTUReaction(0), pAlfaCorr(trans.pAlfaCorr), pAlfaAdd(trans.pAlfaAdd), pAlfaSPNoCorr(trans.pAlfaSPNoCorr), sigmaRad0(trans.sigmaRad0) {} ;

// Assigment operator - New!
// data is owned by the grid and needs to be updated elsewhere if necessay
// CTDReaction and CTUReaction are reset - Have to be updated elsewhere
TransBF& TransBF::operator=(const TransBF& trans) {
  if (&trans!=this) {
    data=trans.data;
    pRD=trans.pRD; 
    pRU=trans.pRU; 
    pdRDdN=trans.pdRDdN; 
    pdRUdN=trans.pdRUdN; 
    pCD=trans.pCD;
    pCU=trans.pCU;
    pCNU=trans.pCNU; 
    pCTD=trans.pCTD; 
    pCTU=trans.pCTU; 
    //pNonThermal=trans.pNonThermal; 
    pProcesses=trans.pProcesses;
    pCoolingSP=trans.pCoolingSP; 
    pGammaOriginal=trans.pGammaOriginal; 
    pGammaModified=trans.pGammaModified;
    pGammaOriginalOSP=trans.pGammaOriginalOSP;
    pGammaModifiedOSP=trans.pGammaModifiedOSP; 
    pGammaSTOriginal=trans.pGammaSTOriginal; 
    pGammaSTModified=trans.pGammaSTModified;
    pGammaSTOriginalOSP=trans.pGammaSTOriginalOSP; 
    pGammaSTModifiedOSP=trans.pGammaSTModifiedOSP; 
    pAlfaSPOriginal=trans.pAlfaSPOriginal;
    pAlfaSPModified=trans.pAlfaSPModified;
    pAlfaSPOriginalOSP=trans.pAlfaSPOriginalOSP;
    pAlfaSPModifiedOSP=trans.pAlfaSPModifiedOSP;
    pAlfaSTOriginal=trans.pAlfaSTOriginal;
    pAlfaSTModified=trans.pAlfaSTModified;
    pAlfaSTOriginalOSP=trans.pAlfaSTOriginalOSP;
    pAlfaSTModifiedOSP=trans.pAlfaSTModifiedOSP;
    pOnTheSpotNew=trans.pOnTheSpotNew;
    pROSP=trans.pROSP;
    pGOSP=trans.pGOSP;
    pRNet=trans.pRNet;
    pPhi=trans.pPhi;
    pNuCutOff=trans.pNuCutOff;
    pEtaBoost=trans.pEtaBoost;
    pKappaBoost=trans.pKappaBoost;
    etaP=trans.etaP; 
    etaPTable=trans.etaPTable;
    CTDReaction.resize(0);
    CTUReaction.resize(0);
    pAlfaCorr=trans.pAlfaCorr;
    pAlfaAdd=trans.pAlfaAdd;
    pAlfaSPNoCorr=trans.pAlfaSPNoCorr;
    sigmaRad0=trans.sigmaRad0; // New - Testing!
  }
}

// Destructor - Testing!
TransBF::~TransBF() {
  // Testing!
  //if (upperState!=NULL and lowerState!=NULL) if (upperState->iState()!=NULL and lowerState->iState()!=NULL) if (upperState->iState()->atom()!=NULL) if (upperState->iState()->atom()->cell()!=NULL) if (upperState->iState()->atom()->cell()->master()) cout<<"TransBF::~TransBF: cell: "<<upperState->iState()->atom()->cell()->index()<<", "<<upperState->iState()->name()<<":"<<upperState->getIndex()<<"->"<<lowerState->iState()->name()<<":"<<lowerState->getIndex()<<"\n";
}

// Get transition type
TransClass TransBF::type() const {
  return transClassBF;
}

// Get transition energy
double TransBF::Xi() const {
  return data->Xi;
}

// Get transition frequency
double TransBF::nu() const {
  return Xi()/h;
}

// Get detailed balance method - New!
DBMethod TransBF::dBMethod() const {
  return data->dBMethod;
}

// Get downward radiative rate
double TransBF::RD() {
 
  // If update requested
  if (pRD==update) {
    double ne=upperState->iState()->atom()->cell()->ne();
    pRD=alfaSP()*upperState->n()*ne;
  }

  if (!isnumber(pRD) or pRD<0) cout<<"TransBF::RD: "<<upperState->iState()->name()<<":"<<upperState->getIndex()<<"->"<<lowerState->iState()->name()<<":"<<lowerState->getIndex()<<", cell: "<<upperState->iState()->atom()->cell()->index()<<", Warning: Invalid rate: "<<pRD<<", nUpper: "<<upperState->n()<<", ne: "<<upperState->iState()->atom()->cell()->ne()<<", alfaSP: "<<alfaSP()<<"\n";

  return pRD;
}

// Get upward radiative rate
double TransBF::RU() {
  
  // If update requested
  if (pRU==update) {
    pRU=gammaST()*lowerState->n();
  }

  if (!isnumber(pRU) or pRU<0) cout<<"TransBF::RU: "<<lowerState->iState()->name()<<":"<<lowerState->getIndex()<<"->"<<upperState->iState()->name()<<":"<<upperState->getIndex()<<", cell: "<<upperState->iState()->atom()->cell()->index()<<", Warning: Invalid rate: "<<pRU<<", nLower: "<<lowerState->n()<<", gammaST: "<<gammaST()<<"\n";

  return pRU;
}

// Get downward radiative rate number derivative
// Note: Inculdes stimulated emission
double TransBF::dRDdN(BState* bState) {

  // If update requested
  if (pdRDdN==update) {
    double ne=upperState->iState()->atom()->cell()->ne();
    pdRDdN=dAlfaSTdN(bState)*upperState->n()*ne;
    //if (includeNe) pdRdN+=alfa()*upperState->n()*cell->dNedN(bState);
    if (bState==upperState) pdRDdN+=alfa()*ne;
  }

  if (!isnumber(pdRDdN)) cout<<"TransBF::dRDdN: "<<upperState->iState()->name()<<":"<<upperState->getIndex()<<"->"<<lowerState->iState()->name()<<":"<<lowerState->getIndex()<<", cell: "<<upperState->iState()->atom()->cell()->index()<<", Warning: Invalid rate number derivative: "<<pdRDdN<<"\n";

  return pdRDdN;
}

// Get upward radiative rate number derivative
// Note: not corrected for stimulated emission
double TransBF::dRUdN(BState* bState) {
  
  // If update requested
  if (pdRUdN==update) {
    pdRUdN=dGammadN(bState)*lowerState->n();
    if (bState==lowerState) pdRUdN+=gamma();
  }

  if (!isnumber(pdRUdN)) cout<<"TransBF::dRUdN: "<<lowerState->iState()->name()<<":"<<lowerState->getIndex()<<"->"<<upperState->iState()->name()<<":"<<upperState->getIndex()<<", cell: "<<upperState->iState()->atom()->cell()->index()<<", Warning: Invalid rate number derivative: "<<pdRUdN<<"\n";

  return pdRUdN;
}

// Get net (upward) radiative rate number derivative - New!
double TransBF::dRNetdN(BState* bState, bool includeOSP, bool dNedN, bool expdN) {
  double _dRNetdN=0;
  Cell* cell=upperState->iState()->atom()->cell();

  // If requested
  if (expdN) {
    // Add explicit terms with respect to upper/lower state
    if (bState==lowerState) _dRNetdN+=gamma(false,includeOSP);
    if (bState==upperState) _dRNetdN-=(alfa(noT,false,includeOSP)+alfaAdd())*cell->ne();
  }

  // If requested - Add implicit electron number density terms
  if (dNedN) _dRNetdN-=(alfa(noT,false,includeOSP)+alfaAdd())*upperState->n()*cell->dNedN(bState);
  
  return _dRNetdN;
}

// Get downward collisional rate
double TransBF::CD(double T) {
  double _CD;

  // Set cell pointer
  Cell* cell=upperState->iState()->atom()->cell();

  // If update requested or temperature supplied
  if (pCD==update or (T!=noT and T!=cell->T())) {

    // Get temperature if not supplied
    if (T==noT) T=cell->T();

    // Calculate LTE number density ratio
    double xLTE=nIonLTE(1.0,cell->ne(),T,Xi(),lowerState->g(),upperState->g());
    
    // If non-zero lower-state number density - Calculate collisional rate using detailed balance
    if (lowerState->n()!=0) _CD=CU(T)*(upperState->n()/lowerState->n())*(1/xLTE);
    // Try this to avoid division also by xLTE=0 - Testing!
    //if (lowerState->n()!=0 and xLTE!=0) _CD=CU(T)*(upperState->n()/lowerState->n())*(1/xLTE);
    // Else - Calculate collisional rate assuming LTE
    else _CD=CU(T);

    // Calculate collisional rate - Testing !
    //_CD=cD(T)*cell->ne()*upperState->n();

    // If requested - Update
    if (pCD==update) pCD=_CD;
  }
  // Else - Get collisional rate
  else _CD=pCD;

  if (!isnumber(_CD) or _CD<0) cout<<"TransBF::CD: "<<upperState->iState()->name()<<":"<<upperState->getIndex()<<"->"<<lowerState->iState()->name()<<":"<<lowerState->getIndex()<<", cell: "<<upperState->iState()->atom()->cell()->index()<<", Warning: Invalid rate: "<<_CD<<"\n";

  // Testing!!!
  //if (upperState->iState()->atom()->cell()->index()==36 and fabs(upperState->iState()->C())>3)  cout<<"TransBF::CD: "<<upperState->iState()->name()<<":"<<upperState->getIndex()<<"->"<<lowerState->iState()->name()<<":"<<lowerState->getIndex()<<", cell: "<<upperState->iState()->atom()->cell()->index()<<", CD: "<<_CD<<", xLTE: "<<nIonLTE(1.0,cell->ne(),cell->T(),Xi(),lowerState->g(),upperState->g())<<", nLower: "<<lowerState->n()<<", nUpper: "<<upperState->n()<<"\n";

  return _CD;
}

// Get upward collisional rate
double TransBF::CU(double T) {
  double _CU;

  // Set cell pointer
  Cell* cell=upperState->iState()->atom()->cell();  

  // If update requested or temperature supplied
  if (pCU==update or (T!=noT and T!=cell->T())) {
    
    // Get temperature if not supplied
    if (T==noT) T=cell->T();

    // Calculate collisional rate
    _CU=sigmaColl(T)*lowerState->n();

    // Testing!!!
    //if (upperState->iState()->atom()->cell()->index()==36 and fabs(upperState->iState()->C())>3)  cout<<"TransBF::CU(1): "<<upperState->iState()->name()<<":"<<upperState->getIndex()<<"->"<<lowerState->iState()->name()<<":"<<lowerState->getIndex()<<", cell: "<<upperState->iState()->atom()->cell()->index()<<", CU: "<<_CU<<", sigmaColl: "<<sigmaColl(T)<<", nLower: "<<lowerState->n()<<"\n";

    // Calculate collisional rate - Testing !
    //_CU=cU(T)*lowerState->n();

    // If requested - Update
    if (pCU==update) pCU=_CU;
  }
  // Else - Get collisional rate
  else _CU=pCU;

  if (!isnumber(_CU) or _CU<0) cout<<"TransBF::CU: "<<lowerState->iState()->name()<<":"<<lowerState->getIndex()<<"->"<<upperState->iState()->name()<<":"<<upperState->getIndex()<<", cell: "<<upperState->iState()->atom()->cell()->index()<<", Warning: Invalid rate: "<<_CU<<"\n";

  // Testing!!!
  //if (upperState->iState()->atom()->cell()->index()==36 and fabs(upperState->iState()->C())>3)  cout<<"TransBF::CU(2): "<<upperState->iState()->name()<<":"<<upperState->getIndex()<<"->"<<lowerState->iState()->name()<<":"<<lowerState->getIndex()<<", cell: "<<upperState->iState()->atom()->cell()->index()<<", CU: "<<_CU<<"\n";

  return _CU;
}

// Get net (upward) collisional rate
double TransBF::CNet(double T) {
  return CU(T)-CD(T);
}

// Get net (upward) collisional rate number derivative - New!
double TransBF::dCNetdN(BState* bState, double T, bool dNedN, bool expdN) {
  double _dCNetdN=0;
  Cell* cell=upperState->iState()->atom()->cell();

  // If requested
  if (expdN) {
    // Add explicit terms with respect to upper/lower state
    if (bState==lowerState) _dCNetdN+=CU(T)/lowerState->n();
    if (bState==upperState) _dCNetdN+=CD(T)/upperState->n(); // Should be -=, right?!?
  }
    
  // If requested- Add electron number density terms
  // Note: Downward collisional rates quadratic in ne.
  if (dNedN) _dCNetdN+=(CU(T)-2*CD(T))*(cell->dNedN(bState)/cell->ne());

  return _dCNetdN;
}

// Get updward non-thermal collisional rate
double TransBF::CNU() const {

  if (pCNU<0) cout<<"TransBF::CNU: "<<lowerState->iState()->name()<<":"<<lowerState->getIndex()<<"->"<<upperState->iState()->name()<<":"<<upperState->getIndex()<<", Warning: Negative rate: "<<pCNU<<"\n";

  return pCNU;
}

// Set updward non-thermal collisional rate
void TransBF::CNU(const double& _CNU) {

  pCNU=_CNU;
}

// Get downward charge-transfer rate - New!
double TransBF::CTD() {
  int i;

  // If update requested
  if (pCTD==update) {
    pCTD=0;
    //for (i=0;i<CTDReaction.size();i++) pCTD+=CTDReaction[i]->rate();
    for (i=0;i<CTDReaction.size();i++) pCTD+=CTDReaction[i]->forwardRate();
    for (i=0;i<CTUReaction.size();i++) pCTD+=CTUReaction[i]->backwardRate();
    if (lowerState->iState()->name().compare("HeI")==0 and lowerState->getIndex()==0 and upperState->iState()->name().compare("HeII")==0 and upperState->getIndex()==0) {
    }
  }

  if (!isnumber(pCTD) or pCTD<0) cout<<"TransBF::CTD: "<<upperState->iState()->name()<<":"<<upperState->getIndex()<<"->"<<lowerState->iState()->name()<<":"<<lowerState->getIndex()<<", cell: "<<upperState->iState()->atom()->cell()->index()<<", Warning: Invalid rate: "<<pCTD<<"\n";

  return pCTD;
}

// Get upward charge-transfer rate - New!
double TransBF::CTU() {
  int i;

  // If update requested
  if (pCTU==update) {
    pCTU=0;
    //for (i=0;i<CTUReaction.size();i++) pCTU+=CTUReaction[i]->rate();
    for (i=0;i<CTUReaction.size();i++) pCTU+=CTUReaction[i]->forwardRate();
    for (i=0;i<CTDReaction.size();i++) pCTU+=CTDReaction[i]->backwardRate();
    if (lowerState->iState()->name().compare("HeI")==0 and lowerState->getIndex()==0 and upperState->iState()->name().compare("HeII")==0 and upperState->getIndex()==0) {
    }
  }

  if (!isnumber(pCTU) or pCTU<0) cout<<"TransBF::CTU: "<<lowerState->iState()->name()<<":"<<lowerState->getIndex()<<"->"<<upperState->iState()->name()<<":"<<upperState->getIndex()<<", cell: "<<upperState->iState()->atom()->cell()->index()<<", Warning: Invalid rate: "<<pCTU<<"\n";

  return pCTU;
}

// Get net (upward) charge-transfer rate
double TransBF::CTNet() {
  return CTU()-CTD();
}

// Get net (upward) charge-transfer rate number derivative - New!
double TransBF::dCTNetdN(BState* bState) {
  double _dCTNetdN=0;
  BState *lowerState,*upperState;
  Cell* cell=upperState->iState()->atom()->cell();
  int i;

  // Get number derivative with respect to upper/lower state
  if (bState==lowerState) _dCTNetdN+=CTU()/lowerState->n();
  if (bState==upperState) _dCTNetdN-=CTD()/upperState->n();

  // For each upward CT transition
  for (i=0;i<CTUReaction.size();i++) {
    // Set pointer to upper state of corresponding downward CT transition
    upperState=CTUReaction[i]->downTrans->upperState;
    // Add contribution to number derivative with respect to this state
    if (bState==upperState) _dCTNetdN+=CTUReaction[i]->rate()/upperState->n();
    // Needs update for reaction direction!!!
  }

  // For each downward CT reaction
  for (i=0;i<CTDReaction.size();i++) {
    // Set pointer to lower state of corresponding upward bound-free transition
    lowerState=CTDReaction[i]->upTrans->lowerState;
    // Add contribution to number derivative with respect to this state
    if (bState==lowerState) _dCTNetdN-=CTDReaction[i]->rate()/lowerState->n();
    // Needs update for reaction direction!!!
  }

  return _dCTNetdN;
}

// Get processes flags - New!
void TransBF::getProcesses() {
  pProcesses=data->processes;
}

// Process is active - New!
bool TransBF::active(const ProcessType& processType) {
  return pProcesses&processType;
}

// Activate process - New!
void TransBF::activate(const ProcessType& processType) {  
  pProcesses=pProcesses|processType;
}

// De-activate process - New!
void TransBF::deActivate(const ProcessType& processType) {
  pProcesses=pProcesses^processType;
}

// Get total downward rate (per particle)
// Including stimulated emission
double TransBF::gammaD(bool collisional, bool chargeTransfer, bool includeOSP) { // Changed!

  double pGammaD;
  double ne=upperState->iState()->atom()->cell()->ne();
  pGammaD=alfa(noT,false,includeOSP)*ne;
  pGammaD+=alfaAdd()*ne; // New!
  if (collisional) pGammaD+=CD()/upperState->n();
  if (chargeTransfer) pGammaD+=CTD()/upperState->n(); // New!

  if (!isnumber(pGammaD) or pGammaD<0) cout<<"TransBF::gammaD: "<<upperState->iState()->name()<<":"<<upperState->getIndex()<<"->"<<lowerState->iState()->name()<<":"<<lowerState->getIndex()<<", cell: "<<upperState->iState()->atom()->cell()->index()<<", Warning: Invalid rate: "<<pGammaD<<", alfa: "<<alfa(noT,false,includeOSP)<<", alfaSP: "<<alfaSP(noT,false,includeOSP)<<", alfaST: "<<alfaST(noT,false,includeOSP)<<", ne: "<<ne<<", CD: "<<CD()<<", upperState->n: "<<upperState->n()<<"\n";

  // Testing!!!
  //if (upperState->iState()->atom()->cell()->index()==36 and upperState->iState()->atom()->name().compare("Al")==0) cout<<"TransBF::gammaD: "<<upperState->iState()->name()<<":"<<upperState->getIndex()<<"->"<<lowerState->iState()->name()<<":"<<lowerState->getIndex()<<", cell: "<<upperState->iState()->atom()->cell()->index()<<", pGammaD: "<<pGammaD<<", alfa: "<<alfa(noT,false,includeOSP)<<", alfaSP: "<<alfaSP(noT,false,includeOSP)<<", alfaST: "<<alfaST(noT,false,includeOSP)<<", ne: "<<ne<<", CD: "<<CD()<<", upperState->n: "<<upperState->n()<<"\n";
  //cout<<"TransBF::gammaD: "<<upperState->iState()->name()<<":"<<upperState->getIndex()<<"->"<<lowerState->iState()->name()<<":"<<lowerState->getIndex()<<", cell: "<<upperState->iState()->atom()->cell()->index()<<", pGammaD: "<<pGammaD<<"\n";

  return pGammaD;
}

// Get total upward rate (per particle)
// Excluding stimulated emission
double TransBF::gammaU(bool collisional, bool nonThermal, bool chargeTransfer, bool includeOSP) { // Changed!

  double pGammaU;
  pGammaU=gamma(false,includeOSP);
  if (collisional) pGammaU+=CU()/lowerState->n();
  if (nonThermal) pGammaU+=CNU()/lowerState->n();
  if (chargeTransfer) pGammaU+=CTU()/lowerState->n(); // New!

  if (!isnumber(pGammaU) or pGammaU<0) cout<<"TransBF::gammaU: "<<lowerState->iState()->name()<<":"<<lowerState->getIndex()<<"->"<<upperState->iState()->name()<<":"<<upperState->getIndex()<<", Warning: Invalid rate: "<<pGammaU<<"\n";

  // Testing!!!
  //if (upperState->iState()->atom()->cell()->index()==36 and upperState->iState()->atom()->name().compare("Al")==0) cout<<"TransBF::gammaU: "<<lowerState->iState()->name()<<":"<<lowerState->getIndex()<<"->"<<upperState->iState()->name()<<":"<<upperState->getIndex()<<", cell: "<<upperState->iState()->atom()->cell()->index()<<", pGammaU: "<<pGammaU<<", gamma: "<<gamma(false,includeOSP)<<", CU: "<<CU()<<", lowerState->n: "<<lowerState->n()<<"\n";

  return pGammaU;
}

// Get total downward rate (per particle) number derivative
// Including stimulated emission
double TransBF::dGammaDdN(BState* bState) {

  double pdGammaDdN;
  double ne=upperState->iState()->atom()->cell()->ne();

  pdGammaDdN=dAlfaSTdN(bState)*ne;

  if (!isnumber(pdGammaDdN)) cout<<"TransBF::dGammaDdN: "<<upperState->iState()->name()<<":"<<upperState->getIndex()<<"->"<<lowerState->iState()->name()<<":"<<lowerState->getIndex()<<", cell: "<<upperState->iState()->atom()->cell()->index()<<", Warning: Invalid rate: "<<pdGammaDdN<<"\n";

  return pdGammaDdN;
}

// Get total upward rate (per particle) number derivative
// Excluding stimulated emission
double TransBF::dGammaUdN(BState* bState) {

  double pdGammaUdN;
  
  pdGammaUdN=dGammadN(bState);

  if (!isnumber(pdGammaUdN)) cout<<"TransBF::gammadGammaUdN: "<<lowerState->iState()->name()<<":"<<lowerState->getIndex()<<"->"<<upperState->iState()->name()<<":"<<upperState->getIndex()<<", Warning: Invalid rate: "<<pdGammaUdN<<"\n";

  return pdGammaUdN;
}

// Get radiative cross-section
//double TransBF::sigmaRad(const double& _nu, bool correct) const { // Changed!
double TransBF::sigmaRad(const double& _nu, bool correct) { // Changed!

  double sigma;

  // If below edge frequency - Return zero
  if (_nu<nu()) return 0.0;

  // If radiative cross-section available - Get it
  if (data->sigmaRad!=NULL) sigma=(*data->sigmaRad)(_nu);
  // Else - Calculate radiative cross-section using hydrogenic approximation
  else {
    //sigma=BoundFree::sigmaRadH(lowerState->iState()->C(),lowerState->iState()->Xi0(),lowerState->E(),_nu);
    // Testing!!!
    if (sigmaRad0==update) sigmaRad0=BoundFree::sigmaRadH(lowerState->iState()->C(),lowerState->iState()->Xi0(),lowerState->E(),nu());
    sigma=sigmaRad0*pow(_nu/nu(),-3);
  }

  if (!isnumber(sigma) or sigma<0) cout<<"TransBF::sigmaRad: "<<lowerState->iState()->name()<<":"<<lowerState->getIndex()<<"->"<<upperState->iState()->name()<<":"<<upperState->getIndex()<<", Warning: Invalid cross-section: "<<sigma<<"\n";

  if (correct) sigma*=sigmaCorr(_nu); // Changed
  return sigma;
}

// Get radiative cross-section
//double TransBF::sigmaRad(FrequencyBinIterator _nu, EdgeType edgeType, bool correct) const { // Changed!
double TransBF::sigmaRad(FrequencyBinIterator _nu, EdgeType edgeType, bool correct) { // Changed!
  return sigmaRad(nuEdge(_nu,edgeType),correct); // Changed!
}

// Get radiative cross-section (corrected for stimulated emission)
double TransBF::sigmaRadST(const double& _nu) {

  // Testing
  if (alfaSimple()) return 0;

  double T=upperState->iState()->atom()->cell()->T();
  double ne=upperState->iState()->atom()->cell()->ne();

  //return sigmaRad(_nu)*(1-(upperState->n()/lowerState->n())*phi(T)*ne*exp(-(h*_nu)/(k*T)));
  // Testing - Avoid negative cross-section (similar to Lucy 2003)
  // Hmm, is this really correct?
  return sigmaRad(_nu)*(1-MIN((upperState->n()/lowerState->n())*phi(T)*ne,1.0)*exp(-(h*_nu)/(k*T)));
  //return sigmaRad(_nu)*(1-MIN((upperState->n()/lowerState->n())*phi(T)*ne,1.0)*bMFactor(_nu));
}

// Get radiative cross-section (corrected for stimulated emission)
double TransBF::sigmaRadST(FrequencyBinIterator _nu, EdgeType edgeType) {

  // Testing
  if (alfaSimple()) return 0;

  Cell* cell=upperState->iState()->atom()->cell();

  double T=cell->T();
  double ne=cell->ne();

  // Testing - Avoid negative cross-section (similar to Lucy 2003)
  // Hmm, is this really correct?
  return sigmaRad(_nu,edgeType)*(1-MIN((upperState->n()/lowerState->n())*phi(T)*ne,1.0)*cell->bMFactor(_nu,edgeType));
}

// Get collisional cross-section
// Not a cross-section - rate per particle !!!
//double TransBF::sigmaColl(double T) const {
double TransBF::sigmaColl(double T) {
  
  double sigma;
  if (T==noT) T=upperState->iState()->atom()->cell()->T();
  double ne=upperState->iState()->atom()->cell()->ne();
  double C=lowerState->iState()->C();

  // If collisional cross-section available - Get it
  //if (data->sigmaColl!=NULL) sigma=(*data->sigmaColl)(T)*exp(-Xi()/(k*T))*ne;
  if (data->sigmaColl!=NULL) sigma=(*data->sigmaColl)(T)*expLB(-Xi()/(k*T))*ne; // Changed! - Testing!
  // Else - If collisional strength available - Calculate cross-section
  else if (data->omega!=NULL) sigma=BoundFree::sigmaColl(Xi(),lowerState->g(),(*data->omega)(T),T,ne);
  // Else - Calculate collisional cross-section using Seaton approximation
  else sigma=BoundFree::sigmaCollS(C,Xi(),sigmaRad(nu()),T,ne);

  if (sigma<0) cout<<"TransBF::sigmaColl: "<<lowerState->iState()->name()<<":"<<lowerState->getIndex()<<"->"<<upperState->iState()->name()<<":"<<upperState->getIndex()<<", Warning: Negative rate: "<<sigma<<"\n";

  // Testing!!!
  //if (upperState->iState()->atom()->cell()->index()==36 and fabs(upperState->iState()->C())>3) cout<<"TransBF::sigmaColl: "<<upperState->iState()->name()<<":"<<upperState->getIndex()<<"->"<<lowerState->iState()->name()<<":"<<lowerState->getIndex()<<", cell: "<<upperState->iState()->atom()->cell()->index()<<", sigma: "<<sigma<<", sigmaColl: "<<BoundFree::sigmaCollS(C,Xi(),sigmaRad(nu()),T,ne)<<", sigmaRad: "<<sigmaRad(nu())<<"\n";

  return sigma;
}

// Calculate photo-ionization rate coefficient (per frequency and mean intensity)
//double TransBF::gammaNu(const double& _nu, bool modified) const {
double TransBF::gammaNu(const double& _nu, bool modified) {

  // Testing
  if (alfaSimple()) return 0;

  double pGammaNu,E=h*_nu;
  
  pGammaNu=4*pi*sigmaRad(_nu);
  if (modified) pGammaNu=pGammaNu*MAX(1/Xi()-1/E,0);
  else pGammaNu=pGammaNu/E;
  
  if (pGammaNu<0) cout<<"TransBF::gammaNu: "<<lowerState->iState()->name()<<":"<<lowerState->getIndex()<<"->"<<upperState->iState()->name()<<":"<<upperState->getIndex()<<", Warning: Negative rate coefficient: "<<pGammaNu<<", modified: "<<modified<<", nu: "<<_nu<<", nu0: "<<nu()<<", Xi: "<<Xi()<<", E: "<<E<<", (1/Xi-1/E): "<<MAX(1/Xi()-1/E,0)<<", sigmaRad(nu): "<<sigmaRad(_nu)<<"\n";

  return pGammaNu;
}

// Calculate photo-ionization rate coefficient (per frequency and mean intensity)
//double TransBF::gammaNu(FrequencyBinIterator _nu, EdgeType edgeType, bool modified) const {
double TransBF::gammaNu(FrequencyBinIterator _nu, EdgeType edgeType, bool modified) {

  // Testing
  if (alfaSimple()) return 0;

  double pGammaNu,E=h*nuEdge(_nu,edgeType);
  
  pGammaNu=4*pi*sigmaRad(_nu,edgeType);
  if (modified) pGammaNu=pGammaNu*MAX(1/Xi()-1/E,0);
  else pGammaNu=pGammaNu/E;
  
  if (pGammaNu<0) cout<<"TransBF::gammaNu: "<<lowerState->iState()->name()<<":"<<lowerState->getIndex()<<"->"<<upperState->iState()->name()<<":"<<upperState->getIndex()<<", Warning: Negative rate coefficient: "<<pGammaNu<<", modified: "<<modified<<", nu: "<<nuEdge(_nu,edgeType)<<", nu0: "<<nu()<<", Xi: "<<Xi()<<", E: "<<E<<", (1/Xi-1/E): "<<MAX(1/Xi()-1/E,0)<<", sigmaRad(nu): "<<sigmaRad(_nu,edgeType)<<"\n";

  return pGammaNu;
}

// Photo-ionization rate coefficient integrand
class GammaIntegrand : public JIntegrand {
public:
  //const TransBF* trans;
  TransBF* trans;
  const bool modified;
  //GammaIntegrand(const TransBF* _trans, const bool& _modified) : trans(_trans), modified(_modified) {};
  GammaIntegrand(TransBF* _trans, const bool& _modified) : trans(_trans), modified(_modified) {};
  // The virtual prefix is missing, right?
  double operator()(double nu) const { return trans->gammaNu(nu,modified); }
  double operator()(FrequencyBinIterator nu, EdgeType edgeType=noEdge) const { return trans->gammaNu(nu,edgeType,modified); }
};

// Get photo-ionization coefficient
double TransBF::gamma(bool modified, bool includeOSP) {

  // Testing
  if (alfaSimple()) return 0;

  // If on-the-spot approximation - Return zero for ground-state transitions
  if (upperState->iState()->atom()->onTheSpot() and lowerState->E()==0) return 0.0;

  double _gamma=0,_gammaOSP=0;
  double relTol=1e-2,absTol=0;
  bool useIterator=true;
  int nIter=0;
  GammaIntegrand integrand(this,modified);

  // If update requested
  if ((pGammaOriginal==update and (not modified)) or (pGammaModified==update and modified)) {

    // Calculate photo-ionization coefficient    
    //if (upperState->iState()->atom()->cell()->index()==36 and upperState->iState()->atom()->name().compare("Fe")==0) cout<<"TransBF::gamma: cell: "<<upperState->iState()->atom()->cell()->index()<<", "<<upperState->iState()->name()<<":"<<upperState->getIndex()<<"->"<<lowerState->iState()->name()<<":"<<lowerState->getIndex()<<", call JIntegrate\n";
    _gamma=upperState->iState()->atom()->cell()->JIntegrate(integrand,nu(),nuInf,nIter,relTol,absTol,false,useIterator,false);

    if (modified) pGammaModified=_gamma;
    else pGammaOriginal=_gamma;
  }
  else
    if (modified) _gamma=pGammaModified;
    else _gamma=pGammaOriginal;

  // If include OSP region
  if (includeOSP) {

    // If update requested
    if ((pGammaOriginalOSP==update and (not modified)) or (pGammaModifiedOSP==update and modified)) {

      // Adjust absolute tolerence
      absTol=MAX(absTol,fabs(_gamma)*relTol);
      
      // Calculate photo-ionization coefficient
      //if (upperState->iState()->atom()->cell()->index()==36) cout<<"TransBF::gamma: cell: "<<upperState->iState()->atom()->cell()->index()<<", "<<upperState->iState()->name()<<":"<<upperState->getIndex()<<"->"<<lowerState->iState()->name()<<":"<<lowerState->getIndex()<<", call JIntegrate(OSP)\n";
      _gammaOSP=upperState->iState()->atom()->cell()->JIntegrate(integrand,nu(),nuInf,nIter,relTol,absTol,true,useIterator,true);
      // Testing!!!
      //stringstream test;
      //if (upperState->iState()->atom()->cell()->index()==36 and upperState->iState()->name().compare("FeIII")==0) { test.str(""); test<<"cell: "<<upperState->iState()->atom()->cell()->index()<<", "<<upperState->iState()->name()<<":"<<upperState->getIndex()<<"->"<<lowerState->iState()->name()<<":"<<lowerState->getIndex(); }
      //_gammaOSP=upperState->iState()->atom()->cell()->JIntegrate(integrand,nu(),nuInf,nIter,relTol,absTol,true,useIterator,true,test.str());
      
      if (modified) pGammaModifiedOSP=_gammaOSP;
      else pGammaOriginalOSP=_gammaOSP;
    }
    else
      if (modified) _gammaOSP=pGammaModifiedOSP;
      else _gammaOSP=pGammaOriginalOSP;
  }

  if (includeOSP) return _gamma+_gammaOSP;
  else return _gamma;
}

// Photo-ionization rate coefficient number derivative integrand
class DGammadNIntegrand : public NuIntegrand {
public:
  //const TransBF* trans;
  TransBF* trans;
  const bool modified;
  //DGammadNIntegrand(const TransBF* _trans, const bool& _modified) : trans(_trans), modified(_modified) {};
  DGammadNIntegrand(TransBF* _trans, const bool& _modified) : trans(_trans), modified(_modified) {};
  // The virtual prefix is missing, right?
  double operator()(double nu) const { return trans->gammaNu(nu,modified); }
  double operator()(FrequencyBinIterator nu, EdgeType edgeType=noEdge) const { return 0.0; };
};

// Get photo-ionization coefficient number derivative
double TransBF::dGammadN(BState* bState, bool modified) {

  // Testing
  if (alfaSimple()) return 0;

  // Set cell pointer
  Cell* cell=upperState->iState()->atom()->cell();
    
  double pdGammadN;
  double tol=1e-2;
  double nuMin=nu(), nuMax=nuInf;
  // Testing
  int nIter=0;

  DGammadNIntegrand integrand(this,modified);

  pdGammadN=cell->dJdNIntegrate(integrand,bState,nuMin,nuMax,nIter,tol);

  return pdGammadN;
}

// Calculate (corrected) photo-ionization rate coeffiecient (per frequency and mean intensity)
double TransBF::gammaSTNu(const double& _nu, bool modified) {

  // Testing
  if (alfaSimple()) return 0;

  double pGammaSTNu,E=h*_nu;

  pGammaSTNu=4*pi*sigmaRadST(_nu);
  if (modified) pGammaSTNu=pGammaSTNu*MAX(1/Xi()-1/E,0);
  else pGammaSTNu=pGammaSTNu/E;

  if (!isnumber(pGammaSTNu) or pGammaSTNu<0) cout<<"TransBF::gammaSTNu: "<<lowerState->iState()->name()<<":"<<lowerState->getIndex()<<"->"<<upperState->iState()->name()<<":"<<upperState->getIndex()<<", Warning: Invalid rate coefficient: "<<pGammaSTNu<<"\n";

  return pGammaSTNu;
}

// Calculate (corrected) photo-ionization rate coeffiecient (per frequency and mean intensity)
double TransBF::gammaSTNu(FrequencyBinIterator _nu, EdgeType edgeType, bool modified) {

  // Testing
  if (alfaSimple()) return 0;

  double pGammaSTNu,E=h*nuEdge(_nu,edgeType);

  pGammaSTNu=4*pi*sigmaRadST(_nu,edgeType);
  if (modified) pGammaSTNu=pGammaSTNu*MAX(1/Xi()-1/E,0);
  else pGammaSTNu=pGammaSTNu/E;

  if (!isnumber(pGammaSTNu) or pGammaSTNu<0) cout<<"TransBF::gammaSTNu: "<<lowerState->iState()->name()<<":"<<lowerState->getIndex()<<"->"<<upperState->iState()->name()<<":"<<upperState->getIndex()<<", Warning: Invalid rate coefficient: "<<pGammaSTNu<<"\n";

  return pGammaSTNu;
}

// (Corrected) photo-ionization rate coefficient integrand
class GammaSTIntegrand : public JIntegrand {
public:
  TransBF* trans;
  const bool modified;
  GammaSTIntegrand(TransBF* _trans, const bool& _modified) : trans(_trans), modified(_modified) {};
  // The virtual prefix is missing, right?
  double operator()(double nu) const { return trans->gammaSTNu(nu,modified); }
  double operator()(FrequencyBinIterator nu, EdgeType edgeType=noEdge) const { return trans->gammaSTNu(nu,edgeType,modified); };
};

// Get (corrected) photo-ionization coefficient
double TransBF::gammaST(bool modified, bool includeOSP) {

  // Testing
  if (alfaSimple()) return 0;

  // If on-the-spot approximation - Return zero for ground-state transitions
  if (upperState->iState()->atom()->onTheSpot() and lowerState->E()==0) return 0.0;

  double _gammaST=0,_gammaSTOSP=0;
  double relTol=1e-2,absTol=0;
  bool useIterator=true;
  int nIter=0;
  GammaSTIntegrand integrand(this,modified);

  // If update requested
  if ((pGammaSTOriginal==update and (not modified)) or (pGammaSTModified==update and modified)) {

    // Calculate (corrected) photo-ionization coefficient
    _gammaST=upperState->iState()->atom()->cell()->JIntegrate(integrand,nu(),nuInf,nIter,relTol,absTol,false,useIterator,false);

    if (modified) pGammaSTModified=_gammaST;
    else pGammaSTOriginal=_gammaST;
  }
  else
    if (modified) _gammaST=pGammaSTModified;
    else _gammaST=pGammaSTOriginal;

  // If include OSP region 
  if (includeOSP) {

    // If update requested
    if ((pGammaSTOriginalOSP==update and (not modified)) or (pGammaSTModifiedOSP==update and modified)) {

      // Adjust absolute tolerence
      absTol=MAX(absTol,fabs(_gammaST)*relTol);
      
      // Calculate (corrected) photo-ionization coefficient
      _gammaSTOSP=upperState->iState()->atom()->cell()->JIntegrate(integrand,nu(),nuInf,nIter,relTol,absTol,true,useIterator,true);
      
      if (modified) pGammaSTModifiedOSP=_gammaSTOSP;
      else  pGammaSTOriginalOSP=_gammaSTOSP;
    }
    else
       if (modified) _gammaSTOSP=pGammaSTModifiedOSP;
       else _gammaSTOSP=pGammaSTOriginalOSP;
  }

  if (includeOSP) return _gammaST+_gammaSTOSP;
  else return _gammaST;
}

// Get (corrected) photo-ionization coefficient ratio
double TransBF::gammaSTRatio() {

  double gammaSTOriginal,gammaSTModified,pGammaSTRatio;
    
  gammaSTOriginal=gammaST(false);
  gammaSTModified=gammaST(true);

  // Calculate (corrected) photo-ionization coefficient ratio
  //if (gammaSTOriginal!=0 and gammaSTModified!=0) pGammaSTRatio=gammaSTOriginal/gammaSTModified;
  if (gammaSTOriginal!=0 and gammaSTModified!=0) pGammaSTRatio=gammaSTOriginal/(gammaSTOriginal+gammaSTModified);
  // Hack to take care of the (rare) case of zero coefficients !!!
  else pGammaSTRatio=1.0;

  return pGammaSTRatio;
}

// Get (spontaneous) recombination rate coefficient (per frequency)
double TransBF::alfaSPNu(const double& _nu, double T, bool modified, bool correct) { // Changed!

  // Testing
  if (alfaSimple()) return 0;

  double pAlfaSPNu,E=h*_nu;
  if (T==noT) T=upperState->iState()->atom()->cell()->T();

  pAlfaSPNu=4*pi*phi(T)*sigmaRad(_nu,correct)*(2*h*pow(_nu,3)*pow(c,-2))*exp(-E/(k*T)); // Changed!
  if (modified) pAlfaSPNu=pAlfaSPNu*MAX(1/Xi()-1/E,0);
  else pAlfaSPNu=pAlfaSPNu/E;

  if (correct) pAlfaSPNu*=alfaSPCorr(_nu); // Changed!
  return pAlfaSPNu;
}

// Get (spontaneous) recombination rate coefficient (per frequency)
double TransBF::alfaSPNu(FrequencyBinIterator _nu, EdgeType edgeType, double T, bool modified, bool correct) { // Changed!

  // Testing
  if (alfaSimple()) return 0;

  Cell* cell=upperState->iState()->atom()->cell();

  // Need to take edgeType into account here !!!
  double pAlfaSPNu,E=h*nuEdge(_nu,edgeType);
  if (T==noT) T=cell->T();

  // Testing - T==cell->T() assumed!!!
  pAlfaSPNu=4*pi*phi(T)*sigmaRad(_nu,edgeType,correct)*cell->bBFactor(_nu,edgeType)*cell->bMFactor(_nu,edgeType); // Changed!
  if (modified) pAlfaSPNu=pAlfaSPNu*MAX(1/Xi()-1/E,0);
  else pAlfaSPNu=pAlfaSPNu/E;

  if (correct) pAlfaSPNu*=alfaSPCorr(nuEdge(_nu,edgeType)); // Changed!
  return pAlfaSPNu;
}

// (Spontaneous) recombination rate coefficient integrand
class AlfaSPIntegrand {
public:
  TransBF* trans;
  const double T;
  const bool modified;
  const bool correct;
  AlfaSPIntegrand(TransBF* _trans, const double& _T, const bool& _modified, const bool& _correct) : trans(_trans), T(_T), modified(_modified), correct(_correct) {}; // Changed!
  //double operator()(double nu) const { return trans->alfaSPNu(nu,T,modified); }
  double operator()(double nu, EdgeType edgeType=noEdge) const { return trans->alfaSPNu(nuEdge(nu,edgeType),T,modified,correct); }
  double operator()(FrequencyBinIterator nu, EdgeType edgeType=noEdge) const { return trans->alfaSPNu(nu,edgeType,T,modified,correct); }
};

// Get (spontaneous) recombination rate coefficient
// Should be updated along the same lines as alfaST!
double TransBF::alfaSP(double T, bool modified, bool includeOSP, bool correct) { // Changed!
  double pAlfaSP=0; // Moved!

  // If on-the-spot approximation - Return zero for ground-state transitions
  if (upperState->iState()->atom()->onTheSpot() and lowerState->E()==0) return 0.0;

  // Set cell pointer
  Cell* cell=upperState->iState()->atom()->cell();

  // Testing
  if (alfaSimple() and (not cell->config->alfaSimpleEmit)) return 0;

  // If update requested or temperature supplied
  if ((pAlfaSPOriginalOSP==update and (not modified) and includeOSP and correct) or (pAlfaSPModifiedOSP==update and modified and includeOSP and correct) or (pAlfaSPOriginal==update and (not modified) and (not includeOSP) and correct) or (pAlfaSPModified==update and modified and (not includeOSP) and correct) or (pAlfaSPNoCorr==update and (not modified) and includeOSP and (not correct)) or ((modified or (not includeOSP)) and (not correct)) or (T!=noT and T!=cell->T())) { // Changed!
  
    // Testing - Should be fine with the new definition of modified
    double tol=1e-2;
    int nIterMin=1,nIterMax=4;
    double nuMin=nu(),nuMax=1e2*nu(),nuMinOSP=cell->onTheSpotNu().min;
    if (not includeOSP) nuMax=MIN(nuMax,nuMinOSP);
    int nIter=0;
    if (T==noT) T=cell->T();
    //bool useIterator=true; // Removed!
    bool useIterator=(T==cell->T())?true:false; // Changed!
    AlfaSPIntegrand integrand(this,T,modified,correct); // Changed!

    // If any frequency range left to integrate over
    if (nuMax>nuMin) {
      // If radiative cross-section available - New!
      if (data->sigmaRad!=NULL or (not cell->config->alfaSPH)) {
	// Calculate (spontaneous) recombination rate coefficient - Changed!
	if (cell->nuOld.size()==0) pAlfaSP=Integrator<AlfaSPIntegrand>::romberg(integrand,nuMin,nuMax,nIter,tol);
	else {
	  pAlfaSP=BFIntegrator<AlfaSPIntegrand>::trapezoidal(integrand,nuMin,nuMax,cell->nuOld,nIter,tol,0.0,useIterator,"",nIterMin,nIterMax);
	  if (nIter>nIterMax) cout<<"TransBF::alfaSP: "<<upperState->iState()->name()<<":"<<upperState->getIndex()<<"->"<<lowerState->iState()->name()<<":"<<lowerState->getIndex()<<", cell: "<<upperState->iState()->atom()->cell()->index()<<", Warning: Maximum number of iterations exceeded\n";
	}
      }
      // Else - Calculate hydrogenic recombination rate - New!
      else pAlfaSP=alfaSPH(T,nuMin,nuMax,modified,correct);
    }
    // Else - Set (spontaneous) recombination rate coefficient to zero
    else pAlfaSP=0;

    // If requested - Update
    if (correct)
      if (modified) {
	if (includeOSP and pAlfaSPModifiedOSP==update) pAlfaSPModifiedOSP=pAlfaSP; // Changed!
	else if ((not includeOSP) and pAlfaSPModified==update) pAlfaSPModified=pAlfaSP; // Changed!
      }
      else {
	if (includeOSP and pAlfaSPOriginalOSP==update) pAlfaSPOriginalOSP=pAlfaSP; // Changed!
	else if ((not includeOSP) and pAlfaSPOriginal==update) pAlfaSPOriginal=pAlfaSP; // Changed!
      }
    else if ((not modified) and includeOSP) pAlfaSPNoCorr=pAlfaSP;
  }
  // Else - Get (spontaneous) recombination rate - Changed!
  else 
    if (correct)
      if (modified) {
	if (includeOSP) pAlfaSP=pAlfaSPModifiedOSP; // Changed!
	else pAlfaSP=pAlfaSPModified; // Changed!
      }
      else {
	if (includeOSP) pAlfaSP=pAlfaSPOriginalOSP; // Changed!
	else pAlfaSP=pAlfaSPOriginal; // Changed!
      }
    else if ((not modified) and includeOSP) pAlfaSP=pAlfaSPNoCorr;

  return pAlfaSP;
}

// Get (spontaneous) hydrogenic recombination rate coefficient
double TransBF::alfaSPH(double T, double nuMin, double nuMax, bool modified, bool correct) {
  double pAlfaSP=0,alfaSPCorrNormal=1.0,alfaSPCorrOSP=1.0,nuMinOSP;

  // Set cell pointer
  //Cell* cell=upperState->iState()->atom()->cell();

  // If correct - Get correction
  //if (correct) alfaSPCorrNormal=alfaCorr(nuMin);

  // If force detailed balance in OSP region and this region partially overlaps with the integration region
  /*if (cell->config->forceDBalanceOSP and nuMax>cell->onTheSpotNu().min and nuMin<cell->onTheSpotNu().min) {
    nuMinOSP=cell->onTheSpotNu().min;
    // If correct - Get correction in OSP region
    if (correct) alfaSPCorrOSP=alfaSPCorr(nuMinOSP);
    // Calculate hydrogenic recombintation rate coefficient redwards the OSP region
    pAlfaSP=BoundFree::alfaSPH(lowerState->iState()->C(),lowerState->iState()->Xi0(),lowerState->E(),lowerState->g(),upperState->g(),T,nuMin,nuMinOSP,modified)*alfaSPCorrNormal;
    // Calculate and add hydorgenic recombination coefficient in the OSP region
    pAlfaSP+=BoundFree::alfaSPH(lowerState->iState()->C(),lowerState->iState()->Xi0(),lowerState->E(),lowerState->g(),upperState->g(),T,nuMinOSP,nuMax,modified)*alfaSPCorrOSP;
  }
  // Else - Calculate hydrogenic recombination rate coefficent
  else pAlfaSP=BoundFree::alfaSPH(lowerState->iState()->C(),lowerState->iState()->Xi0(),lowerState->E(),lowerState->g(),upperState->g(),T,nuMin,nuMax,modified)*alfaSPCorrNormal;*/

  // Calculate hydrogenic recombination rate coefficent
  pAlfaSP=BoundFree::alfaSPH(lowerState->iState()->C(),lowerState->iState()->Xi0(),lowerState->E(),lowerState->g(),upperState->g(),T,nuMin,nuMax,modified)*alfaSPCorrNormal;

  if (pAlfaSP<0) cout<<"TransBF::alfaSPH: "<<lowerState->iState()->name()<<":"<<lowerState->getIndex()<<"->"<<upperState->iState()->name()<<":"<<upperState->getIndex()<<", Warning: Negative rate: "<<pAlfaSP<<" - Zetting to zero: \n";

  // If correct - Apply correction
  if (correct) pAlfaSP*=alfaCorr();
  
  return pAlfaSP;
}

// Get (stimulated) recombination rate coefficient (per frequency and mean intensity)
double TransBF::alfaSTNu(const double& _nu, double T, bool modified) {

  // Testing
  if (alfaSimple()) return 0;

  double pAlfaSTNu,E=h*_nu;
  if (T==noT) T=upperState->iState()->atom()->cell()->T();

  pAlfaSTNu=4*pi*phi(T)*sigmaRad(_nu)*exp(-E/(k*T));
  if (modified) pAlfaSTNu=pAlfaSTNu*MAX(1/Xi()-1/E,0);
  else pAlfaSTNu=pAlfaSTNu/E;

  if (pAlfaSTNu<0) cout<<"TransBF::alfaSTNu: "<<upperState->iState()->name()<<":"<<upperState->getIndex()<<"->"<<lowerState->iState()->name()<<":"<<lowerState->getIndex()<<", Warning: Negative rate coefficient: "<<pAlfaSTNu<<", modified: "<<modified<<", nu: "<<_nu<<", nu0: "<<nu()<<", Xi: "<<Xi()<<", E: "<<E<<", (1/Xi-1/E): "<<MAX(1/Xi()-1/E,0)<<", sigmaRad(nu): "<<sigmaRad(_nu)<<"\n";

  return pAlfaSTNu;
}

// Get (stimulated) recombination rate coefficient (per frequency and mean intensity)
double TransBF::alfaSTNu(FrequencyBinIterator _nu, EdgeType edgeType, double T, bool modified) {

  // Testing
  if (alfaSimple()) return 0;

  Cell* cell=upperState->iState()->atom()->cell();

  double pAlfaSTNu,E=h*nuEdge(_nu,edgeType);
  if (T==noT) T=cell->T();

  // Testing - T=cell->T() assumed!!!
  pAlfaSTNu=4*pi*phi(T)*sigmaRad(_nu,edgeType)*cell->bMFactor(_nu,edgeType);
  if (modified) pAlfaSTNu=pAlfaSTNu*MAX(1/Xi()-1/E,0);
  else pAlfaSTNu=pAlfaSTNu/E;

  if (pAlfaSTNu<0) cout<<"TransBF::alfaSTNu: "<<upperState->iState()->name()<<":"<<upperState->getIndex()<<"->"<<lowerState->iState()->name()<<":"<<lowerState->getIndex()<<", Warning: Negative rate coefficient: "<<pAlfaSTNu<<", modified: "<<modified<<", nu: "<<nuEdge(_nu,edgeType)<<", nu0: "<<nu()<<", Xi: "<<Xi()<<", E: "<<E<<", (1/Xi-1/E): "<<MAX(1/Xi()-1/E,0)<<", sigmaRad(nu): "<<sigmaRad(_nu,edgeType)<<"\n";

  return pAlfaSTNu;
}

// (Stimulated) recombination rate coefficient integrand
class AlfaSTIntegrand : public JIntegrand {
public:
  TransBF* trans;
  const double T;
  const bool modified;
  AlfaSTIntegrand(TransBF* _trans, const double& _T, const bool& _modified) : trans(_trans), T(_T), modified(_modified) {};
  // The virtual prefix is missing, right?
  double operator()(double nu) const { return trans->alfaSTNu(nu,T,modified); }
  virtual double operator()(FrequencyBinIterator nu, EdgeType edgeType) const { return trans->alfaSTNu(nu,edgeType,T,modified); }
};

// Get (stimulated) recombination rate coefficient
double TransBF::alfaST(double T, bool modified, bool includeOSP) {

  // Testing
  if (alfaSimple()) return 0;

  // If on-the-spot approximation - Return zero for ground-state transitions
  if (upperState->iState()->atom()->onTheSpot() and lowerState->E()==0) return 0.0;

  // Set cell pointer
  Cell* cell=upperState->iState()->atom()->cell();

  double _alfaST=0,_alfaSTOSP=0;
  double relTol=1e-2,absTol=0;
  int nIter=0;
  if (T==noT) T=cell->T();
  bool useIterator=(T==cell->T())?true:false;
  AlfaSTIntegrand integrand(this,T,modified);

  // If update requested or temperature supplied
  if ((pAlfaSTOriginal==update and (not modified)) or (pAlfaSTModified==update and modified) or (T!=cell->T())) {

    // Calculate (stimulated) recombination rate coefficient
    //if (cell->index()==36 and upperState->iState()->atom()->name().compare("Fe")==0) cout<<"TransBF::alfaST: cell: "<<cell->index()<<", "<<upperState->iState()->name()<<":"<<upperState->getIndex()<<"->"<<lowerState->iState()->name()<<":"<<lowerState->getIndex()<<", call JIntegrate(1)\n";
    _alfaST=cell->JIntegrate(integrand,nu(),nuInf,nIter,relTol,absTol,false,useIterator,false);

    // If requested - Update
    if (pAlfaSTOriginal==update and (not modified)) pAlfaSTOriginal=_alfaST;
    else if (pAlfaSTModified==update and modified) pAlfaSTModified=_alfaST;
  }
  // Else
  else 
    // Get (stimulated) recombination rate coefficient
    if (not modified) _alfaST=pAlfaSTOriginal;
    else _alfaST=pAlfaSTModified;

  // If include OSP region
  if (includeOSP) {

    // If update requested or temperature supplied
    if ((pAlfaSTOriginalOSP==update and (not modified)) or (pAlfaSTModifiedOSP==update and modified) or (T!=cell->T())) {

      // Adjust absolute tolerence
      absTol=MAX(absTol,fabs(_alfaST)*relTol);
      
      // Calculate (stimulated) recombination rate coefficient
      //if (cell->index()==36) cout<<"TransBF::alfaST: cell: "<<cell->index()<<", "<<upperState->iState()->name()<<":"<<upperState->getIndex()<<"->"<<lowerState->iState()->name()<<":"<<lowerState->getIndex()<<", call JIntegrate(OSP)\n";
      _alfaSTOSP=cell->JIntegrate(integrand,nu(),nuInf,nIter,relTol,absTol,true,useIterator,true);
      
      // If requested - Update
      if (pAlfaSTOriginalOSP==update and (not modified)) pAlfaSTOriginalOSP=_alfaSTOSP;
      else if (pAlfaSTModifiedOSP==update and modified) pAlfaSTModifiedOSP=_alfaSTOSP;
    }
    // Else
    else
      // Get (stimulated) recombination rate coefficient      
      if (not modified) _alfaSTOSP=pAlfaSTOriginalOSP;
      else _alfaSTOSP=pAlfaSTModifiedOSP;
  }
   
  if (includeOSP) return _alfaST+_alfaSTOSP;
  else return _alfaST;
}

// (Spontaneous) recombination rate coefficient number derivative integrand
class DAlfaSTdNIntegrand : public NuIntegrand {
public:
  TransBF* trans;
  const bool modified;
  DAlfaSTdNIntegrand(TransBF* _trans, const bool& _modified) : trans(_trans), modified(_modified) {};
  // The virtual prefix is missing, right?
  double operator()(double nu) const { return trans->alfaSTNu(nu,noT,modified); }
  double operator()(FrequencyBinIterator nu, EdgeType edgeType=noEdge) const { return 0.0; };
};

// Get (spontaneous) recombination rate coefficient number derivative
double TransBF::dAlfaSTdN(BState* bState, bool modified) {

  // Testing
  if (alfaSimple()) return 0;

  // Set cell pointer
  Cell* cell=upperState->iState()->atom()->cell();
 
  double pdAlfaSTdN;
  double tol=1e-2;
  double nuMin=nu(),nuMax=nuInf;
  // Testing
  int nIter=0;

  DAlfaSTdNIntegrand integrand(this,modified);

  pdAlfaSTdN=cell->dJdNIntegrate(integrand,bState,nuMin,nuMax,nIter,tol);

  return pdAlfaSTdN;
}

// Get (total) recombination rate coefficient (per frequency)
double TransBF::alfaNu(const double& _nu, double T, bool modified) {

  return alfaSPNu(_nu,T,modified)+alfaSTNu(_nu,T,modified);
}

// Get (total) recombination rate coefficient
double TransBF::alfa(double T, bool modified, bool includeOSP, bool correct) {

  return alfaSP(T,modified,includeOSP,correct)+alfaST(T,modified,includeOSP);
  // Testing!!!
  //return alfaSP(T,modified,includeOSP,correct);
}

// Get opacity
double TransBF::kappa(const double& _nu, bool cutOff) {
  double pKappa;

  // Testing
  if (alfaSimple()) return 0;

  // If cut-off - Return zero if above cut-off frequency
  if (cutOff) if (_nu>nuCutOff()) return 0.0;

  // Return zero if below edge frequency
  if (_nu<nu()) return 0.0;  

  // If on-the-spot approximation - Return infinity bluewards edge for ground-state transitions
  if (upperState->iState()->atom()->onTheSpot() and lowerState->E()==0 and _nu>=nu()) return 1e99;

  double rho=upperState->iState()->atom()->cell()->rho();

  pKappa=sigmaRadST(_nu)*lowerState->n()/rho;

  if (!isnumber(pKappa) or pKappa<0) cout<<"TransBF::kappa: "<<upperState->iState()->name()<<":"<<upperState->getIndex()<<"->"<<lowerState->iState()->name()<<":"<<lowerState->getIndex()<<", cell: "<<upperState->iState()->atom()->cell()->index()<<", Warning: Invalid opacity: "<<pKappa<<", nu: "<<_nu<<"\n";

  return pKappa;
}

// Get opacity
double TransBF::kappa(FrequencyBinIterator _nu, EdgeType edgeType, bool cutOff) {
  double pKappa;
  double nuVal=nuEdge(_nu,edgeType);

  // Testing
  if (alfaSimple()) return 0;

  // If cut-off - Return zero if above cut-off frequency
  if (cutOff) if (nuVal>nuCutOff()) return 0.0;

  // Return zero if below edge frequency
  if (nuVal<nu()) return 0.0;

  // If on-the-spot approximation - Return infinity bluewards edge for ground-state transitions
  if (upperState->iState()->atom()->onTheSpot() and lowerState->E()==0 and nuVal>=nu()) return 1e99;

  double rho=upperState->iState()->atom()->cell()->rho();

  pKappa=sigmaRadST(_nu,edgeType)*lowerState->n()/rho;

  if (!isnumber(pKappa) or pKappa<0) cout<<"TransBF::kappa: "<<upperState->iState()->name()<<":"<<upperState->getIndex()<<"->"<<lowerState->iState()->name()<<":"<<lowerState->getIndex()<<", cell: "<<upperState->iState()->atom()->cell()->index()<<", Warning: Invalid opacity: "<<pKappa<<", nu: "<<nuVal<<"\n";

  return pKappa;
}

// Get opacity number derivative
double TransBF::dKappadN(const double& _nu, BState* bState) {
  double pdKappadN;

  // Testing
  if (alfaSimple()) return 0;

  // If on-the-spot approximation - Return zero bluewards edge for ground-state transitions
  if (upperState->iState()->atom()->onTheSpot() and lowerState->E()==0 and _nu>=nu()) return 0.0;

  double T=upperState->iState()->atom()->cell()->T();
  double rho=upperState->iState()->atom()->cell()->rho();
  double ne=upperState->iState()->atom()->cell()->ne();

  // Note that this derivative is not entirely consistent with sigmaRadST used in kappa
  if (bState==lowerState) pdKappadN=sigmaRad(_nu)/rho;
  else pdKappadN=-sigmaRad(_nu)*phi(T)*ne*exp(-(h*_nu)/(k*T));

  if (!isnumber(pdKappadN)) cout<<"TransBF::dKappadN: "<<upperState->iState()->name()<<":"<<upperState->getIndex()<<"->"<<lowerState->iState()->name()<<":"<<lowerState->getIndex()<<", cell: "<<upperState->iState()->atom()->cell()->index()<<", Warning: Invalid opacity number derivative: "<<pdKappadN<<", nu: "<<_nu<<"\n";

  return pdKappadN;
}

// Get emissivity
double TransBF::eta(const double& _nu, bool cutOff) {
  double pEta;

  // Testing
  if (alfaSimple()) return 0;

  // If cut-off - Return zero if above cut-off frequency
  if (cutOff) if (_nu>nuCutOff()) return 0.0;

  // Return zero if below edge frequency
  if (_nu<nu()) return 0.0;  

  double ne=upperState->iState()->atom()->cell()->ne();

  pEta=alfaSPNu(_nu)*upperState->n()*ne*(h*_nu)/(4*pi);

  if (!isnumber(pEta) or pEta<0) cout<<"TransBF::eta: "<<upperState->iState()->name()<<":"<<upperState->getIndex()<<"->"<<lowerState->iState()->name()<<":"<<lowerState->getIndex()<<", cell: "<<upperState->iState()->atom()->cell()->index()<<", Warning: Invalid emissivity: "<<pEta<<", nu: "<<_nu<<"\n";

  return pEta;
}

// Get emissivity
double TransBF::eta(FrequencyBinIterator _nu, EdgeType edgeType, bool cutOff) {
  double pEta;
  double nuVal=nuEdge(_nu,edgeType);

  // Testing
  if (alfaSimple()) return 0;

  // If cut-off - Return zero if above cut-off frequency
  if (cutOff) if (nuVal>nuCutOff()) return 0.0;

  // Return zero if below edge frequency
  if (nuVal<nu()) return 0.0;  

  double ne=upperState->iState()->atom()->cell()->ne();

  pEta=alfaSPNu(_nu,edgeType)*upperState->n()*ne*(h*nuVal)/(4*pi);

  if (!isnumber(pEta) or pEta<0) cout<<"TransBF::eta: "<<upperState->iState()->name()<<":"<<upperState->getIndex()<<"->"<<lowerState->iState()->name()<<":"<<lowerState->getIndex()<<", cell: "<<upperState->iState()->atom()->cell()->index()<<", Warning: Invalid opacity: "<<pEta<<", nu: "<<nuVal<<"\n";

  return pEta;
}

// Get emissivity number derivative
double TransBF::dEtadN(const double& _nu, BState* bState) {
  double pdEtadN;

  // Testing
  if (alfaSimple()) return 0;

  double ne=upperState->iState()->atom()->cell()->ne();

  if (bState==upperState) pdEtadN=alfaSPNu(_nu)*ne*(h*_nu)/(4*pi);
  // Testing - Include derivative of ne
  //if (bState==upperState) pdEtadN=alfaSPNu(_nu)*(h*_nu)*(ne+upperState->n())/(4*pi);
  else pdEtadN=0.0;
  // Maybe also include the derivative of ne here ?
  //else pdEtadN=alfaSPNu(_nu)*(h*_nu)*bState->n()/(4*pi);

  if (!isnumber(pdEtadN) or pdEtadN<0) cout<<"TransBF::dEtadN: "<<upperState->iState()->name()<<":"<<upperState->getIndex()<<"->"<<lowerState->iState()->name()<<":"<<lowerState->getIndex()<<", cell: "<<upperState->iState()->atom()->cell()->index()<<", Warning: Invalid opacity number derivative: "<<pdEtadN<<", nu: "<<_nu<<"\n";

  return pdEtadN;
}

// Draw emission frequency
double TransBF::etaNu() {

  // Get cell
  Cell* cell=upperState->iState()->atom()->cell();

  // If boost mode - Use boost version
  if (cell->config->boost) return etaNuBoost();
  // Else (i.e. normal mode) - Use normal version
  else return etaNuNormal();
}

// Draw emission frequency (intervall version) - New!
// Here we follow Lucy (2003) and assume a v^-3 powerlaw - Fix !
double TransBF::etaNu(double nuMin, double nuMax) {

  nuMin=MAX(nuMin,nu());
  if (nuMin>nuMax) { cout<<"TransBF::etaNu: "<<upperState->iState()->name()<<":"<<upperState->getIndex()<<"->"<<lowerState->iState()->name()<<":"<<lowerState->getIndex()<<", cell: "<<upperState->iState()->atom()->cell()->index()<<", Warning: Invalid frequency range\n"; return nuMin; }

  double T=upperState->iState()->atom()->cell()->T();
  double z=randomize(0,1);
  double pNu=nuMin*(1-(k*T/(h*nuMin))*log(1-(1-exp(-h*(nuMax-nuMin)/(k*T)))*z));

  return pNu;
}

// Draw emission frequency (normal version) - New!
// Here we follow Lucy (2003) and assume a v^-3 powerlaw - Fix !
double TransBF::etaNuNormal() {
  double pNu; // New!

  // If no emission probability table available - New! - Testing!
  if (etaPTable.size()==0) {
    // Draw emission frequency based on v^-3 powerlaw
    double T=upperState->iState()->atom()->cell()->T();
    double z=randomize(0,1);
    pNu=nu()*(1-(k*T/Xi())*log(z));
  }
  // Else - Draw emission frequency based on table - New! - Testing!
  else pNu=etaNuTable(nu(),nuInf);

  return pNu;
}

// Draw emission frequency (boost version) - New!
double TransBF::etaNuBoost() {
  double z,PTot=0;
  int i;

  // Get cell
  Cell* cell=upperState->iState()->atom()->cell();

  // Draw random number betwen 0 and 1
  z=randomize(0,1);

  //cout<<"TransBF::etaNuBoost: "<<upperState->iState()->name()<<":"<<upperState->getIndex()<<"->"<<lowerState->iState()->name()<<":"<<lowerState->getIndex()<<", cell: "<<cell->index()<<", etaP.size: "<<etaP.size()<<", nuBoost.size: "<<cell->nuBoost.size()<<"\n";

  // For each boost bin
  for (i=0;i<cell->nuBoost.size();i++) {

    // If bin drawn - Return emission frequency
    if (z<=PTot+etaP[i]) return etaNu(cell->nuBoost[i].min,cell->nuBoost[i].max);
    // Else increase accumulated probability
    else PTot+=etaP[i];
  }

  cout<<"TransBF::etaNuBoost: "<<upperState->iState()->name()<<":"<<upperState->getIndex()<<"->"<<lowerState->iState()->name()<<":"<<lowerState->getIndex()<<", cell: "<<cell->index()<<", Error: No frequency drawn\n";
  cout<<"TransBF::etaNuBoost: "<<upperState->iState()->name()<<":"<<upperState->getIndex()<<"->"<<lowerState->iState()->name()<<":"<<lowerState->getIndex()<<", cell: "<<cell->index()<<", z: "<<z<<", PTot: "<<PTot<<", i: "<<i<<", etaP.size: "<<etaP.size()<<", nuBoost.size: "<<cell->nuBoost.size()<<", type: "<<type()<<"\n";
  
  // This should never happen!
  return 0.0;
}

// Draw emission frequency (table version) - New!
double TransBF::etaNuTable(double nuMin, double nuMax) {
  FrequencyBinIterator nuBin,nuEdgeBin,nuMinBin,nuMaxBin;
  double etaPTot,PTot,z;
  
  // Get cell
  Cell* cell=upperState->iState()->atom()->cell();
  
  // Get edge bin
  nuEdgeBin=cell->nuOld.find(nu());
  
  // Get interval bins
  nuMinBin=cell->nuOld.find(nuMin);
  nuMaxBin=cell->nuOld.find(nuMax)-1;
  
  // If invalid frequency intervall - Log error and return zero
  if (nuMinBin<nuEdgeBin or nuMinBin>=nuMaxBin) { cout<<"TransBF::etaNuTable: Error: Invalid frequency intervall\n"; return 0.0; }

  // Calculate emission probability in the intervall
  if (nuMinBin>nuEdgeBin) etaPTot=etaPTable[nuMaxBin-nuEdgeBin]-etaPTable[nuMinBin-nuEdgeBin-1];
  else etaPTot=etaPTable[nuMaxBin-nuEdgeBin];

  // Draw random number betwen 0 and 1
  z=randomize(0,1);

  // For each frequency bin in the intervall
  for (nuBin=nuMinBin;nuBin<=nuMaxBin;nuBin++) {

    // Calculate cumulative emission probability
    PTot=etaPTable[nuBin-nuEdgeBin]/etaPTot;

    // If drawn - Return random frequency in the bin
    if (z<=PTot) return randomize(nuBin->min,nuBin->max);
  }

  cout<<"TransBF::etaNuTable: Error: No frequency drawn\n";
  cout<<"TransBF::etaNuTable: z: "<<z<<", PTot: "<<PTot<<"\n";

  // This should never happen!
  return 0.0;
}

// Update emission probality table - New!
// Does not work with alfaSimple and cell->config->alfaSimpleEmit (as eta=0)! Fix!
void TransBF::updateEtaPTable() {
  FrequencyBinIterator nuEdgeBin,nuBin;
  double etaP;
  int i;

  // Get cell
  Cell* cell=upperState->iState()->atom()->cell();

  // If radiative cross-section or frequency table not available - Return
  if (data->sigmaRad==NULL or cell->nuOld.size()==0) return;

  // Get edge bin
  nuEdgeBin=cell->nuOld.find(nu());

  // Reset emission probability table
  etaPTable.resize(0);
  etaPTable.resize(cell->nuOld.end()-nuEdgeBin,0);

  // For each blueward frequency bin
  for (nuBin=nuEdgeBin;nuBin!=cell->nuOld.end();nuBin++) {

    // Calculate emission probability
    if (nuBin->min>=nu()) etaP=(eta(nuBin->min)+eta(nuBin->max))*(nuBin->max-nuBin->min)/2;
    else etaP=0;

    // Sum cumulative emission probability
    if (nuBin==nuEdgeBin) etaPTable[nuBin-nuEdgeBin]=etaP;
    else etaPTable[nuBin-nuEdgeBin]=etaPTable[nuBin-nuEdgeBin-1]+etaP;
  }

  // Normalize cumulative emission probability
  for (i=0;i<etaPTable.size();i++) etaPTable[i]=etaPTable[i]/etaPTable[etaPTable.size()-1];
}

// Get cooling
double TransBF::cooling(double T, bool includeOSP) {

  if (T==noT) T=upperState->iState()->atom()->cell()->T();
  double ne=upperState->iState()->atom()->cell()->ne();

  double pCooling;
  pCooling=alfa(T,true,includeOSP)*upperState->n()*ne*Xi();

  if (!isnumber(pCooling) or pCooling<0) cout<<"TransBF::cooling: "<<lowerState->iState()->name()<<":"<<lowerState->getIndex()<<"->"<<upperState->iState()->name()<<":"<<upperState->getIndex()<<", cell: "<<upperState->iState()->atom()->cell()->index()<<", Warning: Invalid cooling rate: "<<pCooling<<"\n";

  return pCooling;
}

// Get (spontaneous) cooling (per frequency)
double TransBF::coolingSPNu(const double& _nu, double T) {

  double pCoolingSPNu;
  if (T==noT) T=upperState->iState()->atom()->cell()->T();
  double ne=upperState->iState()->atom()->cell()->ne();

  pCoolingSPNu=alfaSPNu(_nu,T,true)*upperState->n()*ne*Xi();

  return pCoolingSPNu;
}

// Get (spontaneous) cooling (per frequency) - New!
double TransBF::coolingSPNu(FrequencyBinIterator _nu, EdgeType edgeType, double T) {

  double pCoolingSPNu;
  if (T==noT) T=upperState->iState()->atom()->cell()->T();
  double ne=upperState->iState()->atom()->cell()->ne();

  pCoolingSPNu=alfaSPNu(_nu,edgeType,T,true)*upperState->n()*ne*Xi();

  return pCoolingSPNu;
}

// (Spontaneous) cooling integrand
class CoolingSPIntegrand {
public:
  TransBF* trans;
  const double T;
  CoolingSPIntegrand(TransBF* _trans, const double& _T) : trans(_trans), T(_T) {};
  double operator()(double nu) const { return trans->coolingSPNu(nu,T); }
  double operator()(FrequencyBinIterator nu, EdgeType edgeType=noEdge) const { return trans->coolingSPNu(nu,edgeType,T); } // New!
};

// Get (spontaneous) cooling
double TransBF::coolingSP(double T) {
  double _coolingSP;

  // If on-the-spot approximation - Return zero for ground-state transitions
  if (upperState->iState()->atom()->onTheSpot() and lowerState->E()==0) return 0.0;

  // Set cell pointer
  Cell* cell=upperState->iState()->atom()->cell();

  // Testing - Changed!
  if (alfaSimple() and (not cell->config->alfaSimpleEmit)) return 0;

  // If update requested or temperature supplied
  if (pCoolingSP==update or (T!=noT and T!=cell->T())) {

    // Get temperature if not supplied
    if (T==noT) T=cell->T();
    // Set tolerance and maximum frequency
    double tol=1e-2;
    int nIterMin=1,nIterMax=4; // New!
    double nuMin=nu(),nuMax=1e2*nu(); // Changed!
    //bool useIterator=true; // New!
    bool useIterator=(T==cell->T())?true:false; // New!
    int nIter=0;

    // If radiative cross-section available - New condition!
    if (data->sigmaRad!=NULL or (not cell->config->alfaSPH)) {
      // Calculate (sponateous) cooling - Changed!
      CoolingSPIntegrand integrand(this,T);
      if (cell->nuOld.size()==0) _coolingSP=Integrator<CoolingSPIntegrand>::romberg(integrand,nuMin,nuMax,nIter,tol);
      else {
	_coolingSP=BFIntegrator<CoolingSPIntegrand>::trapezoidal(integrand,nuMin,nuMax,cell->nuOld,nIter,tol,0.0,useIterator,"",nIterMin,nIterMax);
	if (nIter>nIterMax) cout<<"TransBF::coolingSP: "<<upperState->iState()->name()<<":"<<upperState->getIndex()<<"->"<<lowerState->iState()->name()<<":"<<lowerState->getIndex()<<", cell: "<<upperState->iState()->atom()->cell()->index()<<", Warning: Maximum number of iterations exceeded\n";
      }
    }
    // Else - Calculate (spontaneous) hydrogenic cooling - New!
    else _coolingSP=coolingSPH(T,nuMin,nuMax);

    // If requested - Update
    if (pCoolingSP==update) pCoolingSP=_coolingSP;
  }
  // Else - Get (spontaneous) cooling
  else _coolingSP=pCoolingSP;

  if (!isnumber(_coolingSP) or _coolingSP<0) cout<<"TransBF::coolingSP: "<<lowerState->iState()->name()<<":"<<lowerState->getIndex()<<"->"<<upperState->iState()->name()<<":"<<upperState->getIndex()<<", cell: "<<upperState->iState()->atom()->cell()->index()<<", Warning: Invalid cooling rate: "<<_coolingSP<<"\n";

  return _coolingSP;
}

// Get (spontaneous) hydrogenic cooling
double TransBF::coolingSPH(double T, double nuMin, double nuMax) {
  double _coolingSP;

  // Set cell pointer
  Cell* cell=upperState->iState()->atom()->cell();

  // Calculate (spontaneous) hydrogenic cooling
  _coolingSP=alfaSPH(T,nuMin,nuMax,true)*upperState->n()*cell->ne()*Xi();

  return _coolingSP;
}

// Get heating
//double TransBF::heating(double T) {
double TransBF::heating(double T, bool includeOSP) {

  double pHeating;
  pHeating=gamma(true,includeOSP)*lowerState->n()*Xi();

  if (!isnumber(pHeating) or pHeating<0) cout<<"TransBF::heating: "<<upperState->iState()->name()<<":"<<upperState->getIndex()<<"->"<<lowerState->iState()->name()<<":"<<lowerState->getIndex()<<", cell: "<<upperState->iState()->atom()->cell()->index()<<", Warning: Invalid heating rate: "<<pHeating<<"\n";

  return pHeating;
}

// Net radiative rate integrand 1
class RNetIntegrand1 : public NuIntegrand {
public:
  TransBF* trans;
  const double ne;
  const double nLower;
  const double nUpper;
  RNetIntegrand1(TransBF* _trans) : trans(_trans), ne(_trans->upperState->iState()->atom()->cell()->ne()), nLower(_trans->lowerState->n()), nUpper(_trans->upperState->n()) {};
  // The virtual prefix is missing, right?
  double operator()(double nu) const { return trans->gammaNu(nu)*nLower-trans->alfaSTNu(nu)*ne*nUpper; }
  double operator()(FrequencyBinIterator nu, EdgeType edgeType=noEdge) const { return trans->gammaNu(nu,edgeType)*nLower-trans->alfaSTNu(nu,edgeType)*ne*nUpper; }
};

// Net radiative rate integrand 2
class RNetIntegrand2 : public NuIntegrand {
public:
  TransBF* trans;
  const double ne;
  const double nUpper;
  RNetIntegrand2(TransBF* _trans) : trans(_trans), ne(_trans->upperState->iState()->atom()->cell()->ne()), nUpper(_trans->upperState->n()){};
  // The virtual prefix is missing, right?
  double operator()(double nu) const { return trans->alfaSPNu(nu)*ne*nUpper; }
  double operator()(FrequencyBinIterator nu, EdgeType edgeType=noEdge) const { return trans->alfaSPNu(nu,edgeType)*ne*nUpper; }
};

// Get net (upwared) radiative rate
/*double TransBF::RNet() {

  // Testing
  if (alfaSimple()) return 0;

  // If update requested
  if (pRNet==update) {
    
    // Set cell pointer
    Cell* cell=upperState->iState()->atom()->cell();
   
    double relTol=1e-2,absTol=MAX(fabs(RD()),fabs(RU()))*numeric_limits<double>::epsilon();
    // Testing!
    //double relTol=1e-2,absTol=0;
    bool includeOSP=true;
    bool useIterator=true;
    RNetIntegrand1 integrand1(this);
    RNetIntegrand2 integrand2(this);
    int nIter;
    
    // Calculate net radiative rate in the on-the-spot range
    pRNet=cell->NetIntegrate(integrand1,integrand2,nu(),nuInf,nIter,relTol,absTol,includeOSP,useIterator);
  }

  if (!isnumber(pRNet)) cout<<"TransBF::RNet: "<<upperState->iState()->name()<<":"<<upperState->getIndex()<<"->"<<lowerState->iState()->name()<<":"<<lowerState->getIndex()<<", cell: "<<upperState->iState()->atom()->cell()->index()<<", Warning: Invalid rate: "<<pRNet<<"\n";

  return pRNet;
  }*/

// Get net (upwared) radiative rate
double TransBF::RNet() {

  // Set cell pointer - Moved!
  Cell* cell=upperState->iState()->atom()->cell();

  // Testing - Changed!
  if (alfaSimple() and (not cell->config->alfaSimpleEmit)) return 0;

  // If update requested
  if (pRNet==update) {
    
    // If simple and emit - Update and return (negative of) downward rate - New!
    if (alfaSimple() and cell->config->alfaSimpleEmit) { pRNet=-alfa(noT,false,true)*upperState->n()*cell->ne(); return pRNet; }
 
    // Calculate net radiative rate (excluding the on-the-spot range)
    pRNet=gamma(false,false)*lowerState->n()-alfa(noT,false,false)*upperState->n()*cell->ne();
 
    double relTol=1e-2,absTol=fabs(pRNet)*relTol;
    bool includeOSP=true;
    bool useIterator=true;
    RNetIntegrand1 integrand1(this);
    RNetIntegrand2 integrand2(this);
    int nIter;
    
    // Calculate and add net radiative rate in the on-the-spot range
    //if (cell->index()==36 and upperState->iState()->atom()->name().compare("Fe")==0) cout<<"TransbF::RNet: cell: "<<upperState->iState()->name()<<":"<<upperState->getIndex()<<"->"<<lowerState->iState()->name()<<":"<<lowerState->getIndex()<<", cell: "<<upperState->iState()->atom()->cell()->index()<<", call NetIntegrate\n";
    pRNet+=cell->NetIntegrate(integrand1,integrand2,nu(),nuInf,nIter,relTol,absTol,true,useIterator,true);
  }

  if (!isnumber(pRNet)) cout<<"TransBF::RNet: "<<upperState->iState()->name()<<":"<<upperState->getIndex()<<"->"<<lowerState->iState()->name()<<":"<<lowerState->getIndex()<<", cell: "<<upperState->iState()->atom()->cell()->index()<<", Warning: Invalid rate: "<<pRNet<<"\n";
      
  return pRNet;
}

// Net radiative rate integrand 1
class ROSPIntegrand1 : public NuIntegrand {
public:
  TransBF* trans;
  const double ne;
  const double nLower;
  const double nUpper;
  ROSPIntegrand1(TransBF* _trans) : trans(_trans), ne(_trans->upperState->iState()->atom()->cell()->ne()), nLower(_trans->lowerState->n()), nUpper(_trans->upperState->n()) {};
  // The virtual prefix is missing, right?
  double operator()(double nu) const { return trans->gammaNu(nu)*nLower-trans->alfaSTNu(nu)*ne*nUpper; }
  double operator()(FrequencyBinIterator nu, EdgeType edgeType=noEdge) const { return trans->gammaNu(nu,edgeType)*nLower-trans->alfaSTNu(nu,edgeType)*ne*nUpper; } // Changed!
};

// Net radiative rate integrand 2
class ROSPIntegrand2 : public NuIntegrand {
public:
  TransBF* trans;
  const double ne;
  const double nUpper;
  ROSPIntegrand2(TransBF* _trans) : trans(_trans), ne(_trans->upperState->iState()->atom()->cell()->ne()), nUpper(_trans->upperState->n()){};
  // The virtual prefix is missing, right?
  double operator()(double nu) const { return trans->alfaSPNu(nu)*ne*nUpper; }
  double operator()(FrequencyBinIterator nu, EdgeType edgeType=noEdge) const { return trans->alfaSPNu(nu,edgeType)*ne*nUpper; } // Changed!
};

// Get net radiative rate in the on-the-spot region
double TransBF::ROSP() {

  // Set cell pointer - Moved!
  Cell* cell=upperState->iState()->atom()->cell();

  // Testing - Changed!
  if (alfaSimple() and (not cell->config->alfaSimpleEmit)) return 0;
  
  // If update requested
  if (pROSP==update) {
    
    // If simple and emit - Update and return (negative of) downward rate in the OSP region - New!
    if (alfaSimple() and cell->config->alfaSimpleEmit) { pROSP=-(alfa(noT,false,true)-alfa(noT,false,false))*upperState->n()*cell->ne(); return pROSP; }
   
    //double relTol=1e-2,absTol=MAX(fabs(RD()),fabs(RU()))*numeric_limits<double>::epsilon();
    // Testing
    double relTol=1e-2,absTol=0.0;
    bool useIterator=true; // New!
    //double relTol=1e-4,absTol=MAX(fabs(RD()),fabs(RU()))*numeric_limits<double>::epsilon();
    ROSPIntegrand1 integrand1(this);
    ROSPIntegrand2 integrand2(this);
    
    // Calculate net radiative rate in the on-the-spot range
    pROSP=cell->OSPIntegrate(integrand1,integrand2,nu(),nuInf,relTol,absTol,useIterator); // Changed!
  }

  if (!isnumber(pROSP)) cout<<"TransBF::ROSP: "<<upperState->iState()->name()<<":"<<upperState->getIndex()<<"->"<<lowerState->iState()->name()<<":"<<lowerState->getIndex()<<", cell: "<<upperState->iState()->atom()->cell()->index()<<", Warning: Invalid rate: "<<pROSP<<"\n";

  return pROSP;
}

// Net (OSP) radiative rate number derivative integrand 1
class DROSPdNIntegrand1 : public NuDNIntegrand {
public:
  TransBF* trans;
  const double ne;
  const double nLower;
  const double nUpper;
  DROSPdNIntegrand1(TransBF* _trans) : trans(_trans), ne(_trans->upperState->iState()->atom()->cell()->ne()), nLower(_trans->lowerState->n()), nUpper(_trans->upperState->n()) {};
  // The virtual prefix is missing, right?
  double F(double nu) const { return trans->gammaNu(nu)*nLower-trans->alfaSTNu(nu)*ne*nUpper; }
  double F(FrequencyBinIterator nu, EdgeType edgeType=noEdge) const { return trans->gammaNu(nu,edgeType)*nLower-trans->alfaSTNu(nu,edgeType)*ne*nUpper; } // New!
  double dFdN(double nu, BState* bState) const { 
    double value=0.0;
    if (bState==trans->lowerState) value+=trans->gammaNu(nu);
    if (bState==trans->upperState) value-=trans->alfaSTNu(nu)*ne;
    return value;
  }
  double dFdN(FrequencyBinIterator nu, BState* bState, EdgeType edgeType=noEdge) const { // New!
    double value=0.0;
    if (bState==trans->lowerState) value+=trans->gammaNu(nu,edgeType);
    if (bState==trans->upperState) value-=trans->alfaSTNu(nu,edgeType)*ne;
    return value;
  }
};

// Net (OSP) radiative rate number derivative integrand 2
class DROSPdNIntegrand2 : public NuDNIntegrand {
public:
  TransBF* trans;
  const double ne;
  const double nUpper;
  DROSPdNIntegrand2(TransBF* _trans) : trans(_trans), ne(_trans->upperState->iState()->atom()->cell()->ne()), nUpper(_trans->upperState->n()){};
  // The virtual prefix is missing, right?
  double F(double nu) const { return trans->alfaSPNu(nu)*ne*nUpper; }
  double F(FrequencyBinIterator nu, EdgeType edgeType=noEdge) const { return trans->alfaSPNu(nu,edgeType)*ne*nUpper; } // New!
  double dFdN(double nu, BState* bState) const { 
    double value=0.0;
    if (bState==trans->upperState) value+=trans->alfaSPNu(nu)*ne;
    return value;
  }
  double dFdN(FrequencyBinIterator nu, BState* bState, EdgeType edgeType=noEdge) const { // New!
    double value=0.0;
    if (bState==trans->upperState) value+=trans->alfaSPNu(nu,edgeType)*ne;
    return value;
  }
};

// Get (OSP) net radiative rate number derivative
double TransBF::dROSPdN(BState* bState) {
  double pdROSPdN;

  // Testing
  if (alfaSimple()) return 0;

  // Set cell pointer
  Cell* cell=upperState->iState()->atom()->cell();
   
  double relTol=1e-2;
  bool useIterator=true; // New!
  DROSPdNIntegrand1 integrand1(this);
  DROSPdNIntegrand2 integrand2(this);
    
  // Calculate net radiative rate in the on-the-spot range
  pdROSPdN=cell->dROSPdNIntegrate(integrand1,integrand2,bState,nu(),nuInf,relTol,useIterator); // Changed!

  if (!isnumber(pdROSPdN)) cout<<"TransBF::dROSPdN: "<<upperState->iState()->name()<<":"<<upperState->getIndex()<<"->"<<lowerState->iState()->name()<<":"<<lowerState->getIndex()<<", "<<bState->iState()->name()<<":"<<bState->getIndex()<<", cell: "<<upperState->iState()->atom()->cell()->index()<<", Warning: Invalid rate: "<<pdROSPdN<<"\n";

  return pdROSPdN;
}

// Net radiative heating integrand 1
class GOSPIntegrand1 : public NuIntegrand {
public:
  TransBF* trans;
  const double ne;
  const double nLower;
  const double nUpper;
  GOSPIntegrand1(TransBF* _trans) : trans(_trans), ne(_trans->upperState->iState()->atom()->cell()->ne()), nLower(_trans->lowerState->n()), nUpper(_trans->upperState->n()) {};
  // The virtual prefix is missing, right?
  double operator()(double nu) const { return (trans->gammaNu(nu,true)*nLower-trans->alfaSTNu(nu,true)*ne*nUpper)*trans->Xi(); }
  double operator()(FrequencyBinIterator nu, EdgeType edgeType=noEdge) const { return 0.0; };
};

// Net radiative heating integrand 2
class GOSPIntegrand2 : public NuIntegrand {
public:
  TransBF* trans;
  const double ne;
  const double nUpper;
  GOSPIntegrand2(TransBF* _trans) : trans(_trans), ne(_trans->upperState->iState()->atom()->cell()->ne()), nUpper(_trans->upperState->n())  {};
  // The virtual prefix is missing, right?
  double operator()(double nu) const { return trans->alfaSPNu(nu,true)*ne*nUpper*trans->Xi(); }
  double operator()(FrequencyBinIterator nu, EdgeType edgeType=noEdge) const { return 0.0; };
};

// Get net radiative heating in the on-the-spot region
double TransBF::GOSP() {

  // Set cell pointer - Moved!
  Cell* cell=upperState->iState()->atom()->cell();

  // Testing - Changed!
  if (alfaSimple() and (not cell->config->alfaSimpleEmit)) return 0;

  // If update requested
  if (pGOSP==update) {

    // If simple and emit - Update and return (negative of) cooling in the OSP region - New!
    if (alfaSimple() and cell->config->alfaSimpleEmit) { pGOSP=-(cooling(noT,true)-cooling(noT,false)); return pGOSP; }

    double relTol=0.0,absTol=MAX(fabs(RD()),fabs(RU()))*numeric_limits<double>::epsilon();
    GOSPIntegrand1 integrand1(this);
    GOSPIntegrand2 integrand2(this);

    // Calculate net radiative rate in the on-the-spot range
    pGOSP=cell->OSPIntegrate(integrand1,integrand2,nu(),nuInf,relTol,absTol);
  }

  return pGOSP;
}

// Get LTE number ratio (See Lucy 2003)
double TransBF::phi(double T) {
  double _phi;

  // Set cell pointer
  Cell* cell=upperState->iState()->atom()->cell();

  // If update requested or temperature supplied
  if (pPhi==update or (T!=noT and T!=cell->T())) {

    // Get temperature if not supplied
    if (T==noT) T=cell->T();

    // Calculate LTE number ratio
    _phi=1/nIonLTE(1.0,1.0,T,Xi(),lowerState->g(),upperState->g());
    // Bound phi above to prevent division by zero - Testing !!!
    //_phi=1/(MAX(nIonLTE(1.0,1.0,T,Xi(),lowerState->g(),upperState->g()),1e-200));

    // If requested - Update
    if (pPhi==update) pPhi=_phi;
  }
  // Else - get LTE number ratio
  else _phi=pPhi;

  return _phi;
}

// Set cut-off frequency
void TransBF::nuCutOff(const double& _nuCutOff) {
  pNuCutOff=_nuCutOff;
}

// Get cut-off frequency
double TransBF::nuCutOff() const {
  return pNuCutOff;
}

// Reset derived quantities (to be updated)
// Needs to be modified if forceDBalance is set!
void TransBF::reset(int dependence) {
  if (dependence&(dpT|dpN|dpNe)) pRD=update;
  if (dependence&(dpRad|dpT|dpN|dpNe)) pRU=update;
  if (dependence&(dpRad|dpT|dpN|dpNe)) pdRDdN=update;
  if (dependence&(dpRad|dpT|dpN|dpNe)) pdRUdN=update;
  if (dependence&(dpT|dpN|dpNe)) pCD=update;
  if (dependence&(dpT|dpN|dpNe)) pCU=update;
  if (dependence&(dpT|dpN)) pCTD=update; // New!
  if (dependence&(dpT|dpN)) pCTU=update; // New!
  // Testing !
  //if (dependence&(dpT)) pcD=update;
  //if (dependence&(dpT)) pcU=update;
  if (dependence&(dpT|dpN|dpNe)) pCoolingSP=update;
  if (dependence&dpRad) pGammaOriginal=update;
  if (dependence&dpRad) pGammaModified=update;
  if (dependence&(dpT|dpN|dpNe)) pGammaOriginalOSP=update;
  if (dependence&(dpT|dpN|dpNe)) pGammaModifiedOSP=update;
  if (dependence&(dpRad|dpT|dpNe)) pGammaSTOriginal=update;
  if (dependence&(dpRad|dpT|dpNe)) pGammaSTModified=update;
  if (dependence&(dpT|dpN|dpNe)) pGammaSTOriginalOSP=update;
  if (dependence&(dpT|dpN|dpNe)) pGammaSTModifiedOSP=update;
  // Testing!
  if (dependence&dpT) pAlfaSPOriginal=update;
  if (dependence&dpT) pAlfaSPModified=update;
  // Changed as these depend on alfaCorr!
  //if (dependence&dpT|dpN) pAlfaSPOriginal=update;
  //if (dependence&dpT|dpN) pAlfaSPModified=update;
  // The following two only depend on T, right? Fix!
  //if (dependence&(dpT|dpN|dpNe)) pAlfaSPOriginalOSP=update;
  //if (dependence&(dpT|dpN|dpNe)) pAlfaSPModifiedOSP=update;
  // Testing!
  if (dependence&dpT) pAlfaSPOriginalOSP=update;
  if (dependence&dpT) pAlfaSPModifiedOSP=update;
  if (dependence&(dpRad|dpT)) pAlfaSTOriginal=update;
  if (dependence&(dpRad|dpT)) pAlfaSTModified=update;
  if (dependence&(dpT|dpN|dpNe)) pAlfaSTOriginalOSP=update;
  if (dependence&(dpT|dpN|dpNe)) pAlfaSTModifiedOSP=update;
  if (dependence&(dpRad|dpT|dpN|dpNe)) pROSP=update;
  if (dependence&(dpRad|dpT|dpN|dpNe)) pGOSP=update;
  if (dependence&(dpRad|dpT|dpN|dpNe)) pRNet=update;
  if (dependence&(dpT)) pPhi=update;
  //if (dependence&(dpT|dpN)) pAlfaCorr=update; // New!
  //if (dependence&(dpT|dpN)) pAlfaAdd=update; // New!
  //if (dependence&(dpT|dpN)) pAlfaSPNoCorr=update; // New!
  // Testing!
  if (dependence&dpT) pAlfaCorr=update; // New!
  if (dependence&dpT) pAlfaAdd=update; // New!
  if (dependence&dpT) pAlfaSPNoCorr=update; // New!
}

// Export data
void TransBF::exportStat(YAML::Emitter& data) {
  Cell* cell=lowerState->iState()->atom()->cell();

  data<<YAML::Flow<<YAML::BeginMap;

  // Export upper-state 
  data<<YAML::Key<<"upper"<<YAML::Value;
  data<<YAML::BeginSeq<<upperState->iState()->name()<<upperState->getIndex()<<YAML::EndSeq;

  // Export upper-state wavelength
  data<<YAML::Key<<"lambda"<<YAML::Value<<(c/nu())/AA;

  // Export upper-state typical optical depth
  data<<YAML::Key<<"tau"<<YAML::Value<<kappa(nu())*cell->size()*cell->rho();

  // Export (radiative) heating
  data<<YAML::Key<<"H"<<YAML::Value<<heating(noT,true);

  // Export (radiative) cooling
  data<<YAML::Key<<"C"<<YAML::Value<<cooling(noT,true);

  // Export radiative upward rate
  //data<<YAML::Key<<"RU"<<YAML::Value<<RU();
  data<<YAML::Key<<"RU"<<YAML::Value<<YAML::BeginSeq<<RU()<<gamma()*lowerState->n()<<YAML::EndSeq;

  // Export radiative downward rate
  data<<YAML::Key<<"RD"<<YAML::Value<<YAML::BeginSeq<<RD()<<alfa()*cell->ne()*upperState->n()<<YAML::EndSeq;

  // Export downward (non-emissive) radiative rate correction
  data<<YAML::Key<<"RDAdd"<<YAML::Value<<RDAdd();

  // Export collisional upward rate
  data<<YAML::Key<<"CU"<<YAML::Value<<CU();

  // Export collisional downward rate
  data<<YAML::Key<<"CD"<<YAML::Value<<CD();

  // Export non-thermal (upward) rate
  data<<YAML::Key<<"CNU"<<YAML::Value<<CNU();

  // Export charge-transfer upward rate - New!
  data<<YAML::Key<<"CTU"<<YAML::Value<<CTU();

  // Export charge-transfer downward rate - New!
  data<<YAML::Key<<"CTD"<<YAML::Value<<CTD();

  data<<YAML::EndMap;

  // Testing!!!
  /*if (lowerState->iState()->name().compare("HeI")==0 and lowerState->getIndex()==0) {
    bool onTheSpotNewTest=onTheSpotNew();
    FrequencyBin onTheSpotNuTest=cell->onTheSpotNu();
    bool continuumBBTest=cell->config->continuumBB;
    FrequencyBin onTheSpotNuNone;
    onTheSpotNuNone.min=nuInf; onTheSpotNuNone.max=nuInf;
    onTheSpotNew(false); cell->onTheSpotNu(onTheSpotNuNone); 
    pGammaOriginal=update; pGammaOriginalOSP=update; 
    pAlfaSTOriginal=update; pAlfaSTOriginalOSP=update;
    pAlfaSPOriginal=update; pAlfaSPOriginalOSP=update; 
    pRNet=update;
    cout<<"TransBF::testLog: "<<upperState->iState()->name()<<":"<<upperState->getIndex()<<"->"<<lowerState->iState()->name()<<":"<<lowerState->getIndex()<<", cell: "<<upperState->iState()->atom()->cell()->index()<<", RNet(1): "<<RNet()<<", RU(1): "<<gamma()*lowerState->n()<<", RD(1): "<<alfa()*cell->ne()*upperState->n()<<"\n";
    onTheSpotNew(false); cell->onTheSpotNu(onTheSpotNuNone); cell->config->continuumBB=false;
    pGammaOriginal=update; pGammaOriginalOSP=update; 
    pAlfaSTOriginal=update; pAlfaSTOriginalOSP=update;
    pAlfaSPOriginal=update; pAlfaSPOriginalOSP=update; 
    pRNet=update;
    cout<<"TransBF::testLog: "<<upperState->iState()->name()<<":"<<upperState->getIndex()<<"->"<<lowerState->iState()->name()<<":"<<lowerState->getIndex()<<", cell: "<<upperState->iState()->atom()->cell()->index()<<", RNet(2): "<<RNet()<<", RU(2): "<<gamma()*lowerState->n()<<", RD(2): "<<alfa()*cell->ne()*upperState->n()<<"\n";
    onTheSpotNew(onTheSpotNewTest); cell->onTheSpotNu(onTheSpotNuTest); cell->config->continuumBB=continuumBBTest;
    pGammaOriginal=update; pGammaOriginalOSP=update; 
    pAlfaSTOriginal=update; pAlfaSTOriginalOSP=update;
    pAlfaSPOriginal=update; pAlfaSPOriginalOSP=update; 
    pRNet=update;
    cout<<"TransBF::testLog: "<<upperState->iState()->name()<<":"<<upperState->getIndex()<<"->"<<lowerState->iState()->name()<<":"<<lowerState->getIndex()<<", cell: "<<upperState->iState()->atom()->cell()->index()<<", RNet(OSP): "<<RNet()<<", RU(OSP): "<<gamma()*lowerState->n()<<", RD(OSP): "<<alfa()*cell->ne()*upperState->n()<<"\n";
    }*/
}

// Save state
void TransBF::saveState(YAML::Emitter& data) {
  Cell* cell=lowerState->iState()->atom()->cell();

  data<<YAML::Flow<<YAML::BeginMap;

  // Export upper-state 
  data<<YAML::Key<<"upper"<<YAML::Value;
  data<<YAML::BeginSeq<<upperState->iState()->name()<<upperState->getIndex()<<YAML::EndSeq;

  // Export non-thermal (upward) rate
  data<<YAML::Key<<"CNU"<<YAML::Value<<CNU();

  data<<YAML::EndMap;
}

// Load state
void TransBF::loadState(const YAML::Node& data) {

  // Load non-thermal (upward rate)
  if (data["CNU"]) pCNU=data["CNU"].as<double>();
}

// Get on-the-spot (new) flag
bool TransBF::onTheSpotNew() const  {
  return pOnTheSpotNew;
}

// Set on-the-spot (new) flag
void TransBF::onTheSpotNew(const bool& _onTheSpotNew) {
  pOnTheSpotNew=_onTheSpotNew;
}

// Boost emission - New!
// Does not work with cell->config->etaTable! Fix!
void TransBF::boostEta() {
  double PTot=0;
  int i;

  // Get cell
  Cell* cell=upperState->iState()->atom()->cell();

  // Resize emission probability vector
  etaP.resize(cell->nuBoost.size());

  // If Boltzman factor zero - Set dummy emission probability and boost factor and return - New! - Testing!!!
  if (exp(-h*nu()/(k*cell->T()))==0) { for (i=0;i<cell->nuBoost.size();i++) etaP[i]=1.0/cell->nuBoost.size(); pEtaBoost=1.0; return; }

  // For each boost bin
  for (i=0;i<cell->nuBoost.size();i++) {
    // Calculate emission probability
    // Here we follow Lucy (2003) and assume a v^-3 powerlaw - Fix !
    if (cell->nuBoost[i].max>nu()) etaP[i]=(exp(-h*MAX(cell->nuBoost[i].min,nu())/(k*cell->T()))-exp(-h*cell->nuBoost[i].max/(k*cell->T())))/exp(-h*nu()/(k*cell->T()));
    else etaP[i]=0;
    // Sum emission probability
    PTot+=etaP[i];
  }

  // Normalize emission probability (to be on the safe side)
  for (i=0;i<cell->nuBoost.size();i++) etaP[i]=etaP[i]/PTot;

  // Reset boost factor
  pEtaBoost=0;

  // Calculate boost factor
  for (i=0;i<cell->nuBoost.size();i++) pEtaBoost+=etaP[i]*cell->boost[i];

  if (cell->master() and cell->index()==60) cout<<"TransBF::boostEta: cell: "<<cell->index()<<", "<<upperState->iState()->name()<<":"<<upperState->getIndex()<<"->"<<lowerState->iState()->name()<<":"<<lowerState->getIndex()<<", pEtaBoost: "<<pEtaBoost<<"\n";

  // Adjust emission probability for each boost bin
  for (i=0;i<cell->nuBoost.size();i++) etaP[i]=etaP[i]*cell->boost[i]/etaBoost();

  //if (cell->master()) for (i=0;i<cell->nuBoost.size();i++) cout<<"TransBF::boostEta: cell: "<<cell->index()<<", "<<upperState->iState()->name()<<":"<<upperState->getIndex()<<"->"<<lowerState->iState()->name()<<":"<<lowerState->getIndex()<<", etaP["<<i<<"]: "<<etaP[i]<<"\n";
}

// Boost opacity - New!
void TransBF::boostKappa() {
  double P;

  // Get cell
  Cell* cell=upperState->iState()->atom()->cell();

  // Calculate probability for ionization
  P=gammaSTRatio();

  // Calculate boost factor
  pKappaBoost=P*upperState->markovPBoost()+(1-P)*cell->coolingBoost();

  if (cell->master() and cell->index()==60) cout<<"TransBF::boostKappa: cell: "<<cell->index()<<", "<<upperState->iState()->name()<<":"<<upperState->getIndex()<<"->"<<lowerState->iState()->name()<<":"<<lowerState->getIndex()<<", pKappaBoost: "<<pKappaBoost<<"\n";
}

// Get emissivity boost factor - New!
double TransBF::etaBoost() { return pEtaBoost; }

// Get opacity boost factor - New!
double TransBF::kappaBoost() { return pKappaBoost; }

// Draw bound-free absorption MC channel - New!
MCChannel TransBF::absorptionChannel() {
  double z,P;
  MCChannel channel;

  // Set cell pointer
  Cell* cell=upperState->iState()->atom()->cell();

  // Pick random number between 0 and 1
  z=randomize(0,1);

  // Calculate ionization probability
  P=gammaSTRatio();
  // Changed from this (which is frequency dependent). Check!
  //P=nu()/nuAbs;

  // If boost - Adjust ionization probability
  if (cell->config->boost) P=P*upperState->markovPBoost()/kappaBoost();

  // Select bound-free channel
  if (z<P) channel=cIonization;
  else channel=cHeating;

  return channel;
}

// Draw emission type
int TransBF::etaType() {

  // Set cell pointer
  Cell* cell=upperState->iState()->atom()->cell();

  // If nonzero (downward) charge-transfer rate and charge-transfer emit flag set
  if (CTD()!=0 and cell->config->chargeTransferEmit) {

    // Draw a random number between 0 and 1
    double z=randomize(0,1);

    // Calculate probability for charge-transfer emission
    double P=CTD()/(CTD()+RD());

    // If drawn - Return bound-free charge-transfer process
    if (z<=P) return pBoundFree|pChargeTransfer;
    // Else - Return bound-free radiative (line) process
    else return pBoundFree|pRadiative;
  }
  // Else - Return bound-free radiative (line) process
  else return pBoundFree|pRadiative;
}

// Draw downward charge-transfer reaction
//CTReaction* TransBF::drawCTDReaction() {
CTReaction* TransBF::drawCTReaction(CTDirection& cTDirection) {
  double PTot=0,_CTD=0;
  int i;

  // Set cell pointer
  Cell* cell=upperState->iState()->atom()->cell();

  // Draw a random number between 0 and 1
  double z=randomize(0,1);

  // To be on the safe side - Make sure CTD is up-to-date
  //pCTD=update;
  //for (i=0;i<CTDReaction.size();i++) _CTD+=CTDReaction[i]->rate();
  for (i=0;i<CTDReaction.size();i++) _CTD+=CTDReaction[i]->forwardRate();
  for (i=0;i<CTUReaction.size();i++) _CTD+=CTUReaction[i]->backwardRate();

  // For each downward charge-transfer reaction
  for (i=0;i<CTDReaction.size();i++) {

    // Calculate cumulative probability
    //PTot+=CTDReaction[i]->rate()/_CTD;
    PTot+=CTDReaction[i]->forwardRate()/_CTD;

    // If drawn - Return charge-transfer reaction
    //if (z<=PTot) return CTDReaction[i];
    if (z<=PTot) { cTDirection=ctForward; return CTDReaction[i]; }
  }

  // For each upward charge-transfer reaction
  for (i=0;i<CTUReaction.size();i++) {

    // Calculate cumulative probability
    //PTot+=CTDReaction[i]->rate()/_CTD;
    PTot+=CTUReaction[i]->backwardRate()/_CTD;

    // If drawn - Return charge-transfer reaction
    if (z<=PTot) { cTDirection=ctBackward; return CTUReaction[i]; }
  }

  cout<<"TransBF::drawCTDReaction: cell: "<<cell->index()<<", "<<upperState->iState()->name()<<":"<<upperState->getIndex()<<"->"<<lowerState->iState()->name()<<":"<<lowerState->getIndex()<<", Error: No transition drawn\n";
  cout<<"TransBF::drawCTDReaction: cell: "<<cell->index()<<", "<<upperState->iState()->name()<<":"<<upperState->getIndex()<<"->"<<lowerState->iState()->name()<<":"<<lowerState->getIndex()<<", z: "<<z<<", PTot: "<<PTot<<"\n";
  //for (i=0;i<CTDReaction.size();i++) cout<<"TransBF::drawCTDReaction: cell: "<<cell->index()<<", "<<upperState->iState()->name()<<":"<<upperState->getIndex()<<"->"<<lowerState->iState()->name()<<":"<<lowerState->getIndex()<<", CTDReaction["<<i<<"]->rate: "<<CTDReaction[i]->rate()<<"\n";

  // This should not happen
  return NULL;
}

// *****************************************
// *** New functions for alfa correction ***
// *****************************************

// Testing!
bool TransBF::alfaSimple() const {
  return lowerState->alfaSimple();
}

// Get (spontaneos) emission correction factor - New!
double TransBF::alfaSPCorr(double _nu) const {

  Cell* cell=upperState->iState()->atom()->cell();

  //if (cell->index()==1 and lowerState->iState()->updateCounter()>0) cout<<"TransBF::alfaSPCorr: cell: "<<cell->index()<<", "<<upperState->iState()->name()<<":"<<upperState->getIndex()<<"->"<<lowerState->iState()->name()<<":"<<lowerState->getIndex()<<", updateCounter: "<<lowerState->iState()->updateCounter()<<", Enter\n";

  if (dBMethod()==dbForceAlfa or cell->config->forceDBalance or (cell->config->forceDBalanceOSP and _nu>=cell->onTheSpotNu().min)) return 1.0;
  else return alfaCorr();
}

// Get cross-section correction factor - New!
double TransBF::sigmaCorr(double _nu) const {

  Cell* cell=upperState->iState()->atom()->cell();

  //if (cell->index()==1 and lowerState->iState()->updateCounter()>0) cout<<"TransBF::sigmaCorr: cell: "<<cell->index()<<", "<<upperState->iState()->name()<<":"<<upperState->getIndex()<<"->"<<lowerState->iState()->name()<<":"<<lowerState->getIndex()<<", updateCounter: "<<lowerState->iState()->updateCounter()<<", Enter\n";
  
  if (dBMethod()==dbForceAlfa or cell->config->forceDBalance or (cell->config->forceDBalanceOSP and _nu>=cell->onTheSpotNu().min)) return alfaCorr();
  else return 1.0;
}

// Get (emissive) recombination correction factor - New!
double TransBF::alfaCorr() const {

  //Cell* cell=lowerState->iState()->atom()->cell();
  //if (cell->index()==1 and lowerState->iState()->updateCounter()>0) cout<<"TransBF::alfaCorr: cell: "<<cell->index()<<", "<<upperState->iState()->name()<<":"<<upperState->getIndex()<<"->"<<lowerState->iState()->name()<<":"<<lowerState->getIndex()<<", updateCounter: "<<lowerState->iState()->updateCounter()<<", Enter\n";

  // If requested - Update
  // Note that this call implicitly overrides const specifier (as pAlfaCorr is updated).
  if (pAlfaCorr==update) lowerState->iState()->updateAlfaCorr();

  return pAlfaCorr;
}

// Get (non-emissive) recombination correction term - New!
double TransBF::alfaAdd() {

  //Cell* cell=lowerState->iState()->atom()->cell();
  //if (cell->index()==1) cout<<"TransBF::alfaAdd: cell: "<<cell->index()<<", "<<upperState->iState()->name()<<":"<<upperState->getIndex()<<"->"<<lowerState->iState()->name()<<":"<<lowerState->getIndex()<<", Enter\n";

  // If requested - Update
  if (pAlfaAdd==update) lowerState->iState()->updateAlfaCorr();
  
  return pAlfaAdd;
}

// Get downward radiative rate (non-emissive) correction - New!
double TransBF::RDAdd() {
  double pRDAdd=0;
  double ne=upperState->iState()->atom()->cell()->ne();

  // Get non-emissive correction (if any)
  pRDAdd=alfaAdd()*upperState->n()*ne;

  return pRDAdd;
}

// Update recombination correction (using distributed ionization and bound state total coefficients) - New!
void TransBF::updateAlfaCorr(double alfaSPBStateDist, double alfaSPIStateDist, double alfaDiIStateDist) {
  double alfaSPOriginal,alfaSPCorrected;
  Cell* cell=upperState->iState()->atom()->cell();

  //if (cell->index()==1) cout<<"TransBF::updateAlfaCorr: cell: "<<cell->index()<<", "<<upperState->iState()->name()<<":"<<upperState->getIndex()<<"->"<<lowerState->iState()->name()<<":"<<lowerState->getIndex()<<", Enter\n";

  // Reset (emmissive) correction factor
  pAlfaCorr=1.0;

  // Reset (non-emissive) correction term
  pAlfaAdd=0.0;

  // If dummy update - Return
  if (alfaSPBStateDist==noAlfa and alfaSPIStateDist==noAlfa and alfaDiIStateDist==noAlfa) return;

  // Testing - Changed!
  if (alfaSimple() and (not cell->config->alfaSimpleEmit)) {
    if (alfaSPBStateDist!=noAlfa) pAlfaAdd=alfaSPBStateDist;
    if (alfaSPIStateDist!=noAlfa) pAlfaAdd=alfaSPIStateDist;
    if (alfaDiIStateDist!=noAlfa) pAlfaAdd+=alfaDiIStateDist;
    return;
  }

  // Get original sponateous) recombination coefficient
  alfaSPOriginal=alfaSP(noT,false,true,false);

  // Initialize corrected value with the original one
  alfaSPCorrected=alfaSPOriginal;

  // If (lower) bound state spontaneous recombination coefficient provided - Use this as the corrected value
  if (alfaSPBStateDist!=noAlfa) alfaSPCorrected=alfaSPBStateDist;

  // If ionization state spontaneous recombination coeffiecient provided
  if (alfaSPIStateDist!=noAlfa) {
    // If treated as non-emissive - limit the corrected value
    if (cell->config->alfaCorrectEmit==false) alfaSPCorrected=MIN(alfaSPCorrected,alfaSPIStateDist);
    // Else - Use this as the corrected value
    else alfaSPCorrected=alfaSPIStateDist;
  }

  // Calculate (emissive) correction factor
  pAlfaCorr=alfaSPCorrected/alfaSPOriginal;
  
  // If ionization state di-electronic recombination coeffiecient provided - Add this to non-emissive correction term
  if (alfaDiIStateDist!=noAlfa) pAlfaAdd+=alfaDiIStateDist;

  // If ionization state spontaneous recombination coeffiecient provided
  if (alfaSPIStateDist!=noAlfa) {
    // If treated as non-emissive - Add spontaneous recombination contribution to non-emissive correction term
    if (cell->config->alfaCorrectEmit==false) pAlfaAdd+=(alfaSPIStateDist-alfaSPCorrected);
  }

  //if (lowerState->iState()->name().compare("FeII")==0 and lowerState->getIndex()==0 and cell->index()==22) cout<<"TransBF::updateAlfaCorr: cell: "<<cell->index()<<", "<<upperState->iState()->name()<<":"<<upperState->getIndex()<<"->"<<lowerState->iState()->name()<<":"<<lowerState->getIndex()<<", alfaSPOriginal: "<<alfaSPOriginal<<", alfaSPIStateDist: "<<alfaSPIStateDist<<", alfaDiIStateDist: "<<alfaDiIStateDist<<", alfaSPBStateDist: "<<alfaSPBStateDist<<", alfaCorr: "<<pAlfaCorr<<", alfaAdd: "<<pAlfaAdd<<", upperState->X: "<<upperState->X()<<"\n";
}

// *****************************************
// *****************************************
// *****************************************

// Get memory allocated - New!
double TransBF::mem() const {
  double _mem=sizeof(*this);

  // Add memory allocated for etaP vector
  _mem+=etaP.size()*sizeof(double);

  // Add memory allocated ofr etaPTable vector
  _mem+=etaPTable.size()*sizeof(double);

  return _mem;
}

// Get atom back reference pointer
Atom* Isotope::atom() const {
  return pAtom;
}

// Get density
double Isotope::rho() const {
  return pAtom->cell()->rho()*X();
}

// Get number density
double Isotope::n() const {
  return rho()/m();
}

// Get mass fraction
double Isotope::X() const {
  return pX;
}

// Set mass fraction
void Isotope::X(const double& _X) {
  pX=_X;
}

// Get number of electrons
int Isotope::N() const {
  return data->N;
}

// Get atomic mass
double Isotope::m() const {
  return data->m;
}

// Load Isotope data
void Isotope::load(pstring& isotopePData, IsotopeDataList& isotopeDataList) {
  string value;
  IsotopeData* isotopeData;

  // Get N
  value=isotopePData.getData("N");
  // Error handling
  if (value.size()==0) {cout<<"Isotope::load: Error: Could not get N\n"; return;}
  // Set isotope data pointer
  isotopeData=isotopeDataList.find(atof(value.c_str()));
  // Error handling
  if (isotopeData==NULL)  {cout<<"Isotope::load: Error: could not find isotope with Z="<<pAtom->Z()<<" and N="<<value<<" in atom data\n"; return;}
  data=isotopeData;
  // Get X
  value=isotopePData.getData("X");
  pX=atof(value.c_str());
  // Error handling
  if (value.size()==0) {cout<<"Isotope::load: Error: Could not get X\n"; return;}

  return;
}

// Load Isotope data
void Isotope::loadYAML(const YAML::Node& isotopeData, IsotopeDataList& isotopeDataList) {
  double _N;

  // Get N
  if (isotopeData["N"]) _N=isotopeData["N"].as<double>();
  else {cout<<"Isotope::load: Error: Could not get N\n"; exit(-1);}

  // Set isotope data pointer
  data=isotopeDataList.find(_N);
  if (data==NULL)  {cout<<"Isotope::load: Error: could not find isotope with Z="<<atom()->Z()<<" and N="<<_N<<" in atom data\n"; exit(-1);}

  // Get X
  if (isotopeData["X"]) pX=isotopeData["X"].as<double>();
  else {cout<<"Isotope::load: Error: Could not get X\n"; exit(-1);}

  return;
}

// Load Isotope state
void Isotope::loadState(const YAML::Node& data) {

  // Get X
  if (data["X"]) pX=data["X"].as<double>();
  else {cout<<"Isotope::loadState: Error: Could not get X\n"; exit(-1);}

  return;
}

// Save Isotope data
void Isotope::save(YAML::Emitter& data) const {
  
  // Begin isotope map
  data<<YAML::BeginMap;

  // Save N
  data<<YAML::Key<<"N"<<YAML::Value<<N();

  // Save X
  data<<YAML::Key<<"X"<<YAML::Value<<X();
}

// Save Isotope state
void Isotope::saveState(YAML::Emitter& data) {
  int i;

  // Begin isotope map
  data<<YAML::BeginMap;

  // Save mass fraction
  data<<YAML::Key<<"X"<<YAML::Value<<X();

  // Close isotope map
  data<<YAML::EndMap;
}

// Set back references
void Isotope::setBackReference(Atom* atom) {
  // Set atom back reference pointer
  pAtom=atom;
}

// Get memory allocated - New!
double Isotope::mem() const {
  double _mem=sizeof(*this);

  return _mem;
}

// Constructor
Atom::Atom() : data(NULL), pCell(NULL) { }

// Copy constructor
// data is owned by the grid and needs to be updated elsewhere if necessay
Atom::Atom(const Atom& atom) : data(atom.data), pIndex(atom.pIndex), pCell(atom.pCell), pOnTheSpot(atom.pOnTheSpot), isotope(atom.isotope), iState(atom.iState) { // Changed!
  int i,j,k;
  int iStateIndex,bStateIndex; // Changed!
  TransBF* transBF;

  // For each ionization state
  for (i=0;i<atom.iState.size();i++) {
    // For each bound state
    for (j=0;j<atom.iState[i].bState.size();j++) {

      // Update LTE reference state pointer - New!
      if (atom.iState[i].bState[j].LTERefState!=NULL) {
	//cout<<"Atom::Atom: "<<iState[i].name()<<":"<<iState[i].bState[j].getIndex()<<", Set LTERefState\n";
	iStateIndex=atom.iState[i].bState[j].LTERefState->iState()->index();
	bStateIndex=atom.iState[i].bState[j].LTERefState->index;
	iState[i].bState[j].LTERefState=&(iState[iStateIndex].bState[bStateIndex]);
      }
      else iState[i].bState[j].LTERefState=NULL;

      // Update LTE group pointers - New!
      for (k=0;k<atom.iState[i].bState[j].LTEGroup.size();k++) {
	iStateIndex=atom.iState[i].bState[j].LTEGroup[k]->iState()->index();
	bStateIndex=atom.iState[i].bState[j].LTEGroup[k]->index;
	iState[i].bState[j].LTEGroup.push_back(&(iState[iStateIndex].bState[bStateIndex]));
      }
      
      // For each downward bound-free transition
      for (k=0;k<atom.iState[i].bState[j].downTransBF.size();k++) {
	// Make copy of transition
	transBF = new TransBF(*(atom.iState[i].bState[j].downTransBF[k]));
	// Set bound state pointers
	transBF->lowerState=&(iState[i-1].bState[atom.iState[i].bState[j].downTransBF[k]->lowerState->index]);
	transBF->upperState=&(iState[i].bState[atom.iState[i].bState[j].downTransBF[k]->upperState->index]);
	// Append to the transition lists
	transBF->lowerState->upTransBF.push_back(transBF);
	transBF->upperState->downTransBF.push_back(transBF);
      }
    }
  }
}

// Assigment operator
// data is owned by the grid and needs to be updated elsewhere if necessay
Atom& Atom::operator=(const Atom& atom) {
  int i,j,k;
  int iStateIndex,bStateIndex; // New!
  TransBF* transBF;

  if (&atom!=this)  {
    data=atom.data;
    pIndex=atom.pIndex;
    pCell=atom.pCell;
    pOnTheSpot=atom.pOnTheSpot;
    isotope=atom.isotope;
    iState=atom.iState;

    // For each ionization state
    for (i=0;i<atom.iState.size();i++) {
      // For each bound state
      for (j=0;j<atom.iState[i].bState.size();j++) {
	
	// Update LTE reference state pointer - New!
	if (atom.iState[i].bState[j].LTERefState!=NULL) {
	  //cout<<"Atom::operator=: "<<iState[i].name()<<":"<<iState[i].bState[j].getIndex()<<", Set LTERefState\n";
	  iStateIndex=atom.iState[i].bState[j].LTERefState->iState()->index();
	  bStateIndex=atom.iState[i].bState[j].LTERefState->index;
	  iState[i].bState[j].LTERefState=&(iState[iStateIndex].bState[bStateIndex]);
	}
	else iState[i].bState[j].LTERefState=NULL;

	// Update LTE group pointers - New!
	for (k=0;k<atom.iState[i].bState[j].LTEGroup.size();k++) {
	  iStateIndex=atom.iState[i].bState[j].LTEGroup[k]->iState()->index();
	  bStateIndex=atom.iState[i].bState[j].LTEGroup[k]->index;
	  iState[i].bState[j].LTEGroup.push_back(&(iState[iStateIndex].bState[bStateIndex]));
	}
	
	// For each downward bound-free transition
	for (k=0;k<atom.iState[i].bState[j].downTransBF.size();k++) {
	  // Make copy of transition
	  transBF = new TransBF(*(atom.iState[i].bState[j].downTransBF[k]));
	  // Set bound state pointers
	  transBF->lowerState=&(iState[i-1].bState[atom.iState[i].bState[j].downTransBF[k]->lowerState->index]);
	  transBF->upperState=&(iState[i].bState[atom.iState[i].bState[j].downTransBF[k]->upperState->index]);
	  // Append to the transition lists
	  transBF->lowerState->upTransBF.push_back(transBF);
	  transBF->upperState->downTransBF.push_back(transBF);
	}
      }
    }
  }
}

// Destructor
// data is owned by the grid and has to deleted elsewhere if necessary
Atom::~Atom() {
  // Testing!
  //if (cell()!=NULL) if (cell()->master()) cout<<"Atom::~Atom: cell: "<<cell()->index()<<", "<<name()<<"\n";
}

// Get atom index - New!
int Atom::index() const {
  return pIndex;
}

// Get cell back reference pointer
Cell* Atom::cell() const {
  return pCell;
}

// Get mass fraction
double Atom::X() const {
  int i;
  double pX=0;
  for (i=0;i<isotope.size();i++) {
    pX+=isotope[i].X();
  }
  return pX;
}

// Get density
double Atom::rho() const {
  int i;
  double prho=0;
  for (i=0;i<isotope.size();i++) {
    prho+=isotope[i].rho();
  }
  return prho;
}

// Get number density
double Atom::n() const {
  int i;
  double pn=0;
  for (i=0;i<isotope.size();i++) {
    pn+=isotope[i].n();
  }
  return pn;
}

// Get atomic mass
double Atom::m() const {
  return rho()/n();
}
 
// Get number of protons
double Atom::Z() const {
  return data->Z;
}

// Total bound-free cooling - New!
double Atom::boundFreeCooling(double T) const {
  double C=0;
  int i;
  if (T==noT) T=cell()->T();

  // For each ionization state
  for (i=0; i<iState.size(); i++) C+=iState[i].boundFreeCooling(T);

  return C;
}

// Total bound-free heating - New!
double Atom::boundFreeHeating(double T) const {
  double H=0;
  int i;
  if (T==noT) T=cell()->T();
  
  // For each ionization state
  for (i=0; i<iState.size(); i++) H+=iState[i].boundFreeHeating(T);
  
  return H;
}

// Total collisional cooling
double Atom::collCooling(double T) const {
  double C=0;
  int i;
  if (T==noT) T=cell()->T();

  // For each ionization state
  for (i=0; i<iState.size(); i++) C+=iState[i].collCooling(T);

  return C;
}

// Total collisional heating - New!
double Atom::collHeating(double T) const {
  double H=0;
  int i;
  if (T==noT) T=cell()->T();
  
  // For each ionization state
  for (i=0; i<iState.size(); i++) H+=iState[i].collHeating(T);
  
  return H;
}

// Total charge-transfer cooling - New!
double Atom::CTCooling(double T) const {
  double C=0;
  int i;
  if (T==noT) T=cell()->T();

  // For each ionization state
  for (i=0; i<iState.size(); i++) C+=iState[i].CTCooling(T);

  return C;
}

// Total charge-transfer heating - New!
double Atom::CTHeating(double T) const {
  double H=0;
  int i;
  if (T==noT) T=cell()->T();
  
  // For each ionization state
  for (i=0; i<iState.size(); i++) H+=iState[i].CTHeating(T);
  
  return H;
}

// Update recombination correction - New!
void Atom::updateAlfaCorr() {
  int i;

  // Update recombination correction for each ionization state
  for (i=0; i<iState.size(); i++) iState[i].updateAlfaCorr();  
}

// Get name
string Atom::name() const {
  switch(int(Z())) {
  case 1: return "H";
  case 2: return "He";
  case 3: return "Li";
  case 4: return "Be";
  case 5: return "B";
  case 6: return "C";
  case 7: return "N";
  case 8: return "O";
  case 9: return "F";
  case 10: return "Ne";
  case 11: return "Na";
  case 12: return "Mg";
  case 13: return "Al";
  case 14: return "Si";
  case 15: return "P";
  case 16: return "S";
  case 17: return "Cl";
  case 18: return "Ar";
  case 19: return "K";
  case 20: return "Ca";
  case 21: return "Sc";
  case 22: return "Ti";
  case 23: return "V";
  case 24: return "Cr";
  case 25: return "Mn";
  case 26: return "Fe";
  case 27: return "Co";
  case 28: return "Ni";
  default: return ">Ni";
  }
   
  return ">Ni";
}

// Get on-the-spot flag
bool Atom::onTheSpot() const {
  return pOnTheSpot;
}

// Set on-the-spot flag
void Atom::onTheSpot(const bool& _onTheSpot) {
  pOnTheSpot=_onTheSpot;
}

// Calculate (row-normalized) Markov S matrix
vector<vector<double> > Atom::markovS(const vector<vector<int> >& matrixIndex) {
  vector<vector<double> > S;
  Mat_DP PT,SInv;
  Vec_DP SCol;
  Vec_INT index;
  DP d;
  BState* bState;
  double P,PTot;
  int i,j,k,nStates=0,rowIndex,iStateIndex,bStateIndex; // Changed!
  bool boundFree=cell()->config->boundFreeMacro;
  bool coll=cell()->config->collMacro;
  bool nonThermal=cell()->config->nonThermalMacro;
  bool twoPhoton=cell()->config->twoPhotonMacro; // New!
  bool chargeTransfer=cell()->config->chargeTransferMacro; // New!
  bool useLTEGroup=cell()->config->LTEGroupMacro; // New!
  bool sumLTEGroup=useLTEGroup; // New!

  //cout<<"Atom::markovS: cell: "<<cell()->index()<<", "<<name()<<", Enter\n";

  // Count total number of states - Removed!
  // Unclear if muliple initializers work - Fix!
  //nStates=0; // New!
  //for (i=0;i<matrixIndex.size();i++) nStates+=matrixIndex[i].size(); // Changed!

  // Count total number of states - New!
  countBMatrixIndex(nStates);

  // Initialize
  PT=Mat_DP(0.0,nStates,nStates);
  SInv=Mat_DP(0.0,nStates,nStates);
  S.resize(nStates); for (i=0;i<S.size();i++) S[i].resize(nStates,0.0);
  index=Vec_INT(0.0,nStates);
  
  // For each ionization state
  for (i=0; i<iState.size(); i++)
    // For each bound state
    for (j=0; j<iState[i].bState.size(); j++) {

      // Set bound state pointer
      bState=&(iState[i].bState[j]);

      // For each downward bound-bound transition
      for (k=0; k<bState->downTransBB.size(); k++) {

	// If use LTE group and trans is internal to the group - skip and continue - New!
	if (useLTEGroup and bState->downTransBB[k]->isLTEGroupInternal()) continue;
      
	// Calculate the probability for an internal transition and insert in matrix
	P=bState->downTransBB[k]->RD(twoPhoton)*bState->downTransBB[k]->lowerState->Etot()/bState->D(boundFree,coll,nonThermal,twoPhoton,chargeTransfer,sumLTEGroup); // Changed!
	if (coll) P+=bState->downTransBB[k]->CD()*bState->downTransBB[k]->lowerState->Etot()/bState->D(boundFree,coll,nonThermal,twoPhoton,chargeTransfer,sumLTEGroup); // Changed!
	PT[matrixIndex[i][j]][matrixIndex[i][bState->downTransBB[k]->lowerState->index]]+=P; // Changed!
      }

      // For each upward bound-bound transition
      for (k=0; k<bState->upTransBB.size(); k++) {

	// If use LTE group and trans is internal to the group - skip and continue - New!
	if (useLTEGroup and bState->upTransBB[k]->isLTEGroupInternal()) continue;
      
	// Calculate the probability for an internal transition and insert in matrix
	P=bState->upTransBB[k]->RU()*bState->Etot()/bState->D(boundFree,coll,nonThermal,twoPhoton,chargeTransfer,sumLTEGroup); // Changed!
	if (coll) P+=bState->upTransBB[k]->CU()*bState->Etot()/bState->D(boundFree,coll,nonThermal,twoPhoton,chargeTransfer,sumLTEGroup); // Changed!
	if (nonThermal) P+=bState->upTransBB[k]->CNU()*bState->Etot()/bState->D(boundFree,coll,nonThermal,twoPhoton,chargeTransfer,sumLTEGroup); // Changed!
	PT[matrixIndex[i][j]][matrixIndex[i][bState->upTransBB[k]->upperState->index]]+=P; // Changed!
      }
 
      // If bound-free processes included
      if (boundFree) {

	// For each downward bound-free transition
	for (k=0; k<bState->downTransBF.size(); k++) {

	  // If use LTE group and trans is internal to the group - skip and continue - New!
	  if (useLTEGroup and bState->downTransBF[k]->isLTEGroupInternal()) continue;

	  // Calculate the probability for an internal transition and insert in matrix
	  P=bState->downTransBF[k]->RD()*bState->downTransBF[k]->lowerState->Etot()/bState->D(boundFree,coll,nonThermal,twoPhoton,chargeTransfer,sumLTEGroup); // Changed!
	  P+=bState->downTransBF[k]->RDAdd()*bState->downTransBF[k]->lowerState->Etot()/bState->D(boundFree,coll,nonThermal,twoPhoton,chargeTransfer,sumLTEGroup); // New!
	  if (coll) P+=bState->downTransBF[k]->CD()*bState->downTransBF[k]->lowerState->Etot()/bState->D(boundFree,coll,nonThermal,twoPhoton,chargeTransfer,sumLTEGroup); // Changed!
	  if (chargeTransfer) P+=bState->downTransBF[k]->CTD()*bState->downTransBF[k]->lowerState->Etot()/bState->D(boundFree,coll,nonThermal,twoPhoton,chargeTransfer,sumLTEGroup); // New!
	  PT[matrixIndex[i][j]][matrixIndex[i-1][bState->downTransBF[k]->lowerState->index]]+=P; // Changed!
	}

	// For each upward bound-free transition
	for (k=0; k<bState->upTransBF.size(); k++) {

	  // If use LTE group and trans is internal to the group - skip and continue - New!
	  if (useLTEGroup and bState->upTransBF[k]->isLTEGroupInternal()) continue;
            
	  // Calculate the probability for an internal transition
	  P=bState->upTransBF[k]->RU()*bState->Etot()/bState->D(boundFree,coll,nonThermal,twoPhoton,chargeTransfer,sumLTEGroup); // Changed!
	  if (coll) P+=bState->upTransBF[k]->CU()*bState->Etot()/bState->D(boundFree,coll,nonThermal,twoPhoton,chargeTransfer,sumLTEGroup); // Changed!
	  if (nonThermal) P+=bState->upTransBF[k]->CNU()*bState->Etot()/bState->D(boundFree,coll,nonThermal,twoPhoton,chargeTransfer,sumLTEGroup);// Changed!
	  if (chargeTransfer) P+=bState->upTransBF[k]->CTU()*bState->Etot()/bState->D(boundFree,coll,nonThermal,twoPhoton,chargeTransfer,sumLTEGroup);// New!
	  PT[matrixIndex[i][j]][matrixIndex[i+1][bState->upTransBF[k]->upperState->index]]+=P; // Changed!
	}
      }
    }

  //cout<<"Atom::markovS: cell: "<<cell()->index()<<", "<<name()<<", Invert "<<nStates<<"x"<<nStates<<" matrix\n";

  // Calculate S^-1=(I-PT)
  for (i=0; i<nStates; i++) {
    for (j=0; j<nStates; j++) {
      if (i==j) SInv[i][j]=1-PT[i][j];
      else SInv[i][j]=0-PT[i][j];
    }
  }

  //cout<<"Atom::markovS: cell: "<<cell()->index()<<", "<<name()<<", Matrix inverted\n";

  /*for (i=0;i<SInv.nrows();i++) {
    cout<<"SInv["<<i<<"]: ";
    for (j=0;j<SInv.ncols();j++) cout<<", SInv["<<i<<","<<j<<"]: "<<SInv[i][j];
    cout<<"\n";
    }*/

  // If (SInv) matrix singular - New!
  // Does not work properly - Temporary removed.
  /*if (singular(SInv,matrixIndex,rowIndex,iStateIndex,bStateIndex)) {
    // Log error and exit
    cout<<"Atom::markovS: cell: "<<cell()->index()<<", "<<name()<<", Error: Singular matrix, only zeros in row "<<rowIndex<<" corresponding to "<<iState[iStateIndex].name()<<":"<<iState[iStateIndex].bState[bStateIndex].getIndex()<<"\n"; 
    //exit(-1); 
    }*/
  
  // Compose error info string - New!
  stringstream errorInfo;  errorInfo.str(""); errorInfo<<"Atom::markovS: cell: "<<cell()->index()<<", "<<name()<<", ludcmp Error\n";

  // LU decompose S^-1
  ludcmp(SInv,index,d,errorInfo.str()); //Changed!

  // Calculate each column of S by back-substitution
  for (i=0; i<nStates; i++) {
    SCol=Vec_DP(0.0,nStates);
    SCol[i]=1;
    lubksb(SInv,index,SCol);
    for (j=0; j<nStates; j++) S[j][i]=SCol[j];
  }

  //cout<<"Atom::markovS: cell: "<<cell()->index()<<", "<<name()<<", Matrix return\n";

  return S;
}

// Singular (SInv) matrix check - New!
bool Atom::singular(Mat_DP& SInv, const vector<vector<int> >& matrixIndex, int& rowIndex, int& iStateIndex, int& bStateIndex) {
  int i,j,ii,jj;
  int nZero;
  double SAbs,max;
  bool _singular=false,found;

  for (i=0;i<SInv.nrows();i++) {
    nZero=0;
    max=0.0;
    for (j=0;j<SInv.ncols();j++) if (SInv[i][j]==0) nZero++;
    for (j=0;j<SInv.ncols();j++) if ((SAbs=fabs(SInv[i][j]))>max) max=SAbs;
    cout<<"Atom::singular: "<<cell()->index()<<", "<<name()<<", row: "<<i<<", max: "<<max<<", nZero: "<<nZero<<", ncols: "<<SInv.ncols()<<"\n"; 
    //if (nZero==SInv.ncols()) {
    if (max==0.0) {
      found=false;
      for (ii=0;ii<iState.size();ii++) {
	for (jj=0;jj<iState[ii].bState.size();jj++) if (matrixIndex[ii][jj]==i) { found=true; break; }
	if (found) break;
      }
      rowIndex=i; iStateIndex=ii; bStateIndex=jj;
      _singular=true;
      cout<<"Atom::singular: cell: "<<cell()->index()<<", "<<name()<<", Error: Singular matrix, only zeros in row "<<rowIndex<<" corresponding to "<<iState[iStateIndex].name()<<":"<<iState[iStateIndex].bState[bStateIndex].getIndex()<<"\n"; 
      //cout<<"Atom::singular: "<<cell()->index()<<", "<<name()<<", singular: "<<_singular<<" max: "<<max<<", rowIndex: "<<rowIndex<<", iStateIndex: "<<iStateIndex<<", bStateIndex: "<<bStateIndex<<", found: "<<found<<"\n"; 
      //break;
    }
  }
 
  return _singular;
}

// Update transition probability list
void Atom::updateTransPList() {
  int i,j;

  if (cell()->config->markovTable) updateMarkovPList();
  else {
    for (i=0; i<iState.size(); i++)
      for (j=0; j<iState[i].bState.size(); j++)
	iState[i].bState[j].updateTransPList();
  }
}

// Setup bound state index mapping
vector<vector<int> > Atom::bMatrixIndex(int& nStates) {
  vector<vector<int> > matrixIndex(iState.size());
  int i,j,LTERefIStateIndex,LTERefBStateIndex;

  // Count total number of states and setup matrix index mapping
  for (i=0;i<iState.size();i++) {
    matrixIndex[i].resize(iState[i].bState.size());
    for (j=0;j<iState[i].bState.size();j++) {
      if (not (cell()->config->LTEGroupMacro and iState[i].bState[j].LTERefState!=NULL)) { // New condition!
	matrixIndex[i][j]=nStates;
	nStates++;
      }
    }
  }
  // New block!
  for (i=0;i<iState.size();i++) {
    for (j=0;j<iState[i].bState.size();j++) {
      if (cell()->config->LTEGroupMacro and iState[i].bState[j].LTERefState!=NULL) {
	LTERefIStateIndex=iState[i].bState[j].LTERefState->iState()->index();
	LTERefBStateIndex=iState[i].bState[j].LTERefState->getIndex();       
	matrixIndex[i][j]=matrixIndex[LTERefIStateIndex][LTERefBStateIndex];
      }
    }
  }

  return matrixIndex;
}

// Count number of unique bound state indeces
void Atom::countBMatrixIndex(int& nStates) {
  vector<vector<int> > matrixIndex=bMatrixIndex(nStates);
}

// Update Markov transition probability list
void Atom::updateMarkovPList() {
  vector<vector<double> > S;
  vector<vector<int> > matrixIndex;
  int i,j,nStates=0; // Changed!

  // Count total number of states and setup matrix index mapping - Removed!
  /*nStates=0;
  matrixIndex.resize(iState.size());
  for (i=0;i<iState.size();i++) {
    matrixIndex[i].resize(iState[i].bState.size());
    for (j=0;j<iState[i].bState.size();j++) {
      matrixIndex[i][j]=nStates;
      nStates++;
    }
    }*/

  // Setup bound state index mapping - New!
  matrixIndex=bMatrixIndex(nStates);

  // Calculate markov S matrix
  S=markovS(matrixIndex);

  // Update Markov transition probability list
  updateMarkovPList(S,matrixIndex);
}

// Update Markov transition probability list
//void Atom::updateMarkovPList() {
void Atom::updateMarkovPList(const vector<vector<double> >& S, const vector<vector<int> >& matrixIndex) {
  TransP transP;
  MarkovP markovP;
  MarkovPList markovPList;
  //Mat_DP PT,SInv,S;
  //Vec_DP SCol,RTot;
  //Vec_INT index;
  //DP d;
  vector<double> RTot;
  BState* bState;
  BState* bStateMaster;
  //vector<vector<int> > matrixIndex;
  double P,PTot;
  int i,j,k,l,nStates=0;
  double markovTableLim=cell()->config->markovTableLim;
  bool boost=cell()->config->boost; // New!
  bool boundFree=cell()->config->boundFreeMacro;
  bool coll=cell()->config->collMacro;
  bool nonThermal=cell()->config->nonThermalMacro;
  bool twoPhoton=cell()->config->twoPhotonMacro; // New!
  bool chargeTransfer=cell()->config->chargeTransferMacro; // New!
  bool chargeTransferEmit=cell()->config->chargeTransferEmit; // New!
  float PFloor=numeric_limits<float>::min(); // New! - Testing!
  bool useLTEGroup=cell()->config->LTEGroupMacro; // New!
  bool sumLTEGroup=useLTEGroup; // New!
   
  // Count number of states - Removed!
  // Unclear if muliple initializers work - Fix!
  //nStates=0; // New!
  //for (i=0;i<matrixIndex.size();i++) nStates+=matrixIndex[i].size(); // Changed!

  // Count number of states - New!
  countBMatrixIndex(nStates);

  // Initialize
  RTot.resize(nStates,0.0);

  // Initialize transtion probability lists - Moved!
  for (i=0; i<iState.size(); i++)
    for (j=0; j<iState[i].bState.size(); j++) iState[i].bState[j].transPList.resize(0);
  
  // For each ionization state
  for (i=0; i<iState.size(); i++)
    // For each bound state
    for (j=0; j<iState[i].bState.size(); j++) {

      // Set bound state pointer
      bState=&(iState[i].bState[j]);

      // Set master bound state (which differs in the case of LTE groups)
      if (useLTEGroup and bState->LTERefState!=NULL) bStateMaster=bState->LTERefState;
      else bStateMaster=bState;

      // Initialize transition probability list - Moved
      //bState->transPList.resize(0);

      // For each downward bound-bound transition
      for (k=0; k<bState->downTransBB.size(); k++) {

	// Initialize radiative de-activation probability
	transP.trans=bState->downTransBB[k];
	transP.index=k; // New! - Removed!
	transP.type=radiativeBB;
	transP.upward=false; // New!
	transP.P=(bState->downTransBB[k]->RD(twoPhoton)*bState->downTransBB[k]->Xi())/bState->D(boundFree,coll,nonThermal,twoPhoton,chargeTransfer,sumLTEGroup); // Changed!

	//if (cell()->master() and cell()->index()==36 and iState[i].name().compare("AlI")==0) cout<<"Atom::updateMarkovPList(1): "<<bState->downTransBB[k]->upperState->iState()->name()<<":"<<bState->downTransBB[k]->upperState->getIndex()<<"->"<<bState->downTransBB[k]->lowerState->iState()->name()<<":"<<bState->downTransBB[k]->lowerState->getIndex()<<", P: "<<transP.P<<", RD: "<<bState->downTransBB[k]->RD(twoPhoton)<<", D: "<<bState->D(boundFree,coll,nonThermal,twoPhoton,chargeTransfer)<<"\n";

	RTot[matrixIndex[i][j]]+=transP.P;

	// Add entry to list
	bStateMaster->transPList.push_back(transP);
 
	// If collisional processes included
	if (coll) {

	  // Initialize collisional de-activation probability
	  transP.trans=bState->downTransBB[k];
	  transP.index=k; // New! - Removed!
	  transP.type=collisional;
	  transP.upward=false; // New!
	  transP.P=(bState->downTransBB[k]->CD()*bState->downTransBB[k]->Xi())/bState->D(boundFree,coll,nonThermal,twoPhoton,chargeTransfer,sumLTEGroup); // Changed!

	  //if (cell()->master() and cell()->index()==36 and iState[i].name().compare("AlI")==0) cout<<"Atom::updateMarkovPList(2): "<<bState->downTransBB[k]->upperState->iState()->name()<<":"<<bState->downTransBB[k]->upperState->getIndex()<<"->"<<bState->downTransBB[k]->lowerState->iState()->name()<<":"<<bState->downTransBB[k]->lowerState->getIndex()<<", P: "<<transP.P<<", CD: "<<bState->downTransBB[k]->CD()<<", D: "<<bState->D(boundFree,coll,nonThermal,twoPhoton,chargeTransfer)<<"\n";

	  RTot[matrixIndex[i][j]]+=transP.P;

	  // Add entry to list
	  bStateMaster->transPList.push_back(transP);
	}
      }

      // If bound-free processes included
      if (boundFree) {

	// For each downward bound-free transition
	for (k=0; k<bState->downTransBF.size(); k++) {

	  // Initialize radiative de-activation probability
	  transP.trans=bState->downTransBF[k];
	  transP.index=k; // New! - Removed!
	  transP.type=radiativeBF;
	  transP.upward=false; // New!
	  transP.P=(bState->downTransBF[k]->RD()*bState->downTransBF[k]->Xi())/bState->D(boundFree,coll,nonThermal,twoPhoton,chargeTransfer,sumLTEGroup); // Changed!
	  // Testing!
	  transP.P+=(bState->downTransBF[k]->RDAdd()*bState->downTransBF[k]->Xi())/bState->D(boundFree,coll,nonThermal,twoPhoton,chargeTransfer,sumLTEGroup); // New!
	  //if (chargeTransfer and chargeTransferEmit) transP.P+=(bState->downTransBF[k]->CTD()*bState->downTransBF[k]->Xi())/bState->D(boundFree,coll,nonThermal,twoPhoton,chargeTransfer); // New!

	  //if (cell()->master() and cell()->index()==36 and iState[i].name().compare("AlI")==0) cout<<"Atom::updateMarkovPList(1): "<<bState->downTransBF[k]->upperState->iState()->name()<<":"<<bState->downTransBF[k]->upperState->getIndex()<<"->"<<bState->downTransBF[k]->lowerState->iState()->name()<<":"<<bState->downTransBF[k]->lowerState->getIndex()<<", P: "<<transP.P<<", RD: "<<bState->downTransBF[k]->RD()<<", RDAdd: "<<bState->downTransBF[k]->RDAdd()<<", D: "<<bState->D(boundFree,coll,nonThermal,twoPhoton,chargeTransfer)<<"\n";

	  RTot[matrixIndex[i][j]]+=transP.P;

	  // Add entry to list
	  bStateMaster->transPList.push_back(transP);
	  
	  // If collisional processes included
	  if (coll) {

	    // Initialize collisional de-activation probability
	    transP.trans=bState->downTransBF[k];
	    transP.index=k; // New! - Removed!
	    transP.type=collisional;
	    transP.upward=false; // New!
	    transP.P=(bState->downTransBF[k]->CD()*bState->downTransBF[k]->Xi())/bState->D(boundFree,coll,nonThermal,twoPhoton,chargeTransfer,sumLTEGroup); // Changed!

	    //if (cell()->master() and cell()->index()==36 and iState[i].name().compare("AlI")==0) cout<<"Atom::updateMarkovPList(1): "<<bState->downTransBF[k]->upperState->iState()->name()<<":"<<bState->downTransBF[k]->upperState->getIndex()<<"->"<<bState->downTransBF[k]->lowerState->iState()->name()<<":"<<bState->downTransBF[k]->lowerState->getIndex()<<", P: "<<transP.P<<", CD: "<<bState->downTransBF[k]->CD()<<", D: "<<bState->D(boundFree,coll,nonThermal,twoPhoton,chargeTransfer)<<"\n";

	    RTot[matrixIndex[i][j]]+=transP.P;	

	    // Add entry to list
	    bStateMaster->transPList.push_back(transP);
	  }

	  // If charge-transfer processes included - New!
	  if (chargeTransfer) {

	    // Initialize charge-transfer de-activation probability
	    transP.trans=bState->downTransBF[k];
	    transP.index=k; // New! - Removed!
	    transP.type=chargeTransferBF;
	    transP.upward=false; // New!
	    transP.P=(bState->downTransBF[k]->CTD()*bState->downTransBF[k]->Xi())/bState->D(boundFree,coll,nonThermal,twoPhoton,chargeTransfer,sumLTEGroup); // Changed!

	    //if (cell()->master() and cell()->index()==36 and iState[i].name().compare("AlI")==0) cout<<"Atom::updateMarkovPList(1): "<<bState->downTransBF[k]->upperState->iState()->name()<<":"<<bState->downTransBF[k]->upperState->getIndex()<<"->"<<bState->downTransBF[k]->lowerState->iState()->name()<<":"<<bState->downTransBF[k]->lowerState->getIndex()<<", P: "<<transP.P<<", CTD: "<<bState->downTransBF[k]->CTD()<<", D: "<<bState->D(boundFree,coll,nonThermal,twoPhoton,chargeTransfer)<<"\n";

	    RTot[matrixIndex[i][j]]+=transP.P;	

	    // Add entry to list
	    bStateMaster->transPList.push_back(transP);
	  }
	}
      }
    }

  // For each bound state - Moved!
  for (i=0; i<iState.size(); i++)
    for (j=0; j<iState[i].bState.size(); j++) {
      // Initialize Markov probability lists
      iState[i].bState[j].markovPList.resize(0);
      // Initialize total radiative and collisional de-activation probabilities
      iState[i].bState[j].PRTot=0;
      iState[i].bState[j].PCTot=0;
    }

  // For each ionization state
  for (i=0; i<iState.size(); i++)
    // For each bound state
    for (j=0; j<iState[i].bState.size(); j++) {

      // If use LTE group and not master state - Skip and continue - New!
      if (useLTEGroup and iState[i].bState[j].LTERefState!=NULL) continue;

      // Set bound state pointer
      bState=&(iState[i].bState[j]);
	  
      // Initialize Markov transition probability list
      //bState->markovPList.resize(0); // Moved!
      markovPList.resize(0);
	  
      // For each ionization state
      for (k=0; k<iState.size(); k++)
	// For each bound state
	for (l=0; l<iState[k].bState.size(); l++) {

	  // If use LTE group and not master state - Skip and continue - New!
	  if (useLTEGroup and iState[k].bState[l].LTERefState!=NULL) continue;
	  
	  // Initialize Markov transition probability
	  markovP.nextState=&(iState[k].bState[l]);
	  // Note: This may not work for LTE groups!
	  if (iState[k].bState[l].downTransBB.size()==0 and iState[k].bState[l].downTransBF.size()==0) markovP.P=0.0;
	  else markovP.P=S[matrixIndex[i][j]][matrixIndex[k][l]]*RTot[matrixIndex[k][l]];

	  // Testing!
	  if (!isnumber(markovP.P)) cout<<"Atom::updateMarkovPList: cell: "<<cell()->index()<<", "<<name()<<", "<<iState[i].name()<<":"<<iState[i].bState[j].getIndex()<<"->"<<iState[k].name()<<":"<<iState[k].bState[l].getIndex()<<", Warning: Invalid Markov transition probability: "<<markovP.P<<"\n";
	  
	  // Add Markov transition probability list
	  markovPList.push_back(markovP);
	}

      //if (name().compare("Ti")==0) for (k=0;k<markovPList.size();k++) cout<<"Atom::updateMarkovPList: cell: "<<cell()->index()<<", "<<name()<<", P["<<k<<"]: "<<markovPList[i].P<<"\n";

      //if (name().compare("Ti")==0) for (k=0;k<RTot.size();k++) cout<<"Atom::updateMarkovPList: cell: "<<cell()->index()<<", "<<name()<<", RTot["<<k<<"]: "<<RTot<<"\n";

      // Set transition probability floor to avoid division by zero in the normalization
      for (k=0; k<bState->transPList.size(); k++) bState->transPList[k].P=MAX(bState->transPList[k].P,PFloor); // New! - Testing!
      // Set Markov transition probability floor to avoid division by zero in the normalization
      for (k=0; k<bState->markovPList.size(); k++) bState->markovPList[k].P=MAX(bState->markovPList[k].P,PFloor); // New! - Testing!

      // Normalize transition probability list
      // Unclear if muliple initializers work - Fix!
      PTot=0; for (k=0; k<bState->transPList.size(); k++) PTot+=bState->transPList[k].P; // Changed!
      for (k=0; k<bState->transPList.size(); k++) bState->transPList[k].P=bState->transPList[k].P/PTot;
    
      // Normalize Markov transition probability list
      // Unclear if muliple initializers work - Fix!
      PTot=0; for (k=0; k<markovPList.size(); k++) PTot+=markovPList[k].P; // Changed!
      for (k=0; k<markovPList.size(); k++) markovPList[k].P=markovPList[k].P/PTot;

      // If not boost mode - New!
      if (not boost) {

	// Sort out Markov transitions above limit
	for (k=0; k<markovPList.size(); k++) 
	  if (markovPList[k].P>markovTableLim) bState->markovPList.push_back(markovPList[k]);
	
	// Renormalize Markov transition probability list
	// Unclear if muliple initializers work - Fix!
	PTot=0; for (k=0; k<bState->markovPList.size(); k++) PTot+=bState->markovPList[k].P; // Changed!
	for (k=0; k<bState->markovPList.size(); k++)  bState->markovPList[k].P=bState->markovPList[k].P/PTot;
	if (cell()->master() and PTot<0.99) {
	  cout<<"Atom::updateMarkovPList: cell: "<<cell()->index()<<", "<<name()<<", Warning: Total probability of transitions sorted out: "<<(1.0-PTot)<<", "<<bState->markovPList.size()<<" out of "<<markovPList.size()<<" transitions kept for "<<iState[i].name()<<":"<<iState[i].bState[j].getIndex()<<"\n";
	  // Testing!
	  //for (k=0;k<markovPList.size();k++) cout<<"Atom::updateMarkovPList: cell: "<<cell()->index()<<", "<<name()<<", P["<<k<<"]: "<<markovPList[i].P<<"\n";
	}
      }
      // Else - Copy Markov transition probability list
      else bState->markovPList=markovPList;

      // Calculate total radiative and collisional de-activation probabilities
      // Unclear if muliple initializers work - Fix!
      //bState->PRTot=0; bState->PCTot=0; // New! - Moved!
      for (k=0; k<bState->transPList.size(); k++) // Changed!
	if (bState->transPList[k].type==collisional) bState->PCTot+=bState->transPList[k].P;
	else if (bState->transPList[k].type!=chargeTransferBF) bState->PRTot+=bState->transPList[k].P; // Changed!
      
      // Sort transition probability list in descending probability
      bState->transPList.sort();
      
      // Sort markov transition probability list in descending probability
      bState->markovPList.sort();
    }  
}

// Reset derived quantities (to be updated)
void Atom::reset(int dependence) {
  int i;
  for (i=0;i<iState.size();i++) iState[i].reset(dependence);
}

// Export statistics
void Atom::exportStat(YAML::Emitter& data, const Atom& atomLTE, const double& tauLim) {
  int i;

  data<<YAML::BeginMap;

  // Export name
  data<<YAML::Key<<"name"<<YAML::Value<<name();

  // Export mass fraction
  data<<YAML::Key<<"X"<<YAML::Value<<X();

  // Export bound-free heating-cooling - New!
  data<<YAML::Key<<"boundFree"<<YAML::Value<<YAML::BeginMap;
  data<<YAML::Key<<"H"<<YAML::Value<<boundFreeHeating();
  data<<YAML::Key<<"C"<<YAML::Value<<boundFreeCooling()<<YAML::EndMap;

  // Export collisional heating-cooling - New!
  data<<YAML::Key<<"coll"<<YAML::Value<<YAML::BeginMap;
  data<<YAML::Key<<"H"<<YAML::Value<<collHeating();
  data<<YAML::Key<<"C"<<YAML::Value<<collCooling()<<YAML::EndMap;
  
  // Export data for each ionization state
  data<<YAML::Key<<"iState"<<YAML::Value<<YAML::BeginSeq;
  for (i=0;i<iState.size();i++) iState[i].exportStat(data,atomLTE.iState[i],tauLim);
  data<<YAML::EndSeq;

  data<<YAML::EndMap;
}

// Load Atom data
void Atom::load(pstring& atomPData, AtomDataList& atomDataList, TransBBSelector* transSelector, BStateSelector* bStateSelector) {
  string value;
  pstring isotopePData,iStatePData;
  AtomData* atomData;
  IStateData* iStateData;
  int i,j,k,l,nIsotopes,nIStates; // Changed!

  // Get Z
  value=atomPData.getData("Z");
  // Error handling
  if (value.size()==0) {cout<<"Atom::load: Error: Could not get C\n"; return;}
  // Set atom data pointer
  atomData=atomDataList.find(atof(value.c_str()));
  // Error handling
  if (atomData==NULL)  {cout<<"Atom::load: Error: Could find atom with Z="<<value<<"\n"; return;}
  data=atomData;
  // Get size of isotope array
  nIsotopes=atomPData.getSize("isotope");
  // Error handling
  if (nIsotopes<=0) {cout<<"Atom::load: Error: Isotope list empty\n"; return;}  
  // Resize isotope list
  isotope.resize(nIsotopes);
  for (i=0;i<isotope.size();i++) {
    // Get isotope data
    isotopePData=atomPData.getData("isotope",i);
    // Load isotope data
    isotope[i].load(isotopePData,data->isotope);
  }
  // Get size of ionization state array
  nIStates=atomPData.getSize("iState");
  // Error handling
  if (nIStates<=0) {cout<<"Atom::load: Error: Ionization state list empty\n"; return;}  
  // Resize ionization state list
  iState.resize(nIStates+1);
  // For each ionization state
  for (i=0;i<iState.size()-1;i++) {
    // Get ionization state data
    iStatePData=atomPData.getData("iState",i);
    // Load ionization state data
    if (i==0) iState[i].load(iStatePData,data->iState,NULL,transSelector,bStateSelector);
    else iState[i].load(iStatePData,data->iState,&(iState[i-1]),transSelector,bStateSelector);
    // Set ionization state index - New!
    iState[i].pIndex=i;
  }

  // Add one further ionization state to the list (check this!!!)

  // Set atom back reference pointer (is this necessary?)
  iState[i].pAtom=this;
  // Load (but not from file) the extra ionization state
  iState[i].load(iStatePData,data->iState,&(iState[i-1]),transSelector,bStateSelector,iState[i-1].C()-1);
  // Set ionization state index - New!
  iState[i].pIndex=i;

  // For each bound state - New block!
  for (i=0;i<iState.size();i++)
    for (j=0;j<iState[i].bState.size();j++) {
      // If LTE reference state pointer set
      if (iState[i].bState[j].data->LTERefState!=NULL) {
	iState[i].bState[j].LTERefState=NULL;
	//cout<<"Atom::load: "<<Z()<<":"<<iState[i].C()<<":"<<iState[i].bState[j].getIndex()<<", Set LTERefState\n";
	// For each bound state
	for (k=0;k<iState.size();k++)
	  for (l=0;l<iState[k].bState.size();l++) {
	    // If match set pointer
	    if (iState[k].bState[l].data==iState[i].bState[j].data->LTERefState) {
	      // Set LTE reference pointer
	      iState[i].bState[j].LTERefState=&(iState[k].bState[l]);
	      //cout<<"Atom::load: "<<Z()<<":"<<iState[i].C()<<":"<<iState[i].bState[j].getIndex()<<", i: "<<i<<", j: "<<j<<", LTERefState "<<iState[i].bState[j].LTERefState<<" set\n";
	      // Add state to LTE group - New!
	      iState[k].bState[l].LTEGroup.push_back(&(iState[i].bState[j]));
	    }
	  }
	if (iState[i].bState[j].LTERefState==NULL) cout<<"Atom::load: "<<Z()<<":"<<iState[i].C()<<":"<<iState[i].bState[j].getIndex()<<", i: "<<i<<", j: "<<j<<", Warning: LTERefState "<<iState[i].bState[j].LTERefState<<" not set\n";
      }
      // Else - Set pointer to NULL
      else iState[i].bState[j].LTERefState=NULL;
    }
  
  return;
}

// Reload bound-bound transitions
void Atom::reloadTransBB(TransBBSelector* transSelector) {
  int i;

  // Reload bound-bound transitions for each ionizations state
  for (i=0;i<iState.size();i++) iState[i].reloadTransBB(transSelector);

  return;
}

// Load Atom data - Not up to date!
void Atom::loadYAML(const YAML::Node& atomData, AtomDataList& atomDataList, TransBBSelector* transSelector, BStateSelector* bStateSelector) {
  double _Z;
  TransBF* transBF;
  int i,j,k,l;

  // Get Z
  if (atomData["Z"]) _Z=atomData["Z"].as<double>();
  else {cout<<"Atom::load: Error: Could not get Z\n"; exit(-1);}

  // Set atom data pointer
  data=atomDataList.find(_Z);
  if (data==NULL) {cout<<"Atom::load: Error: Could find atom with Z="<<_Z<<"\n"; exit(-1);}

  // Resize isotope list
  if (atomData["isotope"]) isotope.resize(atomData["isotope"].size());
  else {cout<<"Atom::load: Error: Could not get isotope data\n"; exit(-1);}

  // Load each isotope
  for (i=0;i<isotope.size();i++) isotope[i].loadYAML(atomData["isotope"][i],data->isotope);

  // Resize ionization state list
  if (atomData["iState"]) iState.resize(atomData["iState"].size()+1);
  else {cout<<"Atom::load: Error: Could not get ionization state data\n"; exit(-1);}

  // Load each ionization state
  for (i=0;i<iState.size()-1;i++) iState[i].loadYAML(atomData["iState"][i],data->iState,transSelector,bStateSelector);
  
  // Construct one further ionization state
  iState[i].construct(iState[i-1].C()-1,data->iState,transSelector,bStateSelector);

  // For each ionization state
  for (i=0;i<iState.size();i++) {
    // For each bound state
    for (j=0;j<iState[i].bState.size();j++) {
      // For each downward bound-free transition
      for (k=0;k<iState[i].bState[j].data->downTransBF.size();k++) {

	// Create transition object
	transBF = new TransBF;
	// Set transition data pointer
	transBF->data=iState[i].bState[j].data->downTransBF[k];
	// Get processes flag - New!
	transBF->getProcesses();
	// Set upper state pointer
	transBF->upperState=&(iState[i].bState[j]);
	// For each bound state of the next lower ionization state
	// Should this loop not be over the data->iState[i-1]->bState list instead??
	for (l=0;l<iState[i-1].bState.size();l++) {
	  // If match set lower state pointer
	  if (&(data->iState[i-1].bState[l])==iState[i].bState[j].data->downTransBF[k]->lowerState) transBF->lowerState=&(iState[i-1].bState[l]);
	}
	// Insert (pointer to) transition in upper state downward bound-free transition list
	transBF->upperState->downTransBF.push_back(transBF);
	// Insert (pointer to) transition in lower state upward bound-free transition list
	transBF->lowerState->upTransBF.push_back(transBF);
      }
    }
  }
  
  return;
}

// Load Atom state
void Atom::loadState(const YAML::Node& data, bool loadIState, bool loadBState) {
  int i;
  
  // Load each isotope
  if (data["isotope"])
    if (isotope.size()==data["isotope"].size())
      for (i=0;i<isotope.size();i++) isotope[i].loadState(data["isotope"][i]);
    else {cout<<"Atom::loadState: Error: Wrong number of isotopes\n"; exit(-1);}
  else {cout<<"Atom::loadState: Error: Could not get isotopes\n"; exit(-1);}

  // If requested
  if (loadIState)
    // Load each ionization state
    if (data["iState"]) 
      if (iState.size()==data["iState"].size())  
	for (i=0;i<iState.size();i++) iState[i].loadState(data["iState"][i],loadBState);
      else {cout<<"Atom::loadState: Error: Wrong number of ionization states\n";}
    else {cout<<"Atom::loadState: Error: Could not get ionization states\n"; exit(-1);}
  
  return;
}

// Save Atom data
void Atom::save(YAML::Emitter& data) const {
  int i;

  // Begin atom map
  data<<YAML::BeginMap;

  // Save Z
  data<<YAML::Key<<"Z"<<YAML::Value<<Z();

  // Begin isotope sequence
  data<<YAML::Key<<"isotope"<<YAML::BeginSeq;

  // Save data for each isotope
  for (i=0;i<isotope.size();i++) isotope[i].save(data);

  // End isotope sequence
  data<<YAML::EndSeq;

  // Begin ionization state sequence
  data<<YAML::Key<<"iState"<<YAML::BeginSeq;

  // Save data for each ionization state
  for (i=0;i<iState.size();i++) iState[i].save(data);

  // End isotope sequence
  data<<YAML::EndSeq;

  // End atom map
  data<<YAML::EndMap;
}

// Save Atom state
void Atom::saveState(YAML::Emitter& data) {
  int i;

  // Begin atom map
  data<<YAML::BeginMap;
  
  // Save state for each ionization state
  data<<YAML::Key<<"iState"<<YAML::Value<<YAML::BeginSeq;
  for (i=0;i<iState.size();i++) iState[i].saveState(data);
  data<<YAML::EndSeq;

  // Save state for each isotope
  data<<YAML::Key<<"isotope"<<YAML::Value<<YAML::BeginSeq;
  for (i=0;i<isotope.size();i++) isotope[i].saveState(data);
  data<<YAML::EndSeq;

  // Close atom map
  data<<YAML::EndMap;
}

// Set back references
void Atom::setBackReference(Cell* cell) { 
  int i;
   
  // Set cell back reference pointer
  pCell=cell;
  // For each isotope set back references
  for (i=0;i<isotope.size();i++) isotope[i].setBackReference(this);
  // For each ionization state set back references
  for (i=0;i<iState.size();i++) iState[i].setBackReference(this);
}

// Get memory allocated for bound-free transitions - New!
double Atom::memTransBF() const {
  double _mem=0;
  int i,j,k;

  // For each ionization state
  for (i=0;i<iState.size();i++) {
    // For each bound state
    for (j=0;j<iState[i].bState.size();j++) {
      // Add memory allocated for each (upward) bound-bound transition
      for (k=0;k<iState[i].bState[j].upTransBF.size();k++) _mem+=2*sizeof(TransBF*)+iState[i].bState[j].upTransBF[k]->mem();
    }
  }
  
  return _mem;
}

// Get memory allocated - New!
double Atom::mem() const {
  double _mem=sizeof(*this);
  int i;

  // Add memory allocated for each ionization state
  for (i=0;i<iState.size();i++) _mem+=iState[i].mem();

  // Add memory allocated for each izotope
  for (i=0;i<isotope.size();i++) _mem+=isotope[i].mem();

  // Add memory allocated for bound-free transitions
  _mem+=memTransBF();

  return _mem;
}

// Get memory allocated for probability lists - New!
double Atom::memPList() const {
  double _mem=0;
  int i;

  // For each ionization state
  for (i=0;i<iState.size();i++) {
    // Add memory allocated for probability lists
    _mem+=iState[i].memPList();
  }
  
  return _mem;
}

// Get memory allocated for bound-bound transitions - New!
double Atom::memTransBB() const {
  double _mem=0;
  int i;

  // For each ionization state
  for (i=0;i<iState.size();i++) {
    // Add memory allocated for bound-bound transitions
    _mem+=iState[i].memTransBB();
  }
  
  return _mem;
}
