#include <cstdlib>
#include <iostream>
#include <iomanip>
#include <fstream>
#include <sstream>
#include <string>
#include <math.h>
#include <algorithm>
#include <limits>
#include <omp.h>
#include <boost/archive/binary_oarchive.hpp>
#include <boost/archive/binary_iarchive.hpp>
#include "mc-util.h"
#include "mc-math.h"
#include "mc-physics.h"
#include "common.h"
#include "atomdata.h"
#include "mc.h"

using namespace std;
using namespace util;
using namespace math;
using namespace physics;
using namespace common;
using namespace atomdata;
using namespace mc;

#include "grid.cpp"

//using namespace grid;

// Constructors
PacketEvent::PacketEvent(const PacketEventType& _type, const Vector3D& _position, const double& _t, const double& _E, const double& _nu, const double& _ECMF, const double& _nuCMF, int _cellIndex, int _subCellIndex, int _packetIndex) : type(_type), position(_position), t(_t), E(_E), nu(_nu), ECMF(_ECMF), nuCMF(_nuCMF), cellIndex(_cellIndex), subCellIndex(_subCellIndex), packetIndex(_packetIndex) { }

PacketEvent::PacketEvent() : type(), position(), t(), E(), nu(), ECMF(), nuCMF(), cellIndex(), subCellIndex(), packetIndex() { }

// Destructor
PacketEvent::~PacketEvent() {
}

// Validate data
bool PacketEvent::valid() const {
  bool _valid=true;
  if (!isnumber(E) or E<0) { cout<<"PacketEvent::valid: Warning: Invalid data for packet "<<packetIndex<<" , E: "<<E<<"\n"; _valid=false; }
  if (!isnumber(ECMF) or ECMF<0) { cout<<"PacketEvent::valid: Warning: Invalid data for packet "<<packetIndex<<" , ECMF: "<<ECMF<<"\n"; _valid=false; }
  if (!isnumber(nu) or nu<0) { cout<<"PacketEvent::valid: Warning: Invalid data for packet "<<packetIndex<<" , nu: "<<nu<<"\n"; _valid=false; }
  if (!isnumber(nuCMF) or nuCMF<0) {cout<<"PacketEvent::valid: Warning: Invalid data for packet "<<packetIndex<<" , nuCMF: "<<nuCMF<<"\n"; _valid=false; }
  return _valid;
}

// Testing!!!
void mc::logSysCPU(int i, string label) {
  stringstream header;
  int status;

  header.str(""); 
  header<<"echo \"i: "<<i<<", "<<label<<"\" >> mc-ps-$(hostname).log"; 
  status=system(header.str().c_str());
  status=system("date >> mc-ps-$(hostname).log");
  status=system("./mc-ps-cpu.sh >> mc-ps-$(hostname).log");
}

// Check if continuum event
bool mc::isContinuumEvent(const PacketEventType& eventType) {
  switch(eventType) {
  case evElectronScatter :
  case evComptonScatter :
  case evGreyScatter : // New!
  case evComptonAbs : // New!
  case evDustAbs : // New!
  case evTwoPhotonEmission : // New!
  case evBoundFreeEmission :
  case evBoundFreeAbs :
  case evFreeFreeEmission :
  case evFreeFreeAbs :return true;
  default : return false;
  }
}

// Check if line event
bool mc::isLineEvent(const PacketEventType& eventType) {
  switch(eventType) {
  case evLineScatter :
  case evLineEmission :
  case evLineAbs : return true;
  default : return false;
  }
}

// Check if (true) emission event
bool mc::isEmissionEvent(const PacketEventType& eventType) {
  switch(eventType) {
  case evBoundFreeEmission :
  case evTwoPhotonEmission : // New!
  case evFreeFreeEmission : return true;
  default : return false;
  }
}

// Check if (true) absorption event
bool mc::isTrueAbsEvent(const PacketEventType& eventType) {
  switch(eventType) {
  case evBoundFreeAbs :
  case evFreeFreeAbs : return true;
  default : return false;
  }
}

// Check if absorption event
bool mc::isAbsEvent(const PacketEventType& eventType) {
  switch(eventType) {
  case evBoundFreeAbs :
  case evLineAbs :
  case evFreeFreeAbs : return true;
  default : return false;
  }
}

// Check if scatter event
bool mc::isScatterEvent(const PacketEventType& eventType) {
  switch(eventType) {
  case evElectronScatter :
  case evGreyScatter : // New!
  case evLineScatter : return true;
  default : return false;
  }
}

// Check radioactive emission event
bool mc::isRadioEmissionEvent(const PacketEventType& eventType) {
  switch(eventType) {
  case ev56NiGammaDecay :
  case ev56CoGammaDecay :
  case ev56CoPosDecay :
  case ev57NiGammaDecay :
  case ev57CoGammaDecay :
  case ev57CoPosDecay :
  case ev44TiGammaDecay :
  case ev44TiPosDecay : return true;
  default : return false;
  }
}

string mc::logPacketEventType(const PacketEventType& eventType) {
  switch(eventType) {
  case evNone : return "evNone";
  case evEmission : return "evEmission";
  case evLineScatter: return "evLineScatter";
  case evLineAbs : return "evLineAbs";
  case evLineEmission : return "evLineEmission";
  case evElectronScatter : return "evElectronScatter";
  case evComptonScatter : return "evComptonScatter";
  case evGreyScatter : return "evGreyScatter";
  case evFreeFreeAbs : return "evFreeFreeAbs";
  case evFreeFreeEmission : return "evFreeFreeEmission";
  case evBoundFreeAbs : return "evBoundFreeAbs";
  case evBoundFreeEmission : return "evBoundFreeEmission";
  case evComptonAbs : return "evComptonAbs";
  case evDustAbs : return "evDustAbs";
  case evTwoPhotonEmission : return "evTwoPhotonEmission";
  case evCellCrossing : return "evCellCrossing";
  case ev56NiGammaDecay : return "ev56NiGammaDecay";
  case ev56CoGammaDecay : return "ev56CoGammaDecay";
  case ev56CoPosDecay : return "ev56CoPosDecay";
  case ev57NiGammaDecay : return "ev57NiGammaDecay";
  case ev57CoGammaDecay : return "ev57CoGammaDecay";
  case ev57CoPosDecay : return "ev57CoPosDeacy";
  case ev44TiGammaDecay : return "ev44TiGammDecay";
  case ev44TiPosDecay : return "ev44TiPosDecay";
  case evReflect : return "evReflect";
  case evBoundaryCrossing : return "evBoundaryCrossing";
  case evCellCrossingInner : return "evCellCrossingInner";
  case evCellCrossingOuter : return "evCellCrossingOuter";
  case evSubCellCrossing : return "evSubCellCrossing";
  case evTimeBorderCrossing : return "evTimeBorderCrossing";
  case evDummyEmission : return "evDummyEmission";
  default: return "unknown";
  }
}

// Constructors
PacketEmissionEvent::PacketEmissionEvent(const PacketEvent& event, Trans* trans) { 
  type=event.type;
  cellIndex=event.cellIndex;
  subCellIndex=event.subCellIndex;
  if (trans!=NULL) {
    //cout<<"PacketEmissionEvent::PacketEmissionEvent: 1, packetIndex: "<<event.packetIndex<<", cellIndex: "<<event.cellIndex<<", eventType: "<<event.type<<", trans: "<<trans<<"\n";
    Z=trans->upperState->iState()->atom()->Z();
    C=trans->upperState->iState()->C();
    //cout<<"PacketEmissionEvent::PacketEmissionEvent: 2, packetIndex: "<<event.packetIndex<<", cellIndex: "<<event.cellIndex<<", eventType: "<<event.type<<"\n";
  }
}

PacketEmissionEvent::PacketEmissionEvent() : type(evNone), cellIndex(0), subCellIndex(0), C(0), Z(0) { }

// Copy constructor
PacketEmissionEvent::PacketEmissionEvent(const PacketEmissionEvent& event) : type(event.type), cellIndex(event.cellIndex), subCellIndex(event.subCellIndex), Z(event.Z), C(event.C) { }

// Asigment operator
PacketEmissionEvent& PacketEmissionEvent::operator=(const PacketEmissionEvent& event) {
  type=event.type;
  cellIndex=event.cellIndex;
  subCellIndex=event.subCellIndex;
  Z=event.Z;
  C=event.C;
}

// Destructor
PacketEmissionEvent::~PacketEmissionEvent() {
}

// Constructor
PacketState::PacketState(const Packet& packet) {
  index=packet.index;
  type=packet.type;
  position=packet.position;
  direction=packet.direction;
  t=packet.t;
  E=packet.E;
  nu=packet.nu;
  boostFactor=packet.boostFactor; // New!
  lastEmissionEvent=packet.lastEmissionEvent; // New!
  // Hack
  comptonAbsSigma=packet.config->comptonAbsSigma;
  cellVirtualState=NULL; // New! - Testing!
  if (packet.cell!=NULL) if (packet.cell->type()==cellSub) cellVirtualState = new CellVirtualState((CellVirtual*) packet.cell); // New! - Testing!

  if (!isnumber(E) or !isnumber(nu)) { cout<<"PacketState::PacketState: Invalid packet: index: "<<index<<", E: "<<E<<", nu: "<<nu<<"\n"; exit(1); }
}

// Constructors
PacketConfig::PacketConfig() : steadyState(true), addVTh(false), lineScatter(true), electronScatter(true), comptonScatter(false), greyScatter(false), freeFreeAbs(false), boundFreeAbs(false), comptonAbs(false), comptonAbsSigma(0), dustAbs(false), flourescence(false), collProc(false), nonThermalProc(false), twoPhotonProc(false), chargeTransferProc(false), transTable(false), updateMethod(urPacketHistory), boost(false), convertGPacket(true), cellIndexMin(-1), cellIndexMax(-1), reflectInner(false), PReflectInner(1.0), reflectOuter(false), PReflectOuter(1.0), reflectInnerGP(false), transformKappa(false), bbEPacket(false) { } // Changed!
PacketConfig::PacketConfig(const bool& _steadyState, const bool& _addVTh, const bool& _lineScatter, const bool& _electronScatter, const bool& _comptonScatter, const bool& _greyScatter, const bool& _freeFreeAbs, const bool& _boundFreeAbs, const bool& _comptonAbs, const double& _comptonAbsSigma, const bool& _dustAbs, const bool& _flourescence, const bool& _collProc, const bool& _nonThermalProc, const bool& _twoPhotonProc, const bool& _chargeTransferProc, const bool& _transTable, const UpdateMethod& _updateMethod, const bool& _boost, const bool& _convertGPacket, const int& _cellIndexMin, const int& _cellIndexMax, const bool& _reflectInner, const double& _PReflectInner, const bool& _reflectOuter, const double& _PReflectOuter, const bool& _reflectInnerGP, const bool& _transformKappa, const bool& _bbEPacket) : steadyState(_steadyState), addVTh(_addVTh), lineScatter(_lineScatter), electronScatter(_electronScatter), comptonScatter(_comptonScatter), greyScatter(_greyScatter), freeFreeAbs(_freeFreeAbs), boundFreeAbs(_boundFreeAbs), comptonAbs(_comptonAbs), comptonAbsSigma(_comptonAbsSigma), dustAbs(_dustAbs), flourescence(_flourescence), collProc(_collProc), nonThermalProc(_nonThermalProc), twoPhotonProc(_twoPhotonProc), chargeTransferProc(_chargeTransferProc), transTable(_transTable), updateMethod(_updateMethod), boost(_boost), convertGPacket(_convertGPacket), cellIndexMin(_cellIndexMin), cellIndexMax(_cellIndexMax), reflectInner(_reflectInner), PReflectInner(_PReflectInner), reflectOuter(_reflectOuter), PReflectOuter(_PReflectOuter), reflectInnerGP(_reflectInnerGP), transformKappa(_transformKappa), bbEPacket(_bbEPacket) { } // Changed!
PacketConfig::PacketConfig(const PacketConfig& packetConfig) : steadyState(packetConfig.steadyState), addVTh(packetConfig.addVTh), lineScatter(packetConfig.lineScatter), electronScatter(packetConfig.electronScatter), comptonScatter(packetConfig.comptonScatter), greyScatter(packetConfig.greyScatter), freeFreeAbs(packetConfig.freeFreeAbs), boundFreeAbs(packetConfig.boundFreeAbs), comptonAbs(packetConfig.comptonAbs), comptonAbsSigma(packetConfig.comptonAbsSigma), dustAbs(packetConfig.dustAbs), flourescence(packetConfig.flourescence), collProc(packetConfig.collProc), nonThermalProc(packetConfig.nonThermalProc), twoPhotonProc(packetConfig.twoPhotonProc), chargeTransferProc(packetConfig.chargeTransferProc), transTable(packetConfig.transTable), updateMethod(packetConfig.updateMethod), boost(packetConfig.boost), convertGPacket(packetConfig.convertGPacket), cellIndexMin(packetConfig.cellIndexMin), cellIndexMax(packetConfig.cellIndexMax), reflectInner(packetConfig.reflectInner), PReflectInner(packetConfig.PReflectInner), reflectOuter(packetConfig.reflectOuter), PReflectOuter(packetConfig.PReflectOuter), reflectInnerGP(packetConfig.reflectInnerGP), transformKappa(packetConfig.transformKappa), bbEPacket(packetConfig.bbEPacket) { } // Changed!

// Destructor
PacketConfig::~PacketConfig() {
}

// Constructor
Packet::Packet(PacketConfig* _config) : cell(NULL), packetStat(NULL) { // Changed!
  // Remove this to cure memory leak
  // Fix this in a better way !!!
  //if (_config==NULL) config = new PacketConfig();
  //else config=_config;
}

// Constructor
Packet::Packet(const double& _E, const FrequencyBin& _nu, const Vector3D& _position, const Vector3D& _direction, const double& _t, CellBase* _cell, Grid* _grid, int _index, const PacketType& _type, const PacketEventType& eventType, PacketConfig* _config, double _boostFactor) : packetStat(NULL), testLog(false) { // Changed!

  // Initialize packet config
  if (_config==NULL) config = new PacketConfig();
  else config=_config;
  // Initialize the packet index
  index=_index;
  // Initialize the packet type
  type=_type;
  // Initialize the energy
  ECMF=_E;
  // Initialize position
  position=_position;
  // Initialize direction
  direction=_direction;
  // Initialize time
  t=_t;
  // Initialize the cell
  cell=_cell;
  // Initialize grid
  grid=_grid;
  // Select a random frequency in the comoving frame
  nuCMF=_nu.min+randomize(0,1)*(_nu.max-_nu.min);
  // Select a random optical depth to the next event
  eventTau=-log(randomize(0,1));
  // Initialize the accumulated optical depth
  accumulatedTau=0;
  // Initialize the cell velocity
  v=cell->v(position,t);
  // Initialize the (cosine of the) polar angle about the radius vector
  mu=direction*position/position.norm();
  // Transform to and initialize the energy (in the rest frame)
  E=ECMF/(1-mu*v/c);
  // Transform to and initialize the frequeny (in the rest frame)
  nu=nuCMF/(1-mu*v/c);

  // Select next line (by searching the cell line list)
  selectNextLine();

  // Select next ionization edge (by searching the cell ionization edge list)
  selectNextEdge();

  // Select next boost bin edge (by searching the cell boost bin list) - New!
  selectNextBoostBin();

  // If not k-packet
  if (not (type==kPacket)) {
    // Handle initial packet event
    /*if (config->updateMethod==urOnTheFly) {
      lastEvent=PacketEvent(eventType,position,t,E,nu,ECMF,nuCMF,cell->index(),cell->subIndex(),index);
      lastEventHandled=false;
    }
    else handleEvent(PacketEvent(eventType,position,t,E,nu,ECMF,nuCMF,cell->index(),cell->subIndex(),index));*/
    handleEvent(PacketEvent(eventType,position,t,E,nu,ECMF,nuCMF,cell->index(),cell->subIndex(),index),true); // Changed!
  }
    
  // Flag all distances to be updated
  pBorderDistance=dUpdate;
  pContinuumEventDistance=dUpdate;
  pLineDistance=dUpdate;
  pEdgeDistance=dUpdate;
  pBoostBinDistance=dUpdate; // New!

  // Flag all continuum opacities to be updated
  pElectronScatterKappa=kUpdate;
  pComptonScatterKappa=kUpdate;
  pGreyScatterKappa=kUpdate; // New!
  pFreeFreeAbsKappa=kUpdate;
  pBoundFreeAbsKappa=kUpdate;
  pComptonAbsKappa=kUpdate;
  pDustAbsKappa=kUpdate; // New!
  pContinuumEventKappa=kUpdate;

  // Set the packet terminated flag to false
  pTerminated=false;

  // Initialize boost factor - New!
  //boostFactor=1.0;
  boostFactor=_boostFactor;

  // Select a random optical depth to the next event
  eventTau=-log(randomize(0,1));
  // Initialize the accumulated optical depth
  accumulatedTau=0;

  // If k-packet - This solution is a bit hacky - Improve !
  if (type==kPacket) {

    Trans* trans;
    PacketEventType initialEventType;

    // Process k-packet (to determine new frequency)
    trans=processKPacket(initialEventType);
    // Should probably handle this as an (failed) event
    if (trans==NULL and initialEventType!=evFreeFreeEmission) { reset(); return; }

    // Re-emit packet
    reEmit();

    // Handle packet event
    // Hmm, how did this work with old version of handleEvent?
    //handleEvent(PacketEvent(initialEventType,position,t,E,nu,ECMF,nuCMF,cell->index(),cell->subIndex(),index));
    handleEvent(PacketEvent(initialEventType,position,t,E,nu,ECMF,nuCMF,cell->index(),cell->subIndex(),index),true,trans); // Changed!

    // Select next resonant line
    selectNextLine();
    // Avoid self-absorption
    if ((*line)==trans and !endOfLineList()) line++;

    // Select next ionization edge
    selectNextEdge();

    // Select next boost bin - New!
    selectNextBoostBin();
  }

  if (!isnumber(E) or !isnumber(ECMF) or !isnumber(nu) or !isnumber(nuCMF)) { cout<<"Packet::Packet: Invalid packet: index: "<<index<<", E: "<<E<<", ECMF: "<<ECMF<<", nu: "<<nu<<", nuCMF: "<<nuCMF<<", v: "<<v<<", mu: "<<mu<<", position.norm: "<<position.norm()<<", direction.norm: "<<direction.norm()<<"\n"; exit(1); }

  // Testing!!!
  //logStream.str("");
}

// Constructor
Packet::Packet(const PacketState& packetState, const PacketEventType& eventType, Grid* _grid, PacketConfig* _config) : packetStat(NULL), testLog(false) {

  config=_config;
  index=packetState.index;
  type=packetState.type;
  position=packetState.position;
  direction=packetState.direction;
  t=packetState.t;
  E=packetState.E;
  nu=packetState.nu;
  boostFactor=packetState.boostFactor; // New!
  lastEmissionEvent=packetState.lastEmissionEvent; // New!
  grid=_grid;

  // Hack ! Does this work with threading ?!?
  // Still a hack - but now it should work with threading ...
  //if (type==gPacket) config->comptonAbsSigma=packetState.comptonAbsSigma;
  if (type==gPacket) {
    config = new PacketConfig(*config);
    config->comptonAbsSigma=packetState.comptonAbsSigma;
  }

  // Initialize cell pointer
  cell=NULL;

  // If virtual cell state stored - Select cell (based on virtual cell state)
  if (packetState.cellVirtualState!=NULL) cell=grid->selectCell(packetState.cellVirtualState);

  // If not done yet - Select cell (based on position, direction and time)
  if (cell==NULL) cell=grid->selectCell(position,direction,t);

  // Check if within specified cell range
  if (cell!=NULL) if ((cell->index()<config->cellIndexMin and config->cellIndexMin!=-1 and type!=gPacket) or (cell->index()>config->cellIndexMax and config->cellIndexMax!=-1 and type!=gPacket)) cell=NULL;

  // Return if not on grid - Probably no the proper way to do it - Improve!
  if (not onGrid()) return;
  
  // Select a random optical depth to the next event
  eventTau=-log(randomize(0,1));
  // Initialize the accumulated optical depth
  accumulatedTau=0;

  // Initialize the cell velocity
  v=cell->v(position,t);
  // Initialize the (cosine of the) polar angle about the radius vector
  mu=direction*position/position.norm();
  // Transform to and initialize the energy (in the comoving frame)
  ECMF=E*(1-mu*v/c);
  // Transform to and initialize the frequeny (in the comoving frame)
  nuCMF=nu*(1-mu*v/c);

  // Select next line (by searching the cell line list)
  selectNextLine();

  // Select next ionization edge (by searching the cell ionization edge list)
  selectNextEdge();

  // Select next boost bin (by searching the cell boost bin list) - New!
  selectNextBoostBin();

  // Handle initial packet event
  /*if (config->updateMethod==urOnTheFly) {
    lastEvent=PacketEvent(eventType,position,t,E,nu,ECMF,nuCMF,cell->index(),cell->subIndex(),index);
    lastEventHandled=false;
  }
  else handleEvent(PacketEvent(eventType,position,t,E,nu,ECMF,nuCMF,cell->index(),cell->subIndex(),index));*/
  handleEvent(PacketEvent(eventType,position,t,E,nu,ECMF,nuCMF,cell->index(),cell->subIndex(),index),true); // Changed!
    
  // Flag all distances to be updated
  pBorderDistance=dUpdate;
  pContinuumEventDistance=dUpdate;
  pLineDistance=dUpdate;
  pEdgeDistance=dUpdate;
  pBoostBinDistance=dUpdate; // New!

  // Flag all continuum opacities to be updated
  pElectronScatterKappa=kUpdate;
  pComptonScatterKappa=kUpdate;
  pGreyScatterKappa=kUpdate; // New!
  pFreeFreeAbsKappa=kUpdate;
  pBoundFreeAbsKappa=kUpdate;
  pComptonAbsKappa=kUpdate;
  pDustAbsKappa=kUpdate; // New!
  pContinuumEventKappa=kUpdate;

  // Set the packet terminated flag to false
  pTerminated=false;

  // Initialize split level
  //splitLevel=0;
 
  // Select a random optical depth to the next event
  eventTau=-log(randomize(0,1));
  // Initialize the accumulated optical depth
  accumulatedTau=0;

  //cout<<"Packet::Packet(PacketState): index: "<<index<<", config->lineScatter: "<<config->lineScatter<<"\n";

  // Testing
  //ECMFOrig=ECMF;

  // Testing!!!
  //logStream.str("");
}

// Copy constructor
// config and packetStat are currently not owned by packet so only copy pointer
Packet::Packet(const Packet& packet, bool deep) : index(packet.index), type(packet.type), config(packet.config), nu(packet.nu), nuCMF(packet.nuCMF), E(packet.E), ECMF(packet.ECMF), position(packet.position), direction(packet.direction), t(packet.t), eventTau(packet.eventTau), accumulatedTau(packet.accumulatedTau), grid(packet.grid), cell(packet.cell), line(packet.line), edge(packet.edge), boostBin(packet.boostBin), pBorderDistance(packet.pBorderDistance), pBorderType(packet.pBorderType), pContinuumEventDistance(packet.pContinuumEventDistance), pLineDistance(packet.pLineDistance), pEdgeDistance(packet.pEdgeDistance), pBoostBinDistance(packet.pBoostBinDistance), pElectronScatterKappa(packet.pElectronScatterKappa), pComptonScatterKappa(packet.pComptonScatterKappa), pGreyScatterKappa(packet.pGreyScatterKappa), pFreeFreeAbsKappa(packet.pFreeFreeAbsKappa), pBoundFreeAbsKappa(packet.pBoundFreeAbsKappa), pComptonAbsKappa(packet.pComptonAbsKappa), pDustAbsKappa(packet.pDustAbsKappa), pContinuumEventKappa(packet.pContinuumEventKappa), pTerminated(packet.pTerminated), mu(packet.mu), v(packet.v), lastEvent(packet.lastEvent), lastEventHandled(packet.lastEventHandled), lastEmissionEvent(packet.lastEmissionEvent), boostFactor(packet.boostFactor), packetStat(packet.packetStat), history(0), child(0), ECMFOrig(packet.ECMFOrig), testLog(packet.testLog) { if (deep) { history=packet.history; child=packet.child; }; if (cell!=NULL) if (cell->type()==cellSub) cell = new CellVirtual(*((CellVirtual*) cell)); } // Changed!

// Asigment operator
// config and packetStat are currently not owned by packet so only copy pointer
Packet& Packet::operator=(const Packet& packet) {

  // If present - Delete virtual cell
  if (cell!=NULL) if (cell->type()==cellSub) delete cell;

  index=packet.index;
  type=packet.type;
  config=packet.config;
  nu=packet.nu;
  nuCMF=packet.nuCMF;
  E=packet.E;
  ECMF=packet.ECMF;
  position=packet.position;
  direction=packet.direction;
  t=packet.t;
  eventTau=packet.eventTau;
  accumulatedTau=packet.accumulatedTau;
  grid=packet.grid;
  cell=packet.cell;
  if (cell!=NULL) if (cell->type()==cellSub) cell = new CellVirtual(*((CellVirtual*) cell)); // New!
  line=packet.line;
  edge=packet.edge;
  boostBin=packet.boostBin; // New!
  pBorderDistance=packet.pBorderDistance;
  pBorderType=packet.pBorderType; // New!
  pContinuumEventDistance=packet.pContinuumEventDistance;
  pLineDistance=packet.pLineDistance;
  pEdgeDistance=packet.pEdgeDistance;
  pBoostBinDistance=packet.pBoostBinDistance; // New!
  pElectronScatterKappa=packet.pElectronScatterKappa;
  pComptonScatterKappa=packet.pComptonScatterKappa;
  pGreyScatterKappa=packet.pGreyScatterKappa; // New!
  pFreeFreeAbsKappa=packet.pFreeFreeAbsKappa;
  pBoundFreeAbsKappa=packet.pBoundFreeAbsKappa;
  pComptonAbsKappa=packet.pComptonAbsKappa;
  pDustAbsKappa=packet.pDustAbsKappa; // New!
  pContinuumEventKappa=packet.pContinuumEventKappa;  
  pTerminated=packet.pTerminated;
  mu=packet.mu;
  v=packet.v;
  lastEvent=packet.lastEvent;
  lastEventHandled=packet.lastEventHandled;
  lastEmissionEvent=packet.lastEmissionEvent; // New!
  boostFactor=packet.boostFactor; // New!
  packetStat=packet.packetStat;
  history=packet.history;
  child=packet.child;
  ECMFOrig=packet.ECMFOrig;
  // Testing!!!
  //logStream.str("");
  // Testing!!!
  testLog=packet.testLog;
}

// Destructor
Packet::~Packet() {

  // If present - Delete virtual cell
  //if (cell!=NULL) if (cell->type()==cellSub) delete cell;
}

// Check if packet on grid
bool Packet::onGrid() {
  return (cell!=NULL);
}

// Check if packet terminated
bool Packet::terminated() {
  return pTerminated;
}

// Adjust energy (ignored in normal mode) - New!
void Packet::adjustEnergy() {
  double newBoostFactor,boost;

  // If normal mode - Ignore and return
  if (not config->boost) return;

  // Get new boost factor
  //newBoostFactor=cell->boostFactor(nuCMF);
  // Testing!
  if (type!=gPacket) newBoostFactor=cell->boostFactor(nuCMF);
  else newBoostFactor=cell->comptonKappaBoost();

  // Calculate relative boost factor
  boost=newBoostFactor/boostFactor;

  // Adjust energy
  E=E/boost;
  ECMF=ECMF/boost;

  // If on-the-fly update method - Adjust energy of last event
  // Note: No effect if the initial event have not yet been handled
  if (config->updateMethod==urOnTheFly) { lastEvent.ECMF=lastEvent.ECMF/boost; lastEvent.E=lastEvent.E/boost; }
  // Else - Fix!

  // Update boost factor
  boostFactor=newBoostFactor;

  // Flag boost factor depentent (continuum) opacities to be updated - Testing!
  pFreeFreeAbsKappa=kUpdate;
  pBoundFreeAbsKappa=kUpdate;
  pComptonAbsKappa=kUpdate;
  pContinuumEventKappa=kUpdate;

  // Flag the continuum event distance to be updated - Testing!
  pContinuumEventDistance=dUpdate;

  //if (nuCMF>c/(911*AA) and cell->index()==79) cout<<"Packet::adjustEnergy: index: "<<index<<", cell: "<<cell->index()<<", lambdaCMF: "<<(c/nuCMF)/AA<<", boost: "<<boost<<", boostFactor: "<<boostFactor<<"\n";
}

// Select next resonant line
void Packet::selectNextLine() {
  line=cell->selectNextLine(nu,position,direction,t);
  
  return;
}

// End of line list
bool Packet::endOfLineList() {
  return cell->endOfLineList(line);
}

// Select next ionization edge
void Packet::selectNextEdge() {
  edge=cell->selectNextEdge(nu,position,direction,t);
  // Testing
  pEdgeDistance=dUpdate;
  
  return;
}

// End of ionization edge list
bool Packet::endOfEdgeList() {
  return cell->endOfEdgeList(edge);
}

// Select next boost bin
void Packet::selectNextBoostBin() {
  boostBin=cell->selectNextBoostBin(nu,position,direction,t);
  // Testing
  pBoostBinDistance=dUpdate;
  
  return;
}

// End of boost bin list
bool Packet::endOfBoostBinList() {
  return cell->endOfBoostBinList(boostBin);
}

// Calculate cell border distance
double Packet::borderDistance() {
  if (pBorderDistance==dUpdate) {
    //pBorderDistance=cell->borderDistance(position,direction,t);
    // Make sure to use correct version of borderDistance - Improve!
    if (config->steadyState) pBorderDistance=cell->borderDistance(position,direction,pBorderType); // Changed!
    else pBorderDistance=cell->borderDistance(position,direction,pBorderType,t); // Changed!
    // Testing!!!
    /*if (cell->index()==6) {
      Vector3D step=direction*pBorderDistance;
      Vector3D newPosition=position+step;
      double rInner=((CellSS*) ((Cell*) cell)->grid()->cell[config->cellIndexMin])->rMinT(t);
      if (newPosition.norm()<rInner and pBorderType==btSubCellBorder) cout<<"Packet::borderDistance: index: "<<index<<", cellIndex: "<<cell->index()<<", subCellIndex: "<<cell->subIndex()<<", borderDistance: "<<pBorderDistance<<", borderType: "<<pBorderType<<"\n";
      }*/
    //if (cell!=NULL) if (type!=gPacket and cell->index()==6 and ((Cell*) cell)->grid()->master() and omp_get_thread_num()==0) cout<<"Packet::borderDistance: index: "<<index<<", borderDistance: "<<pBorderDistance<<", borderType: "<<pBorderType<<"\n";
    /*if (cell!=NULL) if (type!=gPacket and cell->index()==6) {
	BorderType borderTypeContainer;
	CellSS* container=((CellSS*) ((Cell*) cell)->grid()->cell[cell->index()]);
	CellSS* cellInner=((CellSS*) ((Cell*) cell)->grid()->cell[config->cellIndexMin]);
	double borderDistanceContainer=container->borderDistance(position,direction,borderTypeContainer,t);
	Vector3D step=direction*pBorderDistance,stepContainer=direction*borderDistanceContainer;
	Vector3D newPosition=position+step,newPositionContainer=position+stepContainer;
	double tNew=t+pBorderDistance/c,tNewContainer=t+borderDistanceContainer/c;
	double rInner=cellInner->rMinT(tNew),rInnerContainer=cellInner->rMinT(tNewContainer);	
	logStream<<"Packet::borderDistance: index: "<<index<<", borderDistance: "<<pBorderDistance<<", borderDistanceContainer: "<<borderDistanceContainer<<", borderType: "<<pBorderType<<", r: "<<position.norm()<<", rNew: "<<newPosition.norm()<<", rNewContainer: "<<newPositionContainer.norm()<<", rInner: "<<rInner<<", rInnerContainer: "<<rInnerContainer<<", tNew: "<<tNew<<", tNewContainer: "<<tNewContainer<<"\n";
	}*/
    // Testing!
    //if (pBorderDistance<0 and cell->index()==6) { cout<<"Packet::borderDistance: index: "<<index<<", Warning, invalid distance: "<<pBorderDistance<<"\n"; }
  }
  return pBorderDistance;
}

// Return cell border type
// Note: Requires a previous call to borderDistance
BorderType Packet::borderType() {
  return pBorderType;
}

// Calculate distance to next resonant line
double Packet::lineDistance() {
  double _mu,_nuCMF;
  if (pLineDistance==dUpdate) {
    if (!endOfLineList() and config->lineScatter) {
      pLineDistance=MAX(cell->lineDistance(line,nuCMF,t),0);
      if (!isnumber(pLineDistance)) { cout<<"Packet::lineDistance: index: "<<index<<", Warning, invalid distance: "<<pLineDistance<<", nuCMF: "<<nuCMF<<", ECMF: "<<ECMF<<", t: "<<t<<", nuLine: "<<(*line)->nu()<<", line.size: "<<((Cell*) cell)->line.size()<<", lineIndex: "<<line-((Cell*) cell)->line.begin()<<"\n"; pLineDistance=1e99; }
      // Testing!
      //if (pLineDistance<0) { cout<<"Packet::lineDistance: index: :"<<index<<", Warning, invalid distance: "<<pLineDistance<<"\n"; }
    }
    else pLineDistance=1e99;
  }
  return pLineDistance;
}

// Calculate distance to next ionization edge
double Packet::edgeDistance() {
  double _mu,_nuCMF;
  if (pEdgeDistance==dUpdate) {
    if (!endOfEdgeList() and config->boundFreeAbs) {
      pEdgeDistance=MAX(cell->edgeDistance(edge,nuCMF,t),0);
      if (!isnumber(pEdgeDistance)) { cout<<"Packet::edgeDistance: index: "<<index<<", Warning, invalid distance: "<<pEdgeDistance<<", nuEgde: "<<(*edge)->nu()<<", edge.size: "<<((Cell*) cell)->edge.size()<<", edgeIndex: "<<edge-((Cell*) cell)->edge.begin()<<"\n"; pEdgeDistance=1e99; }
      // Testing!
      //if (pEdgeDistance<0) { cout<<"Packet::edgeDistance: index: :"<<index<<", Warning, invalid distance: "<<pEdgeDistance<<"\n"; }
    }
    else pEdgeDistance=1e99;
  }
  return pEdgeDistance;
}

// Calculate distance to next boost bin
double Packet::boostBinDistance() {
  if (pBoostBinDistance==dUpdate) {
    if (!endOfBoostBinList() and config->boost) {
      pBoostBinDistance=MAX(cell->boostBinDistance(boostBin,nuCMF,t),0);
      if (!isnumber(pBoostBinDistance)) { cout<<"Packet::boostBinDistance: index: "<<index<<", Warning, invalid distance: "<<pBoostBinDistance<<", cell: "<<cell->index()<<", nuCMF: "<<nuCMF<<", lambdaBin: ["<<(c/boostBin->max)/AA<<":"<<(c/boostBin->min)/AA<<"], nuBin: ["<<boostBin->min<<":"<<boostBin->max<<", nuBoost: ["<<((Cell*) cell)->nuBoost[0].min<<":"<<((Cell* ) cell)->nuBoost[((Cell*) cell)->nuBoost.size()-1].max<<"], nuBoost.size: "<<((Cell*) cell)->nuBoost.size()<<", nuBoostIndex: "<<boostBin-((Cell*) cell)->nuBoost.begin()<<"]\n"; pBoostBinDistance=1e99; }
      // Testing!
      //if (pBoostBinDistance<0) { cout<<"Packet::boostBinDistance: index: :"<<index<<", Warning, invalid distance: "<<pBoostBinDistance<<"\n"; }
    }
    else pBoostBinDistance=1e99;
  }
  return pBoostBinDistance;
}

// Calculate distance to next continuum event
double Packet::continuumEventDistance() {
  if (pContinuumEventDistance==dUpdate) {
    if (testLog and index%100==0) cout<<"Packet::continuumEventDistance: index: "<<index<<", cellIndex: "<<cell->index()<<", Update\n";
    if (continuumEventKappa()!=0 and cell->rho()!=0) {
      pContinuumEventDistance=(eventTau-accumulatedTau)/(continuumEventKappa()*cell->rho());
      if (testLog and index%100==0) cout<<"Packet::continuumEventDistance: index: "<<index<<", cellIndex: "<<cell->index()<<", continuumEventKappa: "<<continuumEventKappa()<<", eventTau: "<<eventTau<<", accumulatedTau: "<<accumulatedTau<<"\n";
      if (!isnumber(pContinuumEventDistance)) { cout<<"Packet::continuumEventDistance: index: :"<<index<<", Warning, invalid distance: "<<pContinuumEventDistance<<"\n"; pContinuumEventDistance=1e99; }
      // Testing!
      //if (pContinuumEventDistance<0) { cout<<"Packet::continuumEventDistance: index: :"<<index<<", Warning, invalid distance: "<<pContinuumEventDistance<<"\n"; }
    }
    else pContinuumEventDistance=1e99;
  }
  if (testLog and index%100==0) cout<<"Packet::continuumEventDistance: index: "<<index<<", cellIndex: "<<cell->index()<<", Return\n";
  return pContinuumEventDistance;
}

// Calculate opacity for electron scattering
double Packet::electronScatterKappa() {
  if (pElectronScatterKappa==kUpdate) pElectronScatterKappa=cell->electronScatterKappa();    
  return pElectronScatterKappa;
}

// Calculate opacity for compton scattering (not supported yet)
double Packet::comptonScatterKappa() {
  if (pComptonScatterKappa==kUpdate) pComptonScatterKappa=cell->comptonScatterKappa(nuCMF);
  return pComptonScatterKappa;
}

// Calculate opacity for grey scattering
double Packet::greyScatterKappa() {
  if (pGreyScatterKappa==kUpdate) pGreyScatterKappa=cell->greyScatterKappa();    
  return pGreyScatterKappa;
}

// Calculate opacity for free-fee absorption
double Packet::freeFreeAbsKappa() {
  if (pFreeFreeAbsKappa==kUpdate) { 
    pFreeFreeAbsKappa=cell->freeFreeAbsKappa(nuCMF);
    // If boost mode - boost opacity - New!
    if (config->boost) pFreeFreeAbsKappa*=MAX(cell->freeFreeKappaBoost()/boostFactor,1.0);
  }
  return pFreeFreeAbsKappa;
}

// Calculate opacity for bound-free absorption
double Packet::boundFreeAbsKappa() {
  //if (pBoundFreeAbsKappa==kUpdate) pBoundFreeAbsKappa=cell->boundFreeAbsKappa(nuCMF);
  if (pBoundFreeAbsKappa==kUpdate) {
    // Prevent negative BF opacities until we understand were they originate from
    pBoundFreeAbsKappa=MAX(cell->boundFreeAbsKappa(nuCMF),0);
    // If boost mode - boost opacity - New!
    if (config->boost) pBoundFreeAbsKappa*=MAX(cell->boundFreeKappaBoost(nuCMF)/boostFactor,1.0);
  }
  return pBoundFreeAbsKappa;
}

// Calculate opacity for compton absorption
double Packet::comptonAbsKappa() {
  if (pComptonAbsKappa==kUpdate) {
    pComptonAbsKappa=cell->comptonAbsKappa(config->comptonAbsSigma);
    // If boost mode - boost opacity - New!
    if (config->boost) pComptonAbsKappa*=MAX(cell->comptonKappaBoost()/boostFactor,1.0);
  }
  return pComptonAbsKappa;
}

// Calculate opacity for dust absorption - New!
double Packet::dustAbsKappa() {
  if (pDustAbsKappa==kUpdate) pDustAbsKappa=cell->dustAbsKappa();
  return pDustAbsKappa;
}

// Calculate continuum event opacity
//double Packet::continuumEventKappa() {
// Testing
double Packet::continuumEventKappa(bool restFrame) {
  double dFactor;

  // Testing!
  //if (((Cell*) cell)->master()) cout<<"Packet::continuumEventKappa(1): "<<omp_get_thread_num()<<", index: "<<index<<", cellIndex: "<<cell->index()<<"\n"; 

  if (pContinuumEventKappa==kUpdate) {
      pContinuumEventKappa=0;
      // Add continuum opacity according to packet type and configuration
      if (type==gPacket) {
	if (config->comptonAbs) pContinuumEventKappa+=comptonAbsKappa();
	if (config->comptonScatter) pContinuumEventKappa+=comptonScatterKappa();
      }
      else {
	// Testing!
	//if (((Cell*) cell)->master()) cout<<"Packet::continuumEventKappa(2): "<<omp_get_thread_num()<<", index: "<<index<<", cellIndex: "<<cell->index()<<"\n"; 
	if (config->electronScatter) pContinuumEventKappa+=electronScatterKappa();
	if (config->greyScatter) pContinuumEventKappa+=greyScatterKappa();
	// Testing!
	//if (((Cell*) cell)->master()) cout<<"Packet::continuumEventKappa(3): "<<omp_get_thread_num()<<", index: "<<index<<", cellIndex: "<<cell->index()<<"\n"; 
	if (config->freeFreeAbs) pContinuumEventKappa+=freeFreeAbsKappa();
	// Testing!
	//if (((Cell*) cell)->master()) cout<<"Packet::continuumEventKappa(4): "<<omp_get_thread_num()<<", index: "<<index<<", cellIndex: "<<cell->index()<<"\n"; 
	if (config->boundFreeAbs) pContinuumEventKappa+=boundFreeAbsKappa();
	// Testing!
	//if (((Cell*) cell)->master()) cout<<"Packet::continuumEventKappa(5): "<<omp_get_thread_num()<<", index: "<<index<<", cellIndex: "<<cell->index()<<"\n"; 
	if (config->dustAbs) pContinuumEventKappa+=dustAbsKappa(); // New!
      }
      if (testLog and index%100==0) cout<<"Packet::continuumEventKappa: index: "<<index<<", cellIndex: "<<cell->index()<<", pContinuumEventKappa: "<<pContinuumEventKappa<<", electronScatterKappa. "<<electronScatterKappa()<<", greyScatterKappa: "<<greyScatterKappa()<<", freeFreeAbsKappa: "<<freeFreeAbsKappa()<<", boundFreeAbsKappa: "<<boundFreeAbsKappa()<<", dustAbsKappa: "<<dustAbsKappa()<<", Updated\n";
    }

  // If transform to rest frame - Calculate doppler factor
  //if (config->transformKappa) {
  // Testing
  if (restFrame and config->transformKappa) {
    dFactor=1-mu*(v/c);
  }
  // Else - Ignore transformation and set doppler factor to 1
  else dFactor=1;

  if (testLog and index%100==0) cout<<"Packet::continuumEventKappa: index: "<<index<<", cellIndex: "<<cell->index()<<", pContinuumEventKappa: "<<pContinuumEventKappa<<", dFactor: "<<dFactor<<", electronScatter. "<<config->electronScatter<<", greyScatter: "<<config->greyScatter<<", freeFreeAbs: "<<config->freeFreeAbs<<", boundFreeAbs: "<<config->boundFreeAbs<<", dustAbs: "<<config->dustAbs<<", Return\n";

  return pContinuumEventKappa*dFactor;
}

// Calculate optical depth of next line
double Packet::lineTau() {
  double tau;
  if (!endOfLineList() and config->lineScatter) {
    tau=cell->lineTau(line);
    // If boost mode - Boost optical depth - New!
    if (config->boost) tau*=MAX((*line)->kappaBoost()/boostFactor,1.0);
  }
  else tau=0;

  return tau;
}

// Calculate the optical depth for continuum events
double Packet::continuumEventTau(const double& distance) {
  return continuumEventKappa()*cell->rho()*distance;
}

// Update packet position
void Packet::updatePosition(const double& distance) {
  //double x,y,z;
  // Testing!!!
  double r,rOld,rInner,tMin,tMax;

  if (testLog and index%100==0) cout<<"Packet::updatePosition: index: "<<index<<", cellIndex: "<<cell->index()<<", distance: "<<distance<<", Enter\n";

  // If zero distance - Ignore and return - New!
  if (distance==0) return;

  // Testing!
  //if (distance<0) { cout<<"Packet::updatePosition: index: :"<<index<<", Warning 1, invalid distance: "<<distance<<", t: "<<t<<"\n"; }

  // Testing!!!
  /*if (type!=gPacket) {
    r=position.norm();
    rInner=((CellSS*) ((Cell*) cell)->grid()->cell[config->cellIndexMin])->rMinT(t);
    if (r<rInner) cout<<"Packet::updatePosition: 1, index: :"<<index<<", cellIndex: "<<cell->index()<<", cellIndexMin: "<<config->cellIndexMin<<", r: "<<r<<", rInner: "<<rInner<<", t: "<<t<<"\n";
    }*/

  // Testing!!!
  /*if (type!=gPacket) {
    r=position.norm();
    rInner=((CellSS*) ((Cell*) cell)->grid()->cell[config->cellIndexMin])->rMinT(t);
    tMin=((CellSS*) ((Cell*) cell)->grid()->cell[config->cellIndexMin])->tMin();
    tMax=((CellSS*) ((Cell*) cell)->grid()->cell[config->cellIndexMin])->tMax();
    if (r<rInner) cout<<"Packet::updatePosition: 1, index: :"<<index<<", cellIndex: "<<cell->index()<<", cellIndexMin: "<<config->cellIndexMin<<", r: "<<r<<", rInner: "<<rInner<<", mu: "<<mu<<", t: "<<t<<", tMin: "<<tMin<<", tMax: "<<tMax<<", distance: "<<distance<<"\n";
    }*/

  // Testing!!!
  //rOld=position.norm();

  // Update the position
  position=position+direction*distance/cell->fExp();
  // If not steady state update the time
  if (!config->steadyState) t+=distance/c;
  // Update the velocity
  v=cell->v(position,t);
  // Update the (cosine of the) polar angle about the radius vector
  mu=direction*position/position.norm();
  // Update the frequency in the comoving frame
  nuCMF=nu*(1-mu*(v/c));
  // Update the energy in the comoving frame
  ECMF=E*(1-mu*(v/c));
  // Update accumulated optical depth
  accumulatedTau+=continuumEventTau(distance);

  // Testing!
  //if (distance<0 and t<0) { cout<<"Packet::updatePosition: index: :"<<index<<", Warning 2, invalid distance: "<<distance<<", t: "<<t<<"\n"; }

  // Flag all distances to be updated
  pBorderDistance=dUpdate;
  pContinuumEventDistance=dUpdate;
  pLineDistance=dUpdate;
  pEdgeDistance=dUpdate;
  pBoostBinDistance=dUpdate; // New!

  if (testLog and index%100==0) cout<<"Packet::updatePosition: index: "<<index<<", cellIndex: "<<cell->index()<<", Return\n";

  // Testing!!!
  //if (cell!=NULL) if (type!=gPacket and cell->index()==6 and ((Cell*) cell)->grid()->master()) cout<<"Packet::updatePosition: index: "<<index<<", r: "<<position.norm()<<", mu: "<<mu<<", t: "<<t<<", distance: "<<distance<<"\n";
  /*if (cell!=NULL) if (type!=gPacket and cell->index()==6) {
      double rInner=((CellSS*) ((Cell*) cell)->grid()->cell[config->cellIndexMin])->rMinT(t);
      logStream<<"Packet::updatePosition: index: "<<index<<", r: "<<position.norm()<<", mu: "<<mu<<", t: "<<t<<", distance: "<<distance<<", rInner: "<<rInner<<"\n";
      }*/
  /*if (type!=gPacket and cell->index()==6) {
    r=position.norm();
    rInner=((CellSS*) ((Cell*) cell)->grid()->cell[config->cellIndexMin])->rMinT(t);
    if (r<rInner) cout<<"Packet::updatePosition: index: "<<index<<", r: "<<position.norm()<<", rOld: "<<rOld<<", mu: "<<mu<<", t: "<<t<<", distance: "<<distance<<"\n";
    }*/
}

// Get continuum event type
PacketEventType Packet::getContinuumEventType() {
  PacketEventType eventType;
  double kappa,kappaChoice;

  //kappaChoice=continuumEventKappa()*randomize(0,1);
  // Testing
  kappaChoice=continuumEventKappa(false)*randomize(0,1);

  // Get continuum event type according to packet type and configuration
  if (type==gPacket) {
    if (config->comptonAbs) {
      kappa=comptonAbsKappa();
      if (kappaChoice<=kappa) return evComptonAbs;
      else kappaChoice-=kappa;
    }
    if (config->comptonScatter) {
      kappa=comptonScatterKappa();
      if (kappaChoice<=kappa) return evComptonScatter;
      else kappaChoice-=kappa;
    }
  }
  else {
    if (config->electronScatter) {
      kappa=electronScatterKappa();
      if (kappaChoice<=kappa) return evElectronScatter;
      else kappaChoice-=kappa;
    }
    // New!
    if (config->greyScatter) {
      kappa=greyScatterKappa();
      if (testLog and index%100==0) cout<<"Packet::getContinuumEventType: index: "<<index<<", cellIndex: "<<cell->index()<<", kappa: "<<kappa<<", kappaChoice: "<<kappaChoice<<", kappaChoice<=kappa: "<<(kappaChoice<=kappa)<<"\n";
      if (kappaChoice<=kappa) return evGreyScatter;
      else kappaChoice-=kappa;
    }
    if (config->freeFreeAbs) {
      kappa=freeFreeAbsKappa();
      if (kappaChoice<=kappa) return evFreeFreeAbs;
      else kappaChoice-=kappa;
    }
    if (config->boundFreeAbs) {
      kappa=boundFreeAbsKappa();
      if (kappaChoice<=kappa) return evBoundFreeAbs;
      else kappaChoice-=kappa;
    }
    // New!
    if (config->dustAbs) {
      kappa=dustAbsKappa();
      if (kappaChoice<=kappa) return evDustAbs;
      else kappaChoice-=kappa;
    }
  }
  
  return eventType;
}

// Get event treatment (absorption/emission) - New!
void Packet::getEventTreatment(const PacketEventType& eventType, bool& emission, bool& absorption) {
  double z,P,boost;

  // No effect in normal mode
  //if (not config->boost) { absorption=true, emission=true; return; }
  if (not config->boost) { absorption=true; emission=true; return; }

  // Switch on event type
  switch(eventType) {
  case evFreeFreeAbs : 
    // Calculate effective boost for free-free absorption
    boost=cell->freeFreeKappaBoost()/boostFactor;
    //if (boost!=1.0) cout<<"Packet::getEventTreatment: index: "<<index<<", cell: "<<cell->index()<<", boost: "<<boost<<", freeFreeKappaBoost: "<<cell->freeFreeKappaBoost()<<", boostFactor: "<<boostFactor<<"\n";
    break;
  case evBoundFreeAbs : 
    // Calculate effective boost for bound-free absorption
    boost=cell->boundFreeKappaBoost(nuCMF)/boostFactor;
    //if (boost!=1.0) cout<<"Packet::getEventTreatment: index: "<<index<<", cell: "<<cell->index()<<", boost: "<<boost<<", boundFreeKappaBoost: "<<cell->boundFreeKappaBoost(nuCMF)<<", boostFactor: "<<boostFactor<<"\n";
    break;
  case evComptonAbs : 
    // Calculate effective boost for compton absorption
    boost=cell->comptonKappaBoost()/boostFactor;
    // Testing!
    //{ absorption=true; emission=true; return; }
    //if (boost!=1.0) cout<<"Packet::getEventTreatment: index: "<<index<<", cell: "<<cell->index()<<", boost: "<<boost<<", comptonKappaBoost: "<<cell->comptonKappaBoost()<<", boostFactor: "<<boostFactor<<"\n";
    break;
  case evLineAbs : 
    // Calculate effective boost for line absorption
    boost=(*line)->kappaBoost()/boostFactor;
    //if (boost!=1.0) cout<<"Packet::getEventTreatment: index: "<<index<<", cell: "<<cell->index()<<", boost: "<<boost<<", line->kappaBoost: "<<(*line)->kappaBoost()<<", boostFactor: "<<boostFactor<<"\n";
    break;
  default : 
    // No effect for other events
    { absorption=true; emission=true; return; }
  }

  // Draw a number between 0 and 1
  z=randomize(0,1);

  // If effective boost less than one
  if (boost<1) {

    // Set absorption to true
    absorption=true;

    // Draw emission
    emission=(z<=boost);
  }
  // Else
  else {

    // Set emission to true
    emission=true;

    // Draw absorption
    absorption=(z<=(1/boost));
  }

  //if (boostFactor==15.0 and (cell->index()==40 or cell->index()==50 or cell->index()==60 or cell->index()==70)) cout<<"Packet::getEventTreatment: index: "<<index<<", cell: "<<cell->index()<<", lambdaCMF: "<<(c/nuCMF)/AA<<", absorption: "<<absorption<<", emission: "<<emission<<", boost: "<<boost<<", boostFactor: "<<boostFactor<<"\n";

  //if ((not absorption) or (not emission)) cout<<"Packet::getEventTreatment: index: "<<index<<", cell: "<<cell->index()<<", absorption: "<<absorption<<", emission: "<<emission<<", boost: "<<boost<<", boostFactor: "<<boostFactor<<"\n";

  //if (not emission) cout<<"Packet::getEventTreatment: index: "<<index<<", cell: "<<cell->index()<<", absorption: "<<absorption<<", emission: "<<emission<<", boost: "<<boost<<", boostFactor: "<<boostFactor<<"\n";
}

// Handle continuum event
//void Packet::continuumEvent(const PacketEventType& eventType) {
void Packet::continuumEvent(const PacketEventType& eventType, bool emission, bool absorption) { // Changed!

  //if (not emission) cout<<"Packet::continuumEvent: index: "<<index<<", cell: "<<cell->index()<<", eventType: "<<eventType<<", emission: "<<emission<<", absorption: "<<absorption<<"\n";

  switch(eventType) {
  case evElectronScatter : electronScatter(); break;
  case evComptonScatter : comptonScatter(); break;
  case evGreyScatter : greyScatter(); break; // New!
  case evFreeFreeAbs : freeFreeAbs(emission,absorption); break;
  case evBoundFreeAbs : boundFreeAbs(emission,absorption); break;
  case evComptonAbs : comptonAbs(emission,absorption); break;
  case evDustAbs : dustAbs(); break; // New!
  }
}

// Handle electron scattering
void Packet::electronScatter() {

  // Re-emit packet
  reEmit();

  // Handle packet event
  handleEvent(PacketEvent(evElectronScatter,position,t,E,nu,ECMF,nuCMF,cell->index(),cell->subIndex(),index));

  return;
}

// Handle compton scattering (not supported yet)
void Packet::comptonScatter() {
  double ECMF;

  // Terminate
  pTerminated=true;

  // Handle packet event
  handleEvent(PacketEvent(evComptonScatter,position,t,E,nu,ECMF,nuCMF,cell->index(),cell->subIndex(),index));

  return;
}

// Handle grey scattering
void Packet::greyScatter() {

  // Re-emit packet
  reEmit();

  // Handle packet event
  handleEvent(PacketEvent(evGreyScatter,position,t,E,nu,ECMF,nuCMF,cell->index(),cell->subIndex(),index));

  return;
}

// Handle free-free absorption
//void Packet::freeFreeAbs() {
void Packet::freeFreeAbs(bool emission, bool absorption) { // Changed!
  PacketEventType eventType;
  Trans* trans;

  // If absorption (New!) - Handle packet event
  if (absorption) handleEvent(PacketEvent(evFreeFreeAbs,position,t,E,nu,ECMF,nuCMF,cell->index(),cell->subIndex(),index));

  // If no emission - Terminate packet and return - New!
  if (not emission) { terminate(); return; }

  // Convert to k-packet
  type=kPacket;

  // Process k-packet (to determine new frequency)
  trans=processKPacket(eventType);
  // Should probably handle this as an (failed) event
  if (trans==NULL and eventType!=evFreeFreeEmission) { reset(); return; }

  // Adjust energy (no effect in normal mode) - New!
  adjustEnergy();

  // Re-emit packet
  reEmit();

  // Handle packet event
  //handleEvent(PacketEvent(eventType,position,t,E,nu,ECMF,nuCMF,cell->index(),cell->subIndex(),index));
  handleEvent(PacketEvent(eventType,position,t,E,nu,ECMF,nuCMF,cell->index(),cell->subIndex(),index),(not absorption),trans); // Changed!

  // Select next resonant line
  selectNextLine();
  // Avoid self-absorption
  if (!endOfLineList()) if ((*line)==trans) line++;

  // Select next ionization edge
  selectNextEdge();

  // Select next boost bin - New!
  selectNextBoostBin();

  return;
}

// Handle bound-free absorption
//void Packet::boundFreeAbs() {
void Packet::boundFreeAbs(bool emission, bool absorption) { // Changed!
  double z,P;
  PacketEventType eventType;
  Trans* trans;
  MCChannel channel;
  
  // If absorption (New!) - Handle packet event
  if (absorption) handleEvent(PacketEvent(evBoundFreeAbs,position,t,E,nu,ECMF,nuCMF,cell->index(),cell->subIndex(),index));

  // If no emission - Terminate packet and return - New!
  if (not emission) { terminate(); return; }

  // Draw bound-free (upward) transition
  trans=cell->boundFreeExcite(nuCMF);
  if (trans==NULL) { 
    cout<<"Packet::boundFreeAbs: index: "<<index<<", cell: "<<cell->index()<<", Warning: no transition drawn, nuCMF: "<<nuCMF<<", lambdaCMF: "<<(c/nuCMF)/AA<<", lymanNu: "<<((Cell*) cell)->atom[0].iState[0].bState[0].upTransBF[0]->nu()<<", lymanLambda: "<<(c/((Cell*) cell)->atom[0].iState[0].bState[0].upTransBF[0]->nu())/AA<<", boundFreeAbsKappa: "<<boundFreeAbsKappa()<<", cell->boundFreeAbsKappa: "<<cell->boundFreeAbsKappa(nuCMF)<<"\n";
    if (!endOfEdgeList()) cout<<"Packet::boundFreeAbs: index: "<<index<<", cell: "<<cell->index()<<", edgeDistance: "<<edgeDistance()<<", cell->edgeDistance: "<<cell->edgeDistance(edge,nuCMF,t)<<", edge->nu: "<<(*edge)->nu()<<", edge->lambda: "<<(c/(*edge)->nu())/AA<<"\n";
    reset(); 
    return; 
  }

  // Handle packet event
  //handleEvent(PacketEvent(evBoundFreeAbs,position,t,E,nu,ECMF,nuCMF,cell->index(),cell->subIndex(),index));

  // Draw absorption channel - New!
  channel=((TransBF*) trans)->absorptionChannel();

  // If ionization drawn - New!
  if (channel==cIonization) {
    // Convert to i-packet
    type=iPacket;
    // Process i-packet (to determine new frequency)
    trans=processIPacket(eventType,trans);
    // Should probably handle this as an (failed) event
    if (trans==NULL and eventType!=evFreeFreeEmission) { reset(); cout<<"Packet::boundFreeAbs: Warning: No transition drawn in processIPacket\n"; return; }
  }
  // Else (i.e. heating)
  else {
    // Convert to k-packet
    type=kPacket;
    // Process k-packet (to determine new frequency)
    trans=processKPacket(eventType);
    // Should probably handle this as an (failed) event
    if (trans==NULL and eventType!=evFreeFreeEmission) { reset(); cout<<"Packet::boundFreeAbs: Warning: No transition drawn in processKPacket\n"; return; }
  }

  // Adjust energy (no effect in normal mode) - New!
  adjustEnergy();

  // Re-emit packet
  reEmit();

  // Handle packet event
  //handleEvent(PacketEvent(eventType,position,t,E,nu,ECMF,nuCMF,cell->index(),cell->subIndex(),index));
  handleEvent(PacketEvent(eventType,position,t,E,nu,ECMF,nuCMF,cell->index(),cell->subIndex(),index),(not absorption),trans); // Changed!

  // Select next resonant line
  selectNextLine();
  // Avoid self-absorption
  if (!endOfLineList()) if ((*line)==trans) line++;

  // Select next ionization edge
  selectNextEdge();

  // Select next boost bin - New!
  selectNextBoostBin();

  return;
}

// Handle Compton absorption
//void Packet::comptonAbs() {
void Packet::comptonAbs(bool emission, bool absorption) { // Changed!
  PacketEventType eventType;
  Trans* trans;

  // If absorption (New!)- Handle packet event - Removed! - Testing!
  //if (absorption) handleEvent(PacketEvent(evComptonAbs,position,t,E,nu,ECMF,nuCMF,cell->index(),cell->subIndex(),index));

  // Terminate g-packet
  pTerminated=true;

  // If convert g-packet and within (r-packet) cell boundaries
  if (config->convertGPacket and (cell->index()>=config->cellIndexMin or config->cellIndexMin==-1) and (cell->index()<=config->cellIndexMax or config->cellIndexMax==-1)) {

    // If no emission - Terminate packet and return - New!
    if (not emission) { terminate(); return; }

    // Adjust energy (no effect in normal mode) - New! - Testing!
    adjustEnergy();

    // Handle packet (absorption) event - New! - Testing!
    handleEvent(PacketEvent(evComptonAbs,position,t,E,nu,ECMF,nuCMF,cell->index(),cell->subIndex(),index));

    // Convert to e-packet
    type=ePacket;
    // Flag continuum event distance and opacity to be updated - New!
    pContinuumEventDistance=dUpdate;
    pContinuumEventKappa=kUpdate;

    // Process e-packet (to determine new frequency)
    trans=processEPacket(eventType);
    // Should probably handle this as an (failed) event
    //testLog=true;
    //cout<<"Packet::comptonAbs: index: "<<index<<", cellIndex: "<<cell->index()<<", eventType: "<<eventType<<"\n";
    //if (trans==NULL and (eventType!=evFreeFreeEmission or eventType!=evEmission)) { reset(); return; }
    if (trans==NULL and (eventType!=evFreeFreeEmission and eventType!=evEmission)) { reset(); return; } // Bug Fix!
    //cout<<"Packet::comptonAbs: index: "<<index<<", cellIndex: "<<cell->index()<<", 1\n";

    // Adjust energy (no effect in normal mode) - New! - Testing!
    adjustEnergy();

    // Re-emit packet
    reEmit();

    // Handle packet (emission) event
    //handleEvent(PacketEvent(eventType,position,t,E,nu,ECMF,nuCMF,cell->index(),cell->subIndex(),index));
    handleEvent(PacketEvent(eventType,position,t,E,nu,ECMF,nuCMF,cell->index(),cell->subIndex(),index),(not absorption),trans); // Changed!

    // Select next resonant line
    selectNextLine();
    // Avoid self-absorption
    if (!endOfLineList()) if ((*line)==trans) line++;

    // Select next ionization edge
    selectNextEdge();

    // Select next boost bin - New!
    selectNextBoostBin();
  }
  // Else
  //else if (absorption) {
  // Changed! Testing!!!
  else {

    // If no emission - Terminate packet and return - New! - Testing!!!
    if (not emission) { terminate(); return; }

    // Adjust energy (no effect in normal mode) - New! - Testing!!!
    adjustEnergy();

    // Handle packet (absorption) event - New! - Testing!
    handleEvent(PacketEvent(evComptonAbs,position,t,E,nu,ECMF,nuCMF,cell->index(),cell->subIndex(),index));
  }

  //cout<<"Packet::comptonAbs: index: "<<index<<", cellIndex: "<<cell->index()<<", 2\n";

  return;
}

// Handle line absorption
//void Packet::lineScatter() {
//void Packet::lineAbs() {
void Packet::lineAbs(bool emission, bool absorption) {
  PacketEventType eventType;
  Trans* trans;

  // If treatment of flourescence is on
  if (config->flourescence) {

    // If absorption (New!) - Handle packet event
    if (absorption) handleEvent(PacketEvent(evLineAbs,position,t,E,nu,ECMF,(*line)->nu(),cell->index(),cell->subIndex(),index));

    // If no emission - Terminate packet and return - New!
    if (not emission) { terminate(); return; }

    // Convert to i-packet
    type=iPacket;

    // Process i-packet (to determine new frequency)
    trans=processIPacket(eventType,*line);
    if (trans==NULL and eventType!=evFreeFreeEmission) { reset(); return; }

    // Adjust energy (no effect in normal mode) - New!
    adjustEnergy();
  }

  // Re-emit packet
  reEmit();

  // Handle packet event
  //if (config->flourescence) handleEvent(PacketEvent(eventType,position,t,E,nu,ECMF,nuCMF,cell->index(),cell->subIndex(),index));
  if (config->flourescence) handleEvent(PacketEvent(eventType,position,t,E,nu,ECMF,nuCMF,cell->index(),cell->subIndex(),index),(not absorption),trans); // Changed!
  else handleEvent(PacketEvent(evLineScatter,position,t,E,nu,ECMF,(*line)->nu(),cell->index(),cell->subIndex(),index));

  // If treatment of flourescence is on
  if (config->flourescence) {
    
    // Select next resonant line
    selectNextLine();
    // Avoid self-absorption
    if (!endOfLineList()) if ((*line)==trans) line++;

    // Select next ionization edge
    selectNextEdge();

    // Select next boost bin - New!
    selectNextBoostBin();
  }
  // Step foward in the line list
  else if (!endOfLineList()) line++;

  return;
}

// Handle dust absorption - New!
void Packet::dustAbs() {
  double ECMF;

  // Terminate
  pTerminated=true;

  // Handle packet event
  handleEvent(PacketEvent(evDustAbs,position,t,E,nu,ECMF,nuCMF,cell->index(),cell->subIndex(),index));

  return;
}

// Re-emit packet isotropically
//void Packet::scatter(double m) {
void Packet::reEmit(double m) {
  double rPosition,thetaPosition,phiPosition,thetaOut,phiOut;
  double muOutCMF,muOut,EOut,nuOut;
  Vector3D directionOut;

  if (testLog and index%100==0) cout<<"Packet::reEmit: index: "<<index<<", cellIndex: "<<cell->index()<<", Enter\n";

  // Select a random (cosine of the) polar angle (about the radius vector) in the comoving frame
  muOutCMF=-1+2*randomize(0,1);
  // Transform the (cosine of the) polar angle to the rest frame
  muOut=(muOutCMF+(v/c))/(1+muOutCMF*(v/c));
  // Select a random equatorial angle (about the radius vector)
  phiOut=2*pi*randomize(0,1);
  // Calculate the polar angle
  thetaOut=acos(muOut);
  // Update the direction
  position.coords(rPosition,thetaPosition,phiPosition,spherical);
  directionOut=Vector3D(1,thetaOut,phiOut,spherical);
  directionOut.rotate(thetaPosition,phiPosition);
  // Calculate the new (rest frame) frequency
  nuOut=nuCMF/(1-muOut*(v/c));
  // Calculate the new (rest frame) energy
  EOut=ECMF/(1-muOut*(v/c));
  // Update the direction
  direction=directionOut;
  // Update the frequency
  nu=nuOut;
  // Update the energy
  E=EOut;
  // Update the (cosine of the) polar angle about the radius vector
  mu=muOut;
  // Select a new random optical depth for the next event
  eventTau=-log(randomize(0,1));
  // Reset the accumulated optical depth
  accumulatedTau=0;

  // Flag border distance to be updated
  pBorderDistance=dUpdate;

  // Flag continuum event distance to be updated (as this depends on eventTau and accumulatedTau) - New!
  pContinuumEventDistance=dUpdate;

  if (testLog and index%100==0) cout<<"Packet::reEmit: index: "<<index<<", cellIndex: "<<cell->index()<<", eventTau: "<<eventTau<<", accumulatedTau: "<<accumulatedTau<<", Return\n";

  return;
}

// Select next cell
void Packet::selectNextCell() {
  cell=cell->selectCellNeighbour(position,direction,t);
  if (cell!=NULL) {
    // Testing!!!
    /*if (cell!=NULL) if (type!=gPacket and cell->index()<=6 and ((Cell*) cell)->grid()->master() and omp_get_thread_num()==0) {
	double rInner=((CellSS*) ((Cell*) cell)->grid()->cell[config->cellIndexMin])->rMinT(t);
	cout<<"Packet::selectNextCell: index: "<<index<<", cellIndex: "<<cell->index()<<", subCellIndex: "<<cell->subIndex()<<", cellIndexMin: "<<config->cellIndexMin<<", rInner: "<<"\n";
	}*/
    //if (cell!=NULL) if (type!=gPacket and cell->index()<=6) logStream<<"Packet::selectNextCell: index: "<<index<<", cellIndex: "<<cell->index()<<", subCellIndex: "<<cell->subIndex()<<", cellIndexMin: "<<config->cellIndexMin<<"\n";
    if (cell->index()<config->cellIndexMin and config->cellIndexMin!=-1 and type!=gPacket) { 
      //cout<<"Packet::selectNextCell: index: "<<index<<", cellIndex: "<<cell->index()<<", subCellIndex: "<<cell->subIndex()<<", cellIndexMin: "<<config->cellIndexMin<<"\n";
      cell=NULL; return; 
    } 
    if (cell->index()>config->cellIndexMax and config->cellIndexMax!=-1 and type!=gPacket) { cell=NULL; return; }
  }
}

// Is inner border crossing - Changed!
// Note: Assumes SS-like geometry
bool Packet::innerBorderCrossing() {

  // If innermost cell and inner border crossing - Return true
  if (cell->index()==config->cellIndexMin and borderType()==btCellBorderInner) return true;
  // Else - Return false
  else return false;
  // Else - Another fucking hack!!!
  /*else 
    if (cell->index()==config->cellIndexMin and borderType()==btSubCellBorder) {
      Grid* grid=((Cell*) cell)->grid();
      CellSS* cellInner=(CellSS*) grid->cell[(config->cellIndexMin!=-1 and type!=gPacket)?config->cellIndexMin:0];
      if (position.norm()<cellInner->rMinT(t)) return true;
      else return false;
    }
    else return false;*/
}

// Is outer border crossing - Changed!
// Note: Assumes SS-like geometry
bool Packet::outerBorderCrossing() {

  // If outermost cell and outer border crossing - Return true
  if (cell->index()==config->cellIndexMax and borderType()==btCellBorderOuter) return true;
  // Else - Return false
  else return false;
  // Else - Another fucking hack!!!
  /*else 
    if (cell->index()==config->cellIndexMax and borderType()==btSubCellBorder) {
      Grid* grid=((Cell*) cell)->grid();
      CellSS* cellOuter=(CellSS*) grid->cell[(config->cellIndexMax!=-1 and type!=gPacket)?config->cellIndexMax:grid->cell.size()-1];
      if (position.norm()>cellOuter->rMaxT(t)) return true;
      else return false;
    }
    else return false;*/
}

// Is g-packet inner border crossing - New!
// Note: Assumes SS-like geometry
bool Packet::innerBorderCrossingGP() {

  // If innermost cell and inner border crossing - Return true
  if (cell->index()==0 and borderType()==btCellBorderInner) return true;
  // Else - Return false
  else return false;
}


// Enter new cell
void Packet::newCell() {
  double smallStep,z;
  double boost;
  PacketEventType eventType; // New!

  // If reflect packets at inner border and not g-packet
  if (config->reflectInner and type!=gPacket) {

    // If inner border crossing
    if (innerBorderCrossing()) {

      // Pick random number
      z=randomize(0,1);

      // If reflection drawn
      if (z<config->PReflectInner) {

	// Reflect
	reflect();

	// Take a small step
	smallStep=fabs(1e-10*position.norm()/mu);	
	updatePosition(smallStep);      
	
	// Return
	return;
      }
    }
  }

  // If reflect g-packets at inner border and g-packet - New!
  if (config->reflectInnerGP and type==gPacket) {

    // If g-packet inner border crossing
    if (innerBorderCrossingGP()) {

      // Pick random number
      z=randomize(0,1);

      // Reflect
      reflect();

      // Take a small step
      smallStep=fabs(1e-10*position.norm()/mu);	
      updatePosition(smallStep);      
      
      // Return
      return;
    }
  }

  // If reflect packets at outer border and not g-packet
  if (config->reflectOuter and type!=gPacket) {

    // If outer border crossing
    if (outerBorderCrossing()) {

      // Pick random number
      z=randomize(0,1);

      // If reflection drawn
      if (z<config->PReflectOuter) {

	// Reflect
	reflect();

	// Take a small step
	smallStep=fabs(1e-10*position.norm()/mu);
	updatePosition(smallStep);      
	
	// Return
	return;
      }
    }
  }

  // Take a small step
  smallStep=fabs(1e-10*position.norm()/mu);
  updatePosition(smallStep);

  // Set (old cell) event type depending on border type - New!
  switch(borderType()) {
  case btCellBorder : { eventType=evCellCrossing; break; }
  case btCellBorderInner : { eventType=evCellCrossingInner; break; }
  case btCellBorderOuter : { eventType=evCellCrossingOuter; break; }
  case btSubCellBorder : { eventType=evSubCellCrossing; break; }
  default : eventType=evTimeBorderCrossing;
  }

  // Handle (old cell) packet event - Changed!
  handleEvent(PacketEvent(eventType,position,t,E,nu,ECMF,nuCMF,cell->index(),cell->subIndex(),index));

  // Select next cell
  selectNextCell();

  // If on grid
  if (onGrid()) {

    // Set (new cell) event type depending on border type - New!
    switch(borderType()) {
    case btCellBorder : { eventType=evCellCrossing; break; }
    case btCellBorderInner : { eventType=evCellCrossingOuter; break; }
    case btCellBorderOuter : { eventType=evCellCrossingInner; break; }
    case btSubCellBorder : { eventType=evSubCellCrossing; break; }
    default : eventType=evTimeBorderCrossing;
    }
    
    // Handle (new cell) packet event - Changed!
    handleEvent(PacketEvent(eventType,position,t,E,nu,ECMF,nuCMF,cell->index(),cell->subIndex(),index));

    //if (((Cell*) cell)->master()) cout<<"Packet::newCell(1): "<<omp_get_thread_num()<<", index: "<<index<<", cellIndex: "<<cell->index()<<", lambda: "<<(c/nu)/AA<<"\n";

    // Select next resonant line (by searching the cell line list)
    selectNextLine();

    //if (((Cell*) cell)->master()) cout<<"Packet::newCell(2): "<<omp_get_thread_num()<<", index: "<<index<<", cellIndex: "<<cell->index()<<", lambda: "<<(c/nu)/AA<<"\n";

    // Select next ionization edge (by searching the cell ionization edge list)
    selectNextEdge();

    //if (((Cell*) cell)->master()) cout<<"Packet::newCell(3): "<<omp_get_thread_num()<<", index: "<<index<<", cellIndex: "<<cell->index()<<", lambda: "<<(c/nu)/AA<<"\n";

    // Select next boost bin (by searching the cell boost bin list) - New!
    selectNextBoostBin();

    //if (((Cell*) cell)->master()) cout<<"Packet::newCell(4): "<<omp_get_thread_num()<<", index: "<<index<<", cellIndex: "<<cell->index()<<", lambda: "<<(c/nu)/AA<<"\n";

    // If packet history update method
    if (config->updateMethod==urPacketHistory) {
      // Set cell and subcell index to no cell
      history[history.size()-1].cellIndex=cell->index();
      history[history.size()-1].subCellIndex=cell->subIndex();
    }
  }
  // Else
  else {

    // If packet history update method
    if (config->updateMethod==urPacketHistory) {
      // Set cell and subcell index to no cell
      history[history.size()-1].cellIndex=noCell;
      history[history.size()-1].subCellIndex=noCell;
    }
  }

  // Flag all distances to be updated
  pBorderDistance=dUpdate;
  pContinuumEventDistance=dUpdate;
  pLineDistance=dUpdate;
  pEdgeDistance=dUpdate;
  pBoostBinDistance=dUpdate; // New!
  
  // Flag all continuum opacities to be updated
  pElectronScatterKappa=kUpdate;
  pComptonScatterKappa=kUpdate;
  pGreyScatterKappa=kUpdate; // New!
  pFreeFreeAbsKappa=kUpdate;
  pBoundFreeAbsKappa=kUpdate;
  pComptonAbsKappa=kUpdate;
  pDustAbsKappa=kUpdate; // New!
  pContinuumEventKappa=kUpdate;

  // If boost mode, on grid and not a g-packet
  if (config->boost and onGrid() and type!=gPacket) {
  // Testing!
  //if (config->boost and onGrid()) {

    // Calculate relative boost factor
    // Testing!
    if (type!=gPacket) boost=cell->boostFactor(nuCMF)/boostFactor;
    else boost=cell->comptonKappaBoost()/boostFactor;

    // If larger than one - Split packet
    if (boost>1) borderSplit();
    // Else - Merge packet
    else borderMerge();
  }

  // If on grid
  //if (onGrid()) {
    
    // Select a new random optical depth for the next event
    //eventTau=-log(randomize(0,1));
    // Reset the accumulated optical depth
    //accumulatedTau=0;
  //}
}

// Enter new boost bin
void Packet::newBoostBin() {
  double small=1e-12;

  // If end of boost bin list - Just return;
  if (endOfBoostBinList()) return;

  // Make sure the comoving frequency is just redwards the bin edge
  nuCMF=boostBin->min*(1.0-small);
  nu=nuCMF/(1-mu*(v/c));

  // Flag the free-free, bound-free and continuum opacity to be updated
  pBoundFreeAbsKappa=kUpdate;
  pFreeFreeAbsKappa=kUpdate;
  pContinuumEventKappa=kUpdate;
  
  // Flag the continuum, line, edge and boost bin distance to be updated
  pContinuumEventDistance=dUpdate;
  pLineDistance=dUpdate;
  pEdgeDistance=dUpdate;
  pBoostBinDistance=dUpdate;

  // Select the next boost bin (by stepping the zone list)
  // Hack for the first bin - Fix!
  //if (boostBin!=((Cell*) cell)->nuBoost.begin()) boostBin--;
  //else boostBin=((Cell*) cell)->nuBoost.end();
  if (not cell->beginOfBoostBinList(boostBin)) boostBin--;
  else boostBin=cell->endOfBoostBinList();
}

// Pass bound-free new edge - New (moved here from propagateRPacket)!
void Packet::boundFreeEdge() {

  // Testing!
  if (cell==NULL) { cout<<"Packet::boundFreeEgde: cell=NULL\n"; exit(0); }
  if (endOfEdgeList()) { cout<<"Packet::boundFreeEdge: endOfEdgeList, edgeDistance: "<<edgeDistance()<<", continuumEventDistance: "<<continuumEventDistance()<<"\n"; exit(0); }

  // Update bound-free opacity
  pBoundFreeAbsKappa=cell->boundFreeAbsKappa((*edge)->nu(),redEdge);

  // If boost mode - Boost opacity - New!
  if (config->boost) pBoundFreeAbsKappa*=MAX(cell->boundFreeKappaBoost((*edge)->nu(),redEdge)/boostFactor,1.0);

  // Flag continuum opacity to be updated
  pContinuumEventKappa=kUpdate;

  // Select the next ionizaton edge (by stepping the zone list)
  if (!endOfEdgeList()) edge++;

  // Flag edge distance to be updated - New!
  pEdgeDistance=dUpdate;
}

// Reflect packet
void Packet::reflect() {
  double rD,thetaD,phiD,rP,thetaP,phiP,muCMF;
  double smallStep;

  // Get direction and position in spherical coordinates
  direction.coords(rD,thetaD,phiD,spherical);
  position.coords(rP,thetaP,phiP,spherical);

  // Transform the (cosine of the) polar angle to the comoving frame
  muCMF=(mu-(v/c))/(1-mu*(v/c));

  // Reverse direction
  muCMF=-muCMF;

  // Transform the (cosine of the) polar angle to the rest frame
  mu=(muCMF+(v/c))/(1+muCMF*(v/c));
  
  // Calculate the new polar angle
  thetaD=acos(mu);

  // Update direction
  direction=Vector3D(1,thetaD,phiD,spherical);
  direction.rotate(thetaP,phiP);

  // Calculate the new (rest frame) frequency
  nu=nuCMF/(1-mu*(v/c));
  // Calculate the new (rest frame) energy
  E=ECMF/(1-mu*(v/c));

  // Handle packet event
  handleEvent(PacketEvent(evReflect,position,t,E,nu,ECMF,nuCMF,cell->index(),cell->subIndex(),index));

  // Select a new random optical depth for the next event
  eventTau=-log(randomize(0,1));

  // Reset the accumulated optical depth
  accumulatedTau=0;

  // Flag border distance to be updated
  pBorderDistance=dUpdate;
}

// Handle packet event
/*void Packet::handleEvent(const PacketEvent& event) {

  // If on-the-fly update method
  if (config->updateMethod==urOnTheFly) {
    // If last event not handled and pointer to packet statistics object provided - Update event statistics
    if ((not lastEventHandled) and packetStat!=NULL) packetStat->updateEvent(lastEvent);
    // Update cell MC statistics
    if (type!=gPacket) cell->updateMCStatistics(lastEvent,event,mCUpdateInfo);
    else cell->updateGMCStatistics(lastEvent,event); 
    // Update last event
    lastEvent=event;
    lastEventHandled=true;
    // If pointer to packet statistics object provided - Update event statistics
    if (packetStat!=NULL) packetStat->updateEvent(event);
  }
  // Else append the event to the cell packet history
  else history.push_back(event);
  }*/

// Handle packet event - Changed!
void Packet::handleEvent(const PacketEvent& event, bool initial, Trans* trans) {
  
  // Testing!!!
  //if (cell!=NULL) if (type!=gPacket and cell->index()==6 and ((Cell*) cell)->grid()->master()) cout<<"Packet::handleEvent: index: "<<index<<", cellIndex: "<<cell->index()<<", subCellIndex: "<<cell->subIndex()<<", eventType: "<<logPacketEventType(event.type)<<"\n";
  //if (cell!=NULL) if (type!=gPacket and cell->index()==6) logStream<<"Packet::handleEvent: index: "<<index<<", cellIndex: "<<cell->index()<<", subCellIndex: "<<cell->subIndex()<<", eventType: "<<logPacketEventType(event.type)<<"\n";

  if (testLog and index%100==0) cout<<"Packet::handleEvent: index: "<<index<<", cellIndex: "<<cell->index()<<", eventType: "<<logPacketEventType(event.type)<<"\n";

  // Testing!
  //if (((Cell*) cell)->master()) cout<<"Packet::handleEvent: "<<omp_get_thread_num()<<", index: "<<index<<", cellIndex: "<<cell->index()<<", eventType: "<<logPacketEventType(event.type)<<"\n";
  
  // If on-the-fly update method
  if (config->updateMethod==urOnTheFly) {
    // If not initial event
    if (not initial) {
      // If last event not handled and pointer to packet statistics object provided - Update event statistics
      if ((not lastEventHandled) and packetStat!=NULL) packetStat->updateEvent(lastEvent);
      // If not dummy emission - New!
      if (event.type!=evDummyEmission) {
	// Update cell MC statistics
	if (type!=gPacket) cell->updateMCStatistics(lastEvent,event,mCUpdateInfo);
	else cell->updateGMCStatistics(lastEvent,event);
      }
      // If pointer to packet statistics object provided - Update event statistics
      if (packetStat!=NULL) packetStat->updateEvent(event);
      // Set last event handled flag
      lastEventHandled=true;
    }
    // Else - unset last event handled flag
    else lastEventHandled=false;
    // Update last event
    lastEvent=event;
    // If (true or line) emission event - Update last emission event
    if (isEmissionEvent(event.type) or event.type==evLineEmission) lastEmissionEvent=PacketEmissionEvent(event,trans);
    }
  // Else append the event to the cell packet history
  else history.push_back(event);
}

// Process i-packet
Trans* Packet::processIPacket(PacketEventType& eventType, Trans* trans) {
  TransType transType;
  double P,nuCMFOld;
  bool deActivate=false;
  int etaType;
  CTReaction* cTReaction; // New!
  CTDirection cTDirection; // New!
  MCChannel channel; // New!

  //bool testLog=false;
  //if (trans->upperState->iState()->atom()->name().compare("He")==0 and cell->index()==1) testLog=true;
  //if (testLog) cout<<"Packet::processIPacket: packetIndex: "<<index<<", Excite: "<<trans->lowerState->iState()->name()<<":"<<trans->lowerState->getIndex()<<"->"<<trans->upperState->iState()->name()<<":"<<trans->upperState->getIndex()<<", twoPhoton: "<<config->twoPhotonProc<<"\n";

  // De-excite using macro-atom machinery
  trans=trans->upperState->deExcite(transType,trans->upperState,deActivate,config->boundFreeAbs,config->collProc,config->nonThermalProc,config->twoPhotonProc,config->chargeTransferProc); // Changed!
  if (trans==NULL and not deActivate) {
    cout<<"Packet::processIPacket: Warning(1): No transition drawn in processIPacket, index: "<<index<<", cell: "<<cell->index()<<"\n";
    return NULL;
  }

  //if (testLog and trans!=NULL) cout<<"Packet::processIPacket: packetIndex: "<<index<<", deExcite: "<<trans->upperState->iState()->name()<<":"<<trans->upperState->getIndex()<<"->"<<trans->lowerState->iState()->name()<<":"<<trans->lowerState->getIndex()<<", transType: "<<transType<<"\n";

  //if (cell->index()==51) cout<<"Packet::processIPacket: cellIndex: "<<cell->index()<<", packetIndex: "<<index<<", trans: "<<trans<<", transType: "<<transType<<", deActivate: "<<deActivate<<"\n";

  // If radiative bound-bound de-excitation
  if (transType==radiativeBB) {
    // Draw emission type - New!
    etaType=((TransBB* )trans)->etaType();
    //if (testLog) cout<<"Packet::processIPacket: packetIndex: "<<index<<", etaType: "<<etaType<<", etaType&pTwoPhoton: "<<(etaType&pTwoPhoton)<<"\n";
    // Backup old comoving frame frequency
    nuCMFOld=nuCMF;
    // Update comoving frequency - Changed!
    if (etaType&pTwoPhoton) nuCMF=((TransBB*) trans)->eta2Nu();
    else nuCMF=trans->nu();
    // Update rest frame frequency
    nu=nu*(nuCMF/nuCMFOld);
    // Set packet event type - Changed!
    if (etaType&pTwoPhoton) eventType=evTwoPhotonEmission;
    else eventType=evLineEmission;
    // Convert to r-packet
    type=rPacket;
    // If comoving frequency changed
    if (nuCMF!=nuCMFOld) {
      // Flag frequency dependent continuum opacities to be updated
      pContinuumEventKappa=kUpdate;
      pFreeFreeAbsKappa=kUpdate;
      pBoundFreeAbsKappa=kUpdate;
    }
  }
  // Else - If radiative bound-free de-excitation
  else if (transType==radiativeBF) {
    // Draw emission type - New! - Removed!
    //etaType=((TransBF* )trans)->etaType();
    // Backup old comoving frame frequency
    nuCMFOld=nuCMF;
    // Update comoving frequency
    nuCMF=((TransBF*) trans)->etaNu();
    // Update rest frame frequency
    nu=nu*(nuCMF/nuCMFOld);
    // Set packet event type - Changed! - Changed back!
    //if (etaType&pChargeTransfer) eventType=evDummyEmission;
    //else eventType=evBoundFreeEmission;
    eventType=evBoundFreeEmission;
    // Convert to r-packet
    type=rPacket;
    // If comoving frequency changed
    if (nuCMF!=nuCMFOld) {
      // Flag frequency dependent continuum opacities to be updated
      pContinuumEventKappa=kUpdate;
      pFreeFreeAbsKappa=kUpdate;
      pBoundFreeAbsKappa=kUpdate;
    }
    // If dummy emission - Mark packet as terminated - New! - Removed!
    //if (eventType==evDummyEmission) pTerminated=true;
  }
  // Else - If charge-transfer bound-free de-excitation - New!
  else if (transType==chargeTransferBF) {
    // Draw charge-transfer reaction
    //cTReaction=((TransBF*) trans)->drawCTDReaction();
    cTReaction=((TransBF*) trans)->drawCTReaction(cTDirection);
    // Draw absorption channel
    //channel=cTReaction->absorptionChannel();
    channel=cTReaction->absorptionChannel(cTDirection);
    // If ionization - Re-process i-packet (using the charge-transfer bound-free excitation)
    //if (channel==cIonization) trans=processIPacket(eventType,cTReaction->upTrans);
    if (channel==cIonization) 
      if (cTDirection==ctForward) trans=processIPacket(eventType,cTReaction->upTrans);
      else trans=processIPacket(eventType,cTReaction->downTrans);
    // Else (i.e. heating)
    else {
      // Convert to k-packet
      type=kPacket;
      // Process k-packet
      processKPacket(eventType,false);
    }
  }
  // Else (i.e. collisional de-excitation)
  else {
    // Convert to k-packet
    type=kPacket;
    // Process k-packet
    trans=processKPacket(eventType,deActivate);
  }

  if (trans==NULL and eventType!=evFreeFreeEmission) cout<<"Packet::processIPacket: Warning(2): No transition drawn in processIPacket, index: "<<index<<", cell: "<<cell->index()<<"\n";

  return trans;
}

// Process k-packet
Trans* Packet::processKPacket(PacketEventType& eventType, bool deActivate) {
  double nuCMFOld;
  Trans* trans;
  MCChannel channel;

  // Draw cooling channel - New!
  channel=cell->coolingChannel(deActivate);

  // If free-free cooling drawn
  if (channel==cFreeFree) {
    // Backup old comoving frame frequency
    nuCMFOld=nuCMF;
    // Update comoving frame frequency
    nuCMF=cell->freeFreeNu();
    // Update rest frame frequency
    nu=nu*(nuCMF/nuCMFOld);
    // Set packet event type
    eventType=evFreeFreeEmission;
    // Convert to r-packet
    type=rPacket;
    // If comoving frequency changed
    if (nuCMF!=nuCMFOld) {
      // Flag frequency dependent continuum opacities to be updated
      pContinuumEventKappa=kUpdate;
      pFreeFreeAbsKappa=kUpdate;
      pBoundFreeAbsKappa=kUpdate;
    }
    // Return
    return NULL;
  }

  // If bound-free cooling drawn
  if (channel==cBoundFree) {
    // Backup old comoving frame frequency
    nuCMFOld=nuCMF;
    // Draw bound-free (downward) transition
    trans=cell->boundFreeDeExcite();
    if (trans==NULL) return NULL;
    // Update comoving frame frequency
    nuCMF=((TransBF*) trans)->etaNu();
    // Update rest frame frequency
    nu=nu*(nuCMF/nuCMFOld);
    // Set packet event type
    eventType=evBoundFreeEmission;
    // Convert to r-packet
    type=rPacket;
    // If comoving frequency changed
    if (nuCMF!=nuCMFOld) {
      // Flag frequency dependent continuum opacities to be updated
      pContinuumEventKappa=kUpdate;
      pFreeFreeAbsKappa=kUpdate;
      pBoundFreeAbsKappa=kUpdate;
    }
    if (trans==NULL) cout<<"Packet::processKPacket: Warning(1): No transition drawn in processIPacket, index: "<<index<<", cell: "<<cell->index()<<"\n";
    // Return
    return trans;
  }

  // If charge-transfer cooling drawn
  if (channel==cChargeTransfer) {
    // Draw charge-transfer transition
    trans=cell->CTExcite();
    if (trans==NULL) return NULL;
    // Convert to i-packet
    type=iPacket;
    // Process i-packet
    trans=processIPacket(eventType,trans);
    if (trans==NULL) cout<<"Packet::processKPacket: Warning(2): No transition drawn in processIPacket, index: "<<index<<", cell: "<<cell->index()<<"\n";
    // Return
    return trans;
  }

  // If collisional cooling drawn
  if (channel==cCollisional) {
    // Draw collisional transition
    trans=cell->collExcite();
    if (trans==NULL) return NULL;
    // Convert to i-packet
    type=iPacket;
    // Process i-packet
    trans=processIPacket(eventType,trans);
    // Return
    if (trans==NULL and eventType!=evFreeFreeEmission) cout<<"Packet::processIPacket: Warning(3): No transition drawn in processIPacket, index: "<<index<<", cell: "<<cell->index()<<"\n";
    return trans;
  }

  if (trans==NULL) cout<<"Packet::processKPacket: Warning(4): No transition drawn in processIPacket, index: "<<index<<", cell: "<<cell->index()<<"\n";

  // This should not happen
  return NULL;
}

// Process e-packet
Trans* Packet::processEPacket(PacketEventType& eventType) {
  Trans* trans;
  MCChannel channel;

  // If requested - Convert e-packet to BB r-packet
  if (config->bbEPacket) return processBBEPacket(eventType);

  // Draw non-thermal channel - New!
  channel=cell->nonThermalChannel();

  // If non-thermal heating drawn - New!
  if (channel==cHeating) {
    // Convert to k-packet
    type=kPacket;
    // Process k-packet
    trans=processKPacket(eventType);
    // Return
    return trans;
  }

  // If non-thermal ionization or excitation drawn - New
  if (channel==cIonization or channel==cExcitation) {
    // Draw non-thermal transition
    trans=cell->nonThermalExcite();
    if (trans==NULL) return NULL;
    // Convert to i-packet
    type=iPacket;
    // Process i-packet
    trans=processIPacket(eventType,trans);
    // Return
    return trans;
  }

  // This should not happen
  return NULL;
}

// Process BB e-packet
Trans* Packet::processBBEPacket(PacketEventType& eventType) {
  double z,nuMin,nuMax,nuCMFOld,B;

  // Backup old comoving frame frequency
  nuCMFOld=nuCMF;
  // Update comoving frame frequency
  z=randomize(0,1);
  nuMin=1e-3*k*MIN(cell->T(),cell->TR())/h;
  nuMax=1e2*k*MAX(cell->T(),cell->TR())/h;
  B=BlackBody::B(nuMin,nuMax,cell->T(),1e-4);
  nuCMF=BlackBody::nuB(z*B,cell->T(),nuMin,1e-3);
  // Update rest frame frequency
  nu=nu*(nuCMF/nuCMFOld);
  // Set packet event type
  eventType=evEmission;
  // Convert to r-packet
  type=rPacket;

  return NULL;
}

// Propagate packet
void Packet::propagate(PacketStat* _packetStat) {
  // Set pointer to packet statistics object
  packetStat=_packetStat;
  // If g-packet - propagate
  if (type==gPacket) propagateGPacket();
  // If r-packet
  if (type==rPacket) {
    // Set the packet terminated flag to false
    pTerminated=false;
    // Propagate r-packet
    propagateRPacket();
  }
}

// Propagate r-packet
void Packet::propagateRPacket() {
  PacketEventType eventType;
  bool emission,absorption;

  // While the packet is on the grid and not terminated
  while (onGrid() and not terminated()) {
    // Testing!
    if (testLog and index%100==0) cout<<"Packet::propagetRPacket(1): index: "<<index<<", cellIndex: "<<cell->index()<<", borderDistance: "<<borderDistance()<<", continuumEventDistance: "<<continuumEventDistance()<<", edgeDistance: "<<edgeDistance()<<", lineDistance: "<<lineDistance()<<", boostBinDistance: "<<boostBinDistance()<<"\n"; 
    // Testing
    //if (((Cell*) cell)->master()) cout<<"Packet::propagetRPacket(0): "<<omp_get_thread_num()<<", index: "<<index<<", cellIndex: "<<cell->index()<<"\n"; 
    //if (((Cell*) cell)->master()) cout<<"Packet::propagetRPacket(1): "<<omp_get_thread_num()<<", index: "<<index<<", cellIndex: "<<cell->index()<<", borderDistance: "<<borderDistance()<<", continuumEventDistance: "<<continuumEventDistance()<<", edgeDistance: "<<edgeDistance()<<", lineDistance: "<<lineDistance()<<", boostBinDistance: "<<boostBinDistance()<<"\n"; 
    // While the distance to the cell border is not the least and the packet is not terminated
    //while ((lineDistance()<borderDistance() | continuumEventDistance()<borderDistance() | edgeDistance()<borderDistance()) and not terminated()) {
    while ((lineDistance()<borderDistance() | continuumEventDistance()<borderDistance() | edgeDistance()<borderDistance()) | boostBinDistance()<borderDistance() and not terminated()) { // New!
      // Testing!
      if (testLog and index%100==0) cout<<"Packet::propagetRPacket(2): index: "<<index<<", cellIndex: "<<cell->index()<<", borderDistance: "<<borderDistance()<<", continuumEventDistance: "<<continuumEventDistance()<<", edgeDistance: "<<edgeDistance()<<", lineDistance: "<<lineDistance()<<", boostBinDistance: "<<boostBinDistance()<<"\n";
      // Testing!
      //if (((Cell*) cell)->master()) cout<<"Packet::propagetRPacket(2): "<<omp_get_thread_num()<<", index: "<<index<<", cellIndex: "<<cell->index()<<", borderDistance: "<<borderDistance()<<", continuumEventDistance: "<<continuumEventDistance()<<", edgeDistance: "<<edgeDistance()<<", lineDistance: "<<lineDistance()<<", boostBinDistance: "<<boostBinDistance()<<"\n";
      // If the distance to the next line is the least
      //if (lineDistance()<continuumEventDistance() and lineDistance()<edgeDistance()) {
      if (lineDistance()<continuumEventDistance() and lineDistance()<edgeDistance() and lineDistance()<boostBinDistance()) { // New!
	// Update the accumulated optical depth
	accumulatedTau+=lineTau();
	// If the accumulated optical depth at the line scatter position is larger than the optical depth for the next event
	if (accumulatedTau+continuumEventTau(lineDistance())>eventTau) {
	  // Update the position
	  updatePosition(lineDistance());
	  // Get event treatment (no effect in normal mode) - New!
	  getEventTreatment(evLineAbs,emission,absorption);
	  //if (not emission) cout<<"Packet::propagateRPacket: index: "<<index<<", cell: "<<cell->index()<<", eventType: "<<evLineAbs<<", emission: "<<emission<<", absorption: "<<absorption<<"\n";
	  // If no absorption - Split and emit child packet - New!
	  if (not absorption) emissionSplit(evLineAbs);
	  // Else - Handle line absorption (allways done in normal mode)
	  else lineAbs(emission);
	  // Testing!
	  //lineAbs();
	}
	// Else select next resonant line (by stepping the zone linelist)
	else if (!endOfLineList()) line++;
	// Flag the line and continuum distances to be updated
	pLineDistance=dUpdate;
	pContinuumEventDistance=dUpdate;
      }
      // Else
      else {
	// If the distance to the next continuum event is the least
	//if (continuumEventDistance()<edgeDistance()) {
	if (continuumEventDistance()<edgeDistance() and continuumEventDistance()<boostBinDistance()) { // New!
	  // Get continuum event type
	  eventType=getContinuumEventType();
	  // Update the position
	  updatePosition(continuumEventDistance());
	  // Get event treatment (no effect in normal mode) - New!
	  getEventTreatment(eventType,emission,absorption);
	  if (testLog and index%100==0) cout<<"Packet::propagetRPacket(3): index: "<<index<<", cellIndex: "<<cell->index()<<", eventType: "<<logPacketEventType(eventType)<<", emission: "<<emission<<", absorption: "<<absorption<<"\n"; 
	  //if (not emission) cout<<"Packet::propagateRPacket: index: "<<index<<", cell: "<<cell->index()<<", eventType: "<<eventType<<", emission: "<<emission<<", absorption: "<<absorption<<"\n";
	  // If no absorption - Split and emit child packet - New!
	  if (not absorption) emissionSplit(eventType);
	  // Else - Handle the continuum event (allways done in normal mode)
	  else continuumEvent(eventType,emission);
	}
	// Else
	else {
	  // If the distance to the next ionization edge is the least - New!
	  if (edgeDistance()<boostBinDistance()) {
	    // Update position
	    updatePosition(edgeDistance());
	    // Pass ionization edge
	    boundFreeEdge();
	  }
	  // Else (i.e. the distance to the next boost bin is the least) - New!
	  else {
	    // Update position
	    updatePosition(boostBinDistance());
	    // Enter new boost bin
	    newBoostBin();
	  }
	}
      }
    }
    // If not terminated
    if (not terminated()) {
      // Update the position
      updatePosition(borderDistance());
      // Enter new cell
      newCell();
      //if (((Cell*) cell)->master()) cout<<"Packet::propagetRPacket(3): "<<omp_get_thread_num()<<", index: "<<index<<", cellIndex: "<<cell->index()<<"\n"; 
    }
  }
  
  return;
}

// Propagate g-packet
void Packet::propagateGPacket() {
  PacketEventType eventType;
  bool emission,absorption;

  // While the packet is on the grid and not terminated
  while (onGrid() and not terminated()) {
    if (testLog and index%100==0) cout<<"Packet::propagetGPacket: index: "<<index<<", cellIndex: "<<cell->index()<<", borderDistance: "<<borderDistance()<<"\n"; 
    //if ((index%10)==0) cout<<"Packet::propagateGPacket: index: "<<index<<", cellIndex: "<<cell->index()<<", subCellIndex: "<<cell->subIndex()<<", borderDistance: "<<borderDistance()<<", continuumEventDistance: "<<continuumEventDistance()<<", r: "<<position.norm()<<"\n";
    // While (the distance to the next continuum event is less than the distance to the cell border) and not terminated
    while (continuumEventDistance()<borderDistance() and not terminated()) {
      // Get continuum event type
      eventType=getContinuumEventType();
      // Update the position
      updatePosition(continuumEventDistance());
      // Get event treatment (no effect in normal mode) - New!
      getEventTreatment(eventType,emission,absorption);
      // If no absorption - Split and emit child packet - New!
      if (not absorption) emissionSplit(eventType);
      // Else - Handle the continuum event (allways done in normal mode)
      else continuumEvent(eventType,emission);
    }
    // If not terminated
    if (not terminated()) {
      // Update the position
      updatePosition(borderDistance());
      // Enter new cell
      newCell();
    }
  }
  
  return;
}

// Reset r-packet at current position, direction and frequency
void Packet::reset() {
  // Flag all distances to be updated
  pBorderDistance=dUpdate;
  pContinuumEventDistance=dUpdate;
  pLineDistance=dUpdate;
  pEdgeDistance=dUpdate;
  pBoostBinDistance=dUpdate; // New!
  // Flag all continuum opacities to be updated
  pElectronScatterKappa=kUpdate;
  pComptonScatterKappa=kUpdate;
  pGreyScatterKappa=kUpdate; // New!
  pFreeFreeAbsKappa=kUpdate;
  pBoundFreeAbsKappa=kUpdate;
  pComptonAbsKappa=kUpdate;
  pDustAbsKappa=kUpdate; // New!
  pContinuumEventKappa=kUpdate;
  // Select a new random optical depth for the next event
  eventTau=-log(randomize(0,1));
  // Reset the accumulated optical depth
  accumulatedTau=0;
  // Select next line (by searching the cell line list)
  selectNextLine();
  // Select next ionization edge (by searching the cell ionization edge list)
  selectNextEdge();
  // Select next boost bin (by searching the cell boost bin list)
  selectNextBoostBin();
}

// Terminate packet
void Packet::terminate() {

  //cout<<"Packet::terminate: index: "<<index<<", cell: "<<cell->index()<<", boostFactor: "<<boostFactor<<"\n";

  // Set terminated flag
  pTerminated=true;
}

// Remove current child packet (keeping the descendents)
void Packet::removeCurrentChild() {
  list<Packet>::iterator currentChild;

  // Get current child
  currentChild=child.end(); currentChild--;

  // Append descendents of current child to mother packet list
  child.splice(child.end(),(*currentChild).child);

  // Erase current child
  child.erase(currentChild);
}

// Split and emit child packet - New!
void Packet::emissionSplit(const PacketEventType& eventType) {
  bool emission=true, absorption=false;

  //if (eventType==evLineAbs) cout<<"Packet::emissionSplit: index: "<<index<<", cell: "<<cell->index()<<", Enter, "<<(*line)->lowerState->iState()->name()<<":"<<(*line)->lowerState->getIndex()<<"->"<<(*line)->upperState->iState()->name()<<":"<<(*line)->upperState->getIndex()<<", lambdaCMF: "<<(c/nuCMF)/AA<<", accumulatedTau: "<<accumulatedTau<<", eventTau: "<<eventTau<<", lineTau: "<<lineTau()<<", lineDistance: "<<lineDistance()<<"\n";

  // Add new child packet to list
  child.push_back(Packet(*this,false));

  // Set (child packet) index - Fix!
  //child.back().index=child.size()-1;

  // Handle (child packet) event
  if (isContinuumEvent(eventType)) child.back().continuumEvent(eventType,emission,absorption);
  if (eventType==evLineAbs) child.back().lineAbs(emission,absorption);

  // Propagate child packet
  child.back().propagate(packetStat);
  
  // If child packet terminated - Remove (keeping the descendents)
  if (child.back().terminated()) removeCurrentChild();

  // Update the (mother packet) accumulated optical depth
  //accumulatedTau=eventTau-accumulatedTau;
  if (eventType==evLineAbs) accumulatedTau=(accumulatedTau-eventTau)-lineTau();
  else accumulatedTau=0;
  // Testing!
  //accumulatedTau=0;
  // Select a new (mother packet) random optical depth
  eventTau=-log(randomize(0,1));

  // If line absorption - Make sure that the (mother packet) line distance function will return zero - Hack !
  if (eventType==evLineAbs) { nuCMF=(*line)->nu(); nu=nuCMF/(1-mu*(v/c)); pLineDistance=dUpdate; }
  // Testing!
  //if (!endOfLineList()) line++;
  
  //if (eventType==evLineAbs) cout<<"Packet::emissionSplit: index: "<<index<<", cell: "<<cell->index()<<", Return, "<<(*line)->lowerState->iState()->name()<<":"<<(*line)->lowerState->getIndex()<<"->"<<(*line)->upperState->iState()->name()<<":"<<(*line)->upperState->getIndex()<<", lambdaCMF: "<<(c/nuCMF)/AA<<", accumulatedTau: "<<accumulatedTau<<", eventTau: "<<eventTau<<", lineTau: "<<lineTau()<<", lineDistance: "<<lineDistance()<<"\n";
}

// Split packet at border crossing
void Packet::borderSplit() {
  double boost,z,P,nChildMean;
  int i,nChild;

  // Calculate relative boost factor
  //boost=cell->boostFactor(nuCMF)/boostFactor;
  // Testing!
  if (type!=gPacket) boost=cell->boostFactor(nuCMF)/boostFactor;
  else boost=cell->comptonKappaBoost()/boostFactor;
    
  // Calculate the average number of child packets
  nChildMean=boost-1;

  // Calculate the integer number of child packets
  nChild=floor(nChildMean);
  nChild+=(randomize(0,1)<=(nChildMean-floor(nChildMean)))?1:0;

  // Adjust energy
  adjustEnergy();

  // For each new child packet
  for (i=0;i<nChild;i++) {

    //if (boost!=1 and cell->index()==40 or cell->index()==50 or cell->index()==60 or cell->index()==70) cout<<"Packet::borderSplit: index: "<<index<<", cell: "<<cell->index()<<", lambdaCMF: "<<(c/nuCMF)/AA<<", boost: "<<boost<<", boostFactor: "<<cell->boostFactor(nuCMF)<<", oldBoostFactor: "<<boostFactor<<"\n";
    
    // Add to list
    child.push_back(Packet(*this,false));
    
    // Set index - Fix!
    //child.back().index=child.size()-1;
 
    // Propagate
    child.back().propagate(packetStat);

    // If terminated - Remove (keeping the descendents)
    if (child.back().terminated()) removeCurrentChild();

 }
}

// Merge packet at border crossing
void Packet::borderMerge() {
  double boost,z,P;

  // Calculate relative boost factor
  //boost=cell->boostFactor(nuCMF)/boostFactor;
  // Testing!
  if (type!=gPacket) boost=cell->boostFactor(nuCMF)/boostFactor;
  else boost=cell->comptonKappaBoost()/boostFactor;

  // Draw a random number between 0 and 1
  z=randomize(0,1);

  // Calculate the probability for termination
  P=1-boost;

  // If drawn - Terminate
  if (z<=P) { 
    //cout<<"Packet::borderMerge: index: "<<index<<", cell: "<<cell->index()<<", Terminate, boost: "<<boostFactor<<"\n";
    terminate();
  }
  // Else - Adjust energy
  else  adjustEnergy();
}

// Validate data
bool Packet::valid() const {
  bool _valid=true;
  if (!isnumber(E)) { cout<<"Packet::valid: Invalid data for packet "<<index<<" , E: "<<E<<"\n"; _valid=false; }
  if (!isnumber(ECMF)) { cout<<"Packet::valid: Invalid data for packet "<<index<<" , ECMF: "<<ECMF<<"\n"; _valid=false; }
  if (!isnumber(nu)) { cout<<"Packet::valid: Invalid data for packet "<<index<<" , nu: "<<nu<<"\n"; _valid=false; }
  if (!isnumber(nuCMF)) {cout<<"Packet::valid: Invalid data for packet "<<index<<" , nuCMF: "<<nuCMF<<"\n"; _valid=false; }
  return _valid;
}

// Constructors
PacketStateList::PacketStateList() { }
PacketStateList::PacketStateList(int size) : vector<PacketState>(size) { }

// Load packet (state) list
void PacketStateList::load(const string& dataFile) {
  ifstream file(dataFile.c_str());
  boost::archive::binary_iarchive archive(file);
  int i;

  // Load packet (state) list from archive
  archive >> (*this);
}


// Save packet (state) list
void PacketStateList::save(const string& dataFile) const {
  ofstream file(dataFile.c_str());
  boost::archive::binary_oarchive archive(file);
  int i;

  cout<<"PacketStateList::save: Enter\n";
  cout<<"PacketStateList::save: File: "<<dataFile<<", this->size: "<<this->size()<<"\n" ;

  // Save packet (state) list to archive
  archive << (*this);
}

// Collect packet (state) list (MPI)
void PacketStateList::collect(boost::mpi::communicator* mpiCom) {
  int nRequest=(mpiCom->rank()==0)?(mpiCom->size()-1):1;
  boost::mpi::request request[nRequest];
  vector<PacketStateList> collected(mpiCom->size()-1);
  int i;

  cout<<"PacketStateList::collect: Process: "<<mpiCom->rank()<<", Enter\n";

  // If not MPI - just return
  if (mpiCom==NULL) return;

  // Wait for all processes to reach this point
  mpiCom->barrier();

  cout<<"PacketStateList::collect: Process: "<<mpiCom->rank()<<", 1\n";

  // If slave - Deliver packet (state) list to master (non-blocking)
  if (mpiCom->rank()!=0) request[0]=mpiCom->isend(0,mPacketList,*this);

  cout<<"PacketStateList::collect: Process: "<<mpiCom->rank()<<", 2\n";

  // If master - Collect packet (state) lists from slaves (non-blocking)
  if (mpiCom->rank()==0) for (i=1;i<mpiCom->size();i++) request[i-1]=mpiCom->irecv(i,mPacketList,collected[i-1]);

  cout<<"PacketStateList::collect: Process: "<<mpiCom->rank()<<", 3\n";

  // Wait for all MPI requests to complete
  boost::mpi::wait_all(request,request+nRequest);

  cout<<"PacketStateList::collect: Process: "<<mpiCom->rank()<<", 4\n";
 
  // Append collected packet lists
  if (mpiCom->rank()==0) for (i=0;i<collected.size();i++) this->insert(this->end(),collected[i].begin(),collected[i].end());

  cout<<"PacketStateList::collect: Process: "<<mpiCom->rank()<<", Return\n";
}

// Distribute packet (state) list (MPI)
void PacketStateList::distribute(boost::mpi::communicator* mpiCom) {
  int nRequest=(mpiCom->rank()==0)?(mpiCom->size()-1):1;
  boost::mpi::request request[nRequest];
  vector<PacketStateList> dist(mpiCom->size());
  int i,rank;

  // If not MPI - just return
  if (mpiCom==NULL) return;

  // Wait for all processes to reach this point
  mpiCom->barrier();

  // If master
  if (mpiCom->rank()==0) {

    // For each packet
    for (i=0;i<this->size();i++) {
      // Calculate process rank
      rank=i%(mpiCom->size());
      // Add packet (state) to process list
      dist[rank].push_back((*this)[i]);
    }

    // Copy master packet (state) list
    (*this)=dist[0];

    // Distribute slave packet (state) lists (non-blocking)
    for (i=1;i<mpiCom->size();i++) request[i-1]=mpiCom->isend(i,mPacketList,dist[i]);
  }

  // If slave - Collect packet (state) list from master (non-blocking)
  if (mpiCom->rank()!=0) request[0]=mpiCom->irecv(0,mPacketList,*this);

  // Wait for all MPI requests to complete
  boost::mpi::wait_all(request,request+nRequest);
}

// Constructor
MCConfig::MCConfig() : packet(NULL) { }
MCConfig::MCConfig(PacketConfig* _packet) : packet(_packet) { }

// Constructor
MCExperiment::MCExperiment(MCConfig* _config) : packetSampler(NULL), packetStat(NULL), mpiCom(NULL) {
  if (_config==NULL) config = new MCConfig();
  else config=_config;
}

// Constructor
MCExperiment::MCExperiment(PacketSampler* _packetSampler, PacketStat* _packetStat, MCConfig* _config, boost::mpi::communicator* _mpiCom) : packetSampler(_packetSampler), packetStat(_packetStat), config(NULL), mpiCom(_mpiCom) {
  int i;

  // Initialize config
  if (_config==NULL) config = new MCConfig();
  else config=_config;  

  // If packet history update method
  if (config->packet->updateMethod==urPacketHistory) {
    // Create packet list
    packet.resize(packetSampler->nPacket);
    // Sample the packets
    for (i=0;i<packet.size();i++) packet[i]=packetSampler->sample();
  }
}

// Destructor
MCExperiment::~MCExperiment() {
  //cout<<"MCExperiment::~MCExperiment\n";
}

// Run the Monte-Carlo experiment
void MCExperiment::run() {
  // If MPI communicator provided - Run on cluster using MPI
  if (mpiCom!=NULL) runCluster();
  // Else - Run on a single computer
  else runSingle();
}

// Run the Monte-Carlo experiment (on a single computer)
// Probably not up-to-date - Check !!!
void MCExperiment::runSingle() {
  int i,j,imin,imax,n,chunkSize=1000,nChunk=ceil(float(packetSampler->nPacket)/float(chunkSize)),nChild=0;
  Packet sPacket;
  Packet* pPacket;

  // For each chunk of packets
  for (n=0;n<nChunk;n++) {

    // Determine packet index range
    imin=chunkSize*n;
    imax=chunkSize*(n+1)<packetSampler->nPacket ? chunkSize*(n+1) : packetSampler->nPacket;

    cout<<"MCExperiment::runSingle: Propagate packets "<<imin<<"-"<<imax<<"\n";

    // For each packet
    #pragma omp parallel for firstprivate(sPacket,pPacket)
    for (i=imin;i<imax;i++) {

      // If on-the-fly update method sample a packet
      if (config->packet->updateMethod==urOnTheFly) { sPacket=packetSampler->sample(i); pPacket=&sPacket; }
      // Else - select packet from list
      else pPacket=&(packet[i]);

      // Propagate packet
      pPacket->propagate(packetStat);

      // If on-the-fly update method
      if (config->packet->updateMethod==urOnTheFly and packetStat!=NULL) {
	// Update packet statistics
	packetStat->update(*pPacket);
	// Update child packet statistics - Not needed, update recursive, right?
	//for (j=0;j<pPacket->child.size();j++) packetStat->update(pPacket->child[j]);
      }
	
      // Count number of child packets
      // Does not work in boost mode!
      #pragma omp critical
      nChild+=pPacket->child.size();
    }
  }

  cout<<"MCExperiment::runSingle, Propagated "<<packetSampler->nPacket+nChild<<" packets\n";
}

// Run the Monte-Carlo experiment (on a cluster using MPI)
// Probably not up-to-date - Check !!!
void MCExperiment::runCluster() {
  int i,j,imin,imax,n,chunkSize=1000,nChunk=ceil(float(packetSampler->nPacket)/float(chunkSize)),nPacket=0,nChild=0;
  Packet sPacket;
  Packet* pPacket;

  // For each chunk of packets
  for (n=0;n<nChunk;n++) {

    // Determine packet index range
    imin=chunkSize*n;
    imax=chunkSize*(n+1)<packetSampler->nPacket ? chunkSize*(n+1) : packetSampler->nPacket;

    if (mpiCom->rank()==0) cout<<"MCExperiment::runCluster: Process: "<<mpiCom->rank()<<", Propagate packets "<<imin<<"-"<<imax<<"\n";

    // For each packet
    #pragma omp parallel for firstprivate(sPacket,pPacket) 
    for (i=imin;i<imax;i++) {

      // If process responsible for the packet
      if ((i%mpiCom->size())==mpiCom->rank()) {

	// If on-the-fly update method sample a packet
	if (config->packet->updateMethod==urOnTheFly) { sPacket=packetSampler->sample(i); pPacket=&sPacket; }
	// Else - select packet from list
	else pPacket=&(packet[i]);
	
	// Propagate packet
	pPacket->propagate(packetStat);
	
	// If on-the-fly update method
	if (config->packet->updateMethod==urOnTheFly and packetStat!=NULL) {
	  // Update packet statistics
	  packetStat->update(*pPacket);
	  // Update child packet statistics - Not needed, update recursive, right?
	  //for (j=0;j<pPacket->child.size();j++) packetStat->update(pPacket->child[j]);
	}

	#pragma omp critical
	{	
	  // Count number of packets
	  nPacket+=1;

	  // Count number of child packets
	  // Does not work in boost mode!
	  nChild+=pPacket->child.size();

	  // If g-packet - Delete config - New!
	  // Hack - The handling of g-packet config needs to be improved !!!
	  if (pPacket->type==gPacket) delete pPacket->config;
	}
      }
    }
  }

  mpiCom->barrier();

  cout<<"MCExperiment::runCluster, Process: "<<mpiCom->rank()<<", Propagated "<<nPacket+nChild<<" packets\n";

  // Synchronize packet statistics
  if (packetStat!=NULL) syncPacketStat();

  // Wait for all processes to reach this point
  mpiCom->barrier();

  if (mpiCom->rank()==0) cout<<"MCExperiment::runCLuster, Packet statistics synchronized\n";
}

// Evolve the Monte-Carlo experiment
void MCExperiment::evolve(const PacketStateList& packetIn, PacketStateList& packetOut, Grid* grid, int chunkSize) { // Changed!
  // If MPI communicator provided - Run on cluster using MPI
  if (mpiCom!=NULL) evolveCluster(packetIn,packetOut,grid,chunkSize); // Changed!
  // Else - Run on a single computer
  else evolveSingle(packetIn,packetOut,grid,chunkSize);
}

// Evolve the Monte-Carlo experiment (on a single computer)
void MCExperiment::evolveSingle(const PacketStateList& packetIn, PacketStateList& packetOut, Grid* grid, int chunkSize) { // Changed!
  int i,j,imin,imax,n,nPacket,nChild=0,nChunk; // Changed!
  Packet sPacket;

  // Calculate total number of packets
  nPacket=packetIn.size();
  if (packetSampler!=NULL) nPacket+=packetSampler->nPacket;

  // Calculate number of chunks
  nChunk=ceil(float(nPacket)/float(chunkSize));

  // For each chunk of packets
  for (n=0;n<nChunk;n++) {

    // Determine packet index range
    imin=chunkSize*n;
    imax=chunkSize*(n+1)<nPacket ? chunkSize*(n+1) : nPacket;

    cout<<"MCExperiment::evolveSingle: Propagate packets "<<imin<<"-"<<imax<<"\n";

    // For each packet
    #pragma omp parallel for firstprivate(sPacket) 
    for (i=imin;i<imax;i++) {

      // If existing packet - Unpack from list
      if (i<packetIn.size()) sPacket=Packet(packetIn[i],evNone,grid,config->packet);
      // Else - Sample new packet
      else sPacket=packetSampler->sample(i-packetIn.size());

      // If not on grid disregard and continue?
      //if (not sPacket.onGrid()) continue;
      
      // Propagate packet
      sPacket.propagate(packetStat);

      // Update packet statistics
      if (packetStat!=NULL) packetStat->update(sPacket);

      #pragma omp critical
      {
	// Store active (mother and child) packets in out list - New!
	storeActivePackets(sPacket,packetOut,nChild,grid);
      }
    }
  }
}

// Evolve the Monte-Carlo experiment (on a cluster)
void MCExperiment::evolveCluster(const PacketStateList& packetIn, PacketStateList& packetOut, Grid* grid, int chunkSize) { // Changed!
  int i,j,imin,imax,n,nPacket,nPacketSample,nChild=0,nChunk; // Changed!
  Packet sPacket;

  // Initiate packet out list
  packetOut.resize(0);

  // Calculate total number of packets
  nPacket=packetIn.size();
  if (packetSampler!=NULL) {
    nPacketSample=ceil(float(packetSampler->nPacket-mpiCom->rank())/float(mpiCom->size()));
    nPacket+=nPacketSample;
  }
  
  // Calculate number of chunks
  nChunk=ceil(float(nPacket)/float(chunkSize));

  // Testing!!!
  //if (mpiCom->rank()==0) logSysCPU(i,"evolveCluster: Begin");
 
  // For each chunk of packets
  for (n=0;n<nChunk;n++) {

    // Determine packet index range
    imin=chunkSize*n;
    imax=chunkSize*(n+1)<nPacket ? chunkSize*(n+1) : nPacket;

    if (mpiCom->rank()==0) cout<<"MCExperiment::evolveCluster: Propagate packets "<<imin<<"-"<<imax<<"\n";
    //cout<<"MCExperiment::evolveCluster: Process: "<<mpiCom->rank()<<", Propagate packets "<<imin<<"-"<<imax<<"\n";

    // For each packet
    #pragma omp parallel for firstprivate(sPacket)
    // Testing!
    //#pragma omp parallel for firstprivate(sPacket) schedule(dynamic,100)
    for (i=imin;i<imax;i++) {

      //if (i>=packetIn.size()) cout<<"MCExperiment::runCluster: Process: "<<mpiCom->rank()<<", Sample packet "<<i<<", packetIn.size: "<<packetIn.size()<<", packetSampler.nPacket: "<<packetSampler->nPacket<<"\n";
      //if (i==0) cout<<"MCExperiment::evolveCluster: Process: "<<mpiCom->rank()<<", num_threads: "<<omp_get_num_threads()<<"\n";

      // Testing!
      //if (mpiCom->rank()==0) cout<<"MCExperiment::evolveCluster: Process: "<<mpiCom->rank()<<", thread_num: "<<omp_get_thread_num()<<", Propagate packet "<<i<<"\n";

      // If existing packet - Unpack from list
      if (i<packetIn.size()) sPacket=Packet(packetIn[i],evNone,grid,config->packet);
      // Else - Sample new packet
      else sPacket=packetSampler->sample((i-packetIn.size())*mpiCom->size()+mpiCom->rank());
      
      // If not on grid disregard and continue?
      //if (not sPacket.onGrid()) continue;

      // Update injected packet statistics
      if (packetStat!=NULL) packetStat->updateInject(sPacket);

      // Testing!
      //onGrid(sPacket,grid);
      
      // Propagate packet
      sPacket.propagate(packetStat);

      // Update packet statistics
      if (packetStat!=NULL) packetStat->update(sPacket);

      #pragma omp critical
      {
	// Store active (mother and child) packets in out list - New!
	storeActivePackets(sPacket,packetOut,nChild,grid);
	
	// If newly sampled g-packet - Delete config - New!
	// Hack - The handling of g-packet config needs to be improved !!!
	//if (i>=packetIn.size() and sPacket.type==gPacket) delete sPacket.config;
	if (sPacket.type==gPacket) delete sPacket.config;
      }
    }
  }

  // Testing!!!
  //if (mpiCom->rank()==0) logSysCPU(i,"evolveCluster: End");

  mpiCom->barrier();

  cout<<"MCExperiment::evolveCluster, Process: "<<mpiCom->rank()<<", Propagated "<<nPacket+nChild<<" packets\n";
  // Testing - New!
  //cout<<"MCExperiment::evolveCluster, Process: "<<mpiCom->rank()<<", nPacket: "<<nPacket<<", nChild: "<<nChild<<", packetStat->nPacket: "<<packetStat->nPacketInject()<<"/"<<packetStat->nPacket()<<", packetStat->EBoost: "<<packetStat->EBoostInject()<<"/"<<packetStat->EBoost()<<"\n";

  // Synchronize packet statistics
  if (packetStat!=NULL) syncPacketStat();

  // Wait for all processes to reach this point
  mpiCom->barrier();

  if (mpiCom->rank()==0) cout<<"MCExperiment::evolveCluster, Packet statistics synchronized\n";
  if (packetStat!=NULL and mpiCom->rank()==0) cout<<"MCExperiment::evolveCluster, Propagated "<<packetStat->nPacketInject()<<"/"<<packetStat->nPacket()<<" packets, energy boost: "<<packetStat->EBoostInject()<<"/"<<packetStat->EBoost()<<", energy boost per packet: "<<packetStat->EBoostInject()/packetStat->nPacketInject()<<"/"<<packetStat->EBoost()/packetStat->nPacket()<<"\n";
}

// Synchronize packet statistics (MPI)
void MCExperiment::syncPacketStat() {
  int nRequest=(mpiCom->rank()==0)?(mpiCom->size()-1):1;
  boost::mpi::request request[nRequest];
  vector<PacketStat> collected(mpiCom->size()-1);
  int i;

  // If slave - Deliver packet statistics to master (non-blocking)
  if (mpiCom->rank()!=0) request[0]=mpiCom->isend(0,mPacketStat,*packetStat);

  // If master - Collect MC statistics from slaves (non-blocking)
  if (mpiCom->rank()==0) for (i=1;i<mpiCom->size();i++) request[i-1]=mpiCom->irecv(i,mPacketStat,collected[i-1]);

  // Wait for all MPI requests to complete
  boost::mpi::wait_all(request,request+nRequest);
 
  // Sum packet statistics
  if (mpiCom->rank()==0) for (i=0;i<collected.size();i++) *packetStat+=collected[i];

  // Broadcast packet statistics
  boost::mpi::broadcast(*mpiCom,*packetStat,0);

  // If present - Synchronize spectral evolution
  if (packetStat->specEvoObs!=NULL) packetStat->specEvoObs->sync(mpiCom);
}

// Determine if packet is (spatially) on grid
// Improve Packet onGrid function and use that instead !!!
bool MCExperiment::onGrid(const Packet& packet, Grid* grid) {
  bool _onGrid;
  CellSS* cellInner;
  CellSS* cellOuter;

  // Get inner and outer cell
  cellInner=((CellSS*) grid->cell[(config->packet->cellIndexMin!=-1 and packet.type!=gPacket)?config->packet->cellIndexMin:0]);
  cellOuter=((CellSS*) grid->cell[(config->packet->cellIndexMax!=-1 and packet.type!=gPacket)?config->packet->cellIndexMax:(grid->cell.size()-1)]);

  // Evaluate on-grid condition
  _onGrid=(packet.position.norm()>=cellInner->rMinT(packet.t) and packet.position.norm()<=cellOuter->rMaxT(packet.t));

  if (packet.type!=gPacket) {

    int process;

    if (mpiCom!=NULL) process=mpiCom->rank();
    else process=0;

    if (packet.position.norm()>cellOuter->rMaxT(packet.t) and (packet.t<grid->tMin() or packet.t>grid->tMax())) cout<<"MCExperiment::evolveCluster, Process: "<<process<<", Warning: Packet "<<packet.index<<" outside both grid and time-intervall\n";
    
    // This warning does not make sense when dustAbs==true - Removed - Improve!
    //if (packet.position.norm()<=cellOuter->rMaxT(packet.t) and (packet.t>=grid->tMin() and packet.t<=grid->tMax())) cout<<"MCExperiment::evolveCluster, Process: "<<process<<", Warning: Packet "<<packet.index<<" inside both grid and time-intervall, r: "<<packet.position.norm()<<", t: "<<packet.t<<", rMax: "<<cellOuter->rMaxT(packet.t)<<", rMin: "<<cellInner->rMinT(packet.t)<<", tMin: "<<grid->tMin()<<", tMax: "<<grid->tMax()<<", type: "<<packet.type<<", lastEvent.type: "<<packet.lastEvent.type<<"\n";
    
    //if (packet.position.norm()<cellInner->rMinT(packet.t)) cout<<"MCExperiment::onGrid, Process: "<<process<<", Warning: Packet "<<packet.index<<" behind inner border of grid, r: "<<packet.position.norm()<<", t: "<<packet.t<<", rMin: "<<cellInner->rMinT(packet.t)<<", rMax: "<<cellOuter->rMaxT(packet.t)<<", tMin: "<<grid->tMin()<<", tMax: "<<grid->tMax()<<"\n";

    if (packet.position.norm()<cellInner->rMinT(packet.t)) {
      cout<<"MCExperiment::onGrid, Process: "<<process<<", Warning: Packet "<<packet.index<<" behind inner border of grid, r: "<<packet.position.norm()<<", mu: "<<packet.mu<<", t: "<<packet.t<<", rMin: "<<cellInner->rMinT(packet.t)<<", rMax: "<<cellOuter->rMaxT(packet.t)<<", tMin: "<<grid->tMin()<<", tMax: "<<grid->tMax()<<", lastEventHandled: "<<packet.lastEventHandled<<", lastEvent.type: "<<packet.lastEvent.type<<", lastEvent.r: "<<packet.lastEvent.position.norm()<<", lastEvent.t: "<<packet.lastEvent.t<<"\n";
      //cout<<packet.logStream.str();
    }
  }

  return _onGrid;
}

// Store active (mother and child) packets in packet state list (recursive) - New!
void MCExperiment::storeActivePackets(Packet& packet, PacketStateList& packetState, int& nChild, Grid* grid) {
  //int i;
  list<Packet>::iterator child;

  // If not terminated 
  if (not packet.terminated()) {
    // If packet on grid - Pack and add to packet state list
    if (onGrid(packet,grid)) {
     packetState.push_back(PacketState(packet));
    }
  }

  // For each child packet - Store active (mother and child) packets in packet state list
  //for (i=0;i<packet.child.size();i++) storeActivePackets(packet.child[i],packetState,nChild,grid);
  for (child=packet.child.begin();child!=packet.child.end();child++) storeActivePackets((*child),packetState,nChild,grid);

  // Sum number of child packets
  nChild++;  
}

// Reset packet sampler
void PacketSampler::reset() {
  packetIndex=0;
}

// Constructor
PacketSamplerBB::PacketSamplerBB(const double& _Rph, const double& _Lph, const int& _nPacket, const int& _nBin, const FrequencyBin& nuLim, const BBSampling& _bbSampling, const BBFrame& _bbFrame, const bool& _bbLimbDark, Grid* _grid, PacketConfig* _packetConfig, int _packetIndexBase) : Rph(_Rph), Lph(_Lph), grid(_grid), bbSampling(_bbSampling), bbFrame(_bbFrame), bbLimbDark(_bbLimbDark), packetConfig(_packetConfig) {
  double EPacket,EBin,ECorr,dlognu,nuMax;
  int i;

  // Initialize number of packets and bins, packet index and packet index base
  nPacket=_nPacket;
  nBin=_nBin;
  packetIndex=0;
  packetIndexBase=_packetIndexBase;

  // Get innermost cell
  for (i=0;i<grid->cell.size();i++) {
    cell = (CellSS*) grid->cell[i];
    if (Rph>=cell->rMin() and Rph<cell->rMax()) break;
  }
  cout <<"PacketSamplerBB::PacketSamplerBB: innermost cell: "<<i<<", rMin: "<<cell->rMin()<<", rMax: "<<cell->rMax()<<", Rph: "<<Rph<<"\n";

  // Get time and time-span from the innermost cell
  t=(cell->tMin()+cell->tMax())/2;
  dt=cell->tMax()-cell->tMin();

  // Calculate the effective temperature at the photosphere
  Teff=pow(Lph/(4*pi*pow(Rph,2)*sigma),1.0/4.0);
  // Calculate the total radiated energy (during the Monte Carlo experiment)
  Eph=Lph*dt;

  // Calculate the (integer) number of packets per frequency bin
  nPacketBin=floor(nPacket/nBin);
  // Recalculate the total number of packets
  nPacket=nPacketBin*nBin;

  // Initialize the frequency bin list
  nu.resize(nBin);
  // Set lower and upper frequency limits
  nu[0].min=nuLim.min;
  nu[nu.size()-1].max=nuLim.max;

  // If blackbody sampling is log frequency (see Lucy 2003)
  if (bbSampling==bbsLogFrequency) {
    // Calculate the frequency spacing
    dlognu=(log(nu[nu.size()-1].max)-log(nu[0].min))/nBin;
    // Calculate frequency bins
    for (i=0;i<nBin-1;i++) {
      // Calculate the upper limit for the bin
      nu[i].max=nu[i].min*exp(dlognu);
      // Set the lower limit of the next bin
      nu[i+1].min=nu[i].max;
    }
  }
  // Else - i.e. if blackbody sampling is equal energy (see Mazzali & Lucy 1993)
  else {
    // Calculate the energy per packet
    EPacket=Eph/nPacket;
    // Calculate the energy per frequency bin
    EBin=EPacket*nPacketBin;
    // Calculate correction for the limited frequency range
    ECorr=BlackBody::F(nuLim.min,nuLim.max,Teff,1e-4)/BlackBody::F(Teff);
    // Calculate the frequency at the maximum
    nuMax=BlackBody::nuMax(Teff);
    // Calculate the bin flux
    FBin=ECorr*EBin/(4*pi*pow(Rph,2)*dt);
    // Calculate the frequency bins
    for (i=0;i<nBin-1;i++) {
      // Calculate the upper limit for the bin
      nu[i].max=BlackBody::nuF(FBin,Teff,nu[i].min,1e-4);
      // Set the lower limit of the next bin
      nu[i+1].min=nu[i].max;
    }
  }

  cout<<"PacketSamplerBB::PackerSamplerBB: created frequency bin list ["<<nu[0].min<<","<<nu[nu.size()-1].max<<"] with "<<nBin<<" bins for Teff="<<Teff<<"\n";
}

// Sample packet from blackbody distribution
Packet PacketSamplerBB::sample(int _packetIndex) {
  double muPosition,thetaPosition,phiPosition;
  double muDirection,thetaDirection,phiDirection;
  double v,mu,EPacket,nuPacket,nuCMF,ECMF,dlognu;
  Vector3D position,direction;
  int i,j;
  // Hack for openmp
  int packetIndex=this->packetIndex;

  // If provided - Set packet index
  if (_packetIndex!=noIndex) packetIndex=_packetIndex;

  // Calculate bin and packet index within the bin
  { i=packetIndex/nPacketBin; j=packetIndex%nPacketBin; }

  // If blackbody sampling is log frequency - Calculate frequency spacing
  if (bbSampling==bbsLogFrequency) dlognu=(log(nu[i].max)-log(nu[i].min))/nPacketBin;

  // Select a random position
  muPosition=-1+2*randomize(0,1);
  thetaPosition=acos(muPosition);
  phiPosition=2*pi*randomize(0,1);
  //position=Vector3D(cell->rMin(),thetaPosition,phiPosition,spherical);
  position=Vector3D(Rph,thetaPosition,phiPosition,spherical);

  // Select a random direction
  if (bbLimbDark) muDirection=sqrt(randomize(0,1));
  else muDirection=randomize(0,1);
  thetaDirection=acos(muDirection);
  phiDirection=2*pi*randomize(0,1);
  direction=Vector3D(1,thetaDirection,phiDirection,spherical);
  direction.rotate(thetaPosition,phiPosition);

  // If blackbody sampling is log frequency
  if (bbSampling==bbsLogFrequency) {
    // Calculate the packet frequency
    nuPacket=0.5*(nu[i].min*exp(j*dlognu)+nu[i].min*exp((j+1)*dlognu));
    // Calculate the packet energy
    // Becomes zero if nuPacket is small !!! Fix. 
    EPacket=Eph*nuPacket*BlackBody::F(nuPacket,Teff)*dlognu/BlackBody::F(Teff);
  }
  // Else - i.e. if blackbody sampling is equal energy (see Mazzali & Lucy 1993)
  else {
    // Calculate the packet frequency
    // This has to be improved, should allways sample the blackbody distribution
    if (i<nBin-1) nuPacket=nu[i].min+(nu[i].max-nu[i].min)*randomize(0,1);
    else nuPacket=BlackBody::nuF(FBin*randomize(0,1),Teff,nu[i].min,1e-4);
    // Calculate the packet energy;
    EPacket=Eph/nPacket;
  }

  // Get velocity from the innermost cell
  v=cell->v(position,t);

  // Calculate the (cosine of the) polar angle about the radius vector
  mu=direction*position/position.norm();
  // Calculate the frequency in the comoving frame
  if (bbFrame==bbfRest) nuCMF=nuPacket*(1-mu*v/c);
  else nuCMF=nuPacket;
  // Calculate the energy in the comoving frame
  if (bbFrame==bbfRest or bbFrame==bbfMixed) ECMF=EPacket*(1-mu*v/c);
  else ECMF=EPacket;

  // Increase the packet index
  packetIndex++;
  // Hack for openmp
  this->packetIndex=packetIndex;

  // Create and return the packet
  // Note: Mixed frame -> Teff apply to the CMF, but Lph to the rest frame - Find better solution !!!
  return Packet(ECMF, FrequencyBin(nuCMF,nuCMF), position, direction, t, cell, grid, packetIndexBase+packetIndex-1, rPacket, evBoundaryCrossing, packetConfig);
}

// Constructor
// Does not work woth containers in boost mode (see below)!
/*PacketSamplerBBL::PacketSamplerBBL(const double& _L, const int& _nPacket, const int& cellIndex, Grid* _grid, const FrequencyBin& _nuLim, PacketConfig* _packetConfig, boost::mpi::communicator* _mpiCom, int _packetIndexBase) : L(_L), grid(_grid), nuLim(_nuLim), packetConfig(_packetConfig), mpiCom(_mpiCom) {
  double dt;
  int i;
  bool master=false;

  if (mpiCom==NULL) master=true;
  else if (mpiCom->rank()==0) master=true;

  // Initialize number of packets, packet index and packet index base
  nPacket=_nPacket;
  packetIndex=0;
  packetIndexBase=_packetIndexBase;

  // Get cell
  cell=grid->cell[cellIndex];

  // Get (centered) time and time-span
  t=grid->t();
  dt=grid->tMax()-grid->tMin();

  // If cell not container - Get temperature - Removed!
  //if (cell->type()!=cellContainer) T=cell->T();
  // Else - Radiation-energy average subcell temperatures - New!
  //else { T=0; for (i=0;i<cell->subCell.size();i++) T+=pow(cell->subCell[i]->T(),4.0); T=pow(T,1.0/4.0); }
  // Get temperature - Changed again!
  T=cell->T();

  // Caclulate the energy of the packets
  if (L>0) EPacket=(L*dt)/nPacket;
  // Hack to deal with negative luminosities - Fix !!!
  else { if (master) cout<<"PacketSamplerBBL::PacketSamplerBBL: Warning: Negative luminosity not supported - Setting number of packets to zero\n", nPacket=0; EPacket=0; }

  if (master) cout <<"PacketSamplerBBL::PacketSamplerBBL: cell: "<<cellIndex<<", rMin: "<<((CellSS*) cell)->rMin()<<"T: "<<T<<", sampling of "<<nPacket<<" packets with energy "<<EPacket<<" initiated\n";

  // If boost mode - boost emissivity (and adjust number of packets) - New!
  // Does not work with containers in boost mode (due to boostEta)!
  if (((Cell*) cell)->config->boost) boostEta();

  if (((Cell*) cell)->config->boost and master) cout <<"PacketSamplerBBL::PacketSamplerBBL: Number of packets adjusted to "<<nPacket<<"\n";
  }

// Sample packet from luminosity (assumed to be diluted blackbody)
// Does not work with containers in boost mode (see below)!
Packet PacketSamplerBBL::sample(int _packetIndex) {
  double nuPacket,B,z,small=1e-10;
  double muPosition,thetaPosition,phiPosition;
  double muDirection,thetaDirection,phiDirection;
  Vector3D position,direction;
  PacketEventType packetEventType;
  CellBase* _cell; // New!
  double boostFactor; // New!
  // Hack for openmp
  int packetIndex=this->packetIndex;
  
  // If provided - Set packet index
  if (_packetIndex!=noIndex) packetIndex=_packetIndex;

  // Set packet event type
  packetEventType=evBoundaryCrossing;

  // Select a random position
  muPosition=-1+2*randomize(0,1);
  thetaPosition=acos(muPosition);
  phiPosition=2*pi*randomize(0,1);
  position=Vector3D(((CellSS*) cell)->rMinT(t)*(1.0+small),thetaPosition,phiPosition,spherical);

  // Select a random direction
  muDirection=randomize(0,1);
  thetaDirection=acos(muDirection);
  phiDirection=2*pi*randomize(0,1);
  direction=Vector3D(1,thetaDirection,phiDirection,spherical);
  direction.rotate(thetaPosition,phiPosition);

  // Select a random BB frequency
  //z=randomize(0,1);
  //B=BlackBody::B(nuLim.min,nuLim.max,T,1e-4);
  //nuPacket=BlackBody::nuB(z*B,T,nuLim.min,1e-3);
  nuPacket=randomNu(); // Changed!

  // Increase packet counter
  packetIndex++;
  // Hack for openmp
  this->packetIndex=packetIndex;

  // Get boost factor - New!
  // Does not work with containers in boost mode (due to cell.boostFactor)!
  boostFactor=((Cell*) cell)->boostFactor(nuPacket);

  // Select cell - New!
  // Note: Selects subcell and creates virtual cell for container
  _cell = cell->selectCell(position,direction,t);

  //cout<<"PacketSamplerBBL::sample: packetIndex: "<<packetIndexBase+packetIndex-1<<", boostFactor: "<<boostFactor<<"\n";

  if (position.norm()<1.0) cout<<"PacketSamplerBBL::sample: index: "<<packetIndexBase+packetIndex-1<<", r: "<<position.norm()<<"\n";

  // Create and return the packet
  //return Packet(EPacket,FrequencyBin(nuPacket,nuPacket),position,direction,t,cell,grid,packetIndexBase+packetIndex-1,rPacket,packetEventType,packetConfig);
  return Packet(EPacket/boostFactor,FrequencyBin(nuPacket,nuPacket),position,direction,t,_cell,grid,packetIndexBase+packetIndex-1,rPacket,packetEventType,packetConfig,boostFactor); // Changed!
  }

// Boost emission probabilities - New!
// Does not work with containers!
void PacketSamplerBBL::boostEta() {
  double B,PTot=0;
  int i;

  // Calculate total blackbody intensity
  B=BlackBody::B(nuLim.min,nuLim.max,T,1e-4);  

  // Resize emissivity boost vector
  etaP.resize(((Cell*) cell)->nuBoost.size());

  // For each boost bin
  for (i=0;i<((Cell*)cell)->nuBoost.size();i++) {
    // Calculate emission probability
    etaP[i]=BlackBody::B(((Cell*) cell)->nuBoost[i].min,((Cell*) cell)->nuBoost[i].max,T,1e-4)/B;
  }

  // Normalize emission probability (to be on the safe side)
  for (i=0;i<((Cell*)cell)->nuBoost.size();i++) PTot+=etaP[i];
  for (i=0;i<((Cell*)cell)->nuBoost.size();i++) etaP[i]=etaP[i]/PTot;
  
  // Initialize emissivity boost factor
  etaBoost=0;

  // Calculate emissivity boost factor
  for (i=0;i<((Cell*) cell)->boost.size();i++) etaBoost+=etaP[i]*((Cell*) cell)->boost[i];

  // Adjust emission probabilities
  for (i=0;i<etaP.size();i++) etaP[i]=etaP[i]*((Cell*) cell)->boost[i]/etaBoost;

  // Adjust number of packets
  nPacket=nPacket*etaBoost;
}

// Select random frequency - New!
// Does not work with containers in boost mode (see below)!
double PacketSamplerBBL::randomNu() {
  // Select version to use
  // Does not work with containers in boost mode (due to randomNuBoost)!
  if (((Cell*) cell)->config->boost) return randomNuBoost();
  else return randomNuNormal();
}

// Select random frequency (intervall version) - New!
double PacketSamplerBBL::randomNu(double nuMin, double nuMax) {
  double z,B,nu;

  // Select a random BB frequency
  z=randomize(0,1);
  B=BlackBody::B(nuMin,nuMax,T,1e-4);
  nu=BlackBody::nuB(z*B,T,nuMin,1e-3);

  return nu;
}

// Select random frequency (normal verions) - New!
double PacketSamplerBBL::randomNuNormal() {
  double z,B,nu;
 
  // Select a random BB frequency
  z=randomize(0,1);
  B=BlackBody::B(nuLim.min,nuLim.max,T,1e-4);
  nu=BlackBody::nuB(z*B,T,nuLim.min,1e-3);

  return nu;
}

// Select random frequency (boost verions) - New!
// Does not work with containers!
double PacketSamplerBBL::randomNuBoost() {
  double z,PTot=0;
  int i;

  // Draw random number between 0 and 1
  z=randomize(0,1);

  // For each boost bin
  for (i=0;i<etaP.size();i++) {

    // If bin drawn - Return emission frequency
    if (z<=PTot+etaP[i]) return randomNu(((Cell*) cell)->nuBoost[i].min,((Cell*) cell)->nuBoost[i].max);
    // Else increase accumulated probability
    else PTot+=etaP[i];
  }

  // This should not happen!
  cout<<"PacketSamplerBBL::randomNuBoost: Error: No frequency drawn, z: "<<z<<", PTot: "<<PTot<<"\n";

  return 0.0;
  }*/

// Constructor - Changed!
// Works with containers in boost mode, right?
PacketSamplerBBL::PacketSamplerBBL(const double& _L, const int& _nPacket, const int& cellIndex, Grid* _grid, const FrequencyBin& nuLim, PacketConfig* _packetConfig, boost::mpi::communicator* _mpiCom, int _packetIndexBase) : L(_L), grid(_grid), packetConfig(_packetConfig), mpiCom(_mpiCom) {
  double dt,LSubCell,fillFactorTot=0; // Changed!
  int i;
  bool master=false;

  if (mpiCom==NULL) master=true;
  else if (mpiCom->rank()==0) master=true;

  // Initialize number of packets, packet index and packet index base
  nPacket=_nPacket;
  packetIndex=0;
  packetIndexBase=_packetIndexBase;

  // Get cell
  cell=grid->cell[cellIndex];

  // Get (centered) time and time-span
  t=grid->t();
  dt=grid->tMax()-grid->tMin();

  // Get temperature
  T=cell->T();

  // Caclulate the energy of the packets
  if (L>0) EPacket=(L*dt)/nPacket;
  // Hack to deal with negative luminosities - Fix !!!
  else { if (master) cout<<"PacketSamplerBBL::PacketSamplerBBL: Warning: Negative luminosity not supported - Setting number of packets to zero\n", nPacket=0; EPacket=0; return; }

  // Initialize cell packet sampler list
  cellPacketSampler.resize(0);

  // Reset number of packets
  nPacket=0;

  // If cell container
  if (cell->type()==cellContainer) {
    // Calculate total fill factor (to the power of 2/3)
    for (i=0;i<cell->subCell.size();i++) fillFactorTot+=pow(cell->subCell[i]->fillFactor(),(2.0/3.0)); 
    // For each sub cell
    for (i=0;i<cell->subCell.size();i++) {
      // Calculate sub cell luminosity
      LSubCell=L*pow(cell->subCell[i]->fillFactor(),(2.0/3.0))/fillFactorTot;
      // Create cell packet sampler and append to list
      cellPacketSampler.push_back(CellPacketSamplerBBL(grid,cell->subCell[i],cell,nuLim,packetConfig,EPacket,LSubCell,T,mpiCom,packetIndexBase+nPacket)); // Changed!
      // Add upp number of cell packets
      nPacket+=cellPacketSampler[cellPacketSampler.size()-1].nPacket;
    }
  }
  // Else
  else {
    // Create cell packet sampler and append to list
    cellPacketSampler.push_back(CellPacketSamplerBBL(grid,cell,NULL,nuLim,packetConfig,EPacket,L,T,mpiCom,packetIndexBase+nPacket)); // Changed!
    // Add upp number of packet cell packets
    nPacket+=cellPacketSampler[cellPacketSampler.size()-1].nPacket;
  }

  if (master) cout<<"PacketSamplerBBL::PacketSamplerBBL: Initited for sampling of "<<nPacket<<" packets with energy "<<EPacket<<"\n";
}  

// Sample packet from luminosity (assumed to be diluted blackbody) - Changed!
// Works with containers in boost mode, right?
Packet PacketSamplerBBL::sample(int _packetIndex) {
  int i,cellPacketIndex;
  double nPacketTot=0;
  // Hack for openmp
  int packetIndex=this->packetIndex;

  // If provided - Set packet index
  if (_packetIndex!=noIndex) packetIndex=_packetIndex;

  // For each cell packet sampler
  for (i=0;i<cellPacketSampler.size();i++) {

    // Calculate total number of packets
    nPacketTot+=cellPacketSampler[i].nPacket;

    // If larger than packet counter
    if (nPacketTot>packetIndex) {

      // Set cell packet index
      if (_packetIndex!=noIndex) cellPacketIndex=packetIndex-(nPacketTot-cellPacketSampler[i].nPacket);
      else cellPacketIndex=noIndex;

      // Increase packet counter
      packetIndex++;
      // Hack for openmp
      this->packetIndex=packetIndex;

      // Sample and return packet
      return cellPacketSampler[i].sample(cellPacketIndex);
    }
  }
}

// Reset
void PacketSamplerBBL::reset() {
  int i;
  PacketSampler::reset();
  for (i=0;i<cellPacketSampler.size();i++) cellPacketSampler[i].reset();
}

// Constructor - New!
// Works with containers in boost mode, right?
CellPacketSamplerBBL::CellPacketSamplerBBL(Grid* _grid, Cell* _cell, CellBase* _container, const FrequencyBin& _nuLim, PacketConfig* _packetConfig, const double& _EPacket, const double& _L, const double& _T, boost::mpi::communicator* _mpiCom, int _packetIndexBase) : grid(_grid), cell(_cell), container(_container), L(_L), T(_T), nuLim(_nuLim), packetConfig(_packetConfig), EPacket(_EPacket), mpiCom(_mpiCom) {
  double dt;
  int i;
  bool master=false;

  if (mpiCom==NULL) master=true;
  else if (mpiCom->rank()==0) master=true;

  // Initialize packet index and packet index base
  packetIndex=0;
  packetIndexBase=_packetIndexBase,

  // Get (centered) time and time-span
  t=cell->t();
  dt=cell->tMax()-cell->tMin();

  // Calculate number of packets
  if (L>0) nPacket=(L*dt)/EPacket; // Changed!
  // Hack to deal with negative luminosities - Fix !!!
  else { if (master) cout<<"CellPacketSamplerBBL::PacketSamplerBBL: Warning: Negative luminosity not supported - Setting number of packets to zero\n", nPacket=0; EPacket=0; return; }

  // If boost mode - boost emissivity (and adjust number of packets)
  // Works with containers in boost mode, right?
  if (((Cell*) cell)->config->boost) boostEta();
}

// Sample packet from luminosity (assumed to be diluted blackbody) - New!
// Works with containers in boost mode, right?
Packet CellPacketSamplerBBL::sample(int _packetIndex) {
  double nuPacket,B,z,small=1e-10;
  double muPosition,thetaPosition,phiPosition;
  double muDirection,thetaDirection,phiDirection;
  Vector3D position,direction;
  PacketEventType packetEventType;
  CellBase* _cell; // New!
  double boostFactor; // New!
  // Hack for openmp
  int packetIndex=this->packetIndex;

  // If provided - Set packet index
  if (_packetIndex!=noIndex) packetIndex=_packetIndex;

  // Set packet event type
  packetEventType=evBoundaryCrossing;

  // Select a random position
  muPosition=-1+2*randomize(0,1);
  thetaPosition=acos(muPosition);
  phiPosition=2*pi*randomize(0,1);
  position=Vector3D(((CellSS*) cell)->rMinT(t)*(1.0+small),thetaPosition,phiPosition,spherical);

  // Select a random direction
  muDirection=randomize(0,1);
  thetaDirection=acos(muDirection);
  phiDirection=2*pi*randomize(0,1);
  direction=Vector3D(1,thetaDirection,phiDirection,spherical);
  direction.rotate(thetaPosition,phiPosition);

  // Select a random BB frequency
  nuPacket=randomNu();

  // Increase packet counter
  packetIndex++;
  // Hack for openmp
  this->packetIndex=packetIndex;

  // Get boost factor - New!
  // Works with containers in boost mode, right?
  boostFactor=((Cell*) cell)->boostFactor(nuPacket);

  // Select cell - New! - Removed!
  // Note: Selects subcell and creates virtual cell for container
  //_cell = cell->selectCell(position,direction,t);

  // If subcell - Create virtual cell - New!
  if (cell->type()==cellSub) _cell = new CellVirtual(container,cell,position,direction,cellBorder,t);
  // Else - Just copy cell pointer - New!
  else _cell=cell;

  return Packet(EPacket/boostFactor,FrequencyBin(nuPacket,nuPacket),position,direction,t,_cell,grid,packetIndexBase+packetIndex-1,rPacket,packetEventType,packetConfig,boostFactor);
}

// Boost emission probabilities - New!
// Work with containers, right?
void CellPacketSamplerBBL::boostEta() {
  double B,PTot=0;
  int i;

  // Calculate total blackbody intensity
  B=BlackBody::B(nuLim.min,nuLim.max,T,1e-4);  

  // Resize emissivity boost vector
  etaP.resize(((Cell*) cell)->nuBoost.size());

  // For each boost bin
  for (i=0;i<((Cell*)cell)->nuBoost.size();i++) {
    // Calculate emission probability
    etaP[i]=BlackBody::B(((Cell*) cell)->nuBoost[i].min,((Cell*) cell)->nuBoost[i].max,T,1e-4)/B;
  }

  // Normalize emission probability (to be on the safe side)
  for (i=0;i<((Cell*)cell)->nuBoost.size();i++) PTot+=etaP[i];
  for (i=0;i<((Cell*)cell)->nuBoost.size();i++) etaP[i]=etaP[i]/PTot;
  
  // Initialize emissivity boost factor
  etaBoost=0;

  // Calculate emissivity boost factor
  for (i=0;i<((Cell*) cell)->boost.size();i++) etaBoost+=etaP[i]*((Cell*) cell)->boost[i];

  // Adjust emission probabilities
  for (i=0;i<etaP.size();i++) etaP[i]=etaP[i]*((Cell*) cell)->boost[i]/etaBoost;

  // Adjust number of packets
  nPacket=nPacket*etaBoost;
}

// Select random frequency - New!
// Does not work with containers in boost mode (see below)!
double CellPacketSamplerBBL::randomNu() {
  // Select version to use
  // Does not work with containers in boost mode (due to randomNuBoost)!
  if (((Cell*) cell)->config->boost) return randomNuBoost();
  else return randomNuNormal();
}

// Select random frequency (intervall version) - New!
double CellPacketSamplerBBL::randomNu(double nuMin, double nuMax) {
  double z,B,nu;

  // Select a random BB frequency
  z=randomize(0,1);
  B=BlackBody::B(nuMin,nuMax,T,1e-4);
  nu=BlackBody::nuB(z*B,T,nuMin,1e-3);

  return nu;
}

// Select random frequency (normal verions) - New!
double CellPacketSamplerBBL::randomNuNormal() {
  double z,B,nu;
 
  // Select a random BB frequency
  z=randomize(0,1);
  B=BlackBody::B(nuLim.min,nuLim.max,T,1e-4);
  nu=BlackBody::nuB(z*B,T,nuLim.min,1e-3);

  return nu;
}

// Select random frequency (boost verions) - New!
// Work with containers, right?
double CellPacketSamplerBBL::randomNuBoost() {
  double z,PTot=0;
  int i;

  // Draw random number between 0 and 1
  z=randomize(0,1);

  // For each boost bin
  for (i=0;i<etaP.size();i++) {

    // If bin drawn - Return emission frequency
    if (z<=PTot+etaP[i]) return randomNu(((Cell*) cell)->nuBoost[i].min,((Cell*) cell)->nuBoost[i].max);
    // Else increase accumulated probability
    else PTot+=etaP[i];
  }

  // This should not happen!
  cout<<"CellPacketSamplerBBL::randomNuBoost: Error: No frequency drawn, z: "<<z<<", PTot: "<<PTot<<"\n";

  return 0.0;
}

// Constructor
// Sampler does not (exactly) conserve energy - EPacket need to be adjusted - Rewrite !!!
// Works with containers in boost mode, right?
PacketSamplerGamma::PacketSamplerGamma(const int& _nPacket, Grid* _grid, PacketConfig* _packetConfig, boost::mpi::communicator* _mpiCom, int _packetIndexBase) : grid(_grid), packetConfig(_packetConfig), mpiCom(_mpiCom) {

  int i,j;
  double dt,E=0;
  bool master=false;

  if (mpiCom==NULL) master=true;
  else if (mpiCom->rank()==0) master=true;

  // Initialize number of packets, packet index and packet index base
  nPacket=_nPacket;
  packetIndex=0;
  packetIndexBase=_packetIndexBase;

  // For each cell
  for (i=0;i<grid->cell.size();i++) {
    // Get the time span of the cell
    dt=grid->cell[i]->tMax()-grid->cell[i]->tMin();
    // Add up the 56Ni gamma-ray energy
    E+=grid->cell[i]->M(28,28)*eGamma56Ni*dt;
    // Add up the 56Co gamma-ray energy
    E+=grid->cell[i]->M(27,29)*eGamma56Co*dt;
    // Add up the 56Co positron energy
    E+=grid->cell[i]->M(27,29)*ePos56Co*dt;
    // Add up the 57Ni gamma-ray energy
    E+=grid->cell[i]->M(28,29)*eGamma57Ni*dt;
    // Add up the 57Co gamma-ray energy
    E+=grid->cell[i]->M(27,30)*eGamma57Co*dt;
    // Add up the 57Co positron energy
    E+=grid->cell[i]->M(27,30)*ePos57Co*dt;
    // Add up the 44Ti gamma-ray energy
    E+=grid->cell[i]->M(22,22)*eGamma44Ti*dt;
    // Add up the 44Ti positron energy
    E+=grid->cell[i]->M(22,22)*ePos44Ti*dt;
  }
  // Caclulate the energy of the packets
  EPacket=E/nPacket;

  // Initialize cell packet sampler list
  cellPacketSampler.resize(0);

  // Reset number of packets
  nPacket=0;

  // For each cell
  for (i=0;i<grid->cell.size();i++) {
    // If cell container
    if (grid->cell[i]->type()==cellContainer) {
      // For each sub cell
      for (j=0;j<grid->cell[i]->subCell.size();j++) {
	// Create cell packet sampler and append to list
	cellPacketSampler.push_back(CellPacketSamplerGamma(grid,grid->cell[i]->subCell[j],grid->cell[i],packetConfig,EPacket,mpiCom,packetIndexBase+nPacket)); // Changed!
	// Add upp number of cell packets
	nPacket+=cellPacketSampler[cellPacketSampler.size()-1].nPacket;
      }
    }
    // Else
    else {
      // Create cell packet sampler and append to list
      cellPacketSampler.push_back(CellPacketSamplerGamma(grid,grid->cell[i],NULL,packetConfig,EPacket,mpiCom,packetIndexBase+nPacket)); // Changed!
      // Add upp number of packet cell packets
      nPacket+=cellPacketSampler[cellPacketSampler.size()-1].nPacket;
    }
  }

  if (master) cout<<"PacketSamplerGamma::PacketSamplerGamma: Initited for sampling of "<<nPacket<<" packets with energy "<<EPacket<<", dt: "<<dt<<"\n";
}  

// Sample gamma-ray packet
// Works with containers in boost mode, right?
Packet PacketSamplerGamma::sample(int _packetIndex) {
  int i,cellPacketIndex;
  double nPacketTot=0;
  // Hack for openmp
  // Is this really threadsafe?
  int packetIndex=this->packetIndex;

  // If provided - Set packet index
  if (_packetIndex!=noIndex) packetIndex=_packetIndex;

  // For each cell packet sampler
  for (i=0;i<cellPacketSampler.size();i++) {

    // Calculate total number of packets
    nPacketTot+=cellPacketSampler[i].nPacket;

    // If larger than packet counter
    if (nPacketTot>packetIndex) {

      // Set cell packet index
      if (_packetIndex!=noIndex) cellPacketIndex=packetIndex-(nPacketTot-cellPacketSampler[i].nPacket);
      else cellPacketIndex=noIndex;

      // Increase packet counter
      packetIndex++;
      // Hack for openmp
      // Is this really threadsafe?
      this->packetIndex=packetIndex;

      // Sample and return packet
      return cellPacketSampler[i].sample(cellPacketIndex);
    }
  }
}

// Constructor
// Works with containers in boost mode, right?
CellPacketSamplerGamma::CellPacketSamplerGamma(Grid* _grid, CellBase* _cell, CellBase* _container, PacketConfig* _packetConfig, const double& _EPacket, boost::mpi::communicator* _mpiCom, int _packetIndexBase) : grid(_grid), cell(_cell), container(_container), packetConfig(_packetConfig), EPacket(_EPacket), mpiCom(_mpiCom) { // Changed!
  double dt;
  bool master=false;

  if (mpiCom==NULL) master=true;
  else if (mpiCom->rank()==0) master=true;

  // Initialize packet index and packet index base
  packetIndex=0;
  packetIndexBase=_packetIndexBase,

  // Set the time to the mean time of the cell
  t=(cell->tMin()+cell->tMax())/2;

  // Get the time span of the cell
  dt=cell->tMax()-cell->tMin();

  // If boost mode - Adjust packet energy - New!
  // For consistency we need to adjust the grid sampler packet energy as well! No effect in practice though ...
  // Works with containers in boost mode, right?
  // Testing!
  if (((Cell*) cell)->config->boost) EPacket=EPacket/((Cell*) cell)->comptonKappaBoost();
  if (((Cell*) cell)->config->boost and master) cout <<"CellPacketSamplerGamma::CellPacketSamplerGamma: Cell: "<<cell->index()<<", Packet energy adjusted using a bost factor of "<<1/((Cell*) cell)->comptonKappaBoost()<<"\n";

  // Calculate the number of 56Ni gamma-ray packets
  nPacketGamma56Ni=(cell->M(28,28)*eGamma56Ni*dt)/EPacket;
  // Calculate the number of 56Co gamma-ray packets
  nPacketGamma56Co=(cell->M(27,29)*eGamma56Co*dt)/EPacket;
  // Calculate the number of 56Co positron packets
  nPacketPos56Co=(cell->M(27,29)*ePos56Co*dt)/EPacket;
  // Calculate the number of 57Ni gamma-ray packets
  nPacketGamma57Ni=(cell->M(28,29)*eGamma57Ni*dt)/EPacket;
  // Calculate the number of 57Co gamma-ray packets
  nPacketGamma57Co=(cell->M(27,30)*eGamma57Co*dt)/EPacket;
  // Calculate the number of 57Co positron packets - Changed!
  //nPacketPos57Co=(cell->M(27,30)*eGamma57Co*dt)/EPacket;
  nPacketPos57Co=(cell->M(27,30)*ePos57Co*dt)/EPacket;
  // Calculate the number of 44Ti gamma-ray packets
  nPacketGamma44Ti=(cell->M(22,22)*eGamma44Ti*dt)/EPacket;
  // Calculate the number of 44Ti positron packets
  nPacketPos44Ti=(cell->M(22,22)*ePos44Ti*dt)/EPacket;
  // Calculate the total number of packets in the cell
  nPacket=nPacketGamma56Ni+nPacketGamma56Co+nPacketPos56Co+nPacketGamma57Ni+nPacketGamma57Co+nPacketPos57Co+nPacketGamma44Ti+nPacketPos44Ti;

  if (master) cout<<"CellPacketSamplerGamma::CellPacketSamplerGamma: Cell: "<<cell->index()<<", Initited for sampling of "<<nPacket<<" packets with energy "<<EPacket<<", dt: "<<dt<<"\n";
  if (master) cout<<"CellPacketSamplerGamma::CellPacketSamplerGamma: Cell: "<<cell->index()<<", nPacketGamma56Ni "<<nPacketGamma56Ni<<", nPacketGamma56Co: "<<nPacketGamma56Co<<", nPacketPos56Co: "<<nPacketPos56Co<<", M(28,28): "<<cell->M(28,28)<<", M(27,29): "<<cell->M(27,29)<<"\n";
}

// Sample gamma-ray packet
// Works with containers in boost mode, right?
Packet CellPacketSamplerGamma::sample(int _packetIndex) {
  double nuPacket;
  Vector3D position,direction;
  PacketConfig* _packetConfig;
  PacketEventType packetEventType;
  CellBase* _cell; // New!
  // Testing!
  double boostFactor; // New!

  // Hack for openmp
  // Is this really threadsafe?
  int packetIndex=this->packetIndex;

  // If provided - Set packet index
  if (_packetIndex!=noIndex) packetIndex=_packetIndex;

  // Create copy of packet configuration
  _packetConfig = new PacketConfig(*packetConfig);
  
  // Store in packet config list
  // Need to solve this - How to delete PacketConfig objects !!!
  //packetConfigList.push_back(packetConfig);

  // If 56Ni gamma-ray packet
  if (packetIndex<nPacketGamma56Ni) {
    nuPacket=nuGamma56Ni;
    _packetConfig->comptonAbsSigma=sigmaGamma56Ni;
    packetEventType=ev56NiGammaDecay;
  }
  // If 56Co gamma-ray packet
  else if (packetIndex<(nPacketGamma56Ni+nPacketGamma56Co)) {
    nuPacket=nuGamma56Co;
    _packetConfig->comptonAbsSigma=sigmaGamma56Co;
    packetEventType=ev56CoGammaDecay;
  }
  // If 56Co positron packet
  else if (packetIndex<(nPacketGamma56Ni+nPacketGamma56Co+nPacketPos56Co)) {
    nuPacket=nuPos56Co;
    _packetConfig->comptonAbsSigma=sigmaPos56Co;
    packetEventType=ev56CoPosDecay;
  }
  // If 57Ni gamma-ray packet
  else if (packetIndex<(nPacketGamma56Ni+nPacketGamma56Co+nPacketPos56Co+nPacketGamma57Ni)) {
    nuPacket=nuGamma57Ni;
    _packetConfig->comptonAbsSigma=sigmaGamma57Ni;
    packetEventType=ev57NiGammaDecay;
  }
  // If 57Co gamma-ray packet
  else if (packetIndex<(nPacketGamma56Ni+nPacketGamma56Co+nPacketPos56Co+nPacketGamma57Ni+nPacketGamma57Co)) {
    nuPacket=nuGamma57Co;
    _packetConfig->comptonAbsSigma=sigmaGamma57Co;
    packetEventType=ev57CoGammaDecay;
  }
  // If 57Co positron packet
  else if (packetIndex<(nPacketGamma56Ni+nPacketGamma56Co+nPacketPos56Co+nPacketGamma57Ni+nPacketGamma57Co+nPacketPos57Co)) {
    nuPacket=nuPos57Co;
    _packetConfig->comptonAbsSigma=sigmaPos57Co;
    packetEventType=ev57CoPosDecay;
  }
  // If 44Ti gamma-ray packet
  else if (packetIndex<(nPacketGamma56Ni+nPacketGamma56Co+nPacketPos56Co+nPacketGamma57Ni+nPacketGamma57Co+nPacketPos57Co+nPacketGamma44Ti)) {
    nuPacket=nuGamma44Ti;
    _packetConfig->comptonAbsSigma=sigmaGamma44Ti;
    packetEventType=ev44TiGammaDecay;
  }
  // If 44Ti positron packet
  else if (packetIndex<nPacket) {
    nuPacket=nuPos44Ti;
    _packetConfig->comptonAbsSigma=sigmaPos44Ti;
    packetEventType=ev44TiPosDecay;
  }

  // Turn compton absorption on (should this be configured elsewhere?)
  _packetConfig->comptonAbs=true;

  // Select a random position
  position=cell->randomPosition(cellInterior);

  // Select a random direction
  direction=cell->randomDirection(cellInterior);

  // Increase packet counter
  packetIndex++;
  // Hack for openmp
  this->packetIndex=packetIndex;

  // Get boost factor - New!
  // Works with containers in boost mode, right?
  // Testing!
  boostFactor=((Cell*) cell)->comptonKappaBoost();

  // If subcell - Create virtual cell - New!
  if (cell->type()==cellSub) _cell = new CellVirtual(container,cell,position,direction,cellInterior,t);
  // Else - Just copy cell pointer - New!
  else _cell=cell;

  // Create and return the packet
  //return Packet(EPacket,FrequencyBin(nuPacket,nuPacket),position,direction,t,cell,grid,packetIndexBase+packetIndex-1,gPacket,packetEventType,_packetConfig);
  // Testing!
  return Packet(EPacket,FrequencyBin(nuPacket,nuPacket),position,direction,t,_cell,grid,packetIndexBase+packetIndex-1,gPacket,packetEventType,_packetConfig,boostFactor); // Changed!
}

// Constructor
// Sampler does not (exactly) conserve energy - EPacket needs to be adjusted - Rewrite !!!
PacketSamplerThermal::PacketSamplerThermal(const int& _nPacket, int cellIndexMin, int cellIndexMax, Grid* _grid, PacketConfig* _packetConfig, int _packetIndexBase) : grid(_grid), packetConfig(_packetConfig) {

  int i,j;
  double t,E=0;

  // Prevent faulty cell range
  cellIndexMin=(cellIndexMin>=0)?cellIndexMin:0; 
  cellIndexMax=(cellIndexMax<=(grid->cell.size()-1))?cellIndexMax:(grid->cell.size()-1);

  // Initialize number of packets, packet index and packet index base
  nPacket=_nPacket;
  packetIndex=0;
  packetIndexBase=_packetIndexBase;

  // Get (centered) time
  t=grid->t();

  // Calculate total thermal energy
  for (i=cellIndexMin;i<=cellIndexMax;i++) E+=grid->cell[i]->E()*grid->cell[i]->rho()*grid->cell[i]->V();

  // Caclulate the energy of the packets
  EPacket=E/nPacket;

  // Initialize cell packet sampler list
  cellPacketSampler.resize(0);

  // Reset number of packets
  nPacket=0;

  // For each cell
  for (i=cellIndexMin;i<=cellIndexMax;i++) {
    // If cell container
    if (grid->cell[i]->type()==cellContainer) {
      // For each sub cell
      for (j=0;j<grid->cell[i]->subCell.size();j++) {
	// Create cell packet sampler and append to list
	cellPacketSampler.push_back(CellPacketSamplerThermal(grid,grid->cell[i]->subCell[j],grid->cell[i],packetConfig,EPacket,packetIndexBase+nPacket)); // Changed!
	// Add upp number of cell packets
	nPacket+=cellPacketSampler[cellPacketSampler.size()-1].nPacket;
      }
    }
    // Else
    else {
      // Create cell packet sampler and append to list
      cellPacketSampler.push_back(CellPacketSamplerThermal(grid,grid->cell[i],NULL,packetConfig,EPacket,packetIndexBase+nPacket)); // Changed!
      // Add upp number of packet cell packets
      nPacket+=cellPacketSampler[cellPacketSampler.size()-1].nPacket;
    }
  }

  cout<<"PacketSamplerThermal::PacketSamplerThermal: Initited for sampling of "<<nPacket<<" packets with energy "<<EPacket<<"\n";
}  

// Sample thermal packet
Packet PacketSamplerThermal::sample(int _packetIndex) {
  int i,cellPacketIndex;
  double nPacketTot=0;
  // Hack for openmp
  int packetIndex=this->packetIndex;

  // If provided - Set packet index
  if (_packetIndex!=noIndex) packetIndex=_packetIndex;

  // For each cell packet sampler
  for (i=0;i<cellPacketSampler.size();i++) {

    // Calculate total number of packets
    nPacketTot+=cellPacketSampler[i].nPacket;

    // If larger than packet counter
    if (nPacketTot>packetIndex) {

      // Set cell packet index
      if (_packetIndex!=noIndex) cellPacketIndex=packetIndex-(nPacketTot-cellPacketSampler[i].nPacket);
      else cellPacketIndex=noIndex;

      // Increase packet counter
      packetIndex++;
      // Hack for openmp
      this->packetIndex=packetIndex;

      // Sample and return packet
      return cellPacketSampler[i].sample(cellPacketIndex);
    }
  }
}

// Reset
void PacketSamplerThermal::reset() {
  int i;
  PacketSampler::reset();
  for (i=0;i<cellPacketSampler.size();i++) cellPacketSampler[i].reset();
}

// Constructor
CellPacketSamplerThermal::CellPacketSamplerThermal(Grid* _grid, CellBase* _cell, CellBase* _container, PacketConfig* _packetConfig, const double& _EPacket, int _packetIndexBase) : grid(_grid), cell(_cell), container(_container), packetConfig(_packetConfig), EPacket(_EPacket) { // Changed!
  double dt;

  // Initialize packet index and packet index base
  packetIndex=0;
  packetIndexBase=_packetIndexBase,

  // Set the time to the minimum time of the cell
  t=cell->tMin();

  // Calculate the number of thermal packets
  nPacket=(cell->E()*cell->rho()*cell->V())/EPacket;
}

// Sample thermal packet
Packet CellPacketSamplerThermal::sample(int _packetIndex) {
  double nuPacket;
  Vector3D position,direction;
  PacketEventType packetEventType;
  CellBase* _cell; // New!
  // Hack for openmp
  int packetIndex=this->packetIndex;

  // If provided - Set packet index
  if (_packetIndex!=noIndex) packetIndex=_packetIndex;

  nuPacket=-1;
  packetEventType=evNone;

  // Select a random position
  position=cell->randomPosition(cellInterior,t);

  // Select a random direction
  direction=cell->randomDirection(cellInterior);

  // Increase packet counter
  packetIndex++;
  // Hack for openmp
  this->packetIndex=packetIndex;

  // If subcell - Create virtual cell - New!
  if (cell->type()==cellSub) _cell = new CellVirtual(container,cell,position,direction,cellInterior,t);
  // Else - Just copy cell pointer - New!
  else _cell=cell;

  // Create and return the packet
  return Packet(EPacket,FrequencyBin(nuPacket,nuPacket),position,direction,t,_cell,grid,packetIndexBase+packetIndex-1,kPacket,packetEventType,packetConfig); // Changed!
}

// Constructor
// Sampler does not (exactly) conserve energy - EPacket needs to be adjusted - Rewrite !!!
// Works with containers in boost mode, right?
PacketSamplerBBThermal::PacketSamplerBBThermal(const int& _nPacket, int cellIndexMin, int cellIndexMax, Grid* _grid, const FrequencyBin& nuLim, PacketConfig* _packetConfig, boost::mpi::communicator* _mpiCom, int _packetIndexBase, double _expFactorWeight, bool _LCorr, bool _eRadioCorr) : grid(_grid), packetConfig(_packetConfig), mpiCom(_mpiCom), expFactorWeight(_expFactorWeight), LCorr(_LCorr), eRadioCorr(_eRadioCorr) {
// Changed! - Testing!
//PacketSamplerBBThermal::PacketSamplerBBThermal(const int& _nPacket, int cellIndexMin, int cellIndexMax, Grid* _grid, const FrequencyBin& nuLim, PacketConfig* _packetConfig, boost::mpi::communicator* _mpiCom, int _packetIndexBase, vector<double> _dEdt) : grid(_grid), packetConfig(_packetConfig), mpiCom(_mpiCom), dEdt(_dEdt) {

  int i,j;
  double expFactor,t,E=0;
  //double ECell; // New!
  bool master=false;

  if (mpiCom==NULL) master=true;
  else if (mpiCom->rank()==0) master=true;

  // Prevent faulty cell range
  cellIndexMin=(cellIndexMin>=0)?cellIndexMin:0; 
  cellIndexMax=(cellIndexMax<=(grid->cell.size()-1))?cellIndexMax:(grid->cell.size()-1);

  // Initialize number of packets, packet index and packet index base
  nPacket=_nPacket;
  packetIndex=0;
  packetIndexBase=_packetIndexBase;

  // Get (centered) time
  t=grid->t();

  // Calculate total thermal energy
  //for (i=cellIndexMin;i<=cellIndexMax;i++) E+=grid->cell[i]->E()*grid->cell[i]->rho()*grid->cell[i]->V();
  // Take dilution factor into account - Hack!!!
  //for (i=cellIndexMin;i<=cellIndexMax;i++) E+=grid->cell[i]->W()*grid->cell[i]->E()*grid->cell[i]->rho()*grid->cell[i]->V();
  // Changed again!
  for (i=cellIndexMin;i<=cellIndexMax;i++) E+=ECell(i)*grid->cell[i]->V();

  // Caclulate the energy of the packets
  expFactor=(grid->tMin()+expFactorWeight*(grid->tMax()-grid->tMin()))/grid->tMin();
  EPacket=E*expFactor/nPacket;
  // Changed! - Testing!
  //EPacket=E/nPacket;

  // Initialize cell packet sampler list
  cellPacketSampler.resize(0);

  // Reset number of packets
  nPacket=0;

  // For each cell
  for (i=cellIndexMin;i<=cellIndexMax;i++) {
    // If cell container
    if (grid->cell[i]->type()==cellContainer) {
      // For each sub cell
      for (j=0;j<grid->cell[i]->subCell.size();j++) {
	// Create cell packet sampler and append to list
	//cellPacketSampler.push_back(CellPacketSamplerBBThermal(grid,grid->cell[i]->subCell[j],grid->cell[i],nuLim,packetConfig,EPacket,ECell(i,j),dEdr(i,j),mpiCom,packetIndexBase+nPacket,expFactorWeight)); // Changed!
	cellPacketSampler.push_back(CellPacketSamplerBBThermal(grid,grid->cell[i]->subCell[j],grid->cell[i],nuLim,packetConfig,EPacket,ECell(i,j),dEdrBack(i,j),dEdrFront(i,j),mpiCom,packetIndexBase+nPacket,expFactorWeight)); // Changed! - Testing!
	//cellPacketSampler.push_back(CellPacketSamplerBBThermal(grid,grid->cell[i]->subCell[j],grid->cell[i],nuLim,packetConfig,EPacket,ECell(i,j),dEdrBack(i,j),dEdrFront(i,j),mpiCom,packetIndexBase+nPacket)); // Changed! - Testing!
	// Add upp number of cell packets
	nPacket+=cellPacketSampler[cellPacketSampler.size()-1].nPacket;
      }
    }
    // Else
    else {
      // Create cell packet sampler and append to list
      //cellPacketSampler.push_back(CellPacketSamplerBBThermal(grid,grid->cell[i],NULL,nuLim,packetConfig,EPacket,ECell(i),dEdr(i),mpiCom,packetIndexBase+nPacket,expFactorWeight)); // Changed!
      cellPacketSampler.push_back(CellPacketSamplerBBThermal(grid,grid->cell[i],NULL,nuLim,packetConfig,EPacket,ECell(i),dEdrBack(i),dEdrFront(i),mpiCom,packetIndexBase+nPacket,expFactorWeight)); // Changed! - Testing!
      //cellPacketSampler.push_back(CellPacketSamplerBBThermal(grid,grid->cell[i],NULL,nuLim,packetConfig,EPacket,ECell(i),dEdrBack(i),dEdrFront(i),mpiCom,packetIndexBase+nPacket)); // Changed! - Testing!
      // Add upp number of packet cell packets
      nPacket+=cellPacketSampler[cellPacketSampler.size()-1].nPacket;
    }
  }

  if (master) cout<<"PacketSamplerBBThermal::PacketSamplerBBThermal: Initited for sampling of "<<nPacket<<" packets with energy "<<EPacket<<"\n";
}  

// Sample thermal packet
// Works with containers in boost mode, right?
Packet PacketSamplerBBThermal::sample(int _packetIndex) {
  int i,cellPacketIndex;
  double nPacketTot=0;
  // Hack for openmp
  int packetIndex=this->packetIndex;

  // If provided - Set packet index
  if (_packetIndex!=noIndex) packetIndex=_packetIndex;

  // For each cell packet sampler
  for (i=0;i<cellPacketSampler.size();i++) {

    // Calculate total number of packets
    nPacketTot+=cellPacketSampler[i].nPacket;

    // If larger than packet counter
    if (nPacketTot>packetIndex) {

      // Set cell packet index
      if (_packetIndex!=noIndex) cellPacketIndex=packetIndex-(nPacketTot-cellPacketSampler[i].nPacket);
      else cellPacketIndex=noIndex;

      // Increase packet counter
      packetIndex++;
      // Hack for openmp
      this->packetIndex=packetIndex;

      // Sample and return packet
      return cellPacketSampler[i].sample(cellPacketIndex);
    }
  }
}

// Reset
void PacketSamplerBBThermal::reset() {
  int i;
  PacketSampler::reset();
  for (i=0;i<cellPacketSampler.size();i++) cellPacketSampler[i].reset();
}

// Calculate energy (per volume)
double PacketSamplerBBThermal::ECell(const int& i, int j) const {
  CellSS* cell;
  double E,dE;
  bool master=false;

  if (mpiCom==NULL) master=true;
  else if (mpiCom->rank()==0) master=true;

  //if (j==noCell) cell=((CellSS*) grid->cell[i]);
  //else cell=((CellSS*) grid->cell[i]->subCell[j]);
  // New! - Testing!
  cell=((CellSS*) grid->cell[i]);

  // Calculate energy (per volome)
  E=cell->E()*cell->rho();
  //if (master) cout<<"PacketSamplerBBThermal::ECell: cell: "<<cell->index()<<", E: "<<E*cell->V()<<"\n";
  // If not one - Take dilution factor into account
  if (cell->W()!=1) E*=cell->W()*pow(cell->TR()/cell->T(),4);
  //if (master) cout<<"PacketSamplerBBThermal::ECell: cell: "<<cell->index()<<", E: "<<E*cell->V()<<", W: "<<cell->W()<<", T: "<<cell->T()<<", TR: "<<cell->TR()<<", WCorr: "<<cell->W()*pow(cell->TR()/cell->T(),4)<<"\n";
  // If requested - correct for luminosity at cell borders - New!
  // Removed! - Testing!
  if (LCorr) E-=MIN(expFactorWeight*(cell->LBack()-cell->LFront())*(cell->tMax()-cell->tMin())/cell->V(),E);
  //if (master) cout<<"PacketSamplerBBThermal::ECell: cell: "<<cell->index()<<" E: "<<E*cell->V()<<", LCorr: "<<LCorr<<"expFactorWeight: "<<expFactorWeight<<", LBack: "<<cell->LBack()<<", LFront: "<<cell->LFront()<<", dt: "<<(cell->tMax()-cell->tMin())<<", dE: "<<expFactorWeight*(cell->LBack()-cell->LFront())*(cell->tMax()-cell->tMin())<<"\n";
  // If requested - correct for radioactive deposition - New!
  // Removed! - Testing!
  if (eRadioCorr) E-=MIN(expFactorWeight*(cell->eRadio()*cell->rho())*(cell->tMax()-cell->tMin()),E);
  //if (master) cout<<"PacketSamplerBBThermal::ECell: cell: "<<cell->index()<<" E: "<<E*cell->V()<<", eRadioCorr: "<<eRadioCorr<<"expFactorWeight: "<<expFactorWeight<<", eRadio: "<<cell->eRadio()<<", rho: "<<cell->rho()<<", dt: "<<(cell->tMax()-cell->tMin())<<", dE: "<<expFactorWeight*(cell->eRadio()*cell->rho())*(cell->tMax()-cell->tMin())*cell->V()<<"\n";

  // Testing!
  //if (master) cout<<"PacketSamplerBBThermal::ECell: cell: "<<cell->index()<<", E: "<<E<<", LCorr: "<<MIN(0.5*(cell->LBack()-cell->LFront())*(cell->tMax()-cell->tMin())/cell->V(),E)<<", eRadioCorr: "<<MIN(0.5*(cell->eRadio()*cell->rho())*(cell->tMax()-cell->tMin()),E)<<", expFactor(0.5): "<<(grid->tMin()+0.5*(grid->tMax()-grid->tMin()))/grid->tMin()<<", expFactor(1.0): "<<grid->tMax()/grid->tMin()<<"\n";
  //if (master and dEdt.size()!=0) cout<<"PacketSamplerBBThermal::ECell: cell: "<<cell->index()<<", dEdtCorr: "<<dEdt[i]*cell->rho()*(cell->t()-cell->tMin())<<", dEdt: "<<dEdt[i]<<", rho: "<<cell->rho()<<", dt: "<<(cell->t()-cell->tMin())<<", W: "<<cell->W()<<"\n";

  // If provided - Scale to tMin using energy time derivative - New! - Testing!
  //if (dEdt.size()!=0) E-=MIN(dEdt[i]*cell->rho()*(cell->t()-cell->tMin()),E);

  // If subcell - Multiply by filling factor - New! - Testing!
  //if (j!=noCell) E*=cell->subCell[j]->fillFactor(); // Removed!

  return E;
}

// Calculate energy (per volume) gradient
double PacketSamplerBBThermal::dEdr(const int& i, int j) const { // Changed!
  double dEBack,dEFront,drBack,drFront,dEdrBack,dEdrFront,dEdr;
  //double ECell,ECellNext,ECellPrev,ESubCell; // New!
  bool master=false;

  if (mpiCom==NULL) master=true;
  else if (mpiCom->rank()==0) master=true;

  // If one-cell grid - return zero gradient 
  if (grid->cell.size()==1) return 0;

  //if (master) cout<<"PacketSamplerBBThermal::dEdr: cell: "<<i<<", subCell: "<<j<<"\n";

  if (i>0) {
    //dEBack=(grid->cell[i]->E()*grid->cell[i]->rho()-grid->cell[i-1]->E()*grid->cell[i-1]->rho());
    // Take dilution factor into account - Hack!!!
    //dEBack=(grid->cell[i]->W()*grid->cell[i]->E()*grid->cell[i]->rho()-grid->cell[i-1]->W()*grid->cell[i-1]->E()*grid->cell[i-1]->rho());
    // Changed again!
    dEBack=(ECell(i)-ECell(i-1));
    drBack=(((CellSS*) grid->cell[i])->rMin()+((CellSS*) grid->cell[i])->rMax())/2-(((CellSS*) grid->cell[i-1])->rMin()+((CellSS*) grid->cell[i-1])->rMax())/2;
    dEdrBack=dEBack/drBack;
  }

  if (i<grid->cell.size()-1) {
    //dEFront=(grid->cell[i+1]->E()*grid->cell[i+1]->rho()-grid->cell[i]->E()*grid->cell[i]->rho());
    // Take dilution factor into account - Hack!!!
    //dEFront=(grid->cell[i+1]->W()*grid->cell[i+1]->E()*grid->cell[i+1]->rho()-grid->cell[i]->W()*grid->cell[i]->E()*grid->cell[i]->rho());
    // Changed again!
    dEFront=(ECell(i+1)-ECell(i));
    drFront=(((CellSS*) grid->cell[i+1])->rMin()+((CellSS*) grid->cell[i+1])->rMax())/2-(((CellSS*) grid->cell[i])->rMin()+((CellSS*) grid->cell[i])->rMax())/2;
    dEdrFront=dEFront/drFront;
  }

  //if (i==0) dEdr=dEdrFront;
  //else if (i==(grid->cell.size()-1)) dEdr=dEdrBack;
  //else dEdr=(dEdrBack+dEdrFront)/2;
  // Testing!
  if (i==(grid->cell.size()-1)) dEdr=dEdrBack;
  else dEdr=dEdrFront;

  // If sub subcell index provided - Scale gradient to subcell energy - New!
  //if (j!=noCell) dEdr=dEdr*((grid->cell[i]->subCell[j]->W()*grid->cell[i]->subCell[j]->E()*grid->cell[i]->subCell[j]->rho())/(grid->cell[i]->W()*grid->cell[i]->E()*grid->cell[i]->rho()));
  // Changed!
  //if (j!=noCell) dEdr=dEdr*ECell(i,j)/ECell(i);
  // If sub subcell - Multiply by filling factor - Changed! - Testing!
  //if (j!=noCell) dEdr*=grid->cell[i]->subCell[j]->fillFactor(); // Removed!

  //if (master) cout<<"PacketSamplerBBThermal::dEdr, Return\n";

  return dEdr;
}

// Calculate energy (per volume) gradient
double PacketSamplerBBThermal::dEdrBack(const int& i, int j) const { // Changed!
  double dE,dr,dEdr;

  // If one-cell grid - return zero gradient 
  if (grid->cell.size()==1) return 0;

  if (i>0) {
    dE=(ECell(i)-ECell(i-1));
    dr=(((CellSS*) grid->cell[i])->rMin()+((CellSS*) grid->cell[i])->rMax())/2-(((CellSS*) grid->cell[i-1])->rMin()+((CellSS*) grid->cell[i-1])->rMax())/2;
    dEdr=dE/dr;
  }
  else {
    dE=(ECell(i+1)-ECell(i));
    dr=(((CellSS*) grid->cell[i+1])->rMin()+((CellSS*) grid->cell[i+1])->rMax())/2-(((CellSS*) grid->cell[i])->rMin()+((CellSS*) grid->cell[i])->rMax())/2;
    dEdr=dE/dr;
  }

  // If sub subcell index provided - Scale gradient to subcell energy - New!
  //if (j!=noCell) dEdr=dEdr*ECell(i,j)/ECell(i);
  // If sub subcell - Multiply by filling factor - New! - Testing!
  //if (j!=noCell) dEdr*=grid->cell[i]->subCell[j]->fillFactor(); // Removed!

  return dEdr;
}

// Calculate energy (per volume) gradient
double PacketSamplerBBThermal::dEdrFront(const int& i, int j) const { // Changed!
  double dE,dr,dEdr;

  // If one-cell grid - return zero gradient 
  if (grid->cell.size()==1) return 0;

  if (i<grid->cell.size()-1) {
    dE=(ECell(i+1)-ECell(i));
    dr=(((CellSS*) grid->cell[i+1])->rMin()+((CellSS*) grid->cell[i+1])->rMax())/2-(((CellSS*) grid->cell[i])->rMin()+((CellSS*) grid->cell[i])->rMax())/2;
    dEdr=dE/dr;
  }
  else  {
    dE=(ECell(i)-ECell(i-1));
    dr=(((CellSS*) grid->cell[i])->rMin()+((CellSS*) grid->cell[i])->rMax())/2-(((CellSS*) grid->cell[i-1])->rMin()+((CellSS*) grid->cell[i-1])->rMax())/2;
    dEdr=dE/dr;
  }

  // If sub subcell index provided - Scale gradient to subcell energy - New!
  //if (j!=noCell) dEdr=dEdr*ECell(i,j)/ECell(i);
  // If sub subcell - Multiply by filling factor - New! - Testing!
  //if (j!=noCell) dEdr*=grid->cell[i]->subCell[j]->fillFactor(); // Removed!

  return dEdr;
}

// Constructor
// Works with containers in boost mode, right?
//CellPacketSamplerBBThermal::CellPacketSamplerBBThermal(Grid* _grid, Cell* _cell, CellBase* _container, const FrequencyBin& _nuLim, PacketConfig* _packetConfig, const double& _EPacket, const double& _E, const double& _dEdr, boost::mpi::communicator* _mpiCom, int _packetIndexBase, double expFactorWeight) : grid(_grid), cell(_cell), container(_container), nuLim(_nuLim), packetConfig(_packetConfig), EPacket(_EPacket), E(_E), dEdr(_dEdr), mpiCom(_mpiCom) { // Changed!
CellPacketSamplerBBThermal::CellPacketSamplerBBThermal(Grid* _grid, Cell* _cell, CellBase* _container, const FrequencyBin& _nuLim, PacketConfig* _packetConfig, const double& _EPacket, const double& _E, const double& _dEdrBack, const double& _dEdrFront, boost::mpi::communicator* _mpiCom, int _packetIndexBase, double expFactorWeight) : grid(_grid), cell(_cell), container(_container), nuLim(_nuLim), packetConfig(_packetConfig), EPacket(_EPacket), E(_E), dEdrBack(_dEdrBack), dEdrFront(_dEdrFront), mpiCom(_mpiCom) { // Changed! - Testing!
  //CellPacketSamplerBBThermal::CellPacketSamplerBBThermal(Grid* _grid, Cell* _cell, CellBase* _container, const FrequencyBin& _nuLim, PacketConfig* _packetConfig, const double& _EPacket, const double& _E, const double& _dEdrBack, const double& _dEdrFront, boost::mpi::communicator* _mpiCom, int _packetIndexBase) : grid(_grid), cell(_cell), container(_container), nuLim(_nuLim), packetConfig(_packetConfig), EPacket(_EPacket), E(_E), dEdrBack(_dEdrBack), dEdrFront(_dEdrFront), mpiCom(_mpiCom) { // Changed! - Testing!
  //double expFactor,dt,dr,rCenter,rCenterSub,VSub,E,PSum=0;
  double expFactor,dt,dr,rCenter,rCenterSub,VSub,PSum=0;
  double ECell; // New!
  int nSub=10,i;
  bool master=false;

  if (mpiCom==NULL) master=true;
  else if (mpiCom->rank()==0) master=true;

  // Initialize packet index and packet index base
  packetIndex=0;
  packetIndexBase=_packetIndexBase,

  // Set the time to the minimum time of the cell
  t=cell->tMin();

  // Calculate the number of thermal packets
  // Removed! Testing!
  expFactor=(cell->tMin()+expFactorWeight*(cell->tMax()-cell->tMin()))/cell->tMin();
  //nPacket=(cell->E()*cell->rho()*cell->V())*expFactor/EPacket;
  // Take dilution factor into account - Hack!!!
  //nPacket=(cell->W()*cell->E()*cell->rho()*cell->V())*expFactor/EPacket;
  // Changed again!
  nPacket=(E*cell->V())*expFactor/EPacket;
  // And again! - Testing!
  //nPacket=(E*cell->V())/EPacket;

  // If cell not container - Get temperature - Changed!
  if (cell->type()!=cellContainer) T=cell->T();
  // Else - Radiation-energy average subcell temperatures - New!
  else { T=0; for (i=0;i<cell->subCell.size();i++) T+=pow(cell->subCell[i]->T(),4.0); T=pow(T,1.0/4.0); }
  // Note - Should use TR above if W!=1, right?
  
  // Get energy (per volume)
  //E=cell->E()*cell->rho();
  // Take dilution factor into account - Hack!!!
  //E=cell->W()*cell->E()*cell->rho();
  // Removed!

  // Calculate center
  rCenter=(((CellSS*) cell)->rMin()+((CellSS*) cell)->rMax())/2;

  // Calculate dr
  dr=(((CellSS*) cell)->rMax()-((CellSS*) cell)->rMin())/nSub;

  // Initialize subdivisions
  rMinSub.resize(nSub);
  rMaxSub.resize(nSub);
  PSub.resize(nSub);

  // For each subdivision
  for (i=0;i<nSub;i++) {
    // Calculate borders
    rMinSub[i]=((CellSS*) cell)->rMin()+i*dr;
    rMaxSub[i]=((CellSS*) cell)->rMin()+(i+1)*dr;
    // Calculate center
    rCenterSub=(rMinSub[i]+rMaxSub[i])/2;
    // Calculate volume
    VSub=(4.0/3.0)*pi*(pow(rMaxSub[i],3)-pow(rMinSub[i],3));
    // Calculate probability
    //PSub[i]=(E+dEdr*(rCenterSub-rCenter))*VSub;
    // Testing!
    if (rCenterSub<rCenter) PSub[i]=(E+dEdrBack*(rCenterSub-rCenter))*VSub;
    else PSub[i]=(E+dEdrFront*(rCenterSub-rCenter))*VSub;
    // Sum probability
    PSum+=PSub[i];
  }

  // Normalize probability
  for (i=0;i<nSub;i++) PSub[i]=PSub[i]/PSum;

  if (master) cout<<"CellPacketSamplerBBThermal::CellPacketSamplerBBThermal: cellIndex: "<<cell->index()<<", nPacket: "<<nPacket<<", E: "<<nPacket*EPacket<<", T: "<<cell->T()<<", W: "<<cell->W()<<"\n";

  // If boost mode - boost emissivity (and adjust number of packets) - New!
  // Works with containers in boost mode, right?
  if (((Cell*) cell)->config->boost) boostEta();
}

// Sample thermal packet
// Does not work with containers in boost mode (see below)
Packet CellPacketSamplerBBThermal::sample(int _packetIndex) {
  double nuPacket,B,z;
  Vector3D position,direction;
  PacketEventType packetEventType;
  CellBase* _cell; // New!
  double boostFactor; // New!
  // Hack for openmp
  int packetIndex=this->packetIndex;

  // If provided - Set packet index
  if (_packetIndex!=noIndex) packetIndex=_packetIndex;

  nuPacket=-1;
  packetEventType=evEmission;

  // Select a random position
  //position=cell->randomPosition(cellInterior,t);
  position=randomPosition();

  // Select a random direction
  direction=cell->randomDirection(cellInterior);

  // Select a random BB frequency
  /*z=randomize(0,1);
  B=BlackBody::B(nuLim.min,nuLim.max,T,1e-4);
  nuPacket=BlackBody::nuB(z*B,T,nuLim.min,1e-3);*/
  nuPacket=randomNu(); // Changed!

  // Increase packet counter
  packetIndex++;
  // Hack for openmp
  this->packetIndex=packetIndex;

  // Get boost factor - New!
  // Works with containers in boost mode, right?
  boostFactor=((Cell*) cell)->boostFactor(nuPacket);

  // If subcell - Create virtual cell - New!
  if (cell->type()==cellSub) _cell = new CellVirtual(container,cell,position,direction,cellInterior,t);
  // Else - Just copy cell pointer - New!
  else _cell=cell;

  // Create and return the packet
  //return Packet(EPacket,FrequencyBin(nuPacket,nuPacket),position,direction,t,cell,grid,packetIndexBase+packetIndex-1,rPacket,packetEventType,packetConfig);
  return Packet(EPacket/boostFactor,FrequencyBin(nuPacket,nuPacket),position,direction,t,_cell,grid,packetIndexBase+packetIndex-1,rPacket,packetEventType,packetConfig,boostFactor); // Changed!
}

// Sample position of thermal packet
Vector3D CellPacketSamplerBBThermal::randomPosition() const {
  double r,theta,phi,expFactor,z,PTot=0;
  int i;

  // Select random number
  z=randomize(0,1);

  // For each subdivision
  for (i=0;i<PSub.size();i++) {
    // Calculate cumulative probability
    PTot+=PSub[i];
    // If subdivision drawn - Break
    if (z<=PTot) break;
  }

  // Calculate expansion factor
  expFactor=t/((cell->tMin()+cell->tMax())/2);
  // Draw random radius (within subdivision) and scale with expansion factor
  r=pow(pow(rMinSub[i],3)+(pow(rMaxSub[i],3)-pow(rMinSub[i],3))*randomize(0,1),1.0/3.0)*expFactor;
  // Draw random polar angle
  theta=acos(2*randomize(0,1)-1);
  // draw random equatorial angle
  phi=2*pi*randomize(0,1);

  return Vector3D(r,theta,phi,spherical);
}

// Boost emission probabilities - New!
// Works with containers, right?
void CellPacketSamplerBBThermal::boostEta() {
  double B,PTot=0;
  int i;

  // Calculate total blackbody intensity
  B=BlackBody::B(nuLim.min,nuLim.max,T,1e-4);  

  // Resize emissivity boost vector
  etaP.resize(((Cell*) cell)->nuBoost.size());

  // For each boost bin
  for (i=0;i<((Cell*)cell)->nuBoost.size();i++) {
    // Calculate emission probability
    etaP[i]=BlackBody::B(((Cell*) cell)->nuBoost[i].min,((Cell*) cell)->nuBoost[i].max,T,1e-4)/B;
  }

  // Normalize emission probability (to be on the safe side)
  for (i=0;i<((Cell*)cell)->nuBoost.size();i++) PTot+=etaP[i];
  for (i=0;i<((Cell*)cell)->nuBoost.size();i++) etaP[i]=etaP[i]/PTot;

  // Initialize emissivity boost factor
  etaBoost=0;

  // Calculate emissivity boost factor
  for (i=0;i<((Cell*) cell)->boost.size();i++) etaBoost+=etaP[i]*((Cell*) cell)->boost[i];

  // Adjust emission probabilities
  for (i=0;i<etaP.size();i++) etaP[i]=etaP[i]*((Cell*) cell)->boost[i]/etaBoost;

  // Adjust number of packets
  nPacket=nPacket*etaBoost;
}

// Select random frequency - New!
// Does not work with containers in boost mode (see below)!
double CellPacketSamplerBBThermal::randomNu() {
  // Select version to use
  // Does not work woth containers in boost mode (due to randomNuBoost)!
  if (((Cell*) cell)->config->boost) return randomNuBoost();
  else return randomNuNormal();
}

// Select random frequency (intervall version) - New!
double CellPacketSamplerBBThermal::randomNu(double nuMin, double nuMax) {
  double z,B,nu;

  // Select a random BB frequency
  z=randomize(0,1);
  B=BlackBody::B(nuMin,nuMax,T,1e-4);
  nu=BlackBody::nuB(z*B,T,nuMin,1e-3);

  return nu;
}

// Select random frequency (normal verions) - New!
double CellPacketSamplerBBThermal::randomNuNormal() {
  double z,B,nu;
 
  // Select a random BB frequency
  z=randomize(0,1);
  B=BlackBody::B(nuLim.min,nuLim.max,T,1e-4);
  nu=BlackBody::nuB(z*B,T,nuLim.min,1e-3);

  return nu;
}

// Select random frequency (boost verions) - New!
// Works with containers, right?
double CellPacketSamplerBBThermal::randomNuBoost() {
  double z,PTot=0;
  int i;

  // Draw random number between 0 and 1
  z=randomize(0,1);

  // For each boost bin
  for (i=0;i<etaP.size();i++) {

    // If bin drawn - Return emission frequency
    if (z<=PTot+etaP[i]) return randomNu(((Cell*) cell)->nuBoost[i].min,((Cell*) cell)->nuBoost[i].max);
    // Else increase accumulated probability
    else PTot+=etaP[i];
  }

  // This should not happen!
  cout<<"PacketSamplerBBL::randomNuBoost: Error: No frequency drawn, z: "<<z<<", PTot: "<<PTot<<"\n";

  return 0.0;
}

// Constructor
// Sampler does not (exactly) conserve energy - EPacket needs to be adjusted - Rewrite !!!
PacketSamplerI::PacketSamplerI(const int& _nPacket, int cellIndexMin, int cellIndexMax, Grid* _grid, PacketConfig* _packetConfig, boost::mpi::communicator* _mpiCom, int _packetIndexBase, double _expFactorWeight, bool _LCorr, bool _eRadioCorr) : grid(_grid), packetConfig(_packetConfig), mpiCom(_mpiCom), expFactorWeight(_expFactorWeight), LCorr(_LCorr), eRadioCorr(_eRadioCorr) {

  int i,j;
  double expFactor,t,E=0;
  bool master=false;

  if (mpiCom==NULL) master=true;
  else if (mpiCom->rank()==0) master=true;

  // Prevent faulty cell range
  cellIndexMin=(cellIndexMin>=0)?cellIndexMin:0; 
  cellIndexMax=(cellIndexMax<=(grid->cell.size()-1))?cellIndexMax:(grid->cell.size()-1);

  // Initialize number of packets, packet index and packet index base
  nPacket=_nPacket;
  packetIndex=0;
  packetIndexBase=_packetIndexBase;

  // Get (centered) time
  t=grid->t();

  // Calculate total (radiative) energy
  //for (i=cellIndexMin;i<=cellIndexMax;i++) E+=(4*pi/c)*grid->cell[i]->J*grid->cell[i]->V();
  for (i=cellIndexMin;i<=cellIndexMax;i++) E+=ECell(i)*grid->cell[i]->V(); // Changed!

  // Caclulate the energy of the packets
  expFactor=(grid->tMin()+expFactorWeight*(grid->tMax()-grid->tMin()))/grid->tMin();
  EPacket=E*expFactor/nPacket;

  if (master) cout<<"PacketSamplerI::PacketSamplerI: EPacket: "<<EPacket<<", nPacket"<<nPacket<<", expFactor: "<<expFactor<<", E: "<<E<<", tMin: "<<grid->tMin()<<", tMax: "<<grid->tMax()<<", expFactorWeight: "<<expFactorWeight<<"\n";

  // Initialize cell packet sampler list
  cellPacketSampler.resize(0);

  // Reset number of packets
  nPacket=0;

  // For each cell
  for (i=cellIndexMin;i<=cellIndexMax;i++) {
    // If cell container
    if (grid->cell[i]->type()==cellContainer) {
      // For each sub cell
      for (j=0;j<grid->cell[i]->subCell.size();j++) {
	// Create cell packet sampler and append to list
	cellPacketSampler.push_back(CellPacketSamplerI(grid,grid->cell[i]->subCell[j],grid->cell[i],packetConfig,EPacket,ECell(i,j),dEdr(i,j),mpiCom,packetIndexBase+nPacket,expFactorWeight)); // Changed!
	// Add upp number of cell packets
	nPacket+=cellPacketSampler[cellPacketSampler.size()-1].nPacket;
      }
    }
    // Else
    else {
      // Create cell packet sampler and append to list
      cellPacketSampler.push_back(CellPacketSamplerI(grid,grid->cell[i],NULL,packetConfig,EPacket,ECell(i),dEdr(i),mpiCom,packetIndexBase+nPacket,expFactorWeight)); // changed!
      // Add upp number of packet cell packets
      nPacket+=cellPacketSampler[cellPacketSampler.size()-1].nPacket;
    }
  }

  if (master) cout<<"PacketSamplerI::PacketSamplerI: Initited for sampling of "<<nPacket<<" packets with energy "<<EPacket<<"\n";
}  

// Sample packet
Packet PacketSamplerI::sample(int _packetIndex) {
  int i,cellPacketIndex;
  double nPacketTot=0;
  // Hack for openmp
  int packetIndex=this->packetIndex;

  // If provided - Set packet index
  if (_packetIndex!=noIndex) packetIndex=_packetIndex;

  // For each cell packet sampler
  for (i=0;i<cellPacketSampler.size();i++) {

    // Calculate total number of packets
    nPacketTot+=cellPacketSampler[i].nPacket;

    // If larger than packet counter
    if (nPacketTot>packetIndex) {

      // Set cell packet index
      if (_packetIndex!=noIndex) cellPacketIndex=packetIndex-(nPacketTot-cellPacketSampler[i].nPacket);
      else cellPacketIndex=noIndex;

      // Increase packet counter
      packetIndex++;
      // Hack for openmp
      this->packetIndex=packetIndex;

      // Sample and return packet
      return cellPacketSampler[i].sample(cellPacketIndex);
    }
  }
}

// Reset
void PacketSamplerI::reset() {
  int i;
  PacketSampler::reset();
  for (i=0;i<cellPacketSampler.size();i++) cellPacketSampler[i].reset();
}

// Calculate (radiative) energy - New!
double PacketSamplerI::ECell(const int& i, int j) const {
  CellSS* cell;
  double E;
  bool master=false;

  if (mpiCom==NULL) master=true;
  else if (mpiCom->rank()==0) master=true;

  if (j==noCell) cell=((CellSS*) grid->cell[i]);
  else cell=((CellSS*) grid->cell[i]->subCell[j]);

  // Calculate (radiative) energy (per volome)
  E=(4*pi/c)*cell->J;
  // If requested - correct for luminosity at cell borders - New!
  if (LCorr) E-=MIN(expFactorWeight*(cell->LBack()-cell->LFront())*(cell->tMax()-cell->tMin())/cell->V(),E);
  // If requested - correct for radioactive deposition - New!
  if (eRadioCorr) E-=MIN(expFactorWeight*(cell->eRadio()*cell->rho())*(cell->tMax()-cell->tMin()),E);

  return E;
}

// Calculate (radiative) energy gradient
double PacketSamplerI::dEdr(const int& i, int j) const { // Changed!
  double dEBack,dEFront,drBack,drFront,dEdrBack,dEdrFront,dEdr;

  // If one-cell grid - return zero gradient 
  if (grid->cell.size()==1) return 0;

  if (i>0) {
    //dEBack=(4*pi/c)*(grid->cell[i]->J-grid->cell[i-1]->J);
    dEBack=(ECell(i)-ECell(i-1)); // Changed!
    drBack=(((CellSS*) grid->cell[i])->rMin()+((CellSS*) grid->cell[i])->rMax())/2-(((CellSS*) grid->cell[i-1])->rMin()+((CellSS*) grid->cell[i-1])->rMax())/2;
    dEdrBack=dEBack/drBack;
  }

  if (i<grid->cell.size()-1) {
    //dEFront=(4*pi/c)*(grid->cell[i+1]->J-grid->cell[i]->J);
    dEFront=(ECell(i+1)-ECell(i)); // Changed!
    drFront=(((CellSS*) grid->cell[i+1])->rMin()+((CellSS*) grid->cell[i+1])->rMax())/2-(((CellSS*) grid->cell[i])->rMin()+((CellSS*) grid->cell[i])->rMax())/2;
    dEdrFront=dEFront/drFront;
  }

  if (i==0) dEdr=dEdrFront;
  else if (i==(grid->cell.size()-1)) dEdr=dEdrBack;
  else dEdr=(dEdrBack+dEdrFront)/2;

  // if sub subcell index provided - Scale gradient to subcell energy
  //if (j!=noCell) dEdr=dEdr*(grid->cell[i]->subCell[j]->J/grid->cell[i]->J);
  if (j!=noCell) dEdr=dEdr*(ECell(i,j)/ECell(i)); // Changed!

  return dEdr;
}

// Constructor
CellPacketSamplerI::CellPacketSamplerI(Grid* _grid, Cell* _cell, Cell* _container, PacketConfig* _packetConfig, const double& _EPacket, const double& _E, const double& _dEdr, boost::mpi::communicator* _mpiCom, int _packetIndexBase, double expFactorWeight) : grid(_grid), cell(_cell), container(_container), packetConfig(_packetConfig), EPacket(_EPacket), E(_E), dEdr(_dEdr), mpiCom(_mpiCom) { // Changed!
  double expFactor,dt,dr,rCenter,rCenterSub,VSub,PSum=0;
  int nSub=10,i;
  bool master=false;

  if (mpiCom==NULL) master=true;
  else if (mpiCom->rank()==0) master=true;

  // Initialize packet index and packet index base
  packetIndex=0;
  packetIndexBase=_packetIndexBase,

  // Set time to the minimum time of the cell
  t=cell->tMin();

  // Calculate the number of packets
  expFactor=(cell->tMin()+expFactorWeight*(cell->tMax()-cell->tMin()))/cell->tMin();
  //nPacket=((4*pi/c)*cell->J*cell->V())*expFactor/EPacket;
  nPacket=(E*cell->V())*expFactor/EPacket; // Changed!

  // Get (radiative) energy (per volume)
  //E=(4*pi/c)*cell->J;

  // Calculate center
  rCenter=(((CellSS*) cell)->rMin()+((CellSS*) cell)->rMax())/2;

  // Calculate dr
  dr=(((CellSS*) cell)->rMax()-((CellSS*) cell)->rMin())/nSub;

  // Initialize subdivisions
  rMinSub.resize(nSub);
  rMaxSub.resize(nSub);
  PSub.resize(nSub);

  // For each subdivision
  for (i=0;i<nSub;i++) {
    // Calculate borders
    rMinSub[i]=((CellSS*) cell)->rMin()+i*dr;
    rMaxSub[i]=((CellSS*) cell)->rMin()+(i+1)*dr;
    // Calculate center
    rCenterSub=(rMinSub[i]+rMaxSub[i])/2;
    // Calculate volume
    VSub=(4.0/3.0)*pi*(pow(rMaxSub[i],3)-pow(rMinSub[i],3));
    // Calculate probability
    PSub[i]=(E+dEdr*(rCenterSub-rCenter))*VSub;
    // Sum probability
    PSum+=PSub[i];
  }

  // Normalize probability
  for (i=0;i<nSub;i++) PSub[i]=PSub[i]/PSum;
  
  if (master) cout<<"CellPacketSamplerI::CellPacketSamplerI: cellIndex: "<<cell->index()<<", nPacket: "<<nPacket<<", E: "<<nPacket*EPacket<<", T: "<<cell->T()<<"\n";
}

// Sample packet
Packet CellPacketSamplerI::sample(int _packetIndex) {
  double nu,theta,phi,rPosition,thetaPosition,phiPosition;
  Vector3D position,direction;
  PacketEventType packetEventType;
  CellBase* _cell; // New!
  // Hack for openmp
  int packetIndex=this->packetIndex;

  // If provided - Set packet index
  if (_packetIndex!=noIndex) packetIndex=_packetIndex;

  // Set packet event type
  packetEventType=evEmission;

  // Select a random position
  position=randomPosition();

  // Select a random frequency and polar angle (about the radius vector)
  randomNuAndTheta(nu,theta,cell->v(position,t));

  // Select a random equatorial angle (about the radius vector)
  phi=2*pi*randomize(0,1);

  // Calculate direction vector
  position.coords(rPosition,thetaPosition,phiPosition,spherical);
  direction=Vector3D(1,theta,phi,spherical);
  if (!isnumber(direction.norm())) { cout<<"CellPacketSamplerI::sample: Invalid direction 1: norm: "<<direction.norm()<<"\n"; exit(1); }
  direction.rotate(thetaPosition,phiPosition);
  if (!isnumber(direction.norm())) { cout<<"CellPacketSamplerI::sample: Invalid direction 2: norm: "<<direction.norm()<<"\n"; exit(1); }

  // Increase packet counter
  packetIndex++;
  // Hack for openmp
  this->packetIndex=packetIndex;

  // If subcell - Create virtual cell - New!
  if (cell->type()==cellSub) _cell = new CellVirtual(container,cell,position,direction,cellInterior,t);
  // Else - Just copy cell pointer - New!
  else _cell=cell;

  if (!isnumber(EPacket) or !isnumber(nu)) { cout<<"CellPacketSamplerI::sample: Invalid packet: index: "<<packetIndex<<", E: "<<EPacket<<", nu: "<<nu<<"\n"; exit(1); }

  // Create and return the packet
  return Packet(EPacket,FrequencyBin(nu,nu),position,direction,t,_cell,grid,packetIndexBase+packetIndex-1,rPacket,packetEventType,packetConfig); // Changed!
}

// Sample position of packet
Vector3D CellPacketSamplerI::randomPosition() const {
  double r,theta,phi,expFactor,z,PTot=0;
  int i;

  // Select random number
  z=randomize(0,1);

  // For each subdivision
  for (i=0;i<PSub.size();i++) {
    // Calculate cumulative probability
    PTot+=PSub[i];
    // If subdivision drawn - Break
    if (z<=PTot) break;
  }

  // Calculate expansion factor
  expFactor=t/((cell->tMin()+cell->tMax())/2);
  // Draw random radius (within subdivision) and scale with expansion factor
  r=pow(pow(rMinSub[i],3)+(pow(rMaxSub[i],3)-pow(rMinSub[i],3))*randomize(0,1),1.0/3.0)*expFactor;
  // Draw random polar angle
  theta=acos(2*randomize(0,1)-1);
  // draw random equatorial angle
  phi=2*pi*randomize(0,1);

  double PTot2=0;
  for (i=0;i<PSub.size();i++) PTot2+=PSub[i];

  return Vector3D(r,theta,phi,spherical);
}

// Get random frequency and polar angle
void CellPacketSamplerI::randomNuAndTheta(double& nu, double& theta, const double& v) const {
  double mu,muCMF,z,PTot,JTot,ITot;
  int i,j;

  // Select random number
  z=randomize(0,1);

  // Calculate frequency summed J (to be on the safe side)
  // Unclear if muliple initializers work - Fix!
  JTot=0; for (i=0;i<cell->nu.size();i++) JTot+=cell->Jnu[i]; // Changed!

  // For each frequency bin
  // Unclear if muliple initializers work - Fix!
  PTot=0; // New!
  for (i=0;i<cell->nu.size();i++) { // Changed!
    // Calculate cumulative probability
    PTot+=cell->Jnu[i]/JTot;
    if (cell->J==0) {cerr<<"CellPacketSamplerI:randomNuAndTheta: Error: Zero J, cell: "<<cell->index()<<"\n"; exit(1); }
    // If boost - Adjust probability
    //if (cell->config->boost) P=P*cell->boostFactor((cell->nu[i].min+cell->nu[i].max)/2)/boostFactor;
    // If frequency bin drawn - Break
    if (z<=PTot) break;
  }

  if (i==cell->nu.size()) {cerr<<"CellPacketSamplerI:randomNuAndTheta: Error: No frequency bin selected, cell: "<<cell->index()<<", z: "<<z<<", PTot: "<<PTot<<"\n"; exit(1); }

  // Calculate angle summed I (to be on the safe side)
  // Unclear if muliple initializers work - Fix!
  ITot=0; for (j=0;j<cell->mu.size();j++) ITot+=cell->Inu[i][j]; // Changed!

  // For each angle bin
  // Unclear if muliple initializers work - Fix!
  PTot=0; // New!
  for (j=0;j<cell->mu.size();j++) { // Changed!
    // Calculate cumulative probability
    PTot+=cell->Inu[i][j]/ITot;
    if (cell->Jnu[i]==0) {cerr<<"CellPacketSamplerI:randomNuAndTheta: Error: Zero Jnu, cell: "<<cell->index()<<", frequency bin: "<<i<<"\n"; exit(1); }
    // If frequency bin drawn - Break
    if (z<=PTot) break;
  }

  if (j==cell->mu.size()) {cerr<<"CellPacketSamplerI:randomNuAndTheta: Error: No angle bin selected, cell: "<<cell->index()<<", frequency bin: "<<i<<", z: "<<z<<", PTot: "<<PTot<<"\n"; exit(1); }

  // Select a random frequency in the bin
  nu=cell->nu[i].min+randomize(0,1)*(cell->nu[i].max-cell->nu[i].min);

  // Select a random (cosine of) polar ange in the bin
  muCMF=MAX(cell->mu[j].min,-1.0)+randomize(0,1)*(MIN(cell->mu[j].max,1.0)-MAX(cell->mu[j].min,-1.0));

  if (!isnumber(muCMF)) { cout<<"CellPacketSamplerI::randomNuAndTheta: Invalid muCMF: "<<muCMF<<"\n"; exit(1); }

  // Transform (cosine of) polar angle to the rest frame
  mu=(muCMF+(v/c))/(1+muCMF*(v/c));

  if (!isnumber(mu)) { cout<<"CellPacketSamplerI::randomNuAndTheta: Invalid mu: "<<mu<<"\n"; exit(1); }

  // Calculate polar angle
  theta=acos(mu);

  if (!isnumber(theta)) { cout<<"CellPacketSamplerI::randomNuAndTheta: Invalid theta: "<<theta<<", mu: "<<mu<<", muCMF: "<<muCMF<<"\n"; exit(1); }
}

PacketSamplerBBGamma::PacketSamplerBBGamma(const double& Rph, const double& Lph, const int& _nPacketBB, const int& _nPacketGamma, const int& nBin, const FrequencyBin& nuLim, const BBSampling& bbSampling, const BBFrame& bbFrame, const bool& bbLimbDark, Grid* grid, PacketConfig* packetConfig, int _packetIndexBase) : nPacketGamma(_nPacketGamma), nPacketBB(_nPacketBB) {

  // Initialise number of packets, packet index and packet index base
  nPacket=nPacketGamma+nPacketBB;
  packetIndex=0;
  packetIndexBase=_packetIndexBase;

  // Initialize gamma-ray/lepton packet samples
  packetSamplerGamma=PacketSamplerGamma(nPacketGamma,grid,packetConfig,NULL,packetIndexBase);

  // Initialize blackbody packet sampler
  packetSamplerBB=PacketSamplerBB(Rph,Lph,nPacketBB,nBin,nuLim,bbSampling,bbFrame,bbLimbDark,grid,packetConfig,packetIndexBase+nPacketGamma);

  // Set frequency bin
  nu=packetSamplerBB.nu;

  // Recalculate numer of packets
  nPacketGamma=packetSamplerGamma.nPacket;
  nPacketBB=packetSamplerBB.nPacket;
  nPacket=nPacketGamma+nPacketBB;
}

Packet PacketSamplerBBGamma::sample(int _packetIndex) {
  int gammaPacketIndex,bbPacketIndex;
  // Hack for openmp
  int packetIndex=this->packetIndex;

  // If provided - Set packet index
  if (_packetIndex!=noIndex) packetIndex=_packetIndex;

  // If gamma packet
  if (packetIndex<nPacketGamma) {

    // Set gamma sampler packet index
    if (_packetIndex!=noIndex) gammaPacketIndex=packetIndex;
    else gammaPacketIndex=noIndex;

    // Increase packet counter
    packetIndex++;
    // Hack for openmp
    this->packetIndex=packetIndex;

    // Sample and return packet
    return packetSamplerGamma.sample(gammaPacketIndex);
  }
  else {

    // Set BB sampler packet index
    if (_packetIndex!=noIndex) bbPacketIndex=packetIndex-nPacketGamma;
    else bbPacketIndex=noIndex;

    // Increase packet counter
    packetIndex++;
    // Hack for openmp
    this->packetIndex=packetIndex;

    // Sample and return packet
    return packetSamplerBB.sample(bbPacketIndex);
  }
}

PacketSamplerBBLGamma::PacketSamplerBBLGamma(const double& L, const int& _nPacketBB, const int& _nPacketGamma, const FrequencyBin& nuLim, const int& cellIndex, Grid* grid, PacketConfig* packetConfig, boost::mpi::communicator* _mpiCom, int _packetIndexBase) : nPacketGamma(_nPacketGamma), nPacketBB(_nPacketBB), mpiCom(_mpiCom) {

  // Initialise number of packets, packet index and packet index base
  nPacket=nPacketGamma+nPacketBB;
  packetIndex=0;
  packetIndexBase=_packetIndexBase;

  // Initialize gamma-ray/lepton packet samples
  packetSamplerGamma=PacketSamplerGamma(nPacketGamma,grid,packetConfig,mpiCom,packetIndexBase);

  // Initialize blackbody packet sampler
  packetSamplerBB=PacketSamplerBBL(L,nPacketBB,cellIndex,grid,nuLim,packetConfig,mpiCom,packetIndexBase+nPacketGamma);

  // Recalculate numer of packets
  nPacketGamma=packetSamplerGamma.nPacket;
  nPacketBB=packetSamplerBB.nPacket;
  nPacket=nPacketGamma+nPacketBB;
}

Packet PacketSamplerBBLGamma::sample(int _packetIndex) {
  int gammaPacketIndex,bbPacketIndex;
  // Hack for openmp
  int packetIndex=this->packetIndex;

  // If provided - Set packet index
  if (_packetIndex!=noIndex) packetIndex=_packetIndex;

  // If gamma packet
  if (packetIndex<nPacketGamma) {

    // Set gamma sampler packet index
    if (_packetIndex!=noIndex) gammaPacketIndex=packetIndex;
    else gammaPacketIndex=noIndex;

    // Increase packet counter
    packetIndex++;
    // Hack for openmp
    this->packetIndex=packetIndex;

    // Sample and return packet
    return packetSamplerGamma.sample(gammaPacketIndex);
  }
  else {

    // Set BBL sampler packet index
    if (_packetIndex!=noIndex) bbPacketIndex=packetIndex-nPacketGamma;
    else bbPacketIndex=noIndex;

    // Increase packet counter
    packetIndex++;
    // Hack for openmp
    this->packetIndex=packetIndex;

    // Sample and return packet
    return packetSamplerBB.sample(bbPacketIndex);
  }
}

PacketEventStat::PacketEventStat() : nEScatter(0), nLScatter(0), nGScatter(0), nLAbs(0), nLEmission(0), nFFAbs(0), nFFEmission(0), nBFAbs(0), nBFEmission(0), n2PEmission(0), nCCrossing(0), nCICrossing(0), nCOCrossing(0), nTCrossing(0), nReflect(0), nEmission(0), ECEmission(0), nNone(0), ECNone(0), nBCrossing(0), nComptonAbs(0), nRadioEmission(0), n56NiGamma(0), n56CoGamma(0), n56CoPos(0), n57NiGamma(0), n57CoGamma(0), n57CoPos(0), n44TiGamma(0), n44TiPos(0), nDummyEmission(0) {} // Changed!

// Update packet event statistics
void PacketEventStat::update(const PacketEvent& event) {

  // Remove pragma omp atomic - Testing!!!
  switch(event.type) {
  case evElectronScatter :
    #pragma omp atomic
    nEScatter++;
    break;
  case evLineScatter :
    #pragma omp atomic
    nLScatter++;
    break;
  case evGreyScatter : // New!
    #pragma omp atomic
    nGScatter++;
    break;
  case evLineAbs :
    #pragma omp atomic
    nLAbs++;
    break;
  case evLineEmission :
    #pragma omp atomic
    nLEmission++;
    break;
  case evFreeFreeAbs :
    #pragma omp atomic
    nFFAbs++;
    break;
  case evFreeFreeEmission :
    #pragma omp atomic
    nFFEmission++;
    break;
  case evBoundFreeAbs :
    #pragma omp atomic
    nBFAbs++;
    break;
  case evBoundFreeEmission :
    #pragma omp atomic
    nBFEmission++;
    break;
  case evTwoPhotonEmission : // New!
    #pragma omp atomic
    n2PEmission++;
    break;
  case evCellCrossingInner : // New!
    #pragma omp atomic
    nCICrossing++;
    #pragma omp atomic
    nCCrossing++;
    break;
  case evCellCrossingOuter : // New!
    #pragma omp atomic
    nCOCrossing++;
    #pragma omp atomic
    nCCrossing++;
    break;
  case evCellCrossing :
    #pragma omp atomic
    nCCrossing++;
    break;
  case evTimeBorderCrossing : // New!
    #pragma omp atomic
    nTCrossing++;
    break;
  case evReflect :
    #pragma omp atomic
    nReflect++;
    break;
  case evEmission :
    // Testing
    #pragma omp atomic
    nEmission++;
    #pragma omp atomic
    ECEmission+=event.ECMF;
    break;
  case evNone :
    // Testing
    #pragma omp atomic
    nNone++;
    #pragma omp atomic
    ECNone+=event.ECMF;
    break;
  case evBoundaryCrossing :
    #pragma omp atomic
    nBCrossing++;
    break;
  case evComptonAbs :
    #pragma omp atomic
    nComptonAbs++;
    break;
  case ev56NiGammaDecay :
    #pragma omp atomic
    n56NiGamma++;
    #pragma omp atomic
    nRadioEmission++;
    break;
  case ev56CoGammaDecay :
    #pragma omp atomic
    n56CoGamma++;
    #pragma omp atomic
    nRadioEmission++;
    break;
  case ev56CoPosDecay :
    #pragma omp atomic
    n56CoPos++;
    #pragma omp atomic
    nRadioEmission++;
    break;
  case ev57NiGammaDecay :
    #pragma omp atomic
    n57NiGamma++;
    #pragma omp atomic
    nRadioEmission++;
    break;
  case ev57CoGammaDecay :
    #pragma omp atomic
    n57CoGamma++;
    #pragma omp atomic
    nRadioEmission++;
    break;
  case ev57CoPosDecay :
    #pragma omp atomic
    n57CoPos++;
    #pragma omp atomic
    nRadioEmission++;
    break;
  case ev44TiGammaDecay :
    #pragma omp atomic
    n44TiGamma++;
    #pragma omp atomic
    nRadioEmission++;
    break;
  case ev44TiPosDecay :
    #pragma omp atomic
    n44TiPos++;
    #pragma omp atomic
    nRadioEmission++;
    break;
  case evDummyEmission : // New!
    #pragma omp atomic
    nDummyEmission++;
    break;
  default :
    break;
  }
}

// Export packet event statistics
void PacketEventStat::exportStat(YAML::Emitter& data) {

  // Begin packet event statistics map
  data<<YAML::Flow<<YAML::BeginMap;

  // Export number of electron scatterings
  data<<YAML::Key<<"nEScatter"<<YAML::Value<<nEScatter;

  // Export number of line scatterings
  data<<YAML::Key<<"nLScatter"<<YAML::Value<<nLScatter;

  // Export number of grey scatterings - New!
  data<<YAML::Key<<"nGScatter"<<YAML::Value<<nGScatter;

  // Export number of line absorptions
  data<<YAML::Key<<"nLAbs"<<YAML::Value<<nLAbs;

  // Export number of line emissions
  data<<YAML::Key<<"nLEmission"<<YAML::Value<<nLEmission;

  // Export number of free-free absorptions
  data<<YAML::Key<<"nFFAbs"<<YAML::Value<<nFFAbs;

  // Export number of free-free emissions
  data<<YAML::Key<<"nFFEmission"<<YAML::Value<<nFFEmission;

  // Export number of bound-free absorptions
  data<<YAML::Key<<"nBFAbs"<<YAML::Value<<nBFAbs;

  // Export number of bound-free emissions
  data<<YAML::Key<<"nBFEmission"<<YAML::Value<<nBFEmission;

  // Export number of two-photon emissions - New!
  data<<YAML::Key<<"n2PEmission"<<YAML::Value<<n2PEmission;

  // Export number of cell border crossings
  data<<YAML::Key<<"nCCrossing"<<YAML::Value<<nCCrossing;

  // Export number of cell inner border crossings - New!
  data<<YAML::Key<<"nCICrossing"<<YAML::Value<<nCICrossing;

  // Export number of cell outer border crossings - New!
  data<<YAML::Key<<"nCOCrossing"<<YAML::Value<<nCOCrossing;

  // Export number of time border crossings - New!
  data<<YAML::Key<<"nTCrossing"<<YAML::Value<<nTCrossing;

  // Export number of inner border reflections
  data<<YAML::Key<<"nReflect"<<YAML::Value<<nReflect;

  // Testing
  data<<YAML::Key<<"nEmission"<<YAML::Value<<nEmission;
  data<<YAML::Key<<"ECEmission"<<YAML::Value<<ECEmission;
  data<<YAML::Key<<"nNone"<<YAML::Value<<nNone;
  data<<YAML::Key<<"ECNone"<<YAML::Value<<ECNone;
  data<<YAML::Key<<"nBCrossing"<<YAML::Value<<nBCrossing;
  data<<YAML::Key<<"nComptonAbs"<<YAML::Value<<nComptonAbs;
  data<<YAML::Key<<"nRadioEmission"<<YAML::Value<<nRadioEmission;
  data<<YAML::Key<<"n56NiGamma"<<YAML::Value<<n56NiGamma;
  data<<YAML::Key<<"n56CoGamma"<<YAML::Value<<n56CoGamma;
  data<<YAML::Key<<"n56CoPos"<<YAML::Value<<n56CoPos;
  data<<YAML::Key<<"n57NiGamma"<<YAML::Value<<n57NiGamma;
  data<<YAML::Key<<"n57CoGamma"<<YAML::Value<<n57CoGamma;
  data<<YAML::Key<<"n57CoPos"<<YAML::Value<<n57CoPos;
  data<<YAML::Key<<"n44TiGamma"<<YAML::Value<<n44TiGamma;
  data<<YAML::Key<<"n44TiPos"<<YAML::Value<<n44TiPos;

  // Export number of dummy emissions - New!
  data<<YAML::Key<<"nDummyEmission"<<YAML::Value<<nDummyEmission;

  // Close packet event statistics map
  data<<YAML::EndMap;
}

// Sum statistics
void PacketEventStat::operator+=(const PacketEventStat& eventStat) {
  nEScatter+=eventStat.nEScatter;
  nLScatter+=eventStat.nLScatter;
  nGScatter+=eventStat.nGScatter; // New!
  nLAbs+=eventStat.nLAbs;
  nLEmission+=eventStat.nLEmission;
  nFFAbs+=eventStat.nFFAbs;
  nFFEmission+=eventStat.nFFEmission;
  nBFAbs+=eventStat.nBFAbs;
  nBFEmission+=eventStat.nBFEmission;
  n2PEmission+=eventStat.n2PEmission; // New!
  nCCrossing+=eventStat.nCCrossing;
  nCICrossing+=eventStat.nCICrossing; // New!
  nCOCrossing+=eventStat.nCOCrossing; // New!
  nTCrossing+=eventStat.nTCrossing; // New!
  nReflect+=eventStat.nReflect;
  nEmission+=eventStat.nEmission;
  ECEmission+=eventStat.ECEmission;
  nNone+=eventStat.nNone;
  ECNone+=eventStat.ECNone;
  nBCrossing+=eventStat.nBCrossing;
  nComptonAbs+=eventStat.nComptonAbs;
  nRadioEmission+=eventStat.nRadioEmission;
  n56NiGamma+=eventStat.n56NiGamma;
  n56CoGamma+=eventStat.n56CoGamma;
  n56CoPos+=eventStat.n56CoPos;
  n57NiGamma+=eventStat.n57NiGamma;
  n57CoGamma+=eventStat.n57CoGamma;
  n57CoPos+=eventStat.n57CoPos;  
  n44TiGamma+=eventStat.n44TiGamma;
  n44TiPos+=eventStat.n44TiPos;
  nDummyEmission+=eventStat.nDummyEmission; // New!
}

// Reset packet event statistics
void PacketEventStat::reset() {
  nEScatter=0;
  nLScatter=0;
  nGScatter=0;
  nLAbs=0;
  nLEmission=0;
  nFFAbs=0;
  nFFEmission=0;
  nBFAbs=0;
  nBFEmission=0;
  n2PEmission=0; // New!
  nCCrossing=0;
  nCICrossing=0; // New!
  nCOCrossing=0; // New!
  nTCrossing=0; // New!
  nReflect=0;
  nEmission=0;
  ECEmission=0;
  nNone=0;
  ECNone=0;
  nBCrossing=0;
  nComptonAbs=0;
  nRadioEmission=0;
  n56NiGamma=0;
  n56CoGamma=0;
  n56CoPos=0;
  n57NiGamma=0;
  n57CoGamma=0;
  n57CoPos=0;
  n44TiGamma=0;
  n44TiPos=0;
  nDummyEmission=0; // New!
}

// Constructors - New!
Spectrum::Spectrum(Grid* _grid, const FrequencyBinList& _nu, const RefFrame& _refFrame, const double& _tMin, const double& _tMax) : grid(_grid), cellIndexMin(0), cellIndexMax(_grid->cell.size()-1), nu(_nu), refFrame(_refFrame), tMin(_tMin), tMax(_tMax) { 
  // Testing!
  //cout<<"Spectrum::Spectrum: tMin: "<<tMin<<", tMax: "<<tMax<<", refFrame: "<<refFrame<<"\n"; 
  reset(); 
}

Spectrum::Spectrum(Grid* _grid, const int& _cellIndexMin, const int& _cellIndexMax, const FrequencyBinList& _nu, const RefFrame& _refFrame, const double& _tMin, const double& _tMax) : grid(_grid), cellIndexMin(_cellIndexMin), cellIndexMax(_cellIndexMax), nu(_nu), refFrame(_refFrame), tMin(_tMin), tMax(_tMax) { reset(); }

// Reset - New!
void Spectrum::reset() {
  int i,j;

  pL=0;
  Llambda.resize(0);
  Llambda.resize(nu.size(),0);
  LlambdaFreeFree.resize(0);
  LlambdaFreeFree.resize(nu.size(),0);
  LlambdaBoundFree.resize(grid->cell[cellIndexMax]->atom.size());
  for (i=0;i<grid->cell[cellIndexMax]->atom.size();i++) {
    LlambdaBoundFree[i].resize(grid->cell[cellIndexMax]->atom[i].iState.size());
    for (j=0;j<grid->cell[cellIndexMax]->atom[i].iState.size();j++) {
      LlambdaBoundFree[i][j].resize(0);
      LlambdaBoundFree[i][j].resize(nu.size(),0);
    }
  }
  LlambdaBoundBound.resize(grid->cell[cellIndexMax]->atom.size());
  for (i=0;i<grid->cell[cellIndexMax]->atom.size();i++) {
    LlambdaBoundBound[i].resize(grid->cell[cellIndexMax]->atom[i].iState.size());
    for (j=0;j<grid->cell[cellIndexMax]->atom[i].iState.size();j++) {
      LlambdaBoundBound[i][j].resize(0);
      LlambdaBoundBound[i][j].resize(nu.size(),0);
    }
  }
  LlambdaCell.resize(0);
  LlambdaCell.resize(cellIndexMax-cellIndexMin+1);
  for (i=0;i<=(cellIndexMax-cellIndexMin);i++) LlambdaCell[i].resize(nu.size(),0);
  LlambdaSubCell.resize(cellIndexMax-cellIndexMin+1);
  for (i=0;i<=(cellIndexMax-cellIndexMin);i++) {
    if (grid->cell[cellIndexMin+i]->type()==cellContainer) {
      LlambdaSubCell[i].resize(grid->cell[cellIndexMin+i]->subCell.size());
      for (j=0;j<grid->cell[cellIndexMin+i]->subCell.size();j++) {
	LlambdaSubCell[i][j].resize(0);
	LlambdaSubCell[i][j].resize(nu.size(),0);
      }
    }
    else LlambdaSubCell[i].resize(0);
  }
}

// Reset - New!
void Spectrum::reset(Grid* _grid, const int& _cellIndexMin, const int& _cellIndexMax, const FrequencyBinList& _nu, const RefFrame& _refFrame, const double& _tMin, const double& _tMax) {

  grid=_grid;
  cellIndexMin=_cellIndexMin;
  cellIndexMax=_cellIndexMax;
  nu=_nu;
  refFrame=_refFrame;
  tMin=_tMin;
  tMax=_tMax;

  reset();
}

// Synchronize (MPI) - New!
void Spectrum::sync(boost::mpi::communicator* mpiCom) {
  int nRequest=(mpiCom->rank()==0)?(mpiCom->size()-1):1;
  boost::mpi::request request[nRequest];
  vector<Spectrum> collected(mpiCom->size()-1);
  int i;

  // If slave - Deliver spectrum to master (non-blocking)
  if (mpiCom->rank()!=0) request[0]=mpiCom->isend(0,mSpectrum,*this);

  // If master - Collect MC statistics from slaves (non-blocking)
  if (mpiCom->rank()==0) for (i=1;i<mpiCom->size();i++) request[i-1]=mpiCom->irecv(i,mSpectrum,collected[i-1]);

  // Wait for all MPI requests to complete
  boost::mpi::wait_all(request,request+nRequest);
 
  // Sum spectrum
  if (mpiCom->rank()==0) for (i=0;i<collected.size();i++) *this+=collected[i];

  // Broadcast packet statistics
  boost::mpi::broadcast(*mpiCom,*this,0);
}

// Sum - New!
void Spectrum::operator+=(const Spectrum& spectrum) {
  int i,j,k;

  // Sum luminosity
  pL+=spectrum.pL;

  // Sum spectrum
  for (i=0;i<Llambda.size();i++) Llambda[i]+=spectrum.Llambda[i];

  // Sum free-free spectrum
  for (i=0;i<LlambdaFreeFree.size();i++) LlambdaFreeFree[i]+=spectrum.LlambdaFreeFree[i];

  // Sum bound-free spectrum
  for (i=0;i<LlambdaBoundFree.size();i++) 
    for (j=0;j<LlambdaBoundFree[i].size();j++)
      for (k=0;k<LlambdaBoundFree[i][j].size();k++) LlambdaBoundFree[i][j][k]+=spectrum.LlambdaBoundFree[i][j][k];

  // Sum bound-bound spectrum
  for (i=0;i<LlambdaBoundBound.size();i++) 
    for (j=0;j<LlambdaBoundBound[i].size();j++)
      for (k=0;k<LlambdaBoundBound[i][j].size();k++) LlambdaBoundBound[i][j][k]+=spectrum.LlambdaBoundBound[i][j][k];

  // Sum cell spectrum
  for (i=0;i<LlambdaCell.size();i++) 
    for (j=0;j<LlambdaCell[i].size();j++) LlambdaCell[i][j]+=spectrum.LlambdaCell[i][j];

  // Sum subcell spectrum
  for (i=0;i<LlambdaSubCell.size();i++) 
    for (j=0;j<LlambdaSubCell[i].size();j++) 
      for (k=0;k<LlambdaSubCell[i][j].size();k++) LlambdaSubCell[i][j][k]+=spectrum.LlambdaSubCell[i][j][k];
}

// Update - New!
void Spectrum::update(const Packet& packet) {

  // Update luminosity
  updateL(packet);

  // Update spectrum
  updateSpecTotal(packet);

  // Update spectrum per transition type
  updateSpecTrans(packet);

  // Update spectrum per cell
  updateSpecCell(packet);
}

void Spectrum::updateL(const Packet& packet) {
  double dt,t;

  // Exclude and return on invalid data
  if (!packet.valid()) { cout<<"Spectrum::updateL: Warning: Invalid data, packet excluded\n"; return; }

  // Calculate time in the reference frame
  if (refFrame==rfDistantObserver) t=packet.t-(packet.position*packet.direction)/c;
  else t=packet.t;

  // Calculate size of temporal bin
  dt=tMax-tMin;

  if ((packet.position.norm()>=((CellSS*) grid->cell[cellIndexMax])->rMaxT(packet.t)) && (t>=tMin && t<=tMax) && packet.type!=gPacket) {
    #pragma omp atomic
    pL+=packet.E/dt;
  }
}

// Update observed spectrum - New!
void Spectrum::updateSpecTotal(const Packet& packet) {
  double dLambda,dt,t;
  FrequencyBinIterator nuBin;

  // Exclude and return on invalid data
  if (!packet.valid()) { cout<<"PacketStat::updateSpecTotal: Warning: Invalid data, packet excluded\n"; return; }

  // Calculate time in the reference frame
  if (refFrame==rfDistantObserver) t=packet.t-(packet.position*packet.direction)/c;
  else t=packet.t;

  // Calculate size of temporal bin
  dt=tMax-tMin;

  // If outside outermost cell, within temporal bin and not g-packet
  if ((packet.position.norm()>=((CellSS*) grid->cell[cellIndexMax])->rMaxT(packet.t)) && (t>=tMin && t<=tMax) && packet.type!=gPacket) {

    // Find frequency bin
    nuBin=nu.find(packet.nu);
    // If matching bin found
    if (nuBin!=nu.end()) {

      // Calculate size of wavelength bin
      dLambda=(c/nuBin->min-c/nuBin->max)/AA;

      // Sum luminosity per wavelength
      #pragma omp atomic
      Llambda[nuBin-nu.begin()]+=packet.E/(dLambda*dt);
    }
    // If not
    else {
      // How to handle this!
      cout<<"PacketStat::updateSpecTotal: Packet "<<packet.index<<" with frequency "<<packet.nu<<" outside frequency bin list\n";
    }
  }
}

// Update observed spectrum per transition type - New!
void Spectrum::updateSpecTrans(const Packet& packet) {
  double dLambda,dt,t;
  FrequencyBinIterator nuBin;
  //Cell* cell; // New!
  int i,j;

  // Exclude and return on invalid data
  //if (!packet.valid()) { cout<<"PacketStat::updateSpecTrans: Warning: Invalid data, packet excluded\n"; return; }

  // Calculate time in the reference frame
  if (refFrame==rfDistantObserver) t=packet.t-(packet.position*packet.direction)/c;
  else t=packet.t;

  // Calculate size of temporal bin
  dt=tMax-tMin;

  // If outside outermost cell, within temporal bin and not g-packet
  if ((packet.position.norm()>=((CellSS*) grid->cell[cellIndexMax])->rMaxT(packet.t)) && (t>=tMin && t<=tMax) && packet.type!=gPacket) {

    // Find frequency bin
    nuBin=nu.find(packet.nu);
    // If matching bin found
    if (nuBin!=nu.end()) {

      // Calculate size of wavelength bin
      dLambda=(c/nuBin->min-c/nuBin->max)/AA;

      // If last emission event is free-free
      if (packet.lastEmissionEvent.type==evFreeFreeEmission) {
	// Sum luminosity per wavelength
        #pragma omp atomic
        LlambdaFreeFree[nuBin-nu.begin()]+=packet.E/(dLambda*dt);
      }

      // If last emission event is bound-free or bound-bound
      if (packet.lastEmissionEvent.type==evBoundFreeEmission or packet.lastEmissionEvent.type==evLineEmission) {

	// Get outermost cell - New!
	//cell=grid->cell[cellIndexMax];
	// If container - Get first subcell - New!
	//if (cell->type()==cellContainer) cell=cell->subCell[0];

	// For each atom - Changed!
	// Note, the atoms included may vary between cells, so this needs to be improved!!!
	//for (i=0;i<cell->atom.size();i++) {
	for (i=0;i<grid->cell[cellIndexMax]->atom.size();i++) {
	  // If match
	  //if (cell->atom[i].Z()==packet.lastEmissionEvent.Z) {
	  if (grid->cell[cellIndexMax]->atom[i].Z()==packet.lastEmissionEvent.Z) {
	    // For each ion
	    // Note, the ion included may vary between cells, so this needs to be improved!!!
	    //for (j=0;j<cell->atom[i].iState.size();j++) {
	    for (j=0;j<grid->cell[cellIndexMax]->atom[i].iState.size();j++) {
	      // If match
	      //if (cell->atom[i].iState[j].C()==packet.lastEmissionEvent.C) {
	      if (grid->cell[cellIndexMax]->atom[i].iState[j].C()==packet.lastEmissionEvent.C) {
		// If last emission event is bound-free
		if (packet.lastEmissionEvent.type==evBoundFreeEmission) {
		  // Sum luminosity per wavelength
                  #pragma omp atomic
		  LlambdaBoundFree[i][j][nuBin-nu.begin()]+=packet.E/(dLambda*dt);
		}
		// Else (i.e. last emission event is bound-bound)
		else {
		  // Sum luminosity per wavelength
                  #pragma omp atomic
		  LlambdaBoundBound[i][j][nuBin-nu.begin()]+=packet.E/(dLambda*dt);
		}
	      }
	    }
	  }
	}
      }
    }
    // If not
    else {
      // How to handle this!
      //cout<<"PacketStat::updateSpecTrans: Packet "<<packet.index<<" with frequency "<<packet.nu<<" outside frequency bin list\n";
    }
  }
}

// Update observed spectrum per cell - New!
void Spectrum::updateSpecCell(const Packet& packet) {
  double dLambda,dt,t;
  FrequencyBinIterator nuBin;
  int i,j;

  // Exclude and return on invalid data
  //if (!packet.valid()) { cout<<"PacketStat::updateSpecCell: Warning: Invalid data, packet excluded\n"; return; }

  // Calculate time in the reference frame
  if (refFrame==rfDistantObserver) t=packet.t-(packet.position*packet.direction)/c;
  else t=packet.t;

  // Calculate size of temporal bin
  dt=tMax-tMin;

  // If outside outermost cell, within temporal bin and not g-packet
  if ((packet.position.norm()>=((CellSS*) grid->cell[cellIndexMax])->rMaxT(packet.t)) && (t>=tMin && t<=tMax) && packet.type!=gPacket) {

    // Find frequency bin
    nuBin=nu.find(packet.nu);
    // If matching bin found
    if (nuBin!=nu.end()) {

      // Calculate size of wavelength bin
      dLambda=(c/nuBin->min-c/nuBin->max)/AA;

      // For each cell
      for (i=0;i<=(cellIndexMax-cellIndexMin);i++) {
	// If match
	if (grid->cell[cellIndexMin+i]->index()==packet.lastEmissionEvent.cellIndex) {
	  // Sum luminosity per wavelength
          #pragma omp atomic
	  LlambdaCell[i][nuBin-nu.begin()]+=packet.E/(dLambda*dt);
	  // If container
	  if (grid->cell[cellIndexMin+i]->type()==cellContainer) {
	    // For each subcell
	    for (j=0;j<grid->cell[cellIndexMin+i]->subCell.size();j++) {
	      // If match
	      if (grid->cell[cellIndexMin+i]->subCell[j]->index()==packet.lastEmissionEvent.subCellIndex) {
		// Sum luminosity per wavelength
                #pragma omp atomic
		LlambdaSubCell[i][j][nuBin-nu.begin()]+=packet.E/(dLambda*dt);
	      }
	    }
	  }
	}
      }
    }
    // If not
    else {
      // How to handle this!
      //cout<<"PacketStat::updateSpecCell: Packet "<<packet.index<<" with frequency "<<packet.nu<<" outside frequency bin list\n";
    }
  }
}

// Get observed luminosity
double Spectrum::L() {
  return pL;
}

// Export spectrum
void Spectrum::exportSpecTotal(const string& dataFile) {
  ofstream file;
  double lambda;
  int i,j;

  // Open atom data file
  file.open(dataFile.c_str());
  // Error handling
  if (file.fail()) {cerr<<"Error\n"; return;}

  // Write spectrum to file
  for (i=nu.size()-1;i>=0;i--) {
    lambda=((c/nu[i].min+c/nu[i].max)/2)/AA;
    file<<lambda<<" "<<Llambda[i]<<"\n";
  }

  // Close file
  file.close();
  return;
}

// Export spectrum per transition type
void Spectrum::exportSpecTrans(const string& dataFile) {
  ofstream file;
  double lambda;
  //Cell* cell; // New!
  int i,j,k;

  // Open atom data file
  file.open(dataFile.c_str());
  // Error handling
  if (file.fail()) {cerr<<"Error\n"; return;}

  // Get outermost cell - New!
  //cell=grid->cell[cellIndexMax];
  // If container - Get first subcell - New!
  //if (cell->type()==cellContainer) cell=cell->subCell[0];

  // Write header - Changed!
  file<<"#Lambda FF";
  /*for (j=0;j<cell->atom.size();j++)
    for (k=0;k<cell->atom[j].iState.size();k++) {
      file<<" "<<cell->atom[j].iState[k].name()<<"(BF)";
      file<<" "<<cell->atom[j].iState[k].name()<<"(BB)";*/
  for (j=0;j<grid->cell[cellIndexMax]->atom.size();j++)
    for (k=0;k<grid->cell[cellIndexMax]->atom[j].iState.size();k++) {
      file<<" "<<grid->cell[cellIndexMax]->atom[j].iState[k].name()<<"(BF)";
      file<<" "<<grid->cell[cellIndexMax]->atom[j].iState[k].name()<<"(BB)";
    }
  file<<"\n";

  // Write spectrum per transition type to file - Changed!
  for (i=nu.size()-1;i>=0;i--) {
    lambda=((c/nu[i].min+c/nu[i].max)/2)/AA;
    file<<lambda<<" "<<LlambdaFreeFree[i];
    /*for (j=0;j<cell->atom.size();j++)
      for (k=0;k<cell->atom[j].iState.size();k++) {*/
    for (j=0;j<grid->cell[cellIndexMax]->atom.size();j++)
      for (k=0;k<grid->cell[cellIndexMax]->atom[j].iState.size();k++) {
	file<<" "<<LlambdaBoundFree[j][k][i]<<" "<<LlambdaBoundBound[j][k][i];
      }
    file<<"\n";
  }

  // Close file
  file.close();
  return;
}

// Export spectrum per cell - New!
void Spectrum::exportSpecCell(const string& dataFile) {
  ofstream file;
  double lambda;
  int i,j,k;

  // Open atom data file
  file.open(dataFile.c_str());
  // Error handling
  if (file.fail()) {cerr<<"Error\n"; return;}

  // Write header
  file<<"#Lambda Cell:";
  for (i=cellIndexMin;i<=cellIndexMax;i++) {
    file<<" "<<grid->cell[cellIndexMin+i]->index();
    if (grid->cell[cellIndexMin+i]->type()==cellContainer) {
      file<<"SubCell:";
      for (j=0;j<grid->cell[cellIndexMin+i]->subCell.size();j++) file<<" "<<grid->cell[cellIndexMin+i]->subCell[j]->index();
    }
  }
  file<<"\n";

  // Write spectrum per cell to file
  for (i=nu.size()-1;i>=0;i--) {
    lambda=((c/nu[i].min+c/nu[i].max)/2)/AA;
    file<<lambda;
    for (j=0;j<=(cellIndexMax-cellIndexMin);j++) {
      file<<" "<<LlambdaCell[j][i];
      if (grid->cell[cellIndexMin+j]->type()==cellContainer)
	for (k=0;k<grid->cell[cellIndexMin+j]->subCell.size();k++) file<<" "<<LlambdaSubCell[j][k][i];
    }
    file<<"\n";
  }

  // Close file
  file.close();
  return;
}

// Get memory allocated - New!
double Spectrum::mem() const {
  double _mem=sizeof(*this);
  int i,j,k;

  // Add memory allocated for frequency-bin list
  for (i=0;i<nu.size();i++) _mem+=sizeof(FrequencyBin);

  // Add memory allocated for spectrum
  for (i=0;i<Llambda.size();i++) _mem+=sizeof(double);

  // Add memory allocated for free-free spectrum
  for (i=0;i<LlambdaFreeFree.size();i++) _mem+=sizeof(double);

  // Add memory allocated for bound-free spectrum
  for (i=0;i<LlambdaBoundFree.size();i++) 
    for (j=0;j<LlambdaBoundFree[i].size();j++)
      for (k=0;k<LlambdaBoundFree[i][j].size();k++) _mem+=sizeof(double);

  // Add memory allocated for bound-bound spectrum
  for (i=0;i<LlambdaBoundBound.size();i++) 
    for (j=0;j<LlambdaBoundBound[i].size();j++)
      for (k=0;k<LlambdaBoundBound[i][j].size();k++) _mem+=sizeof(double);

  // Add memory allocated for cell spectrum
  for (i=0;i<LlambdaCell.size();i++) 
    for (j=0;j<LlambdaCell[i].size();j++) _mem+=sizeof(double);

  // Add memory allocated for subcell spectrum
  for (i=0;i<LlambdaSubCell.size();i++) 
    for (j=0;j<LlambdaSubCell[i].size();j++) 
      for (k=0;k<LlambdaSubCell[i][j].size();k++) _mem+=sizeof(double);

  return _mem;
}

// Constructors - New!
SpecEvo::SpecEvo(Grid* _grid, const FrequencyBinList& _nu, const RefFrame& _refFrame, const double& tBegin, const double tEnd, const double nBin) : grid(_grid), cellIndexMin(0), cellIndexMax(_grid->cell.size()-1), nu(_nu), refFrame(_refFrame), pLock(false) {
  double tMin,tMax;
  double dlogt=(log10(tEnd)-log10(tBegin))/nBin;
  int i;

  // Testing!
  //cout<<"SpecEvo::SpecEvo: tBegin: "<<tBegin<<", tEnd: "<<tEnd<<" nBin: "<<nBin<<", refFrame: "<<refFrame<<"\n";

  // For each temporal bin
  for (i=0;i<nBin;i++) {
    // Calculate borders
    tMin=tBegin*pow(10,i*dlogt); tMax=tBegin*pow(10,(i+1)*dlogt);
    // Append spectrum to list
    spectrum.push_back(Spectrum(grid,nu,refFrame,tMin,tMax));
    // Append exported and syncronized flags
    exported.push_back(false); synchronized.push_back(false);
  }
}

SpecEvo::SpecEvo(Grid* _grid, const int& _cellIndexMin, const int& _cellIndexMax, const FrequencyBinList& _nu, const RefFrame& _refFrame, const double& tBegin, const double tEnd, const double nBin) : grid(_grid), cellIndexMin(_cellIndexMin), cellIndexMax(_cellIndexMax), nu(_nu), refFrame(_refFrame), pLock(false) {
  double tMin,tMax;
  double dlogt=(log10(tEnd)-log10(tBegin))/nBin;
  int i;

  // For each temporal bin
  for (i=0;i<nBin;i++) {
    // Calculate borders
    tMin=tBegin*pow(10,i*dlogt); tMax=tBegin*pow(10,(i+1)*dlogt);
    // Append spectrum to list
    spectrum.push_back(Spectrum(grid,cellIndexMin,cellIndexMax,nu,refFrame,tMin,tMax));
    // Append exported and synchronized flags
    exported.push_back(false); synchronized.push_back(false);
  }
}

// Reset - New!
void SpecEvo::reset() {
  int i;

  // Reset each spectrum
  for (i=0;i<spectrum.size();i++) spectrum[i].reset();
}

// Get lock-flag
bool SpecEvo::lock() const {
  return pLock;
}

// Set lock-flag
void SpecEvo::lock(const bool& _lock) {
  pLock=_lock;
}

// Synchronize (MPI) - New!
void SpecEvo::sync(boost::mpi::communicator* mpiCom) {
  double tMin;
  bool updated;
  int i;

  // If lock-flag set - Just return
  if (lock()) return;

  // For each spectrum
  for (i=0;i<spectrum.size();i++){

    // Calculate minimum time in reference frame
    if (refFrame==rfDistantObserver) tMin=grid->tMin()-((CellSS*) grid->cell[cellIndexMax])->rMaxT(grid->tMin())/c;
    else tMin=grid->tMin();

    // Determine if spectrum has been updated
    if (tMin>spectrum[i].tMax) updated=true;
    else updated=false;

    // If spectrum updated and not synchronized
    if (updated and not synchronized[i]) { 
      
      // Synchronize (MPI)
      spectrum[i].sync(mpiCom);

      // Mark spectrum as syncronized
      synchronized[i]=true;

      // Wait for all processes to reach this point
      mpiCom->barrier();
    }
  }
}

// Update - New!
void SpecEvo::update(const Packet& packet) {
  double t;
  int i;

  // If lock-flag set - Just return
  if (lock()) return;

  // Calculate time in specified reference frame
  if (refFrame==rfDistantObserver) t=packet.t-(packet.position*packet.direction)/c;
  else t=packet.t;  

  // If outside outer boundary and not g-packet
  if (packet.position.norm()>=((CellSS*) grid->cell[cellIndexMax])->rMaxT(packet.t) and packet.type!=gPacket) {

    // For each spectrum
    for (i=0;i<spectrum.size();i++) {

      // If within temporal bin - Update and break
      if (t>=spectrum[i].tMin and t<=spectrum[i].tMax) { spectrum[i].update(packet); break; }
    }
  }
}

// Export - New!
void SpecEvo::exportUpdated(const string& specDir, const string& prefix) {
  stringstream specTotalFile,specTransFile,specCellFile,specEvoFile;
  double tMin,t;
  bool updated;
  ofstream file;
  int i;

  // Build spectral evolution file name
  specEvoFile.str(""); specEvoFile<<specDir<<"/"<<prefix<<"-evo.dat";

  // Open spectral evolution file
  file.open(specEvoFile.str().c_str(),ofstream::app);
  // Error handling
  if (file.fail()) {cerr<<"Error\n"; return;}

  // Testing!
  //cout<<"SpecEvo::exportUpdated: spectrum.size: "<<spectrum.size()<<"\n";

  // For each spectrum
  for (i=0;i<spectrum.size();i++){

    // Calculate minimum time in reference frame
    if (refFrame==rfDistantObserver) tMin=grid->tMin()-((CellSS*) grid->cell[cellIndexMax])->rMaxT(grid->tMin())/c;
    else tMin=grid->tMin();

    // Determine if spectrum has been updated
    if (tMin>spectrum[i].tMax) updated=true;
    else updated=false;

    // Testing!
    //cout<<"SpecEvo::exportUpdated: i: "<<i<<", tMin: "<<tMin<<", grid.tMin: "<<grid->tMin()<<", spectrum.tMax: "<<spectrum[i].tMax<<", updated: "<<updated<<", exported: "<<exported[i]<<"\n";

    // If spectrum updated and not exported
    if (updated and not exported[i]) {

      // Build (total, transition and cell) spectrum file names
      specTotalFile.str(""); specTotalFile<<specDir<<"/"<<prefix<<"-"<<i<<".dat";
      specTransFile.str(""); specTransFile<<specDir<<"/"<<prefix<<"-trans-"<<i<<".dat";
      specCellFile.str(""); specCellFile<<specDir<<"/"<<prefix<<"-cell-"<<i<<".dat"; 

      // Export (total, transition and cell) spectrum to file
      spectrum[i].exportSpecTotal(specTotalFile.str());
      spectrum[i].exportSpecTrans(specTransFile.str());
      spectrum[i].exportSpecCell(specCellFile.str());

      // Calculate average bin time
      t=(spectrum[i].tMin+spectrum[i].tMax)/2;

      // Write time, luminosity and (total, transition and cell) spectrum file names to spectral evolution file
      file<<t<<"\t"<<spectrum[i].L()<<"\t"<<specTotalFile.str()<<"\t"<<specTransFile.str()<<"\t"<<specCellFile.str()<<"\n";

      // Mark spectrum as exported
      exported[i]=true;
    }
  }

  // Close spectral evolution file
  file.close();
}

// Get memory allocated - New!
double SpecEvo::mem() const {
  double _mem=sizeof(*this);
  int i;

  // Add memory allocated for exported vector
  for (i=0;i<exported.size();i++) _mem+=sizeof(bool);

  // Add memory allocated for synchronized vectpr
  for (i=0;i<synchronized.size();i++) _mem+=sizeof(bool);

  // Add memory allocated for each spectrum
  for (i=0;i<spectrum.size();i++) _mem+=spectrum[i].mem();

  return _mem;
}

// Constructors
PacketStat::PacketStat(Grid* _grid, const bool _createSpectrum, const FrequencyBinList _nu) : grid(_grid), cellIndexMin(0), cellIndexMax(_grid->cell.size()-1), createSpectrum(_createSpectrum), nu(_nu) {
  int i,j;

  pL=0;
  pLC=0;
  pNPacket=0; // New!
  pNPacketInject=0; // New!
  pEBoost=0; // New!
  pEBoostInject=0; //New!
  Llambda.resize(nu.size(),0);
  LlambdaFreeFree.resize(nu.size(),0); // New!
  LlambdaBoundFree.resize(grid->cell[cellIndexMax]->atom.size()); // New!
  for (i=0;i<grid->cell[cellIndexMax]->atom.size();i++) {
    LlambdaBoundFree[i].resize(grid->cell[cellIndexMax]->atom[i].iState.size());
    for (j=0;j<grid->cell[cellIndexMax]->atom[i].iState.size();j++) {
      LlambdaBoundFree[i][j].resize(nu.size(),0);
    }
  }
  LlambdaBoundBound.resize(grid->cell[cellIndexMax]->atom.size()); // New!
  for (i=0;i<grid->cell[cellIndexMax]->atom.size();i++) {
    LlambdaBoundBound[i].resize(grid->cell[cellIndexMax]->atom[i].iState.size());
    for (j=0;j<grid->cell[cellIndexMax]->atom[i].iState.size();j++) {
      LlambdaBoundBound[i][j].resize(nu.size(),0);
    }
  }
  LlambdaCell.resize(cellIndexMax-cellIndexMin+1); // New!
  for (i=0;i<=(cellIndexMax-cellIndexMin);i++) LlambdaCell[i].resize(nu.size(),0); // New!
  LlambdaSubCell.resize(cellIndexMax-cellIndexMin+1); // New!
  for (i=0;i<=(cellIndexMax-cellIndexMin);i++) {
    if (grid->cell[cellIndexMin+i]->type()==cellContainer) {
      LlambdaSubCell[i].resize(grid->cell[cellIndexMin+i]->subCell.size());
      for (j=0;j<grid->cell[cellIndexMin+i]->subCell.size();j++) LlambdaSubCell[i][j].resize(nu.size(),0);
    }
    else LlambdaSubCell[i].resize(0);
  }
  specEvoObs=NULL; // New!
  eventStat.resize(cellIndexMax-cellIndexMin+1);
}

PacketStat::PacketStat(Grid* _grid, const int& _cellIndexMin, const int& _cellIndexMax, const bool _createSpectrum, const FrequencyBinList _nu) : grid(_grid), cellIndexMin(_cellIndexMin), cellIndexMax(_cellIndexMax), createSpectrum(_createSpectrum), nu(_nu) {
  int i,j;

  pL=0;
  pLC=0;
  pNPacket=0; // New!
  pNPacketInject=0; // New!
  pEBoost=0; // New!
  pEBoostInject=0; // New!
  Llambda.resize(nu.size(),0);
  LlambdaFreeFree.resize(nu.size(),0); // New!
  LlambdaBoundFree.resize(grid->cell[cellIndexMax]->atom.size()); // New!
  for (i=0;i<grid->cell[cellIndexMax]->atom.size();i++) {
    LlambdaBoundFree[i].resize(grid->cell[cellIndexMax]->atom[i].iState.size());
    for (j=0;j<grid->cell[cellIndexMax]->atom[i].iState.size();j++) {
      LlambdaBoundFree[i][j].resize(nu.size(),0);
    }
  }
  LlambdaBoundBound.resize(grid->cell[cellIndexMax]->atom.size()); // New!
  for (i=0;i<grid->cell[cellIndexMax]->atom.size();i++) {
    LlambdaBoundBound[i].resize(grid->cell[cellIndexMax]->atom[i].iState.size());
    for (j=0;j<grid->cell[cellIndexMax]->atom[i].iState.size();j++) {
      LlambdaBoundBound[i][j].resize(nu.size(),0);
    }
  }
  LlambdaCell.resize(cellIndexMax-cellIndexMin+1); // New!
  for (i=0;i<=(cellIndexMax-cellIndexMin);i++) LlambdaCell[i].resize(nu.size(),0); // New!
  LlambdaSubCell.resize(cellIndexMax-cellIndexMin+1); // New!
  for (i=0;i<=(cellIndexMax-cellIndexMin);i++) {
    if (grid->cell[cellIndexMin+i]->type()==cellContainer) {
      LlambdaSubCell[i].resize(grid->cell[cellIndexMin+i]->subCell.size());
      for (j=0;j<grid->cell[cellIndexMin+i]->subCell.size();j++) LlambdaSubCell[i][j].resize(nu.size(),0);
    }
    else LlambdaSubCell[i].resize(0);
  }
  specEvoObs=NULL; // New!
  eventStat.resize(cellIndexMax-cellIndexMin+1);
}

// Update packet statistics
//void PacketStat::update(const Packet& packet) {
void PacketStat::update(Packet& packet) { // Changed!
  //int i;
  list<Packet>::iterator child;

  // Update spectrum if requested
  if (createSpectrum) updateSpectrum(packet);

  // Update spectrum per transition type if requested
  if (createSpectrum) updateSpectrumTrans(packet);

  // Update spectrum per cell if requested
  if (createSpectrum) updateSpectrumCell(packet);

  // Update observed spectral evolution if requested - New!
  if (specEvoObs!=NULL) specEvoObs->update(packet);

  // Update observed luminosity
  updateL(packet);

  // Update number of packets - New!
  updateNPacket(packet);

  // Update energy boost - New!
  updateEBoost(packet);

  // If packet history provided - Update event counters
  //if (packet.history.size()>0) for (i=0;i<packet.history.size();i++) updateEventStat(packet.history[i]);

  // Update child packet statistics
  //for (i=0;i<packet.child.size();i++) update(packet.child[i]);
  for (child=packet.child.begin();child!=packet.child.end();child++) update((*child));
}

// Update injected packet statistics
void PacketStat::updateInject(Packet& packet) {
  //int i;
  list<Packet>::iterator child;

  // Update number of packets - New!
  updateNPacketInject(packet);

  // Update energy boost - New!
  updateEBoostInject(packet);

  // Update child packet injection statistics
  for (child=packet.child.begin();child!=packet.child.end();child++) updateInject((*child));
}

// Reset packet statistics
void PacketStat::reset() {
  int i,j;
  pL=0;
  pLC=0;
  pNPacket=0; // New!
  pNPacketInject=0; // New!
  pEBoost=0; // New!
  pEBoostInject=0; // New!
  Llambda.resize(0);
  Llambda.resize(nu.size(),0);
  LlambdaFreeFree.resize(0); // New!
  LlambdaFreeFree.resize(nu.size(),0); // New!
  LlambdaBoundFree.resize(grid->cell[cellIndexMax]->atom.size()); // New!
  for (i=0;i<grid->cell[cellIndexMax]->atom.size();i++) {
    LlambdaBoundFree[i].resize(grid->cell[cellIndexMax]->atom[i].iState.size());
    for (j=0;j<grid->cell[cellIndexMax]->atom[i].iState.size();j++) {
      LlambdaBoundFree[i][j].resize(0);
      LlambdaBoundFree[i][j].resize(nu.size(),0);
    }
  }
  LlambdaBoundBound.resize(grid->cell[cellIndexMax]->atom.size()); // New!
  for (i=0;i<grid->cell[cellIndexMax]->atom.size();i++) {
    LlambdaBoundBound[i].resize(grid->cell[cellIndexMax]->atom[i].iState.size());
    for (j=0;j<grid->cell[cellIndexMax]->atom[i].iState.size();j++) {
      LlambdaBoundBound[i][j].resize(0);
      LlambdaBoundBound[i][j].resize(nu.size(),0);
    }
  }
  LlambdaCell.resize(0); // New!
  LlambdaCell.resize(cellIndexMax-cellIndexMin+1); // New!
  for (i=0;i<=(cellIndexMax-cellIndexMin);i++) LlambdaCell[i].resize(nu.size(),0); // New!
  LlambdaSubCell.resize(cellIndexMax-cellIndexMin+1); // New!
  for (i=0;i<=(cellIndexMax-cellIndexMin);i++) {
    if (grid->cell[cellIndexMin+i]->type()==cellContainer) {
      LlambdaSubCell[i].resize(grid->cell[cellIndexMin+i]->subCell.size());
      for (j=0;j<grid->cell[cellIndexMin+i]->subCell.size();j++) {
	LlambdaSubCell[i][j].resize(0);
	LlambdaSubCell[i][j].resize(nu.size(),0);
      }
    }
    else LlambdaSubCell[i].resize(0);
  }
  for (i=0;i<eventStat.size();i++) eventStat[i].reset();
}

// Reset packet statistics
void PacketStat::reset(Grid* _grid, const int& _cellIndexMin, const int& _cellIndexMax, const bool _createSpectrum, const FrequencyBinList _nu) {
  int i,j;
  grid=_grid;
  cellIndexMin=_cellIndexMin;
  cellIndexMax=_cellIndexMax;
  createSpectrum=_createSpectrum;
  nu=_nu;
  pL=0;
  pLC=0;
  pNPacket=0; // New!
  pNPacketInject=0; // New!
  pEBoost=0; // New!
  pEBoostInject=0; // New!
  Llambda.resize(0);
  Llambda.resize(nu.size(),0);
  LlambdaFreeFree.resize(0); // New!
  LlambdaFreeFree.resize(nu.size(),0); // New!
  LlambdaBoundFree.resize(grid->cell[cellIndexMax]->atom.size()); // New!
  for (i=0;i<grid->cell[cellIndexMax]->atom.size();i++) {
    LlambdaBoundFree[i].resize(grid->cell[cellIndexMax]->atom[i].iState.size());
    for (j=0;j<grid->cell[cellIndexMax]->atom[i].iState.size();j++) {
      LlambdaBoundFree[i][j].resize(0);
      LlambdaBoundFree[i][j].resize(nu.size(),0);
    }
  }
  LlambdaBoundBound.resize(grid->cell[cellIndexMax]->atom.size()); // New!
  for (i=0;i<grid->cell[cellIndexMax]->atom.size();i++) {
    LlambdaBoundBound[i].resize(grid->cell[cellIndexMax]->atom[i].iState.size());
    for (j=0;j<grid->cell[cellIndexMax]->atom[i].iState.size();j++) {
      LlambdaBoundBound[i][j].resize(0);
      LlambdaBoundBound[i][j].resize(nu.size(),0);
    }
  }
  LlambdaCell.resize(0); // New!
  LlambdaCell.resize(cellIndexMax-cellIndexMin+1); // New!
  for (i=0;i<=(cellIndexMax-cellIndexMin);i++) LlambdaCell[i].resize(nu.size(),0); // New!
  LlambdaSubCell.resize(cellIndexMax-cellIndexMin+1); // New!
  for (i=0;i<=(cellIndexMax-cellIndexMin);i++) {
    if (grid->cell[cellIndexMin+i]->type()==cellContainer) {
      LlambdaSubCell[i].resize(grid->cell[cellIndexMin+i]->subCell.size());
      for (j=0;j<grid->cell[cellIndexMin+i]->subCell.size();j++) {
	LlambdaSubCell[i][j].resize(0);
	LlambdaSubCell[i][j].resize(nu.size(),0);
      }
    }
    else LlambdaSubCell[i].resize(0);
  }
  eventStat.resize(0);
  eventStat.resize(cellIndexMax-cellIndexMin+1);
}

// Update observed luminosity
void PacketStat::updateL(const Packet& packet) {
  double dt;

  // Exclude and return on invalid data
  if (!packet.valid()) { cout<<"PacketStat::updateL: Warning: Invalid data, packet excluded\n"; return; }

  dt=grid->tMax()-grid->tMin();

  if ((packet.position.norm()>=((CellSS*) grid->cell[cellIndexMax])->rMaxT(packet.t)) && (packet.t>=grid->tMin() && packet.t<=grid->tMax()) && packet.type!=gPacket) {
    #pragma omp atomic
    pL+=packet.E/dt;
    #pragma omp atomic
    pLC+=packet.ECMF/dt;
  }
}

// Update number of packets - New!
void PacketStat::updateNPacket(Packet& packet) {
  double dt;

  // Exclude and return on invalid data
  if (!packet.valid()) { cout<<"PacketStat::updateNPacket: Warning: Invalid data, packet excluded\n"; return; }

  if (not packet.terminated()) {
    #pragma omp atomic
    pNPacket++;
  }
}

// Update injected number of packets - New!
void PacketStat::updateNPacketInject(Packet& packet) {
  double dt;

  // Exclude and return on invalid data
  if (!packet.valid()) { cout<<"PacketStat::updateNPacket: Warning: Invalid data, packet excluded\n"; return; }

  if (not packet.terminated()) {
    #pragma omp atomic
    pNPacketInject++;
  }
}

// Update energy boost - New!
void PacketStat::updateEBoost(Packet& packet) {
  double dt;

  // Exclude and return on invalid data
  if (!packet.valid()) { cout<<"PacketStat::updateEBoost: Warning: Invalid data, packet excluded\n"; return; }

  if (not packet.terminated()) {
    #pragma omp atomic
    pEBoost+=(1/packet.boostFactor);
  }
}

// Update injected energy boost - New!
void PacketStat::updateEBoostInject(Packet& packet) {
  double dt;

  // Exclude and return on invalid data
  if (!packet.valid()) { cout<<"PacketStat::updateEBoost: Warning: Invalid data, packet excluded\n"; return; }

  if (not packet.terminated()) {
    #pragma omp atomic
    pEBoostInject+=(1/packet.boostFactor);
  }
}

//Update event statistics
void PacketStat::updateEvent(const PacketEvent& event) {

  // Need to check that index is within limits here !!!

  // Update cell event statistics
  eventStat[event.cellIndex-cellIndexMin].update(event);
}

// Update observed spectrum
void PacketStat::updateSpectrum(const Packet& packet) {
  double dLambda, dt;
  FrequencyBinIterator nuBin;

  // Exclude and return on invalid data
  if (!packet.valid()) { cout<<"PacketStat::updateL: Warning: Invalid data, packet excluded\n"; return; }

  if ((packet.position.norm()>=((CellSS*) grid->cell[cellIndexMax])->rMaxT(packet.t)) && (packet.t>=grid->tMin() && packet.t<=grid->tMax()) && packet.type!=gPacket) {
    // Find frequency bin
    nuBin=nu.find(packet.nu);
    // If matching bin found
    if (nuBin!=nu.end()) {
      dLambda=(c/nuBin->min-c/nuBin->max)/AA;
      dt=grid->tMax()-grid->tMin();
      #pragma omp atomic
      Llambda[nuBin-nu.begin()]+=packet.E/(dLambda*dt);
    }
    // If not
    else {
      // How to handle this!
      cout<<"PacketStat::updateSpectrum: Packet "<<packet.index<<" with frequency "<<packet.nu<<" outside frequency bin list\n";
    }
  }
}

// Update observed spectrum per transition type - New!
void PacketStat::updateSpectrumTrans(const Packet& packet) {
  double dLambda,dt;
  FrequencyBinIterator nuBin;
  //Cell* cell; // New!
  int i,j;

  // Exclude and return on invalid data
  //if (!packet.valid()) { cout<<"PacketStat::updateL: Warning: Invalid data, packet excluded\n"; return; }

  dt=grid->tMax()-grid->tMin();

  if ((packet.position.norm()>=((CellSS*) grid->cell[cellIndexMax])->rMaxT(packet.t)) && (packet.t>=grid->tMin() && packet.t<=grid->tMax()) && packet.type!=gPacket) {

    //if (packet.lastEmissionEvent.Z==2 and packet.lastEmissionEvent.C==0) cout<<"PacketStat::updateSpectrumTrans: HeI, 1\n";

    // Find frequency bin
    nuBin=nu.find(packet.nu);
    // If matching bin found
    if (nuBin!=nu.end()) {

      dLambda=(c/nuBin->min-c/nuBin->max)/AA;

      // If last emission event is free-free
      if (packet.lastEmissionEvent.type==evFreeFreeEmission) {
        #pragma omp atomic
        LlambdaFreeFree[nuBin-nu.begin()]+=packet.E/(dLambda*dt);
      }

      // If last emission event is bound-free or bound-bound
      if (packet.lastEmissionEvent.type==evBoundFreeEmission or packet.lastEmissionEvent.type==evLineEmission) {

	// Get outermost cell - New!
	//cell=grid->cell[cellIndexMax];
	// If container - Get first subcell - New!
	//if (cell->type()==cellContainer) cell=cell->subCell[0];

	// For each atom - Changed!
	// Note, the atoms included may vary between cells, so this needs to be improved!!!
	//for (i=0;i<cell->atom.size();i++) {
	for (i=0;i<grid->cell[cellIndexMax]->atom.size();i++) {
	  // If match
	  //if (cell->atom[i].Z()==packet.lastEmissionEvent.Z) {
	  if (grid->cell[cellIndexMax]->atom[i].Z()==packet.lastEmissionEvent.Z) {
	    // For each ion
	    // Note, the ion included may vary between cells, so this needs to be improved!!!
	    //for (j=0;j<cell->atom[i].iState.size();j++) {
	    for (j=0;j<grid->cell[cellIndexMax]->atom[i].iState.size();j++) {
	      // If match
	      //if (cell->atom[i].iState[j].C()==packet.lastEmissionEvent.C) {
	      if (grid->cell[cellIndexMax]->atom[i].iState[j].C()==packet.lastEmissionEvent.C) {
		//if (packet.lastEmissionEvent.Z==2 and packet.lastEmissionEvent.C==0) cout<<"PacketStat::updateSpectrumTrans: HeI, 2\n";
		// If last emission event is bound-free
		if (packet.lastEmissionEvent.type==evBoundFreeEmission) {
                  #pragma omp atomic
		  LlambdaBoundFree[i][j][nuBin-nu.begin()]+=packet.E/(dLambda*dt);
		}
		// Else (i.e. last emission event is bound-bound)
		else {
                  #pragma omp atomic
		  LlambdaBoundBound[i][j][nuBin-nu.begin()]+=packet.E/(dLambda*dt);
		}
	      }
	    }
	  }
	}
      }
    }
    // If not
    else {
      // How to handle this!
      //cout<<"PacketStat::updateSpectrumTrans: Packet "<<packet.index<<" with frequency "<<packet.nu<<" outside frequency bin list\n";
    }
  }
}

// Update observed spectrum per cell - New!
void PacketStat::updateSpectrumCell(const Packet& packet) {
  double dLambda,dt;
  FrequencyBinIterator nuBin;
  int i,j;

  // Exclude and return on invalid data
  //if (!packet.valid()) { cout<<"PacketStat::updateL: Warning: Invalid data, packet excluded\n"; return; }

  dt=grid->tMax()-grid->tMin();

  if ((packet.position.norm()>=((CellSS*) grid->cell[cellIndexMax])->rMaxT(packet.t)) && (packet.t>=grid->tMin() && packet.t<=grid->tMax()) && packet.type!=gPacket) {

    //if (packet.lastEmissionEvent.Z==2 and packet.lastEmissionEvent.C==0) cout<<"PacketStat::updateSpectrumTrans: HeI, 1\n";

    // Find frequency bin
    nuBin=nu.find(packet.nu);
    // If matching bin found
    if (nuBin!=nu.end()) {

      dLambda=(c/nuBin->min-c/nuBin->max)/AA;

      // For each cell
      for (i=0;i<=(cellIndexMax-cellIndexMin);i++) {
	// If match
	if (grid->cell[cellIndexMin+i]->index()==packet.lastEmissionEvent.cellIndex) {
          #pragma omp atomic
	  LlambdaCell[i][nuBin-nu.begin()]+=packet.E/(dLambda*dt);
	  // If container
	  if (grid->cell[cellIndexMin+i]->type()==cellContainer) {
	    // For each subcell
	    for (j=0;j<grid->cell[cellIndexMin+i]->subCell.size();j++) {
	      // If match
	      if (grid->cell[cellIndexMin+i]->subCell[j]->index()==packet.lastEmissionEvent.subCellIndex) {
                #pragma omp atomic
		LlambdaSubCell[i][j][nuBin-nu.begin()]+=packet.E/(dLambda*dt);
	      }
	    }
	  }
	}
      }
    }
    // If not
    else {
      // How to handle this!
      //cout<<"PacketStat::updateSpectrumTrans: Packet "<<packet.index<<" with frequency "<<packet.nu<<" outside frequency bin list\n";
    }
  }
}

// Get observed luminosity
double PacketStat::L() {
  return pL;
}

// Get CMF luminosity
double PacketStat::LC() {
  return pLC;
}

// Get number of packets - New!
int PacketStat::nPacket() {
  return pNPacket;
}

// Get injected number of packets - New!
int PacketStat::nPacketInject() {
  return pNPacketInject;
}

// Get energy boost - New!
double PacketStat::EBoost() {
  return pEBoost;
}

// Get injected energy boost - New!
double PacketStat::EBoostInject() {
  return pEBoostInject;
}

// Export spectrum
void PacketStat::exportSpectrum(const string& dataFile) {
  ofstream file;
  double lambda;
  int i,j;

  // Open atom data file
  file.open(dataFile.c_str());
  // Error handling
  if (file.fail()) {cerr<<"Error\n"; return;}

  // Write spectrum to file
  for (i=nu.size()-1;i>=0;i--) {
    lambda=((c/nu[i].min+c/nu[i].max)/2)/AA;
    file<<lambda<<" "<<Llambda[i]<<"\n";
  }

  // Close file
  file.close();
  return;
}

// Export spectrum per transition type
void PacketStat::exportSpectrumTrans(const string& dataFile) {
  ofstream file;
  double lambda;
  //Cell* cell; // New!
  int i,j,k;

  // Open atom data file
  file.open(dataFile.c_str());
  // Error handling
  if (file.fail()) {cerr<<"Error\n"; return;}

  // Get outermost cell - New!
  //cell=grid->cell[cellIndexMax];
  // If container - Get first subcell - New!
  //if (cell->type()==cellContainer) cell=cell->subCell[0];

  // Write header - Changed!
  file<<"#Lambda FF";
  /*for (j=0;j<cell->atom.size();j++)
    for (k=0;k<cell->atom[j].iState.size();k++) {
      file<<" "<<cell->atom[j].iState[k].name()<<"(BF)";
      file<<" "<<cell->atom[j].iState[k].name()<<"(BB)";*/
  for (j=0;j<grid->cell[cellIndexMax]->atom.size();j++)
    for (k=0;k<grid->cell[cellIndexMax]->atom[j].iState.size();k++) {
      file<<" "<<grid->cell[cellIndexMax]->atom[j].iState[k].name()<<"(BF)";
      file<<" "<<grid->cell[cellIndexMax]->atom[j].iState[k].name()<<"(BB)";
    }
  file<<"\n";

  // Write spectrum per transition type to file - Changed!
  for (i=nu.size()-1;i>=0;i--) {
    lambda=((c/nu[i].min+c/nu[i].max)/2)/AA;
    file<<lambda<<" "<<LlambdaFreeFree[i];
    /*for (j=0;j<cell->atom.size();j++)
      for (k=0;k<cell->atom[j].iState.size();k++) {*/
    for (j=0;j<grid->cell[cellIndexMax]->atom.size();j++)
      for (k=0;k<grid->cell[cellIndexMax]->atom[j].iState.size();k++) {
	file<<" "<<LlambdaBoundFree[j][k][i]<<" "<<LlambdaBoundBound[j][k][i];
      }
    file<<"\n";
  }

  // Close file
  file.close();
  return;
}

// Export spectrum per cell - New!
void PacketStat::exportSpectrumCell(const string& dataFile) {
  ofstream file;
  double lambda;
  int i,j,k;

  // Open atom data file
  file.open(dataFile.c_str());
  // Error handling
  if (file.fail()) {cerr<<"Error\n"; return;}

  // Write header
  file<<"#Lambda Cell:";
  for (i=cellIndexMin;i<=cellIndexMax;i++) {
    file<<" "<<grid->cell[cellIndexMin+i]->index();
    if (grid->cell[cellIndexMin+i]->type()==cellContainer) {
      file<<"SubCell:";
      for (j=0;j<grid->cell[cellIndexMin+i]->subCell.size();j++) file<<" "<<grid->cell[cellIndexMin+i]->subCell[j]->index();
    }
  }
  file<<"\n";

  // Write spectrum per cell to file
  for (i=nu.size()-1;i>=0;i--) {
    lambda=((c/nu[i].min+c/nu[i].max)/2)/AA;
    file<<lambda;
    for (j=0;j<=(cellIndexMax-cellIndexMin);j++) {
      file<<" "<<LlambdaCell[j][i];
      if (grid->cell[cellIndexMin+j]->type()==cellContainer)
	for (k=0;k<grid->cell[cellIndexMin+j]->subCell.size();k++) file<<" "<<LlambdaSubCell[j][k][i];
    }
    file<<"\n";
  }

  // Close file
  file.close();
  return;
}

// Export packet statistics
void PacketStat::exportStat(const string& statFile) {
  ofstream file;
  YAML::Emitter data;
  PacketEventStat totalEventStat;
  int i;

  // Begin packet statistics map
  data<<YAML::BeginMap;

  // Export luminosity
  data<<YAML::Key<<"L"<<YAML::Value<<L();

  // Export CMF luminosity
  data<<YAML::Key<<"LC"<<YAML::Value<<LC();

  // Sum total event statistics
  for (i=0;i<eventStat.size();i++) totalEventStat+=eventStat[i];

  // Export total event statistics
  data<<YAML::Key<<"event"<<YAML::Value;
  totalEventStat.exportStat(data);

  // Begin cell sequence
  data<<YAML::Key<<"cell"<<YAML::BeginSeq;
 
  // Export event statistics for each cell
  for (i=0;i<eventStat.size();i++) {
    data<<YAML::BeginMap;
    data<<YAML::Key<<"index"<<YAML::Value<<cellIndexMin+i;
    data<<YAML::Key<<"event"<<YAML::Value;
    eventStat[i].exportStat(data);
    data<<YAML::EndMap;
  }

  // Close cell sequence
  data<<YAML::EndSeq;
  
  // End packet statistics map
  data<<YAML::EndMap;

  // Open file
  file.open(statFile.c_str());
  if (file.fail()) {cerr<<"PacketStat::export: Could not open "<<statFile<<"\n"; return;}

  // Write data to file
  file<<data.c_str();

  // Close file
  file.close();
}

// Sum statistics
void PacketStat::operator+=(const PacketStat& packetStat) {
  int i,j,k;

  // Sum observed luminosity
  pL+=packetStat.pL;

  // Sum CMF luminosity
  pLC+=packetStat.pLC;

  // Sum number of packets - New!
  pNPacket+=packetStat.pNPacket;

  // Sum injected number of packets - New!
  pNPacketInject+=packetStat.pNPacketInject;

  // Sum energy boost - New!
  pEBoost+=packetStat.pEBoost;

  // Sum injected energy boost - New!
  pEBoostInject+=packetStat.pEBoostInject;
  
  // Sum spectrum
  for (i=0;i<Llambda.size();i++) Llambda[i]+=packetStat.Llambda[i];

  // Sum free-free spectrum
  for (i=0;i<LlambdaFreeFree.size();i++) LlambdaFreeFree[i]+=packetStat.LlambdaFreeFree[i];

  // Sum bound-free spectrum
  for (i=0;i<LlambdaBoundFree.size();i++) 
    for (j=0;j<LlambdaBoundFree[i].size();j++)
      for (k=0;k<LlambdaBoundFree[i][j].size();k++) LlambdaBoundFree[i][j][k]+=packetStat.LlambdaBoundFree[i][j][k];

  // Sum bound-bound spectrum
  for (i=0;i<LlambdaBoundBound.size();i++) 
    for (j=0;j<LlambdaBoundBound[i].size();j++)
      for (k=0;k<LlambdaBoundBound[i][j].size();k++) LlambdaBoundBound[i][j][k]+=packetStat.LlambdaBoundBound[i][j][k];

  // Sum cell spectrum
  for (i=0;i<LlambdaCell.size();i++) 
    for (j=0;j<LlambdaCell[i].size();j++) LlambdaCell[i][j]+=packetStat.LlambdaCell[i][j];

  // Sum subcell spectrum
  for (i=0;i<LlambdaSubCell.size();i++) 
    for (j=0;j<LlambdaSubCell[i].size();j++) 
      for (k=0;k<LlambdaSubCell[i][j].size();k++) LlambdaSubCell[i][j][k]+=packetStat.LlambdaSubCell[i][j][k];

  // Sum event statistics
  for (i=0;i<eventStat.size();i++) eventStat[i]+=packetStat.eventStat[i];
}

// Get memory allocated - New!
double PacketStat::mem() const {
  double _mem=sizeof(*this);
  int i,j,k;

  // Add memory allocated for frequency-bin list
  for (i=0;i<nu.size();i++) _mem+=sizeof(FrequencyBin);

  // Add memory allocated for spectrum
  for (i=0;i<Llambda.size();i++) _mem+=sizeof(double);

  // Add memory allocated for free-free spectrum
  for (i=0;i<LlambdaFreeFree.size();i++) _mem+=sizeof(double);

  // Add memory allocated for bound-free spectrum
  for (i=0;i<LlambdaBoundFree.size();i++) 
    for (j=0;j<LlambdaBoundFree[i].size();j++)
      for (k=0;k<LlambdaBoundFree[i][j].size();k++) _mem+=sizeof(double);

  // Add memory allocated for bound-bound spectrum
  for (i=0;i<LlambdaBoundBound.size();i++) 
    for (j=0;j<LlambdaBoundBound[i].size();j++)
      for (k=0;k<LlambdaBoundBound[i][j].size();k++) _mem+=sizeof(double);

  // Add memory allocated for cell spectrum
  for (i=0;i<LlambdaCell.size();i++) 
    for (j=0;j<LlambdaCell[i].size();j++) _mem+=sizeof(double);

  // Add memory allocated for subcell spectrum
  for (i=0;i<LlambdaSubCell.size();i++) 
    for (j=0;j<LlambdaSubCell[i].size();j++) 
      for (k=0;k<LlambdaSubCell[i][j].size();k++) _mem+=sizeof(double);

  // Add memory event statistics
  for (i=0;i<eventStat.size();i++) _mem+=sizeof(PacketEventStat);

  // Add memory allocated for observed spectral evolution
  if (specEvoObs!=NULL) _mem+=specEvoObs->mem();

  return _mem;
}
